using System;
using System.Xml;
using System.Reflection;
using System.Diagnostics;
using System.Collections;
using Zcu.Mve.Core;
using System.ComponentModel;


namespace Zcu.Mve.Elementals
{
	/// <summary>
	/// Module reads a data object from XML file, which has to be entered
	/// by setting window of the module.
	/// </summary>
	[Zcu.Mve.Core.ModuleInfo(
    "Miroslav Vavruka, Libor Va",
    "Module for reading data object from XML file.",
     Assemblied = "2004-06-19",
     IconName = "XmlLoader.ico",
     Category = "Input"
  )]
	[PortInfo("Output", "Readed data object (IDataObject)")]
	public class XmlLoader : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Name of the file, from which a data will be readed.
		/// </summary>
		private string fileName;

        private bool paramsChanged = true;

        IDataObject data = null;


		/// <summary>
		/// List of the libraries needed to assemblage of data structure saved in file.
		/// </summary>
		private ArrayList readedLibNames;


		/// <summary>
		/// Constructor creates output port of the module and sets implicit (empty) name
		/// of the file, from which data will be readed.
		/// </summary>
		public XmlLoader()
		{
			AddOutPort("Output", typeof (Zcu.Mve.Core.IDataObject));
			this.fileName = "";
		} // XmlLoader()

        private bool reload = false;

        /// <summary>
        /// Deterimines wether or not will the file be reparsed each time data are requested. Set to tru when some other program changes the xml data, otherwise keep the false value to speed up the computation
        /// </summary>
        [Description("Deterimines wether or not will the file be reparsed each time data are requested. Set to tru when some other program changes the xml data, otherwise keep the false value to speed up the computation")]
        public bool Reload
        {
            get { return (reload); }
            set { reload = value; }
        }

		/// <summary>
		/// Get/Set the name of the file, from which data will be readed.
		/// When filename is set, type of data is readed from file and output port is changed
		/// to this type.
		/// </summary>
        [Editor(typeof(System.Windows.Forms.Design.FileNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
		public string FileName
		{
			get
			{
				return this.fileName;
			}
			set
			{
                paramsChanged = true;
				this.fileName = value;
				this.readedLibNames = new ArrayList();
				if (FileName == "")		// nezadano zadne jmeno souboru
				{
					// TODO: Opravit tak, aby bylo mozne pouze odebrat konexi ...
					RemoveOutPort("Output");
					AddOutPort("Output", typeof (Zcu.Mve.Core.DataObject));
					return;
				}
				try
				{
					XmlTextReader reader = new XmlTextReader(FileName);
					int depth = 0;
					string dataTypeName = null;					
					while (reader.Read())
					{
						if (reader.NodeType == XmlNodeType.Element)
						{
							depth++;
							switch (reader.Name)
							{
								case "libraries":
									int libCount = Int32.Parse(reader.GetAttribute("count"));
									int i = 0;
									while (reader.Read() && i < libCount)
									{	// nacteni nazvu vsech knihoven potrebnych k sestaveni datove struktury ulozene v souboru
										if (reader.NodeType == XmlNodeType.Element && reader.Name == "lib")
										{
											string libName = reader.ReadString().Trim();
											libName = libName.Substring(0, libName.Length - 4);
											readedLibNames.Add(libName);
											i++;
										}
									} // while
									break;
								case "data":
									if (depth == 2)
									{
										if (dataTypeName == null)
											dataTypeName = reader.GetAttribute("type");
									}
									break;
							} // switch
						}
						else if (reader.NodeType == XmlNodeType.EndElement)
						{
							depth--;
						}
					} // while
					reader.Close();

					if (dataTypeName == null)
					{
						throw new MveException(this, "Corrupt XML file, data type name not found. Please check the saver used to create the XML, or changes made to the file.");
					}

					Type type = Globals.FindType(readedLibNames, dataTypeName);
					if (type.FullName != GetOutPort("Output").PortType.FullName)
					{
						RemoveOutPort("Output");
						AddOutPort("Output", type);
					}
				} // try
				catch (Exception ex)
				{
					Debug.WriteLine("XmlLoader " + this.Name + ":");
					Debug.WriteLine(ex);
					throw new MveException(this, ex.Message);
				} // catch
			} // set
		} // FileName

		/// <summary>
		/// Main method of the module.
		/// It reads data from file and this data sends to the output.
		/// It finds out type of data and reads them by using the method
		/// ReadData() of relevant data stucture.
		/// </summary>
		public override void Execute()
		{
            if (paramsChanged || reload ||(data == null))
            {
                try
                {
                    XmlTextReader reader = new XmlTextReader(FileName);
                    string dataTypeName = "";
                    while (reader.Read())
                    {	// zde probehne pouze nacteni vlastnich dat, seznam knihoven je jiz nacten!
                        if (reader.NodeType == XmlNodeType.Element && reader.Name == "data")
                        {
                            dataTypeName = reader.GetAttribute("type");
                            if (dataTypeName != "")
                            {
                                IDataObject dataObject = ReadFromDataElement(reader, readedLibNames, dataTypeName);
                                SetOutput("Output", dataObject);
                                paramsChanged = false;
                            }
                            else
                                throw new MveException(this, "Data element found without a proper data type. Check the saver used to create the file, and the reader implementation.");
                        }
                    } // while()
                    reader.Close();
                } // try
                catch (Exception ex)
                {
                    Debug.WriteLine("XmlLoader " + this.Name + ":");
                    Debug.WriteLine(ex);
                    // MFR 04-08-19 Lepsi chybova hlaska.
                    throw new MveException(this, ex.Message);
                } // catch
            }
            else
            {
                SetOutput("Output", data, true);
            }
		} // Execute()

		/// <summary>
		/// Helping method, which reads type of data structure after finding them in given
		/// libraries. It creates an instance and calls its method ReadData(), which reads
		/// data.
		/// </summary>
		/// <param name="reader">Data stream from reading data from XML file.</param>
		/// <param name="libNames">Names of libraries neeeded to assemblage of data structure saved in the file.</param>
		/// <param name="dataTypeName">Name of data structure.</param>
		public static IDataObject ReadFromDataElement(XmlTextReader reader, ArrayList libNames, string dataTypeName)
		{
			if (libNames.Count != 0 && dataTypeName != "")
			{
				Type type = Globals.FindType(libNames, dataTypeName);
				IDataObject data = (IDataObject) Activator.CreateInstance(type, true);

                if (typeof(DataObject).IsAssignableFrom(type))
                {
                    foreach (string libName in libNames) //this.readedLibNames
                  {
                    ((DataObject)data).AddLibName(libName);
                  }
                }
				data.ReadData(reader);
                return ((IDataObject)data);

			}
			else
			{
				Debug.WriteLine("Bad data format in XML file.");
				throw new MveException(null, "Bad data format in XML file.");
			}
            return (null);
		} // ReadFromDataElement()

        // LVA 2005/11/02 - methods replaced by default implementation
        /*

		/// <summary>
		/// Method fills XML element named &lt;config&gt; by file name, from which data will
		/// be readed.
		/// </summary>
		/// <param name="config">Already created XML element named &lt;config&gt;, where filename will be saved.</param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			config.InnerText = this.FileName;
		} // WriteConfig()

		/// <summary>
		/// Reads name of the file from XML element passed by parameter and sets the member
		/// variable with filename, from which data will be readed.
		/// </summary>
    /// <param name="config">Already created XML element named &lt;config&gt;, where filename is saved.</param>
    public override void ReadConfig(System.Xml.XmlElement config)
		{
			// MFR - 20040604 - Pridano oriznuti nacteneho jmena souboru
			this.FileName = config.InnerText.Trim();
		} // ReadConfig()

		/// <summary>
		/// <p>
		/// Enter name of file with data object, which is going to be readed.<br /> You can
		/// choose it by clicking on the button with three dots. Standard window for file
		/// opening is showed.
		/// </p>
		/// <p>
		/// <img src="Elementals/xmlloader.jpg" />
		/// </p>
		/// </summary>
		/// <returns>Instance of setting window of module</returns>
		public override ModuleSetup InvokeSetup()
		{
			return new XmlLoaderSetup(this);
		} // InvokeSetup()
          
         */

  } // class XmlLoader

} // namespace
