using System;
using System.Xml;
using System.Diagnostics;
using Zcu.Mve.Core;
using System.ComponentModel;


namespace Zcu.Mve.Elementals
{
	/// <summary>
	/// Module receives data from input port and saves them into XML file with filename
	/// specified in setting window of module.
	/// </summary>
	[Zcu.Mve.Core.ModuleInfo("Miroslav Vavruka",
    "Module for saving data into XML file.",
     Assemblied = "2004-06-19",
     IconName = "XmlSaver.ico",
     Category = "Output")]

	[PortInfo("Input", "Data to save")]
	public class XmlSaver : Module
	{
		/// <summary>
		/// Name of file, where data will be saved.
		/// </summary>
		private string fileName;

		/// <summary>
		/// determines wether data were already saved or not.
		/// </summary>
		private bool dataSaved = false;

		/// <summary>
		/// Constructor creates input port of the module and sets implicit filename,
		/// where data will be saved.
		/// </summary>
		public XmlSaver()
		{
			AddInPort("Input", typeof (Zcu.Mve.Core.IDataObject));
			fileName = "Default_XmlSave.xml";
            this.PortConnected += new PortConnectingEventHandler(XmlSaver_OnConnectEvent);
            this.PortDisconnected += new PortDisconnectingEventHandler(XmlSaver_OnDisconnectEvent);
		}

        void XmlSaver_OnDisconnectEvent(object source, InPort portTo, out bool allowed)
        {
            if (portTo.Name == "Input")
            {
                RemoveOutPort("Output");
            }
            allowed = true;
        }

        void XmlSaver_OnConnectEvent(object source, OutPort portFrom, InPort portTo, out bool allowed)
        {
            if (portTo.Name == "Input")
            {
                AddOutPort("Output", portFrom.PortType);
            }
            allowed = true;
        } // XmlSaver()

		/// <summary>
    /// Gets/Sets name of file, where data will be saved.
    /// </summary>
        [Editor(typeof(System.Windows.Forms.Design.FileNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
		public string FileName 
		{
			get
			{
				return fileName;
			}
			set
			{
				this.fileName = value;
			}
		} // FileName

		/// <summary>
		/// It writes data received from input port into given file.
		/// Data of given data type writes into XmlElement by using method 
		/// WriteData() relevant data type.
		/// </summary>
		public override void Execute()  
		{
            if (IsAllDataSame()&&dataSaved)
            {
                IDataObject inputData = this.GetInput("Input");
                SetOutput("Output", inputData, true);
            }
            else
            {
                XmlTextWriter writer;
                try
                {
                    writer = new XmlTextWriter(FileName, System.Text.Encoding.Default);
                }
                catch (Exception)
                {
                    throw new Zcu.Mve.Core.MveException(this, "Cannot open file " + FileName + ".");
                }
                try
                {
                    IDataObject inputData = this.GetInput("Input");
                    SetOutput("Output", inputData);

                    writer.Formatting = Formatting.Indented;
                    writer.WriteStartDocument();
                    WriteDataObjectXML(writer, inputData);
                    writer.WriteEndDocument();
                }
                catch (Exception ex)
                {
                    Debug.WriteLine("XmlSaver " + this.Name + ":");
                    Debug.WriteLine(ex);
                    writer.WriteComment("Error ocured while writing XML data file!\r\n Exception message: " + ex.Message);
                    throw new MveException(this, ex.Message);
                }
                finally
                {                    
                    writer.Flush();
                    writer.Close();
                }
				dataSaved = true;
            }
		}

        public static void WriteDataObjectXML(XmlTextWriter writer, IDataObject inputData)
        {
            writer.WriteStartElement("dataobject");		// <dataobject>
            writer.WriteStartElement("libraries");		//    <libraries>
            writer.WriteAttributeString("count", inputData.LibNames.Length.ToString());
            // zapis seznamu knihoven
            foreach (string libName in inputData.LibNames)
            {
                writer.WriteStartElement("lib");		//      <lib>
                writer.WriteString(libName + ".dll");	//         libName.dll
                writer.WriteEndElement();				//      </lib>
            }
            writer.WriteEndElement();					//    </libraries>

            writer.WriteStartElement("data");			// <data type="DataType"> 
            writer.WriteAttributeString("type", inputData.GetType().FullName); // -> Name -> FullName JK - 20/2/2006
            // zapis vlastnich dat
            inputData.WriteData(writer);
            writer.WriteEndElement();					// </data>
            writer.WriteEndElement();					// </dataobject>
            
        } // Execute()

		/// <summary>
		/// Method fills XML element with name &lt;config&gt; by name of file, where data
		/// willbe saved
		/// </summary>
		/// <param name="config">Already created XML element with name "config", which will be filled by filename.</param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			config.InnerText = this.FileName;
		} // WriteConfig()

		/// <summary>
		/// Parameter of method is a XML element, from which module reads name of file,
		/// where data will be saved.
		/// </summary>
		/// <param name="config">Jiz vytvoreny XML element s nazvem "config" obsahujici jmeno vystupniho souboru.</param>
		public override void ReadConfig(System.Xml.XmlElement config)
		{
			this.FileName = config.InnerText;
		} // ReadConfig()

    /// <summary>
    /// <p>
    /// Enter the name of file, where data object will be saved.<br />You can
    /// choose it by clicking on the button with three dots. Standard window for file
    /// saving is showed.
    /// </p>
    /// <p>
    /// <img src="Elementals/xmlsaver.jpg" />
    /// </p>
    /// </summary>
    /// <returns>Instance of setting window of module</returns>
    public override ModuleSetup InvokeSetup()
		{
			return new XmlSaverSetup(this);
		} // InvokeSetup()

	} // class XmlSaver
} // namespace
