﻿using System;
using System.Xml;
using System.Runtime.InteropServices;
using Zcu.Mve.Core;

namespace Zcu.Mve.Numerics
{
	/// <summary>Complex number that consist of real and imaginary part.</summary>
	[Zcu.Mve.Core.DataObject("This structure represents a complex number of real and imaginary part.")]
	[StructLayout(LayoutKind.Sequential, Pack=1)]
	public struct Complex : IDataObject
	{
		/// <summary>Zero: 0+i0.</summary>
		public readonly static Complex Zero = new Complex(0, 0);

        /// <summary>One: 1+i0.</summary>
		public readonly static Complex One = new Complex(1, 0);

        /// <summary>Xml tag for storage pruposes.</summary>
		public readonly static string XmlTag = "cmx";

		#region Data&Accessors...
		/// <summary>Real component.</summary>
		public double Real;
		/// <summary>Imaginary component.</summary>
		public double Imag;

		/// <summary>Squared modulus.</summary>
		public double ModulusSquared { get { return (Real*Real + Imag*Imag); } }
		
        /// <summary>Modulus.</summary>
		public double Modulus { get { return Math.Sqrt(ModulusSquared); } }
		
        /// <summary>Argument.</summary>
		public double Argument
		{
			get
			{
				return Math.Atan2(Imag, Real);
				//double modulus = Modulus;
				//if (modulus < Double.Epsilon)
				//    return 0;
				//else
				//{
				//    double angle = Math.Asin(Imag / modulus);
				//    if (Real < 0) angle = Math.PI - angle;
				//    if (angle < 0) angle += 2 * Math.PI;
				//    return angle;
				//}
			}
		}

		/// <summary>Complex conjugate.</summary>
		public Complex Conjugate { get { return new Complex(Real, -Imag); } }

        /// <summary>
        /// Returns the sign of complex number.
        /// </summary>
        /// <returns>Sign.</returns>
        public Complex Sign { get { return this / this.Modulus; } }

        /// <summary>
        /// Integer power of complex number. It is faster for small exponents. For large exponents
        /// use the Complex version.
        /// </summary>
        /// <param name="exp">Integer exopnent.</param>
        /// <returns>Power of complex number.</returns>
        public Complex Power(int exp)
        {
            if (exp == 0) return Complex.One;
            if (exp < 0)
            {
                return 1/Power(-exp);
            } 
            Complex result = this;
            for (int i = 1; i < exp; i++)
            {
                result *= this;
            }
            return result;
        }

        public Complex Power(Complex exp)
        {
            double r = this.Modulus;
            double t = this.Argument;
            double c = exp.Real;
            double d = exp.Imag;
            Complex result;
            double pow = Math.Pow(r, c) * Math.Exp(-d * t);
            result.Real = pow * Math.Cos(c * t + d * Math.Log(r));
            result.Imag = pow * Math.Sin(c * t + d * Math.Log(r));
            return result;
        }

        public static Complex Power(Complex cbase, Complex exp)
        {
            Complex result = new Complex(cbase.Real,cbase.Imag);
            result.Power(exp);
            return result;
        }

		#endregion

		#region Constructors...
		/// <summary>Constructor.</summary>
		/// <param name="r">Real part.</param>
		/// <param name="i">Imaginary part.</param>
		public Complex(double r, double i) { Real = r; Imag = i; }

		/// <summary>Construct complex number from modulus and argument.</summary>
		/// <param name="modulus">Modulus.</param>
		/// <param name="argument">Argument.</param>
		/// <returns>Complex number.</returns>
		public static Complex FromModArg(double modulus, double argument) { return new Complex(modulus * Math.Cos(argument), modulus * Math.Sin(argument)); }

		#endregion

		/// <summary>String representation of number.</summary>
		public override string ToString()
		{
			if (Imag < 0)
				return String.Format("{0}-i{1}", Real, -Imag);
			else
				return String.Format("{0}+i{1}", Real, Imag);
		}

		#region Operators...
        /// <summary>
        /// Adds two complex numbers.
        /// </summary>
        /// <param name="a">Complex number.</param>
        /// <param name="b">Complex number.</param>
        /// <returns>Complex number.</returns>
		public static Complex operator+(Complex a, Complex b) { return new Complex(a.Real+b.Real, a.Imag+b.Imag); }

        /// <summary>
        /// Subtracts two complex numbers.
        /// </summary>
        /// <param name="a">Complex number.</param>
        /// <param name="b">Complex number.</param>
        /// <returns>Complex number.</returns>
		public static Complex operator-(Complex a, Complex b) { return new Complex(a.Real-b.Real, a.Imag-b.Imag); }

        /// <summary>
        /// Multiplies two complex numbers.
        /// </summary>
        /// <param name="a">Complex number.</param>
        /// <param name="b">Complex number.</param>
        /// <returns>Complex number.</returns>
		public static Complex operator*(Complex a, Complex b) { return new Complex(a.Real*b.Real - a.Imag*b.Imag, a.Imag*b.Real + a.Real*b.Imag); }

        /// <summary>
        /// Multiplies complex number by a real number.
        /// </summary>
        /// <param name="a">Complex number.</param>
        /// <param name="k">Real number.</param>
        /// <returns>Complex number.</returns>
        public static Complex operator *(Complex a, double k) { return new Complex(a.Real * k, a.Imag * k); }

        /// <summary>
        /// Multiplies real number by a complex number.
        /// </summary>
        /// <param name="a">Complex number.</param>
        /// <param name="k">Real number.</param>
        /// <returns>Complex number.</returns>
		public static Complex operator*(double k, Complex a) { return new Complex(a.Real * k, a.Imag * k); }

        /// <summary>
        /// Negates a complex number
        /// </summary>
        /// <param name="a">Complex number.</param>
        /// <returns>Negation of complex number.</returns>
		public static Complex operator-(Complex a) { return new Complex(-a.Real, -a.Imag); }

        /// <summary>
        /// Divides two complex numbers.
        /// </summary>
        /// <param name="a">Complex number.</param>
        /// <param name="b">Complex number.</param>
        /// <returns>Complex number.</returns>
        public static Complex operator /(Complex a, Complex b) { double modSq = b.ModulusSquared; return new Complex((a.Real * b.Real + a.Imag * b.Imag) / modSq, -(a.Real * b.Imag + a.Imag * b.Real) / modSq); }

        /// <summary>
        /// Divides a complex number by a real number.
        /// </summary>
        /// <param name="a">Complex number.</param>
        /// <param name="k">Real number.</param>
        /// <returns>Complex number.</returns>
		public static Complex operator/(Complex a, double k) { return new Complex(a.Real / k, a.Imag / k); }

        /// <summary>
        /// Divides a real number by a complex number.
        /// </summary>
        /// <param name="k">Real number.</param>
        /// <param name="a">Complex number.</param>
        /// <returns>Complex number.</returns>
        public static Complex operator /(double k, Complex a) { double value = a.Real * a.Real + a.Imag * a.Imag; return new Complex((k * a.Real) / value, (-k * a.Imag) / value); }

        /// <summary>
        /// Implicit conversion from double.
        /// </summary>
        /// <param name="real">Real part of complex number</param>
        /// <returns></returns>
        public static implicit operator Complex(double real)
        {
            return new Complex(real, 0);
        }

        #endregion

		#region IDataObjectImplementation...
		/// <summary>Write current data to Xml stream.</summary>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			xmlTextWriter.WriteStartElement(XmlTag);
			xmlTextWriter.WriteAttributeString("r", XmlConvert.ToString(Real));
			xmlTextWriter.WriteAttributeString("i", XmlConvert.ToString(Imag));
			xmlTextWriter.WriteEndElement();
		}

		/// <summary>Read data from Xml stream to current structure.</summary>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			double r = 0, i = 0;
			if (xmlTextReader.MoveToAttribute("r"))
				r = XmlConvert.ToDouble(xmlTextReader.Value);
			if (xmlTextReader.MoveToAttribute("i"))
				i = XmlConvert.ToDouble(xmlTextReader.Value);
			xmlTextReader.MoveToElement();
			xmlTextReader.Read();

			Real = r; Imag = i;
		}

		/// <summary>Deep copy of data object and its subelements.</summary>
		public IDataObject DeepCopy()
		{
			return new Complex(Real, Imag);
		}

		/// <summary>Check the consistency of current data structure.</summary>
		public bool CheckConsistence() { return true; }

		/// <summary>Returns array of assembly names (filename without extension!) that contain definition of current data structure and its sub elments.</summary>
		public string[] LibNames { get { return new string[] { Globals.GetLibraryName(this.GetType()) }; } }
		#endregion
	}
}  