using System;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.Threading;
using System.Windows.Forms;
using System.ComponentModel;

namespace Zcu.Mve.SimpleDirectXRenderer
{
	/// <summary>
	/// Simple DirectX based renderer for Triangle meshes
	/// </summary>
    [ModuleInfo("Libor Va", "Simple DirectX based renderer for Triangle meshes.", IconName = "SDXR.ico")]
    [PortInfo("TriangleMesh", "Input triangle mesh. When displaying an animation, then the mesh should be the frame indicated by the value at the MeshIndex port.")]
    [PortInfo("MeshIndex", "Index of the requested mesh.")]
	public class SimpleDirectXRenderer:Module
	{
		private ManualResetEvent windowDisplayed;

		private bool showHelp;

		private bool sameConnectivity = false;

        private bool showColorAttribute = false;

        /// <summary>
        /// Determines wether an attribute of ColorRGBA type will be used as color source of the object. Set the colorAttributeName to specify which attribute to be used.
        /// </summary>
        [Description("Determines wether an attribute of ColorRGBA type will be used as color source of the object. Set the colorAttributeName to specify which attribute to be used.")]
        public bool ShowColorAttribute
        {
            get
            {
                return (showColorAttribute);
            }
            set
            {
                showColorAttribute = value;
            }
        }

        private bool showVectorAttribute = false;

        /// <summary>
        /// Determines wether an attribute of ColorRGBA type will be used as color source of the object. Set the colorAttributeName to specify which attribute to be used.
        /// </summary>
        [Description("Determines wether an attribute of PointND type will be displayed. Set the vectorAttributeName to specify which attribute to be used.")]
        public bool ShowVectorAttribute
        {
            get
            {
                return (showVectorAttribute);
            }
            set
            {
                showVectorAttribute = value;
            }
        }

        private string colorAttributeName = "";

        
        /// <summary>
        /// Name of the attribute to be rendered as object color. Must be a point attribute identifier of the ColorRGBA type.
        /// </summary>
        [Description("Name of the attribute to be rendered as object color. Must be a point attribute identifier of the ColorRGBA type.")]
        public string ColorAttributeName
        {
            get
            {
                return (colorAttributeName);
            }
            set
            {
                if (value != null)
                {
                    colorAttributeName = value;
                }
                else
                {
                    throw new MveException(this, "Null passed as color attribute name. Use an empty string when color attributa name is to be determined automatically.");
                }
            }
        }

        private string vectorAttributeName = "";


        /// <summary>
        /// Name of the attribute to be rendered as lines. Must be a point attribute identifier of the PointND type.
        /// </summary>
        [Description("Name of the attribute to be rendered as lines. Must be a point attribute identifier of the PointND type.")]
        public string VectorAttributeName
        {
            get
            {
                return (vectorAttributeName);
            }
            set
            {
                if (value != null)
                {
                    vectorAttributeName = value;
                }
                else
                {
                    throw new MveException(this, "Null passed as vector attribute name. Use an empty string when vector attribute name is to be determined automatically.");
                }
            }
        }

		int meshCount = 1;

		private SimpleDirectXRendererWindow renderWindow;

        private Thread windowThread;

        /// <summary>
        /// Number of meshes to be loaded.
        /// </summary>
        [Description("Number of meshes to be loaded.")]
		public int MeshCount
		{
			get{return(meshCount);}
			set
			{
				if (value>0)
					meshCount = value;
				else throw new MveException(this, "Mesh count for SimpleDirectXRenderer must be more than zero!");
			}
		}

        /// <summary>
        /// Determines whether a help text will be initially displayed or not.
        /// </summary>
        [Description("Determines whether a help text will be initially displayed or not.")]
		public bool ShowHelp
		{
			get{return(showHelp);}
			set{showHelp = value;}
		}

        /// <summary>
        /// Determines weher the input meshes have the same connectivity or not. If set to true, the only the connectivity information form the first mesh will be used. All remaining meshes must have the same number of vertices. 
        /// Memory usage is lower when set to true.
        /// </summary>
        [Description("Determines weher the input meshes have the same connectivity or not. If set to true, the only the connectivity information form the first mesh will be used. All remaining meshes must have the same number of vertices. Memory usage is lower when set to true.")]
		public bool SameConnectivity
		{
			get{return(sameConnectivity);}
			set{sameConnectivity = value;}
		}

        /// <summary>
        /// Default constructor.
        /// </summary>
		public SimpleDirectXRenderer()
		{
			AddInPort("TriangleMesh",typeof(TriangleMesh),true,false);
			AddOutPort("MeshIndex", typeof(Integer),false);
			windowDisplayed = new ManualResetEvent(false);
		}

        System.Drawing.Size size = new System.Drawing.Size(800, 600);

        [Description("Gets or sets the size of the drawing window")]
        public System.Drawing.Size Size
        {
            get { return size; }
            set { size = value; }
        }

        /// <summary>
        /// Starts asking for input meshes by setting the output to the number of demnded mesh and asking for input.
        /// When all input meshes are retrieved, then a rendering window is shown.
        /// </summary>
		public override void Execute()
		{         
			windowThread = new Thread(new ThreadStart(ShowRenderWindow));
            AddRunningThread(windowThread);
			windowThread.IsBackground = true;
			windowDisplayed.Reset();
			windowThread.Start();			
			windowDisplayed.WaitOne();
            RemoveRunningThread(windowThread);
		}        

		private void ShowRenderWindow()
		{						
			renderWindow = new SimpleDirectXRendererWindow(this.meshCount, sameConnectivity, showColorAttribute, colorAttributeName, showVectorAttribute, vectorAttributeName,size);
            //renderWindow.TransparencyKey = System.Drawing.Color.FromArgb(0, 0, 255);
			TriangleMesh input = null;
            try
            {
                for (int i = 0; i < meshCount; i++)
                {
                    SetOutput("MeshIndex", new Integer(i), false);
                    UpdateInput("TriangleMesh");
                    input = (TriangleMesh)GetInput("TriangleMesh");
                    renderWindow.SetMesh(i, input);
                    ProgressInfo((float)i / meshCount);
                }
                renderWindow.SetTransforms(input);
                renderWindow.ShowHelp = showHelp;

                windowDisplayed.Set();
                renderWindow.Show();
            }
            catch(ThreadAbortException)
            {
                return;
            }
			while (renderWindow.Created)
			{
				renderWindow.Render();
				Application.DoEvents();
			}
		}
	}
}
