using System;
using System.Windows.Forms;
using Zcu.Mve.Visualization;
using Zcu.Mve.Core;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Drawing;
using System.Threading;
using System.ComponentModel;

namespace Zcu.Mve.SimpleDirectXRenderer
{
	/// <summary>
	/// Summary description for SimpleDirectXRendererWindow.
	/// </summary>

	internal class SimpleStereoDirectXRendererWindow:Form
	{

		bool deviceLost;

		Form frm1;
		Form frm2;

		SwapChain sc1;
		SwapChain sc2;

		Surface ss1;
		Surface ss2;

		Device device = null;
		PresentParameters[] presentParameters = null;
		

		UniformDataArray[] points;

		bool showHelp = true;

		bool autorun = false;

		bool sameCon;
        bool showColor = false;

        bool showNormals = false;

        string colorAttrName = "";

		int meshCount = 0;

		int currentMesh = -1;

		private int makingScreenShots = 1;

		#region Properties

		public int MeshCount
		{
			get{return(meshCount);}
		}

		public bool ShowHelp
		{
			get{return(showHelp);}
			set{showHelp = value;}
		}

		#endregion
		
		VertexBuffer[] vbs;
		IndexBuffer[] ibs, nibs;
		Matrix world = Matrix.Identity;
		Matrix toCenter = Matrix.Identity;
		Matrix transformation = Matrix.Identity;
		Matrix scale = Matrix.Identity;

		float lightAzimuth, lightElevation;

		UniformDataArray[] triangles;
		UniformDataArray[] normals;
        UniformDataArray[] colors;
		Zcu.Mve.Visualization.BoundingBox bb = null;

		bool mouseLeft = false, mouseRight = false, lightMoving = false;
		Point mousePosition;
		bool wireFrame = false;
		bool lighting = true;
		int culling = 1;

		string helpString = "Simple DirectX Renderer\n" + 
							"-----------------------\n\n" + 
							"Left mouse - rotate object\n" +
							"Right mouse - move object\n" +
							"Mouse wheel - scale object\n" +
							"'l' + mouse move - move light\n\n" +

							"    'w'\n" +
							"'a' 's' 'd' - move object\n\n" +

							"'r' 'f' - scale object\n\n" +

							"'t' 'g' - rotate object elevation\n" +
							"'h' 'j' - rotate obejct azimuth\n\n" +

							"'z' - wireframe/solid\n" +
							"'x' - lighting on/off\n" +
							"'c' - culling off/ccw/cw\n" +
                            "'n' - show normals on/off\n" +
							"'v' - help on/off\n\n" +
							"'<' '>' - previous/next frame\n" +
							"'m' - play 25fps";

		Microsoft.DirectX.Direct3D.Font helpFont;

		public SimpleStereoDirectXRendererWindow()
		{
			
		}

		public void SetMesh(int index, Zcu.Mve.Visualization.TriangleMesh mesh)
		{
			Zcu.Mve.Visualization.BoundingBox bb = mesh.BoundingBox;
			this.points[index] = mesh.Points;
			if (sameCon)
			{
				if (index == 0)
					this.triangles[index] = (mesh.GetCells(typeof(Triangle)))[0];
			}
			else
			{
				this.triangles[index] = (mesh.GetCells(typeof(Triangle)))[0];
			}

            if (showColor)
            {
                if (colorAttrName == "")
                {
                    foreach (string attrId in mesh.GetPointAttrsIds())
                    {
                        UniformDataArray attr = mesh.GetPointAttrs(attrId);
                        if (attr.ElemType == typeof(ColorRGBA))
                            colors[index] = attr;
                    }
                    if (colors[index] == null)
                    {
                        throw new MveException(this, "ColorRGBA attribute not found."); // TODO: dodat odkaz na modul
                    }
                }
                else
                {
                    colors[index] = mesh.GetPointAttrs(colorAttrName);
                }
            }

			this.normals[index] = mesh.GetPointAttrs("normals");
			initBuffers(index, bb);
			currentMesh = index;
		}

		public void SetTransforms(Zcu.Mve.Visualization.TriangleMesh mesh)
		{
			this.bb = mesh.BoundingBox;
			this.toCenter = Matrix.Identity*Matrix.Translation((float)-bb.MidPoint[0],(float)-bb.MidPoint[1],(float)-bb.MidPoint[2]);
			float h = (float)(bb.MaxPoint[1]-bb.MinPoint[1]);
			float w = (float)(bb.MaxPoint[0]-bb.MinPoint[0]);
			float ar = ((float)this.ClientSize.Width)/this.ClientSize.Height;
			float k1 = 1/h;
			float k2 = ar/w;
            if (h == 0) scale *= Matrix.Scaling(k2, k2, k2);
            else
            if (w == 0) scale *= Matrix.Scaling(k1, k1, k1);
            else
			if (k1>k2) scale *= Matrix.Scaling(k1,k1,k1); else scale *= Matrix.Scaling(k2,k2,k2);
		}

		/// <summary>
		/// Constructor for multimesh renderer window
		/// </summary>
		/// <param name="meshCount">Number of meshes to be rendered</param>
        /// <param name="sameConnectivity">Determines wether meshes do have same connectivity or not.</param>
        /// <param name="showColorAttribute">Determines wether the a color attribute is to be rendered.</param>
        /// <param name="colorAttributeName">Defines the name of the attribute that will be used as color. Has no meaning if <paramref name="showColorAttribute"/> is set to false</param>
		public SimpleStereoDirectXRendererWindow(int meshCount, bool sameConnectivity, bool showColorAttribute, string colorAttributeName, out bool success, out string error)
		{
            showColor = showColorAttribute;
            this.colorAttrName = colorAttributeName;
			this.meshCount = meshCount;
			sameCon = sameConnectivity;
			this.points = new UniformDataArray[meshCount];
            if (sameCon)
            {
                this.ibs = new IndexBuffer[1];
                this.nibs = new IndexBuffer[1];
            }
            else
            {
                this.ibs = new IndexBuffer[meshCount];
                this.nibs = new IndexBuffer[meshCount];
            }
			this.vbs = new VertexBuffer[meshCount];            
			if (sameCon)
				this.triangles = new UniformDataArray[1];
			else
				this.triangles = new UniformDataArray[meshCount];
			this.normals = new UniformDataArray[meshCount];
            if (showColor)
            {
                this.colors = new UniformDataArray[meshCount];
            }
			
			this.Width = 800;
			this.Height = 600;
			this.Text = "Simple DirectX Renderer for MVE2";
			if (D3DInit())
			{
				initFonts();
				this.Paint += new PaintEventHandler(SimpleDirectXRendererWindow_Paint);
				this.KeyPress += new KeyPressEventHandler(SimpleDirectXRendererWindow_KeyPress);
				this.KeyUp += new KeyEventHandler(SimpleDirectXRendererWindow_KeyUp);
				this.MouseDown += new MouseEventHandler(SimpleDirectXRendererWindow_MouseDown);
				this.MouseUp += new MouseEventHandler(SimpleDirectXRendererWindow_MouseUp);
				this.MouseMove += new MouseEventHandler(SimpleDirectXRendererWindow_MouseMove);
				this.MouseWheel += new MouseEventHandler(SimpleDirectXRendererWindow_MouseWheel);
				success = true;
				error = "";
			}
			else
			{
				success = false;
				error = "Unable to initialize device";
			}
		}

		private void initFonts()
		{
			System.Drawing.Font drawingFont = new System.Drawing.Font(FontFamily.GenericMonospace, 10);
			helpFont = new Microsoft.DirectX.Direct3D.Font(device,drawingFont);
		}

		private bool D3DInit()
		{
			try
			{
				this.frm1 = new Form();
				this.frm2 = new Form();
				this.frm1.Visible = true;
				this.frm2.Visible = true;
				this.frm1.Top = 100;
				this.frm1.Left = 100;
				this.frm2.Top = 100;
				this.frm2.Left = 500;
				this.frm1.Text = "frm1";
				this.frm2.Text = "frm2";

				DisplayMode DispMode = Manager.Adapters[Manager.Adapters.Default.Adapter].CurrentDisplayMode;
				this.presentParameters = new PresentParameters[2];

				this.presentParameters[0] = new PresentParameters();
				this.presentParameters[0].Windowed = true;
				this.presentParameters[0].BackBufferCount = 1;
				this.presentParameters[0].BackBufferFormat = DispMode.Format;
				this.presentParameters[0].BackBufferWidth = 0;
				this.presentParameters[0].BackBufferHeight = 0;
				this.presentParameters[0].SwapEffect = SwapEffect.Flip;
				this.presentParameters[0].EnableAutoDepthStencil = true;
				this.presentParameters[0].AutoDepthStencilFormat = DepthFormat.D16;
				this.presentParameters[0].DeviceWindow = this.frm1;

				this.presentParameters[1] = new PresentParameters();
				this.presentParameters[1].Windowed = true;
				this.presentParameters[1].BackBufferCount = 1;
				this.presentParameters[1].BackBufferFormat = DispMode.Format;
				this.presentParameters[1].BackBufferWidth = 0;
				this.presentParameters[1].BackBufferHeight = 0;
				this.presentParameters[1].SwapEffect = SwapEffect.Flip;
				this.presentParameters[1].EnableAutoDepthStencil = true;
				this.presentParameters[1].AutoDepthStencilFormat = DepthFormat.D16;
				this.presentParameters[1].DeviceWindow = this.frm2;

				this.device = new Device(0, DeviceType.Hardware, this.frm1, CreateFlags.HardwareVertexProcessing | CreateFlags.PureDevice, this.presentParameters);
				this.device.DeviceResizing += new CancelEventHandler(this.DoDeviceResizing);

				this.sc1 = new SwapChain(this.device, this.presentParameters[0]);
				this.sc2 = new SwapChain(this.device, this.presentParameters[1]);

				this.ss1 = this.device.CreateDepthStencilSurface(this.frm1.Width, this.frm1.Height, DepthFormat.D16, MultiSampleType.None, 1, false);
				this.ss2 = this.device.CreateDepthStencilSurface(this.frm2.Width, this.frm2.Height, DepthFormat.D16, MultiSampleType.None, 1, false);

				this.frm1.Resize += new EventHandler(this.frm1_Resize);
				this.frm2.Resize += new EventHandler(this.frm2_Resize);
			}
			catch(DirectXException dxe)
			{
				Console.WriteLine(dxe.Message);
				return(false);
			}
			return(true);
		}

		protected virtual void DoDeviceResizing(object sender, CancelEventArgs e)
		{
			if (presentParameters[0].Windowed) e.Cancel = false;
			else e.Cancel = true;
		}

		protected void frm1_Resize(object sender, System.EventArgs e)
		{
			sc1 = new SwapChain(device, presentParameters[0]);
			ss1 = device.CreateDepthStencilSurface(frm1.Width, frm1.Height, DepthFormat.D16, MultiSampleType.None, 1, false);
		}

		protected void frm2_Resize(object sender, System.EventArgs e)
		{
			sc2 = new SwapChain(device, presentParameters[1]);
			ss2 = device.CreateDepthStencilSurface(frm2.Width, frm2.Height, DepthFormat.D16, MultiSampleType.None, 1, false);
		}

		private void initBuffers(int index, Zcu.Mve.Visualization.BoundingBox bb)
		{            
            PointND diag = (PointND)bb.MaxPoint - (PointND)bb.MinPoint;
            double diagonal = Math.Sqrt(diag.dotProduct(diag));
            double multiplier = diagonal * 0.05;
            int pCount = points[index].Count;  
            if (showColor)
            {                             
                vbs[index] = new VertexBuffer(typeof(CustomVertex.PositionNormalColored),
                    pCount*2,
                    device,
                    Usage.WriteOnly,
                    CustomVertex.PositionNormalColored.Format,
                    Pool.Managed);
                CustomVertex.PositionNormalColored[] pnt = (CustomVertex.PositionNormalColored[])vbs[index].Lock(0, 0);
                for (int i = 0; i < pCount; i++)
                {
                    Point3D point = (Point3D)points[index][i];
                    Normal3D normal = (Normal3D)normals[index][i];
                    pnt[i].Position = new Vector3((float)point.X, (float)point.Y, (float)point.Z);
                    pnt[i].Nx = (float)-normal.Nx;
                    pnt[i].Ny = (float)-normal.Ny;
                    pnt[i].Nz = (float)-normal.Nz;
                    pnt[i].Color = ((ColorRGBA)(colors[index][i])).GetInt();
                }
                for (int i = 0; i < pCount; i++)
                {
                    Point3D point = (Point3D)points[index][i];
                    Normal3D normal = (Normal3D)normals[index][i];
                    pnt[i + pCount].Position = new Vector3((float)(point.X + normal.Nx * multiplier), (float)(point.Y + normal.Ny * multiplier), (float)(point.Z + normal.Nz * multiplier));
                    pnt[i+pCount].Nx = (float)-normal.Nx;
                    pnt[i+pCount].Ny = (float)-normal.Ny;
                    pnt[i+pCount].Nz = (float)-normal.Nz;
                    pnt[i+pCount].Color = ((ColorRGBA)(colors[index][i])).GetInt();
                }
                vbs[index].Unlock();                
            }
            else
            {
                vbs[index] = new VertexBuffer(typeof(CustomVertex.PositionNormal),
                    2*pCount,
                    device,
                    Usage.WriteOnly,
                    CustomVertex.PositionNormalTextured.Format,
                    Pool.Managed);
                CustomVertex.PositionNormal[] pnt = (CustomVertex.PositionNormal[])vbs[index].Lock(0, 0);
                for (int i = 0; i < pCount; i++)
                {
                    Point3D point = (Point3D)points[index][i];
                    Normal3D normal = (Normal3D)normals[index][i];
                    pnt[i].Position = new Vector3((float)point.X, (float)point.Y, (float)point.Z);
                    pnt[i].Nx = (float)-normal.Nx;
                    pnt[i].Ny = (float)-normal.Ny;
                    pnt[i].Nz = (float)-normal.Nz;
                }
                for (int i = 0; i < pCount; i++)
                {
                    Point3D point = (Point3D)points[index][i];
                    Normal3D normal = (Normal3D)normals[index][i];
                    pnt[i + pCount].Position = new Vector3((float)(point.X + normal.Nx * multiplier), (float)(point.Y + normal.Ny * multiplier), (float)(point.Z + normal.Nz * multiplier));
                    pnt[i+pCount].Nx = (float)-normal.Nx;
                    pnt[i+pCount].Ny = (float)-normal.Ny;
                    pnt[i+pCount].Nz = (float)-normal.Nz;
                }
                vbs[index].Unlock();                
            }
            if ((sameCon && (index == 0)) || (!sameCon))
            {
                ibs[index] = new IndexBuffer(typeof(ushort), triangles[index].Count * 3, device, Usage.WriteOnly, Pool.Managed);
                ushort[] t = (ushort[])ibs[index].Lock(0, 0);
                for (ushort i = 0; (i < triangles[index].Count)&&(i<ushort.MaxValue); i++)
                {
                    Triangle triangle = (Triangle)triangles[index][i];
                    t[3 * i] = (ushort)triangle.V1;
                    t[3 * i + 1] = (ushort)triangle.V2;
                    t[3 * i + 2] = (ushort)triangle.V3;
                }
                ibs[index].Unlock();

                nibs[index] = new IndexBuffer(typeof(ushort), pCount*2, device, Usage.WriteOnly, Pool.Managed);
                t = (ushort[])nibs[index].Lock(0, 0);
                for (ushort i = 0; (i < pCount) && (i < ushort.MaxValue); i++)
                {
                    t[2 * i] = i;
                    t[2 * i + 1] = (ushort)(i + pCount);
                }
                nibs[index].Unlock();
            }
		}

		private void setupDevice()
		{
			if (wireFrame)
				device.RenderState.FillMode = FillMode.WireFrame;			
			else
				device.RenderState.FillMode = FillMode.Solid;
			if (lighting)
				device.RenderState.Lighting = true;
			else
				device.RenderState.Lighting = false;

			device.RenderState.CullMode = (Cull)culling;
			device.RenderState.NormalizeNormals = true;
			device.Lights[0].Type = LightType.Directional;
			device.Lights[0].Position = new Vector3(10,10,10);
			//device.Lights[0].Direction = new Vector3(1,0,-1);
			device.Lights[0].Direction = new Vector3((float)(Math.Cos(-lightAzimuth)*Math.Cos(lightElevation)),
													(float)(Math.Sin(lightElevation)),
													(float)(Math.Sin(-lightAzimuth)*Math.Cos(lightElevation)));
			device.Lights[0].Diffuse = Color.FromArgb(128,128,128);
			device.Lights[0].Enabled = true;
		}

		protected override void OnPaintBackground(PaintEventArgs pevent)
		{
			
		}


		public void Render()
		{
			if (deviceLost)
				{
					try 
					{ 
						device.TestCooperativeLevel(); 
					} 
					catch (DeviceLostException) 
					{ 
						deviceLost = true;
						return;
					} 
					catch (DeviceNotResetException) 
					{ 
						try 
						{
							device.Reset(this.presentParameters);						 
							deviceLost = false; 
						} 
						catch (DeviceLostException) 
						{	 
							deviceLost = true;
							return;
						} 
					} 
				}

				if (frm1.Created && !sc1.Disposed && makingScreenShots == 1)
				{
					Surface s1 = this.sc1.GetBackBuffer(0, BackBufferType.Mono);
					this.device.SetRenderTarget(0, s1);
					this.device.DepthStencilSurface = this.ss1;
					s1.ReleaseGraphics();
				}
				else
					if (this.frm2.Created && !this.sc2.Disposed && makingScreenShots == 2)
					{
						Surface s2 = this.sc2.GetBackBuffer(0, BackBufferType.Mono);
						this.device.SetRenderTarget(0, s2);
						this.device.DepthStencilSurface = this.ss2;
						s2.ReleaseGraphics();
					}
			

			device.Clear(ClearFlags.Target|ClearFlags.ZBuffer,System.Drawing.Color.Black,1.0f,0);
			device.BeginScene();

			setupDevice();

			device.Transform.Projection = 
				Matrix.PerspectiveFovLH( (float)Math.PI/4, 
				(float)this.ClientSize.Width/this.ClientSize.Height, 
				0.01f, 50f);
			Matrix view = Matrix.Identity;
			view *= Matrix.Translation(0,0,5);
			
			switch (makingScreenShots)
			{
				case 1: view*= Matrix.Translation(0.04f,0.0f,0.0f);
					break;
				case 2: view *= Matrix.Translation(-0.04f, 0.0f, 0.0f);
					break;
			}

			device.Transform.View = view;

			device.Transform.World = toCenter*scale*transformation;

			if (currentMesh>-1)
			{
				if (autorun)
				{
					currentMesh = (Environment.TickCount/40)%meshCount;
				}
				device.SetStreamSource(0, vbs[currentMesh], 0);		
                if (showColor)
				    device.VertexFormat = CustomVertex.PositionNormalColored.Format;
                else
                    device.VertexFormat = CustomVertex.PositionNormal.Format;
				device.Indices = ibs[sameCon?0:currentMesh];

				Material mat = new Material();
				mat.Diffuse = Color.FromArgb(255,255,255);

				device.Material = mat;
                device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, points[sameCon ? 0 : currentMesh].Count, 0, triangles[sameCon ? 0 : currentMesh].Count);

                if (showNormals)
                {
                    mat = new Material();
                    mat.Diffuse = Color.FromArgb(255, 0, 0);
                    device.Material = mat;

                    device.Indices = nibs[sameCon ? 0 : currentMesh];
                    device.DrawIndexedPrimitives(PrimitiveType.LineList, 0, 0, 2 * points[sameCon ? 0 : currentMesh].Count, 0, points[sameCon ? 0 : currentMesh].Count);
                }
			}

			if (showHelp)
			{				
				string meshInfo = "\n\n"+points[0].Count+" vertices\n"+triangles[0].Count+" triangles";
				helpFont.DrawText(null,helpString+"\n\n"+currentMesh.ToString()+meshInfo,new Rectangle(10,10,this.ClientSize.Width-10, this.ClientSize.Height-10) ,DrawTextFormat.NoClip,Color.FromArgb(255,255,255));
			}			

			device.EndScene();


			
				try
				{
					if (makingScreenShots == 1)
						sc1.Present();
					else
						sc2.Present();
				}
				catch (DeviceLostException)
				{
					this.deviceLost = true;
					return;
				}
						

			switch (makingScreenShots)
			{
				case 1:
					/*surface = device.GetBackBuffer(0, 0, BackBufferType.Mono);
					
					SurfaceLoader.Save("left.bmp", ImageFileFormat.Bmp, surface);
					*/
					makingScreenShots++;
					break;
				case 2:
					/*if (bmp != null) bmp.Dispose();
					surface = device.GetBackBuffer(0, 0, BackBufferType.Mono);
					SurfaceLoader.Save("right.bmp", ImageFileFormat.Bmp, surface);					
					bmp = new Bitmap("right.bmp");
					//RenderWindow r = new RenderWindow();
					r.pictureBox1.Image = bmp;*/
					makingScreenShots = 1;					
					break;
			}			
		}

		private void Application_Idle(object sender, EventArgs e)
		{
			this.Invalidate();
		}

		private void SimpleDirectXRendererWindow_Paint(object sender, PaintEventArgs e)
		{
			this.Render();
		}

		private void SimpleDirectXRendererWindow_KeyPress(object sender, KeyPressEventArgs e)
		{
			switch (e.KeyChar)
			{
				case 'a': transformation *= Matrix.Translation(-0.1f,0,0);
					break;
				case 's': transformation *= Matrix.Translation(0,-0.1f,0);
					break;
				case 'd': transformation *= Matrix.Translation(0.1f,0,0);
					break;
				case 'w': transformation *= Matrix.Translation(0,0.1f,0);
					break;
				case 'r': transformation *= Matrix.Scaling(1.1f,1.1f,1.1f);
					break;
				case 'f': transformation *= Matrix.Scaling(0.9f,0.9f,0.9f);
					break;
				case 't': transformation *= Matrix.RotationX(0.1f);
					break;
				case 'g': transformation *= Matrix.RotationX(-0.1f);
					break;
				case 'h': transformation *= Matrix.RotationY(-0.1f);
					break;
				case 'j': transformation *= Matrix.RotationY(0.1f);
					break;
				case 'z': wireFrame = !wireFrame;
					break;
				case 'x': lighting = !lighting;
					break;
				case 'v': showHelp = !showHelp;
					break;
				case 'l': lightMoving = true;
					break;
				case 'c': culling++;
					if (culling==4) culling = 1;
					break;
				case ',':
					currentMesh--;
					if (currentMesh<0) currentMesh = meshCount-1;
					break;
				case '.':
					currentMesh++;
					if (currentMesh == MeshCount) currentMesh = 0;
					break;
				case 'm': autorun = !autorun;
					break;
                case 'n': showNormals = !showNormals;
                    break;
			}
		}		

		private void SimpleDirectXRendererWindow_MouseDown(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left) mouseLeft = true;
			if (e.Button == MouseButtons.Right) mouseRight = true;
		}

		private void SimpleDirectXRendererWindow_MouseUp(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left) mouseLeft = false;
			if (e.Button == MouseButtons.Right) mouseRight = false;
		}

		private void SimpleDirectXRendererWindow_MouseMove(object sender, MouseEventArgs e)
		{
			if (mouseLeft)
			{
				int dx = mousePosition.X-e.X;
				int dy = mousePosition.Y-e.Y;
				transformation*=Matrix.RotationY(dx/100f);
				transformation*=Matrix.RotationX(dy/100f);
			}
			if (mouseRight)
			{
				int dx = e.X-mousePosition.X;
				int dy = mousePosition.Y-e.Y;
				transformation*=Matrix.Translation(dx/100f,dy/100f,0);
				
			}
			if (lightMoving)
			{
				int dx = mousePosition.X-e.X;
				int dy = mousePosition.Y-e.Y;
				lightAzimuth += dx/100f;
				lightElevation += dy/100f;				
			}
			mousePosition = new Point(e.X,e.Y);
		}

		private void SimpleDirectXRendererWindow_MouseWheel(object sender, MouseEventArgs e)
		{
			int dw = e.Delta;
			float coef;
			coef = (float)Math.Pow(1.001,dw);
			scale*= Matrix.Scaling(coef,coef,coef);
		}

		private void SimpleDirectXRendererWindow_KeyUp(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.L) lightMoving = false;
		}

		private void device_DeviceLost(object sender, EventArgs e)
		{
			if (helpFont!=null) helpFont.Dispose();
		}

		private void device_DeviceReset(object sender, EventArgs e)
		{
			initFonts();
		}

		private void device_Disposing(object sender, EventArgs e)
		{
			helpFont.Dispose();
		}

        private void InitializeComponent()
        {
            this.SuspendLayout();
            // 
            // SimpleStereoDirectXRendererWindow
            // 
            this.ClientSize = new System.Drawing.Size(284, 264);
            this.Name = "SimpleStereoDirectXRendererWindow";
            this.FormClosing += new System.Windows.Forms.FormClosingEventHandler(this.SimpleStereoDirectXRendererWindow_FormClosing);
            this.ResumeLayout(false);

        }

        private void SimpleStereoDirectXRendererWindow_FormClosing(object sender, FormClosingEventArgs e)
        {
            this.device.Dispose();
        }
	}
}
