using System;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.ComponentModel;

namespace Zcu.Mve.Testing
{
	/// <summary>
	/// A module used for downsampling of MveArrays of Scalar values. Downsampling is performed in selected dimension by selected amount.
	/// To downsample in multiple dimensions use multiple modules.
	/// Averaging is used to compute the output values.
	/// </summary>
	[ModuleInfo("Libor Va", "Module for downsampling MveArray of Scalar values.", Category = "VolumeVisualisation", IconName = "ArrayDownsampler.ico")]
	[PortInfo("Input array", "MveArray of any kind of data and any dimension to be downsampled.")]
	[PortInfo("Output array", "MveArray of downsampled input data.")]
	public class ArrayDownsampler:Module
	{
		int selectedDimension = 0;
		int downsampleFactor = 3;

		bool paramChanged = true;

		MveArray output = null;

		/// <summary>
		/// Dimension to downsample. Value 0 represents the first dimension etc.
		/// If the value is lower than the dimension of the input data then an exception is thrown.
		/// </summary>
		/// 
		[Description("Integer value that sets the dimension to be downsampled.")]
		public int SelectedDimension
		{
			get{return(selectedDimension);}
			set
			{
				if (value<0)
				{
					throw new MveException(this, "Selected dimension cannot be lower than zero.");
				}
				else
				{
					selectedDimension = value;
					paramChanged = true;
				}
			}
		}

		/// <summary>
		/// Downsample factor, i.e. number of samples that will form one output value. Values lower than zero have no meaning and are not accepted.
		/// </summary>
		/// 
		[Description("Integer value that sets how many input samples will make one output sample")]
		public int DownsampleFactor
		{
			get{return(downsampleFactor);}
			set
			{
				if (value<1)
				{
					throw new MveException(this, "Downsample factor cannot be lower than 1.");
				}
				else
				{
					downsampleFactor = value;
					paramChanged = true;
				}
			}
		}

		public ArrayDownsampler()
		{
			AddInPort("Input array",typeof(MveArray));
			AddOutPort("Output array",typeof(MveArray));
		}

		/// <summary>
		/// helper function for generating coordinates of arbitrary ND array
		/// </summary>
		/// <param name="coordinates">array of coordinates to increase</param>
		/// <param name="dimensions">array of maximums of coordinate values</param>
		/// <param name="position">index of coordinate to be increased</param>
		private void incraseCoordinate(int[] coordinates, int[] dimensions, int position)
		{
			coordinates[position]++;
			if (coordinates[position]==dimensions[position])
			{
				if (position!=0)
				{
					coordinates[position] = 0;
					incraseCoordinate(coordinates,dimensions,position-1);
				}				
				if (position == 1)
				{
					ProgressInfo((float)coordinates[0]/dimensions[0]);
				}
			}
		}

		public override void Execute()
		{
			if ((paramChanged)||(!IsDataSame("Input array"))||(output == null))
			{

				MveArray input = (MveArray)GetInput("Input array");

				if (input.ElemType!=typeof(Scalar))
				{
					throw new MveException(this, "ArrayDownsampler is only applicable to Scalar arrays");
				}

				int[] inputDimensions = input.Lengths;
				if (selectedDimension>=inputDimensions.Length)
				{
					throw new MveException(this, "Selected downsample dimension is lower than number of dimensions of input MveArray.");
				}
				inputDimensions[selectedDimension]/=downsampleFactor;
				output = new MveArray(input.ElemType,inputDimensions);

				int[] coordinates = new int[inputDimensions.Length];
				int[] localCoordinates = new int[inputDimensions.Length];
				while (coordinates[0]<inputDimensions[0])
				{
					Scalar sum = new Scalar(0.0);
					localCoordinates = (int[])coordinates.Clone();
					localCoordinates[selectedDimension]*=downsampleFactor;
					for (int i = 0;i<downsampleFactor;i++)
					{
						sum+=(Scalar)input[localCoordinates];
						localCoordinates[selectedDimension]++;
					}
					output[coordinates]=(Scalar)(sum/downsampleFactor);
					incraseCoordinate(coordinates,output.Lengths,output.Lengths.Length-1);
				}

				SetOutput("Output array",output);

				paramChanged = false;
			} // array downsampled

			else // inputs weren't changed
			{
				SetOutput("Output array",output,true);
			}
		}//execute

	}
}
