using System;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.Drawing;
using System.ComponentModel;


namespace Zcu.Mve.Testing
{
	/// <summary>
	/// This is a module useful when loading a series of bitmap images containig volume data. It produces a MveArray of either one color value, or RGB triplet.
	/// Use the properties StartIndex, EndIndex, Prefix and Suffix to specify input files (filenames will be generated from the <code>prefix number suffix</code> pattern, where number changes from startIndex to endIndex, no leading zeroes are added.
	/// Any bitmap format that can be loaded by System.Drawing.Bitmap constructor can be used as input.
	/// Use the DataType property to specify what data should be used. You can use one of following values:
	/// DataType values that will produce array of Mve.Numerics.Scalar:
	/// R for values of red component (0-1)
	/// G for values of green component (0-1)
	/// B for values of blue component (0-1)
	/// H for values of hue (0-360 degrees)
	/// S for values of saturation (0-1)
	/// V for values of brightness (value) (0-1)
	/// </summary>
	
	[ModuleInfo("Libor Va", "Module for loading volume data from a series of images", Category = "Loader", IconName = "VolumeImageLoader.ico")]
	[PortInfo("VolumeArray","A 3D MveArray that contains the images where first two dimensions represent X and Y directions in the inputs, third being the index of the image")]
	public class VolumeImageLoader:Module
	{
		int startIndex = 0, endIndex = 10;
		string prefix = "file", suffix = ".jpg";
		string dataType = "V";
		bool paramsChanged = true;
		bool alwaysReload = false;
		MveArray output = null;

		[Description("Defines wether the data should be reloaded even when no filename generating parameteres are changed. Set to true if the data in the files may change between the runs of the map, otherwise set to false to speed the process up significantly.")]
		public bool AlwaysReload
		{
			get{return(alwaysReload);}
			set
			{
				alwaysReload = value;
				paramsChanged = true;
			}
		}

		/// <summary>
		/// Property that sets the first index that will be used to generate filenames of the input files 
		/// </summary>
		
		[Description("Starting index for the <prefix><number><suffix> pattern. This index will be used.")]
		[Category("File name")]
		public int StartIndex
		{
			get{return(startIndex);}
			set
			{
				startIndex = value;
				paramsChanged = true;
			}
		}

		/// <summary>
		/// Property that sets the last index that will be used to generate filenames of the input files 
		/// </summary>
		[Description("Ending index for the <prefix><number><suffix> pattern. This index will be used.")]
		[Category("File name")]
		public int EndIndex
		{
			get{return(endIndex);}
			set
			{
				endIndex = value;
				paramsChanged = true;
			}
		}

		/// <summary>
		/// Property that sets a string used as prefix when generating input filenames.
		/// </summary>
		[Description("Filename prefix for the <prefix><number><suffix> pattern. Should also contain full path.")]
		[Category("File name")]
		public string Prefix
		{
			get{return(prefix);}
			set
			{
				prefix = value;
				paramsChanged = true;
			}
		}

		/// <summary>
		/// Property that sets a string used as suffix when generating input filenames.
		/// </summary>
		[Description("Filename suffix for the <prefix><number><suffix> pattern. Should also contain file extension.")]
		[Category("File name")]
		public string Suffix
		{
			get{return(suffix);}
			set
			{
				suffix = value;
				paramsChanged = true;
			}
		}

		/// <summary>
		/// The DataType property is used to specify what data should be loaded. You can use one of following values:
		/// dataType values that will produce array of Mve.Numerics.Scalar:
		/// R for values of red component (0-1)
		/// G for values of green component (0-1)
		/// B for values of blue component (0-1)
		/// H for values of hue (0-360 degrees)
		/// S for values of saturation (0-1)
		/// V for values of brightness (value) (0-1)
		/// </summary>
		[Description("Specifies data to be loaded. Use R, G or B for channel values, H S or V for Hue, Saturation or Brightness, or RGB for color triplet.")]
		[Category("Data Type")]
		public string DataType
		{
			get{return(dataType);}
			set
			{
				if ((value == "R")||(value == "G")||(value == "B")||
					(value == "H")||(value == "S")||(value == "V")||
					(value == "RGB"))
				{
					dataType = value;
					paramsChanged = true;
				}
				else
					throw new MveException(this, "DataType must be one of following: R, G, B, H, S, V or RGB.");
			}
		}		

		public VolumeImageLoader()
		{
			this.AddOutPort("VolumeArray",typeof(MveArray));
		}

		public override void Execute()
		{
			if ((paramsChanged)||(alwaysReload)||(output == null))
			{

				if (endIndex<startIndex)
				{
					throw new MveException(this, "EndIndex cannot be lower than StartIndex!");
				}

				Bitmap bmp = null;

				try
				{
					bmp = new Bitmap(prefix+startIndex+suffix);
				}
				catch (Exception e)
				{
					throw new MveException(this, "Loading of bitmap failed. Message:"+e.Message);
				}
				
				switch (dataType)
				{
					case "R":
						output = new MveArray(typeof(Scalar),bmp.Width,bmp.Height,endIndex-startIndex+1);
						bmp.Dispose();
						for (int z = startIndex;z<=endIndex;z++)
						{
							ProgressInfo((float)z/(endIndex-startIndex));
							try
							{
								bmp = new Bitmap(prefix+z+suffix);
							}
							catch (Exception e)
							{
								throw new MveException(this, "Loading of bitmap failed. Message:"+e.Message);
							}
							for (int x = 0;x<bmp.Width;x++)
								for (int y = 0;y<bmp.Height;y++)
									output[x,y,z-startIndex] = new Scalar(bmp.GetPixel(x,y).R/255.0);
							bmp.Dispose();
						}
						break;
					case "G":
						output = new MveArray(typeof(Scalar),bmp.Width,bmp.Height,endIndex-startIndex+1);
						bmp.Dispose();
						for (int z = startIndex;z<=endIndex;z++)
						{
							ProgressInfo((float)z/(endIndex-startIndex));
							try
							{
								bmp = new Bitmap(prefix+z+suffix);
							}
							catch (Exception e)
							{
								throw new MveException(this, "Loading of bitmap failed. Message:"+e.Message);
							}
							for (int x = 0;x<bmp.Width;x++)
								for (int y = 0;y<bmp.Height;y++)
									output[x,y,z-startIndex] = new Scalar(bmp.GetPixel(x,y).G/255.0);
							bmp.Dispose();
						}
						break;
					case "B":
						output = new MveArray(typeof(Scalar),bmp.Width,bmp.Height,endIndex-startIndex+1);
						bmp.Dispose();
						for (int z = startIndex;z<=endIndex;z++)
						{
							ProgressInfo((float)z/(endIndex-startIndex));
							try
							{
								bmp = new Bitmap(prefix+z+suffix);
							}
							catch (Exception e)
							{
								throw new MveException(this, "Loading of bitmap failed. Message:"+e.Message);
							}
							for (int x = 0;x<bmp.Width;x++)
								for (int y = 0;y<bmp.Height;y++)
									output[x,y,z-startIndex] = new Scalar(bmp.GetPixel(x,y).B/255.0);
							bmp.Dispose();
						}
						break;
					case "H":
						output = new MveArray(typeof(Scalar),bmp.Width,bmp.Height,endIndex-startIndex+1);
						bmp.Dispose();
						for (int z = startIndex;z<=endIndex;z++)
						{
							ProgressInfo((float)z/(endIndex-startIndex));
							try
							{
								bmp = new Bitmap(prefix+z+suffix);
							}
							catch (Exception e)
							{
								throw new MveException(this, "Loading of bitmap failed. Message:"+e.Message);
							}
							for (int x = 0;x<bmp.Width;x++)
								for (int y = 0;y<bmp.Height;y++)
									output[x,y,z-startIndex] = new Scalar(bmp.GetPixel(x,y).GetHue());
							bmp.Dispose();
						}
						break;
					case "S":
						output = new MveArray(typeof(Scalar),bmp.Width,bmp.Height,endIndex-startIndex+1);
						bmp.Dispose();
						for (int z = startIndex;z<=endIndex;z++)
						{
							ProgressInfo((float)z/(endIndex-startIndex));
							try
							{
								bmp = new Bitmap(prefix+z+suffix);
							}
							catch (Exception e)
							{
								throw new MveException(this, "Loading of bitmap failed. Message:"+e.Message);
							}
							for (int x = 0;x<bmp.Width;x++)
								for (int y = 0;y<bmp.Height;y++)
									output[x,y,z-startIndex] = new Scalar(bmp.GetPixel(x,y).GetSaturation());
							bmp.Dispose();
						}
						break;
					case "V":
						output = new MveArray(typeof(Scalar),bmp.Width,bmp.Height,endIndex-startIndex+1);
						bmp.Dispose();
						for (int z = startIndex;z<=endIndex;z++)
						{
							ProgressInfo((float)z/(endIndex-startIndex));
							try
							{
								bmp = new Bitmap(prefix+z+suffix);
							}
							catch (Exception e)
							{
								throw new MveException(this, "Loading of bitmap failed. Message:"+e.Message);
							}
							for (int x = 0;x<bmp.Width;x++)
								for (int y = 0;y<bmp.Height;y++)
									output[x,y,z-startIndex] = new Scalar(bmp.GetPixel(x,y).GetBrightness());
							bmp.Dispose();
						}
						break;
					case "RGB":
						output = new MveArray(typeof(ColorRGBA),bmp.Width,bmp.Height,endIndex-startIndex+1);
						bmp.Dispose();
						for (int z = startIndex;z<=endIndex;z++)
						{
							ProgressInfo((float)z/(endIndex-startIndex));
							try
							{
								bmp = new Bitmap(prefix+z+suffix);
							}
							catch (Exception e)
							{
								throw new MveException(this, "Loading of bitmap failed. Message:"+e.Message);
							}
							for (int x = 0;x<bmp.Width;x++)
								for (int y = 0;y<bmp.Height;y++)
								{
									Color c = bmp.GetPixel(x,y);
									output[x,y,z-startIndex] = new ColorRGBA(c.R,c.G,c.B,c.A);
								}
							bmp.Dispose();
						}
						break;
				}//switch
				SetOutput("VolumeArray",output);			
				paramsChanged = false;
			} //reload was needed
			else
			{
				SetOutput("VolumeArray",output,true);			
			}
		}//Execute
	}
}
