using System;
using System.Windows.Forms;
using Microsoft.DirectX.Direct3D;

using Zcu.Mve.UnstrGridRendering.Legend;

namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// Abstract class which extend BaseRenderer capabilities. Class adds: switching between window modes, 
    /// saving screens into images, legend visualization and basic picker. Class define two abstract methods:
    /// UpdateScene() and DrawScene(). These methods are called once per rendering loop, where first one is 
    /// used for rendering objects update and second one for object rendering.
    /// </summary>
    public abstract class AdvancedRenderer : BaseRenderer
    {
        #region Fields

        private Camera _camera;
        private Visualization.UniformDataArray _uda;

        private bool _enablePicker;

        protected BasePicker _picker;
        protected TextureSaver _tSaver;
        protected LegendRenderer _lRenderer;

        #endregion

        #region Properties

        public bool PickerEnabled
        {
            set { _enablePicker = value; }
            get { return _enablePicker; }
        }

        public BasePicker Picker
        {
            set { _picker = (value == null) ? null : _picker; }
            get { return _picker; }
        }

        #endregion

        public AdvancedRenderer(Camera camera, Visualization.UniformDataArray uda, RendererParameters rendererParameters)
            : base(rendererParameters)
        {
            _camera = camera;
            _uda = uda;

            _enablePicker = rendererParameters.PickerEnabled;

            base.WindowCreated += new WindowCreatedEventHandler(AdvancedRenderer_WindowCreated);
            base.RenderingLoopStart += new RenderingStartEventHandler(AdvancedRenderer_RenderingLoopStart);
            base.RenderingLoopEnd += new RenderingEndEventHandler(AdvancedRenderer_RenderingLoopEnd);
        }

        protected abstract void DrawScene(Device device);

        protected abstract void UpdateScene(Device device);

        protected override void Render(Device device)
        {
            // abstract method
            UpdateScene(device);

            //_picker.Update(device.Transform.Projection);

            Surface origSurf = null;

            try
            {
                origSurf = device.GetRenderTarget(0);

                // abstract method, draw 3d scene
                DrawScene(device);

                // draw legend
                if (_lRenderer != null) _lRenderer.Render();

                // save target to texture
                if (_tSaver.toSave()) _tSaver.SaveToFile(origSurf, _rParams.PathSaveto, device);

                device.SetRenderTarget(0, origSurf);

                origSurf.Dispose();
            }

            catch (DeviceLostException)
            {
                //surface must be released
                if ((origSurf != null) && !origSurf.Disposed) origSurf.Dispose();
            }

            _runtime.Device.Present();
        }

        /// <summary>
        /// Base window events: KeyDown & ResizeEnd
        /// </summary>
        protected virtual void HookWindowEvents()
        {
            _window.KeyDown += new KeyEventHandler(_window_KeyDown);
            _window.ResizeEnd += new EventHandler(_window_ResizeEnd);

            // hook evets to picker
            if (_enablePicker) HookEventsToPicker(_window);
        }

        /// <summary>
        /// Hook events to picker.
        /// </summary>
        /// <param name="form"></param>
        protected void HookEventsToPicker(Form form)
        {
            form.MouseUp += new MouseEventHandler(form_MouseUp);
            form.MouseDown += new MouseEventHandler(form_MouseDown);
            form.MouseMove += new MouseEventHandler(form_MouseMove);

            form.KeyDown += new KeyEventHandler(form_KeyDown);
            form.KeyUp += new KeyEventHandler(form_KeyUp);
        }

        private void _window_ResizeEnd(object sender, EventArgs e)
        {
            if (_runtime != null)
            {
                // update window size
                _wndRestoreRect = new System.Drawing.Rectangle(_window.Location, _window.ClientSize);
                // update backbuffer
                _runtime.DeviceResize(_wndRestoreRect.Size);
            }
        }

        /// <summary>
        /// Device independent classes
        /// </summary>
        private void AdvancedRenderer_RenderingLoopStart()
        {
            // create legend renderer
            _lRenderer = new LegendRenderer(_runtime.Device);

            // assign legend content
            _lRenderer.Composer.Add(_uda);
        }

        private void AdvancedRenderer_RenderingLoopEnd()
        {
            if (_lRenderer != null) _lRenderer.Dispose();

            _lRenderer = null;
            _tSaver = null;
        }

        /// <summary>
        /// Form independent classes
        /// </summary>
        private void AdvancedRenderer_WindowCreated()
        {
            _tSaver = new TextureSaver();
            _picker = new BasePicker(_window.ClientSize, _camera);
            
            HookEventsToPicker(_window);
            HookWindowEvents();
        }

        /// <summary>
        /// Key events, which are responsible for saving surface to texture
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void _window_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.R:
                    {
                        _tSaver.RenderTotexture = !_tSaver.RenderTotexture;
                        break;
                    }

                case Keys.S:
                    {
                        _tSaver.SnapShot = true;
                        break;
                    }
            }
        }

        #region Picker

        private void form_MouseMove(object sender, MouseEventArgs e)
        {
            if (_enablePicker && _picker != null) _picker.OnMouseUp(e);
        }

        private void form_KeyUp(object sender, KeyEventArgs e)
        {
            if (_enablePicker && _picker != null) _picker.OnKeyUp(e);
        }

        private void form_KeyDown(object sender, KeyEventArgs e)
        {
            if (_enablePicker && _picker != null) _picker.OnKeyDown(e);
        }

        private void form_MouseDown(object sender, MouseEventArgs e)
        {
            if (_enablePicker && _picker != null) _picker.OnMouseDown(e);
        }

        private void form_MouseUp(object sender, MouseEventArgs e)
        {
            if (_enablePicker && _picker != null) _picker.OnMouseUp(e);
        }

        #endregion
    }
}
