using System;
using System.Threading;
using System.Diagnostics;
using System.Windows.Forms;

using Zcu.Mve.UnstrGridRendering.Legend;

using Microsoft.DirectX.Direct3D;

namespace Zcu.Mve.UnstrGridRendering
{
    public enum WindowMode { Windowed, FullScreeen };

    public class WindowModeEvetnArgs : EventArgs
    {
        private WindowMode wndMode;
        private System.Drawing.Size size;

        public WindowModeEvetnArgs(System.Drawing.Size size, WindowMode wndMode)
        {
            this.size = size;
            this.wndMode = wndMode;
        }
    }

    public delegate void RenderingStartEventHandler();
    public delegate void RenderingEndEventHandler();
    public delegate void WindowCreatedEventHandler();
    public delegate void WindowModeChangeEventHandler(WindowModeEvetnArgs e);
    public delegate void OnIdleEventHandler();

    /// <summary>
    /// Abstract class which create rendering loop and rendering context (window). Class define several events
    /// which are invoke when given object is created or given event occure. These event are:
    /// WindowCreated       - rendering window is created, window events can be hooked
    /// RenderingLoopStart  - device is created, device independent classes can be initialized
    /// RenderingLoopEnd    - rendering finish, release resources
    /// WindowModeChanging  - window mode change (fullscreen / windowed)
    /// </summary>
    public abstract class BaseRenderer
    {
        protected RendererParameters _rParams;

        private bool _fullScreen;
        private bool _doRendering;
        private bool _wndModeChange;

        protected System.Drawing.Rectangle _wndRestoreRect;
        
        /// <summary>
        /// Define monitor used for fullscreen mode
        /// </summary>
        private Screen _screen;

        protected Form _window;
        protected Runtime _runtime;

        public event RenderingStartEventHandler RenderingLoopStart;
        public event RenderingEndEventHandler RenderingLoopEnd;
        public event WindowModeChangeEventHandler WindowModeChanging;
        public event WindowCreatedEventHandler WindowCreated;
        public event OnIdleEventHandler OnIdle;

        public BaseRenderer(RendererParameters rendererParameters)
        {
            _rParams = rendererParameters;

            _fullScreen = false;
            _doRendering = false;
            _wndModeChange = false;

            if (_rParams.Screen == ScreenTarget.Primary) _screen = Screen.AllScreens[0];

            if (_rParams.Screen == ScreenTarget.Secondary && Screen.AllScreens.Length >= 2) _screen = Screen.AllScreens[1];
        }

        /// <summary>
        /// Rendering loop termination.
        /// </summary>
        public void StopRendering()
        {
            _doRendering = false;
        }

        /// <summary>
        /// Base renderer initialization - device is created etc.
        /// </summary>
        public virtual void Run()
        {
            // rendering context
            _window = new Form();

            if (WindowCreated != null) WindowCreated();

            _window.Text = _rParams.Text;
            _window.FormBorderStyle = FormBorderStyle.Sizable;
            _window.ClientSize = new System.Drawing.Size(_rParams.Width, _rParams.Height);
            _window.MinimumSize = _window.Bounds.Size;
            _wndRestoreRect = _window.Bounds;

            _window.FormClosing += new FormClosingEventHandler(window_FormClosing);

            // device creation point
            _runtime = new Runtime(_window.Handle, _rParams.Screen, _rParams.Width, _rParams.Height);

            _runtime.WindowModeChange += new WindowModeChangedEventHandler(_runtime_WindowModeChange);

            Render();

            _window.Dispose();
            _window.Close();

            if (_runtime != null) _runtime.Dispose();
        }

        protected abstract void Render(Device device);

        protected void ChangeWindowMode()
        {
            _fullScreen = !_fullScreen;

            _wndModeChange = true;
        }

        /// <summary>
        /// Rendering loop, block current thread.
        /// </summary>
        protected void Render()
        {
            _window.Show();

            if (RenderingLoopStart != null) RenderingLoopStart();

            _doRendering = true;

            while (_doRendering && _window.Created)
            {
                try
                {
                    Application.DoEvents();

                    // device lost occure, no rendering is possible
                    if (!_runtime.CheckDeviceState()) { continue; }

                    // windowed or fullscreen mode
                    if (_wndModeChange) { _wndModeChange = false; OnWindowModeChange(); continue; }

                    if (_runtime.Device != null) Render(_runtime.Device);

                    if (OnIdle != null) OnIdle();

                    Thread.Sleep(1);
                }

                // any exception which occure in rendering process
                // typical exceptions: DEVICE LOST, DEVICE NOT RESET, GRAPHICS EXCEPTION
                catch (Exception ex)
                {
                    // set window focus
                    _window.Activate();

                    _runtime.CheckDeviceState();

                    Thread.Sleep(100);

                    if (_doRendering) Debug.WriteLine(ex.Message);
                }
            }

            if (RenderingLoopEnd != null) RenderingLoopEnd();
        }

        private void OnWindowModeChange()
        {
            if (_fullScreen)
            {
                _wndRestoreRect = new System.Drawing.Rectangle(_window.Location, _window.ClientSize);

                _window.TopMost = true;
                _window.Bounds = _screen.Bounds;
                _window.FormBorderStyle = FormBorderStyle.None;
                _window.ClientSize = _screen.Bounds.Size;

                // update present parameters
                _runtime.SwitchToFullscreen(_screen.Bounds.Width, _screen.Bounds.Height);

                if (WindowModeChanging != null)
                    WindowModeChanging(new WindowModeEvetnArgs(_window.Size, WindowMode.FullScreeen));
            }
            else
            {
                _window.TopMost = false;
                _window.FormBorderStyle = FormBorderStyle.Sizable;
                _window.Location = _wndRestoreRect.Location;
                _window.ClientSize = _wndRestoreRect.Size;

                // update present parameters
                _runtime.SwitchToWindowed(_wndRestoreRect.Width, _wndRestoreRect.Height);

                if (WindowModeChanging != null)
                    WindowModeChanging(new WindowModeEvetnArgs(_window.Size, WindowMode.Windowed));
            }
        }

        private void _runtime_WindowModeChange(WindowMode windowMode)
        {
            if (_window != null)
            {
                _window.Location = _wndRestoreRect.Location;
                _window.ClientSize = _wndRestoreRect.Size;
            }
        }
        
        /// <summary>
        /// Stop rendering loop.
        /// </summary>
        private void window_FormClosing(object sender, FormClosingEventArgs e)
        {
            _doRendering = false;
        }
    }
}
