using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Zcu.Mve.Visualization;
using System.Drawing;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// Helper class for D3DUTSceneObject.
  /// It provides functionality of drawing scene objects bounding box.
  /// </summary>
  public class UGridBoundingBox : IDisposable
  {
    #region Fields
    private VertexBuffer vertexBuffer;
    private IndexBuffer indexBuffer;
    private Device device;
    private BoundingBox boundingBox;
    private long lmtBoundingBox;
    #endregion

    /// <summary>
    /// Constructor, initializes object.
    /// </summary>
    public UGridBoundingBox()
    {
      this.lmtBoundingBox = 0;
      this.device = null;
    }

    /// <summary>
    /// Update scene objects bounding box when device or scene objects gridSize has changed.
    /// </summary>
    /// <param name="device">Rendering device.</param>
    /// <param name="unstructuredGrid">Unstructured grid.</param>
    public void Update(Device device, UnstrGrid unstructuredGrid)
    {
      if (unstructuredGrid.Points == null)
      {
        vertexBuffer = null;
        indexBuffer = null;
        return;
      }

      if ((lmtBoundingBox != unstructuredGrid.Points.ModificationTime)
      || (this.device != device))
      {
        this.boundingBox = unstructuredGrid.BoundingBox;
        this.device = device;

        createVertexBuffer();
        createIndexBuffer();

        lmtBoundingBox = unstructuredGrid.Points.ModificationTime;
      }
    }

    /// <summary>
    /// Draw scene objects bounding box. This method should be used between device.BeginScene() and device.EndScene()
    /// </summary>
    public void Draw()
    {
      if (vertexBuffer == null)
        return;

      setRenderState();
            
      device.VertexFormat = CustomVertex.PositionColored.Format;
      device.SetStreamSource(0, vertexBuffer, 0);

      device.Indices = indexBuffer;
      device.DrawIndexedPrimitives(PrimitiveType.LineList, 0, 0, 8, 0, 12);
      device.DrawPrimitives(PrimitiveType.PointList, 0, 8);
    }

    /// <summary>
    /// Set device render state for drawing bounding box.
    /// </summary>
    private void setRenderState()
    {
      device.RenderState.Lighting = false;
      device.RenderState.ColorVertex = true;
      device.RenderState.PointSize = 3f;
    }

    /// <summary>
    /// Create vertex buffer of bounding box vertices.
    /// </summary>
    private void createVertexBuffer()
    {
      vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionColored), 8, device, Usage.WriteOnly, CustomVertex.PositionColored.Format, Pool.Managed);

      CustomVertex.PositionColored[] vertices = new CustomVertex.PositionColored[8];
      int i = 0;

      vertices[i++] = new CustomVertex.PositionColored((float) boundingBox.MinPoint[0],
                                                       (float) boundingBox.MinPoint[1],
                                                       (float) boundingBox.MinPoint[2],
                                                       Color.Blue.ToArgb());

      vertices[i++] = new CustomVertex.PositionColored((float) boundingBox.MinPoint[0],
                                                       (float) boundingBox.MinPoint[1],
                                                       (float) boundingBox.MaxPoint[2],
                                                       Color.Green.ToArgb());

      vertices[i++] = new CustomVertex.PositionColored((float) boundingBox.MaxPoint[0],
                                                       (float) boundingBox.MinPoint[1],
                                                       (float) boundingBox.MaxPoint[2],
                                                       Color.Green.ToArgb());

      vertices[i++] = new CustomVertex.PositionColored((float) boundingBox.MaxPoint[0],
                                                       (float) boundingBox.MinPoint[1],
                                                       (float) boundingBox.MinPoint[2],
                                                       Color.Green.ToArgb());

      vertices[i++] = new CustomVertex.PositionColored((float) boundingBox.MaxPoint[0],
                                                       (float) boundingBox.MaxPoint[1],
                                                       (float) boundingBox.MaxPoint[2],
                                                       Color.Red.ToArgb());

      vertices[i++] = new CustomVertex.PositionColored((float) boundingBox.MaxPoint[0],
                                                       (float) boundingBox.MaxPoint[1],
                                                       (float) boundingBox.MinPoint[2],
                                                       Color.Green.ToArgb());

      vertices[i++] = new CustomVertex.PositionColored((float) boundingBox.MinPoint[0],
                                                       (float) boundingBox.MaxPoint[1],
                                                       (float) boundingBox.MinPoint[2],
                                                       Color.Green.ToArgb());

      vertices[i++] = new CustomVertex.PositionColored((float) boundingBox.MinPoint[0],
                                                       (float) boundingBox.MaxPoint[1],
                                                       (float) boundingBox.MaxPoint[2],
                                                       Color.Green.ToArgb());

      vertexBuffer.SetData(vertices, 0, LockFlags.None);
    }

    /// <summary>
    /// Create index buffer.
    /// </summary>
    private void createIndexBuffer()
    {
      indexBuffer = new IndexBuffer(typeof(short), 24, device, Usage.WriteOnly, Pool.Managed);

      short[] indices = new short[] { 0, 1, 1, 2, 2, 3, 3, 0, 4, 5, 5, 6, 6, 7, 7, 4, 0, 6, 1, 7, 2, 4, 3, 5 };

      indexBuffer.SetData(indices, 0, LockFlags.None);
    }

    #region IDisposable Members
    /// <summary>
    /// Dispose vertex and index buffers. 
    /// </summary>
    public void Dispose()
    {
      if (vertexBuffer != null)
        vertexBuffer.Dispose();

      if (indexBuffer != null)
        indexBuffer.Dispose();
    }

    #endregion
  }
}
