using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;

using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using Zcu.Mve.Core;
using System.Globalization;

namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// Type of projection enumearation.
    /// </summary>
    public enum ProjectionType
    {
        /// <summary>
        /// Perspective projection.
        /// </summary>
        Perspective = 0,
        /// <summary>
        /// Orthogonal projection.
        /// </summary>
        Orthogonal = 1,
        /// <summary>
        /// User projection - transform matrix will be provided by user.
        /// </summary>
        User = 2
    }

    /// <summary>
    /// Represents camera that captures the Scene. Important settings are position, direction and projection type.  
    /// </summary>
    [DataObject("Represents camera. Camera captures the Scene. Important settings are position, direction and projection type.")]
    public class Camera : DataObject
    {
        #region Fields

        private Vector3D position;
        private Vector3D target;
        private Vector3D upVector;

        private float zNearPlane;
        private float zFarPlane;

        private ProjectionType projectionType;
        private float fov;
        private float scale;

        private Transform3D projectionTransform;

        #endregion

        #region Properties
        /// <summary>
        /// Camera position.
        /// </summary>
        public Vector3D Position
        {
            get
            {
                return position;
            }
            set
            {
                position = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Camera target, point where Camera is looking to.
        /// </summary>
        public Vector3D Target
        {
            get
            {
                return target;
            }
            set
            {
                target = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Camera up direction vector. Defines the up direction of the current world, usually [0, 1, 0].
        /// </summary>
        public Vector3D UpVector
        {
            get
            {
                return upVector;
            }
            set
            {
                upVector = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Z-value of the near view plane. Objects that are closer will be clipped out.
        /// </summary>
        public float ZNearPlane
        {
            get
            {
                return zNearPlane;
            }
            set
            {
                zNearPlane = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Z-value of the far view plane. Objects that are beyond will be clipped out.
        /// </summary>
        public float ZFarPlane
        {
            get
            {
                return zFarPlane;
            }
            set
            {
                zFarPlane = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Projection type. Ie. Perspective, Orthogonal,...
        /// </summary>
        public ProjectionType ProjectionType
        {
            get
            {
                return projectionType;
            }
            set
            {
                projectionType = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Field of view.
        /// </summary>
        public float Fov
        {
            get
            {
                return fov;
            }
            set
            {
                fov = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Scale. Scale is used when orthogonal projection is selected. 
        /// </summary>
        public float Scale
        {
            get
            {
                return scale;
            }
            set
            {
                scale = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Projection transform.
        /// </summary>
        public Transform3D ProjectionTransform
        {
            get
            {
                return projectionTransform;
            }
            set
            {
                projectionTransform = value;

                UpdateModificationTime();
            }
        }
        #endregion

        /// <summary>
        /// Default constructor. Initializes Camera object to default values.
        /// </summary>
        public Camera()
        {
            this.position = new Vector3D(0, 0, 1);
            this.target = new Vector3D(0, 0, 0);
            this.upVector = new Vector3D(0, 1, 0);

            this.zNearPlane = 0.1f;
            this.zFarPlane = 100f;

            this.projectionType = ProjectionType.Perspective;
            this.fov = (float)Math.PI / 4f;
            this.scale = 1f;

            this.projectionTransform = new Transform3D();

            UpdateModificationTime();
        }

        /// <summary>
        /// Converts current object to its string description.
        /// </summary>
        /// <returns>Resulting string description.</returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder(this.GetType().FullName + ":\r\n");

            str.Append("Position: " + Position + "\r\n");
            str.Append("Target: " + Target + "\r\n");
            str.Append("UpVector: " + UpVector + "\r\n");
            str.Append("ZNearPlane: " + ZNearPlane + "\r\n");
            str.Append("ZFarPlane: " + ZFarPlane + "\r\n");
            str.Append("ProjectionType: " + ProjectionType + "\r\n");
            str.Append("FOV: " + Fov + "\r\n");
            str.Append("Scale: " + Scale + "\r\n");
            str.Append("ProjectionTransform: " + ProjectionTransform + "\r\n");

            str.Append("\r\n");

            return str.ToString();
        }

        /// <summary>
        /// Check if data of this object are consistent.
        /// </summary>
        /// <returns>Returns true if data are consistent.</returns>
        public override bool CheckConsistence()
        {
            bool returnValue = true;

            returnValue &= this.position.CheckConsistence();
            returnValue &= this.target.CheckConsistence();
            returnValue &= this.upVector.CheckConsistence();

            returnValue &= this.projectionTransform.CheckConsistence();

            return returnValue;
        }

        /// <summary>
        /// Performs deep copy of current object.
        /// </summary>
        /// <returns>Deep copy of Camera object.</returns>
        public override IDataObject DeepCopy()
        {
            Camera c = new Camera();

            c.position = (Vector3D)this.position.DeepCopy();
            c.target = (Vector3D)this.target.DeepCopy();
            c.upVector = (Vector3D)this.upVector.DeepCopy();

            c.zNearPlane = this.zNearPlane;
            c.zFarPlane = this.zFarPlane;

            c.projectionType = this.projectionType;
            c.fov = this.fov;
            c.scale = this.scale;

            c.projectionTransform = (Transform3D)this.projectionTransform;

            return c;
        }

        /// <summary>
        /// Gets names of all assemblies used by this class.
        /// </summary>
        public override string[] LibNames
        {
            get
            {
                AddLibName(this.GetType());
                AddLibName(typeof(Vector3D));
                AddLibName(typeof(ProjectionType));
                AddLibName(typeof(Transform3D));

                return base.LibNames;
            }
        }

        /// <summary>
        /// Reads data of this object.
        /// </summary>
        /// <param name="xmlTextReader">Reader.</param>
        public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
        {
            while (xmlTextReader.Read())
            {
                if (xmlTextReader.NodeType == XmlNodeType.Element)
                {
                    switch (xmlTextReader.Name)
                    {
                        case "position":
                            this.position = (Vector3D)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
                            this.position.ReadData(xmlTextReader);
                            break;
                        case "lookat":
                            this.target = (Vector3D)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
                            this.target.ReadData(xmlTextReader);
                            break;
                        case "upvector":
                            this.upVector = (Vector3D)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
                            this.upVector.ReadData(xmlTextReader);
                            break;
                        case "znearplane":
                            this.zNearPlane = FromString.GetFloat(xmlTextReader.ReadString(), 1f);
                            break;
                        case "zfarplane":
                            this.zFarPlane = FromString.GetFloat(xmlTextReader.ReadString(), 100f);
                            break;
                        case "projectiontype":
                            this.projectionType = FromString.GetProjectionType(xmlTextReader.ReadString(), ProjectionType.Perspective);
                            break;
                        case "fov":
                            this.fov = FromString.GetFloat(xmlTextReader.ReadString(), (float)Math.PI / 4f);
                            break;
                        case "scale":
                            this.scale = FromString.GetFloat(xmlTextReader.ReadString(), 1f);
                            break;
                        case "projectionTransform":
                            this.projectionTransform = (Transform3D)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
                            this.projectionTransform.ReadData(xmlTextReader);
                            break;
                        default:
                            break;
                    } // switch
                } // if

                if (xmlTextReader.NodeType == XmlNodeType.EndElement)
                {
                    if (xmlTextReader.Name == "camera")
                        break;
                }
            } // while

            UpdateModificationTime();
        }

        /// <summary>
        /// Writes data of this object.
        /// </summary>
        /// <param name="xmlTextWriter">Writer.</param>
        public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
        {
            xmlTextWriter.WriteStartElement("camera");

            xmlTextWriter.WriteStartElement("position");
            xmlTextWriter.WriteAttributeString("type", this.position.GetType().FullName);
            this.position.WriteData(xmlTextWriter);
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("lookat");
            xmlTextWriter.WriteAttributeString("type", this.target.GetType().FullName);
            this.target.WriteData(xmlTextWriter);
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("upvector");
            xmlTextWriter.WriteAttributeString("type", this.upVector.GetType().FullName);
            this.upVector.WriteData(xmlTextWriter);
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("znearplane");
            xmlTextWriter.WriteAttributeString("type", this.zNearPlane.GetType().FullName);
            xmlTextWriter.WriteString(this.zNearPlane.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("zfarplane");
            xmlTextWriter.WriteAttributeString("type", this.zFarPlane.GetType().FullName);
            xmlTextWriter.WriteString(this.zFarPlane.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("projectiontype");
            xmlTextWriter.WriteAttributeString("type", this.projectionType.GetType().FullName);
            xmlTextWriter.WriteString(this.projectionType.ToString());
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("fov");
            xmlTextWriter.WriteAttributeString("type", this.fov.GetType().FullName);
            xmlTextWriter.WriteString(this.fov.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("scale");
            xmlTextWriter.WriteAttributeString("type", this.scale.GetType().FullName);
            xmlTextWriter.WriteString(this.scale.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("projectiontransform");
            xmlTextWriter.WriteAttributeString("type", this.projectionTransform.GetType().FullName);
            projectionTransform.WriteData(xmlTextWriter);
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteEndElement();
        }
    }
}
