using System;
using System.Collections.Generic;
using System.Text;

using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.ComponentModel;
using System.Xml;
using System.Globalization;
using System.Diagnostics;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// Source of Camera object.
  /// </summary>
  [ModuleInfo("Jan Kaiser", "Source of Camera object.", IconName = "CameraModule.ico")]
  [PortInfo("outCamera","Resulting Camera.")]
  public class CameraModule : Module
  {
    private bool dataChanged = true;
    private Camera camera;

    private bool showInProjectionTransform;
    
    private bool showInPosition;
    private bool showInTarget;
    private bool showInUpVector;

    #region Properties
    /// <summary>
    /// Camera position.
    /// </summary>
    [Category("Camera")]
    [Description("Camera position.")]
    public new Vector3D Position
    {
      get
      {
        return camera.Position;
      }
      set
      {
        camera.Position = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Camera target, point where Camera is looking to.
    /// </summary>
    [Category("Camera")]
    [Description("Camera target, point where Camera is looking to.")]
    public Vector3D Target
    {
      get
      {
        return camera.Target;
      }
      set
      {
        camera.Target = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Camera up direction vector. Defines the up direction of the current world, usually [0, 1, 0].
    /// </summary>
    [Category("Camera")]
    [Description("Camera up direction vector. Defines the up direction of the current world, usually [0, 1, 0].")]
    public Vector3D UpVector
    {
      get
      {
        return camera.UpVector;
      }
      set
      {
        camera.UpVector = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Z-value of the near view plane. Objects that are closer will be clipped out.
    /// </summary>
    [Category("Projection")]
    [DefaultValue(1f)]
    [Description("Z-value of the near view plane. Objects that are closer will be clipped out.")]
    public float ZNearPlane
    {
      get
      {
        return camera.ZNearPlane;
      }
      set
      {
        camera.ZNearPlane = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Z-value of the far view plane. Objects that are beyond will be clipped out.
    /// </summary>
    [Category("Projection")]
    [DefaultValue(100f)]
    [Description("Z-value of the far view plane. Objects that are beyond will be clipped out.")]
    public float ZFarPlane
    {
      get
      {
        return camera.ZFarPlane;
      }
      set
      {
        camera.ZFarPlane = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Selection of what type of projection transform will be used. There is internal perspective
    /// and orthogonal projection. When User projection transform is selected, user defined
    /// projection transform matrix is used.
    /// </summary>
    [Category("Projection")]
    [DefaultValue(ProjectionType.Perspective)]
    [Description("Selection of what type of projection transform will be used. There is "
                + "internal perspective and orthogonal projection. When User projection "
                + "transform is selected, user defined projection transform matrix is used.")]
    public ProjectionType ProjectionType
    {
      get
      {
        return camera.ProjectionType;
      }
      set
      {
        if (value == ProjectionType.User)
        {
          if (!showInProjectionTransform)
          {
            AddInPort("inProjectionTransform", typeof(Transform3D), true);
            showInProjectionTransform = true;
          }
        }

        camera.ProjectionType = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Field of view in the y direction, in radians.
    /// </summary>
    [Category("Perspective projection")]
    [DefaultValue((float) Math.PI / 4f)]
    [Description("Field of view in the y direction, in radians.")]
    public float Fov
    {
      get
      {
        return camera.Fov;
      }
      set
      {
        camera.Fov = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Scale of orthogonally projected scene.
    /// </summary>
    [Category("Orthogonal projection")]
    [DefaultValue(1f)]
    [Description("Scale of orthogonally projected scene.")]
    public float Scale
    {
      get
      {
        return camera.Scale;
      }
      set
      {
        camera.Scale = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Show input port for projection transform matrix.
    /// It is used when User projection is selected.
    /// </summary>
    [Category("Show input port")]
    [DefaultValue(false)]
    [Description("Show input port for projection transform matrix. It is used when User projection is selected.")]
    public bool inProjectionTransform
    {
      get
      {
        return showInProjectionTransform;
      }
      set
      {
        if (value)
        {
          if (!showInProjectionTransform)
            AddInPort("inProjectionTransform", typeof(Transform3D), true);
        }
        else
        {
          if (showInProjectionTransform)
            RemoveInPort("inProjectionTransform");
        }

        showInProjectionTransform = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Show input port for camera position vector.
    /// When port is shown internal module settings of this property are ignored.
    /// </summary>
    [Category("Show input port")]
    [DefaultValue(false)]
    [Description("Show input port for camera position vector. " 
                +"When port is shown internal module settings of this property are ignored.")]
    public bool inPosition
    {
      get
      {
        return showInPosition;
      }
      set
      {
        if (value)
        {
          if (!showInPosition)
            AddInPort("inPosition", typeof(Vector3D), true);
        }
        else
        {
          if (showInPosition)
            RemoveInPort("inPosition");
        }

        showInPosition = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Show input port for camera target vector.
    /// When port is shown internal module settings of this property are ignored.
    /// </summary>
    [Category("Show input port")]
    [DefaultValue(false)]
    [Description("Show input port for camera target vector. "
                +"When port is shown internal module settings of this property are ignored.")]
    public bool inTarget
    {
      get
      {
        return showInTarget;
      }
      set
      {
        if (value)
        {
          if (!showInTarget)
            AddInPort("inTarget", typeof(Vector3D), true);
        }
        else
        {
          if (showInTarget)
            RemoveInPort("inTarget");
        }

        showInTarget = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Show input port for camera up vector.
    /// When port is shown internal module settings of this property are ignored.
    /// </summary>
    [Category("Show input port")]
    [DefaultValue(false)]
    [Description("Show input port for camera up vector. "
                +"When port is shown internal module settings of this property are ignored.")]
    public bool inUpVector
    {
      get
      {
        return showInUpVector;
      }
      set
      {
        if (value)
        {
          if (!showInUpVector)
            AddInPort("inUpVector", typeof(Vector3D), true);
        }
        else
        {
          if (showInUpVector)
            RemoveInPort("inUpVector");
        }

        showInUpVector = value;
        dataChanged = true;
      }
    }
    #endregion

    /// <summary>
    /// Default constructor. Initializes module and adds output port.
    /// </summary>
    public CameraModule()
    {
      this.showInPosition = false;
      this.showInTarget = false;
      this.showInUpVector = false;
      this.showInProjectionTransform = false;

      AddOutPort("outCamera", typeof(Camera));

      camera = new Camera();
    }

    /// <summary>
    /// Exposes created/updated Camera object on output port.
    /// </summary>
    public override void Execute()
    {
      if (showInProjectionTransform)
      {
        if (!IsDataSame("inProjectionTransform"))
        {
          camera.ProjectionTransform = (Transform3D) GetInput("inProjectionTransform");
          dataChanged = true;
        }
      }

      if (showInPosition)
      {
        if (!IsDataSame("inPosition"))
        {
          camera.Position = (Vector3D) GetInput("inPosition");
          dataChanged = true;
        }
      }

      if (showInTarget)
      {
        if (!IsDataSame("inTarget"))
        {
          camera.Target = (Vector3D) GetInput("inTarget");
          dataChanged = true;
        }
      }
      
      if (showInUpVector)
      {
        if (!IsDataSame("inUpVector"))
        {
          camera.UpVector = (Vector3D) GetInput("inUpVector");
          dataChanged = true;
        }
      }

      if (dataChanged == true)
      {
        SetOutput("outCamera", camera, false);
        dataChanged = false;
      }
      else
      {
        SetOutput("outCamera", camera, true);
      }
    }

    /// <summary>
    /// Writes Xml configuration of current module.
    /// </summary>
    /// <param name="config">XmlElement from map config file which belongs to current module.</param>
    public override void WriteConfig(System.Xml.XmlElement config)
    {
      XmlElement xmlProperties = config.OwnerDocument.CreateElement("Properties");

      XmlElement xmlPosition = config.OwnerDocument.CreateElement("Position");
      xmlPosition.SetAttribute("type", camera.Position.GetType().FullName);
      xmlPosition.InnerText = camera.Position.X.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                            + camera.Position.Y.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                            + camera.Position.Z.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlPosition);

      XmlElement xmlLookAt = config.OwnerDocument.CreateElement("Target");
      xmlLookAt.SetAttribute("type", camera.Target.GetType().FullName);
      xmlLookAt.InnerText = camera.Target.X.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                          + camera.Target.Y.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                          + camera.Target.Z.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlLookAt);

      XmlElement xmlUpVector = config.OwnerDocument.CreateElement("UpVector");
      xmlUpVector.SetAttribute("type", camera.UpVector.GetType().FullName);
      xmlUpVector.InnerText = camera.UpVector.X.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                            + camera.UpVector.Y.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                            + camera.UpVector.Z.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlUpVector);

      XmlElement xmlZNearPlane = config.OwnerDocument.CreateElement("ZNearPlane");
      xmlZNearPlane.SetAttribute("type", camera.ZNearPlane.GetType().FullName);
      xmlZNearPlane.InnerText = camera.ZNearPlane.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlZNearPlane);

      XmlElement xmlZFarPlane = config.OwnerDocument.CreateElement("ZFarPlane");
      xmlZFarPlane.SetAttribute("type", camera.ZFarPlane.GetType().FullName);
      xmlZFarPlane.InnerText = camera.ZFarPlane.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlZFarPlane);

      XmlElement xmlProjectionType = config.OwnerDocument.CreateElement("ProjectionType");
      xmlProjectionType.SetAttribute("type", camera.ProjectionType.GetType().FullName);
      xmlProjectionType.InnerText = camera.ProjectionType.ToString();
      xmlProperties.AppendChild(xmlProjectionType);

      XmlElement xmlFov = config.OwnerDocument.CreateElement("Fov");
      xmlFov.SetAttribute("type", camera.Fov.GetType().FullName);
      xmlFov.InnerText = camera.Fov.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlFov);

      XmlElement xmlScale = config.OwnerDocument.CreateElement("Scale");
      xmlScale.SetAttribute("type", camera.Scale.GetType().FullName);
      xmlScale.InnerText = camera.Scale.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlScale);

      XmlElement xmlShowInProjectionTransform = config.OwnerDocument.CreateElement("inProjectionTransform");
      xmlShowInProjectionTransform.SetAttribute("type", inProjectionTransform.GetType().FullName);
      xmlShowInProjectionTransform.InnerText = inProjectionTransform.ToString();
      xmlProperties.AppendChild(xmlShowInProjectionTransform);

      XmlElement xmlShowInPosition = config.OwnerDocument.CreateElement("inPosition");
      xmlShowInPosition.SetAttribute("type", inPosition.GetType().FullName);
      xmlShowInPosition.InnerText = inPosition.ToString();
      xmlProperties.AppendChild(xmlShowInPosition);

      XmlElement xmlShowInTarget = config.OwnerDocument.CreateElement("inTarget");
      xmlShowInTarget.SetAttribute("type", inTarget.GetType().FullName);
      xmlShowInTarget.InnerText = inTarget.ToString();
      xmlProperties.AppendChild(xmlShowInTarget);

      XmlElement xmlShowInUpVector = config.OwnerDocument.CreateElement("inUpVector");
      xmlShowInUpVector.SetAttribute("type", inUpVector.GetType().FullName);
      xmlShowInUpVector.InnerText = inUpVector.ToString();
      xmlProperties.AppendChild(xmlShowInUpVector);

      config.AppendChild(xmlProperties);
    }

    /// <summary>
    /// Reads Xml configuration of curent module.
    /// </summary>
    /// <param name="config">XmlElement from map config file which belongs to current module.</param>
    public override void ReadConfig(System.Xml.XmlElement config)
    {
      foreach (XmlElement xeConfig in config.ChildNodes)
      {
        if (xeConfig.Name == "Properties")
        {
          foreach (XmlElement xeProperties in xeConfig.ChildNodes)
          {
            switch (xeProperties.Name)
            {
              case "Position":
                camera.Position = FromString.GetVector3D(xeProperties.InnerText);
                break;
              case "Target":
                camera.Target = FromString.GetVector3D(xeProperties.InnerText);
                break;
              case "UpVector":
                camera.UpVector = FromString.GetVector3D(xeProperties.InnerText);
                break;
              case "ZNearPlane":
                camera.ZNearPlane = FromString.GetFloat(xeProperties.InnerText, 1f);
                break;
              case "ZFarPlane":
                camera.ZFarPlane = FromString.GetFloat(xeProperties.InnerText, 100f);
                break;
              case "ProjectionType":
                camera.ProjectionType = FromString.GetProjectionType(xeProperties.InnerText, ProjectionType.Perspective);
                break;
              case "Fov":
                camera.Fov = FromString.GetFloat(xeProperties.InnerText, (float) Math.PI / 4f);
                break;
              case "Scale":
                camera.Scale = FromString.GetFloat(xeProperties.InnerText, 1f);
                break;
              case "inProjectionTransform":
                inProjectionTransform = FromString.GetBool(xeProperties.InnerText, false);
                break;
              case "inPosition":
                inPosition = FromString.GetBool(xeProperties.InnerText, false);
                break;
              case "inTarget":
                inTarget = FromString.GetBool(xeProperties.InnerText, false);
                break;
              case "inUpVector":
                inUpVector = FromString.GetBool(xeProperties.InnerText, false);
                break;
              default:
                Debug.WriteLine("Unreaded value:" + xeProperties.Name + " - " + xeProperties.InnerText);
                break;
            } // switch
          } // foreach
        } // if
      } // foreach
    }

    /// <summary>
    /// Creates new instance of module and copy all necessary member variables.
    /// This method is used by MapEditor in Copy/Paste process.
    /// Without overriding it returns implicit instance of current module.
    /// </summary>
    public override Module DeepCopy()
    {
      CameraModule cameraSource = (CameraModule) base.DeepCopy();

      cameraSource.camera = (Camera) this.camera.DeepCopy();

      return cameraSource;
    }
  }
}
