using System;
using System.Collections.Generic;
using System.Text;
using dx3d = Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// Class holds Camera properties for D3D object representation.
    /// It also performs transition from Camera to D3DCamera. 
    /// </summary>
    public class CameraWrapper
    {
        #region Fields

        private Camera camera;
        private SceneWrapper scene;
        private RendererParameters rendererParameters;

        private dx3d.Matrix projectionMatrix;
        private dx3d.Matrix viewMatrix;

        private System.Drawing.Size wndSize;
        private float traverse;
        private float zFarPlane;
        private float zNearPlane;
        private dx3d.Vector3 target;

        private int mouseX, mouseY;
        private float traverseCoefficient;

        #endregion

        public dx3d.Matrix ViewMatrix
        {
            get { return viewMatrix; }
            set { viewMatrix = value; }
        }

        public dx3d.Matrix ProjectionMatrix
        {
            get { return projectionMatrix; }
        }

        /// <summary>
        /// Constructor, takes references to parent objects.
        /// </summary>
        /// <param name="camera">Camera.</param>
        /// <param name="scene">Scene.</param>
        /// <param name="rendererParameters">Renderer parameters.</param>
        public CameraWrapper(Camera camera, SceneWrapper scene, RendererParameters rendererParameters)
        {
            this.camera = camera;
            this.scene = scene;
            this.rendererParameters = rendererParameters;

            this.mouseX = 0;
            this.mouseY = 0;
            this.traverseCoefficient = 1;

            projectionMatrix = dx3d.Matrix.PerspectiveFovLH((float)Math.PI * 0.25f, (float)(wndSize.Width / (float)wndSize.Height), 0.1f, 100.0f);
        }

        /// <summary>
        /// Update camera settings - recompute view and projection matrix.
        /// </summary>
        /// <param name="runtime">D3DUT runtime.</param>
        public void Update(System.Drawing.Size wndSize)
        {
            this.wndSize = wndSize;

            if (rendererParameters.UseInternalCamera || rendererParameters.UseInternalVirtualTrackball)
                computeCameraParameters();

            updateProjectionMatrix();

            updateViewMatrix();
        }

        /// <summary>
        /// Update camera position. This is heavily used in cooperation with virtual trackball. 
        /// </summary>
        /// <param name="mouseX">Horizontal mouse position.</param>
        /// <param name="mouseY">Vertical mouse position.</param>
        /// <param name="traverseCoefficient">Traverse coefficient. It controls the distance of camera from origin.</param>
        public void UpdatePosition(int mouseX, int mouseY, float traverseCoefficient)
        {
            this.mouseX = mouseX;
            this.mouseY = mouseY;
            this.traverseCoefficient = traverseCoefficient;

            computeCameraParameters();

            updateViewMatrix();
        }

        /// <summary>
        /// This method should be called during rendering between device.BeginScene() and device.EndScene().
        /// It sets camera settings in device render state.
        /// </summary>
        /// <param name="device">Rendering device.</param>
        public void SetCamera(Device device)
        {
            if (rendererParameters.UseInternalCamera)
            {
                device.Transform.Projection = dx3d.Matrix.PerspectiveFovLH((float)Math.PI / 4f, wndSize.Width / (float)wndSize.Height, zNearPlane, zFarPlane);
            }
            else
            {
                switch (camera.ProjectionType)
                {
                    case ProjectionType.Orthogonal:
                        if (rendererParameters.UseInternalVirtualTrackball)
                        {
                            device.Transform.Projection = dx3d.Matrix.OrthoLH(camera.Scale * traverseCoefficient * wndSize.Width / (float)wndSize.Height, camera.Scale * traverseCoefficient, camera.ZNearPlane, camera.ZFarPlane);
                        }
                        else
                        {
                            device.Transform.Projection = dx3d.Matrix.OrthoLH(camera.Scale * wndSize.Width / (float)wndSize.Height, camera.Scale, camera.ZNearPlane, camera.ZFarPlane);
                        }
                        break;

                    case ProjectionType.Perspective:
                        device.Transform.Projection = dx3d.Matrix.PerspectiveFovLH(camera.Fov, wndSize.Width / (float)wndSize.Height, camera.ZNearPlane, camera.ZFarPlane);
                        break;

                    case ProjectionType.User:
                        device.Transform.Projection = projectionMatrix;
                        break;
                }
            }

            device.Transform.View = viewMatrix;
        }

        /// <summary>
        /// Update projection matrix.
        /// </summary>
        private void updateProjectionMatrix()
        {
            projectionMatrix = new dx3d.Matrix();

            projectionMatrix.M11 = camera.ProjectionTransform[0, 0];
            projectionMatrix.M12 = camera.ProjectionTransform[0, 1];
            projectionMatrix.M13 = camera.ProjectionTransform[0, 2];
            projectionMatrix.M14 = camera.ProjectionTransform[0, 3];

            projectionMatrix.M21 = camera.ProjectionTransform[1, 0];
            projectionMatrix.M22 = camera.ProjectionTransform[1, 1];
            projectionMatrix.M23 = camera.ProjectionTransform[1, 2];
            projectionMatrix.M24 = camera.ProjectionTransform[1, 3];

            projectionMatrix.M31 = camera.ProjectionTransform[2, 0];
            projectionMatrix.M32 = camera.ProjectionTransform[2, 1];
            projectionMatrix.M33 = camera.ProjectionTransform[2, 2];
            projectionMatrix.M34 = camera.ProjectionTransform[2, 3];

            projectionMatrix.M41 = camera.ProjectionTransform[3, 0];
            projectionMatrix.M42 = camera.ProjectionTransform[3, 1];
            projectionMatrix.M43 = camera.ProjectionTransform[3, 2];
            projectionMatrix.M44 = camera.ProjectionTransform[3, 3];
        }

        /// <summary>
        /// Update view matrix.
        /// </summary>
        private void updateViewMatrix()
        {
            dx3d.Vector3 cameraPosition;
            dx3d.Vector3 cameraTarget;
            dx3d.Vector3 cameraUpVector;

            if (!rendererParameters.UseInternalVirtualTrackball)
            {
                cameraPosition = new dx3d.Vector3((float)camera.Position.X, (float)camera.Position.Y, (float)camera.Position.Z);
                cameraTarget = new dx3d.Vector3((float)camera.Target.X, (float)camera.Target.Y, (float)camera.Target.Z);
                cameraUpVector = new dx3d.Vector3((float)camera.UpVector.X, (float)camera.UpVector.Y, (float)camera.UpVector.Z);

                viewMatrix = dx3d.Matrix.LookAtLH(cameraPosition, cameraTarget, cameraUpVector);
            }
            else
            {
                float angleX = (float)(mouseY * Math.PI / 180f);
                float angleY = (float)(mouseX * Math.PI / 180f);

                while (angleX > (2 * Math.PI))
                    angleX = (float)(angleX - (2 * Math.PI));

                while (angleX < 0)
                    angleX = (float)(angleX + (2 * Math.PI));

                while (angleY > (2 * Math.PI))
                    angleY = (float)(angleY - (2 * Math.PI));

                while (angleY < 0)
                    angleY = (float)(angleY + (2 * Math.PI));


                cameraPosition.X = (float)Math.Sin(angleY) * (float)Math.Cos(angleX);
                cameraPosition.Z = ((float)-Math.Cos(angleY) * (float)Math.Cos(angleX));
                cameraPosition.Y = (float)Math.Sin(angleX);

                cameraUpVector.X = (float)Math.Sin(angleY) * (float)Math.Cos(angleX + 0.2f) - cameraPosition.X;
                cameraUpVector.Z = ((float)-Math.Cos(angleY) * (float)Math.Cos(angleX + 0.2f) - cameraPosition.Z);
                cameraUpVector.Y = (float)Math.Sin(angleX + 0.2f) - cameraPosition.Y;

                viewMatrix = dx3d.Matrix.Translation(-target.X, -target.Y, -target.Z);

                if (camera.ProjectionType == ProjectionType.Orthogonal)
                {
                    viewMatrix *= dx3d.Matrix.LookAtLH(cameraPosition, new dx3d.Vector3(0, 0, 0), cameraUpVector);
                }
                else
                {
                    viewMatrix *= dx3d.Matrix.LookAtLH(traverse * cameraPosition, new dx3d.Vector3(0, 0, 0), cameraUpVector);
                }
            }
        }

        /// <summary>
        /// Compute camera core parameters.
        /// </summary>
        private void computeCameraParameters()
        {
            target = new dx3d.Vector3((float)scene.MidPoint.X, (float)scene.MidPoint.Y, (float)scene.MidPoint.Z);
            traverse = traverseCoefficient * (float)((Math.Abs(this.scene.MaxPoint.Z - this.scene.MinPoint.Z) + (Math.Abs(this.scene.MaxPoint.Y - this.scene.MidPoint.Y) / Math.Tan(camera.Fov / 2))));
            float tmpZFarPlane = (float)(2 * (traverse + Math.Abs(this.scene.MaxPoint.Z - this.scene.MinPoint.Z)));

            if (tmpZFarPlane > camera.ZFarPlane)
                zFarPlane = tmpZFarPlane;
            else
                zFarPlane = 50f;

            zNearPlane = zFarPlane * 0.0001f;

            //Console.WriteLine("T:" + target.x + ", " + target.y + ", " + target.Z);
            //Console.WriteLine("traverse:" + traverse);
            //Console.WriteLine("zFar:" + camera.ZFarPlane);
        }
    }
}
