using System;
using System.Collections.Generic;
using System.Text;

using Zcu.Mve.Visualization;
using Zcu.Mve.Visualization.Processing;
using Zcu.Mve.Visualization.Sources;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using System.Globalization;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// This static class provides methods for converting strings to enumerations or simple data types.
  /// This feature becames wery handy when reading/writing modules or data objetcs settings.
  /// </summary>
  public static class FromString
  {
    /// <summary>
    /// Converts String to RenderElements enum type.
    /// </summary>
    /// <param name="renderElements">RenderElements as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Render elements as RenderElements.</returns>
    public static RenderElements GetRenderElements(string renderElements, RenderElements defaultValue)
    {
      switch(renderElements.Trim())
      {
        case "Points":
          return RenderElements.Points;
        case "Cells":
          return RenderElements.Cells;
        default:
          return defaultValue;
      }
    }

    /// <summary>
    /// Converts String to VertexAttribute enum type.
    /// </summary>
    /// <param name="vertexAttribute">Vertex attribute as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Vertex attribute as VertexAttribute.</returns>
    public static VertexAttributes GetVertexAttribute(string vertexAttribute, VertexAttributes defaultValue)
    {
      switch (vertexAttribute.Trim())
      {
        case "None":
          return VertexAttributes.None;
        case "TextureCoord":
          return VertexAttributes.TextureCoord;
        case "ColorRGBA":
          return VertexAttributes.ColorRGBA;
        case "ColorRGBATextureCoord":
          return VertexAttributes.ColorRGBATextureCoord;
        case "Normal3D":
          return VertexAttributes.Normal3D;
        case "Normal3DColorRGBA":
          return VertexAttributes.Normal3DColorRGBA;
        case "Normal3DTextureCoord":
          return VertexAttributes.Normal3DTextureCoord;
        default:
          return defaultValue;
      }
    }

    /// <summary>
    /// Converts String to FillMode enum type. 
    /// </summary>
    /// <param name="fillMode">Fill mode as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Fill mode as FillMode.</returns>
    public static FillMode GetFillMode(string fillMode, FillMode defaultValue)
    {
      switch (fillMode.Trim())
      {
        case "Point":
          return FillMode.Point;
        case "WireFrame":
          return FillMode.WireFrame;
        case "Solid":
          return FillMode.Solid;
        default:
          return defaultValue;
      }
    }

    /// <summary>
    /// Converts String to Cull enum type. 
    /// </summary>
    /// <param name="cull">Cull as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Cull as Cull.</returns>
    public static Cull GetCull(string cull, Cull defaultValue)
    {
      switch (cull.Trim())
      {
        case "None":
          return Cull.None;
        case "Clockwise":
          return Cull.Clockwise;
        case "CounterClockwise":
          return Cull.CounterClockwise;
        default:
          return defaultValue;
      }
    }

    /// <summary>
    /// Converts String to ShadeMode enum type.
    /// </summary>
    /// <param name="shadeMode">Shade mode as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Shade mode as ShadeMode.</returns>
    public static ShadeMode GetShadeMode(string shadeMode, ShadeMode defaultValue)
    {
      switch (shadeMode.Trim())
      {
        case "Flat":
          return ShadeMode.Flat;
        case "Gouraud":
          return ShadeMode.Gouraud;
        default:
          return defaultValue;
      }
    }

    /// <summary>
    /// Converts String to TransformOperation enum type.
    /// </summary>
    /// <param name="transformOperation">Transform operation as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Transform operation as TransformOperation.</returns>
    public static TransformOperation GetTransformOperation(string transformOperation, TransformOperation defaultValue)
    {
      switch (transformOperation.Trim())
      {
        case "Multiply":
          return TransformOperation.Multiply;
        case "Replace":
          return TransformOperation.Replace;
        default:
          return defaultValue;
      }
    }

    /// <summary>
    /// Converts String to LightType enum type.
    /// </summary>
    /// <param name="lightType">Light type as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Light type as LightType.</returns>
    public static LightType GetLightType(string lightType, LightType defaultValue)
    {
      switch (lightType.Trim())
      {
        case "Directional":
          return LightType.Directional;
        case "Point":
          return LightType.Point;
        case "Spot":
          return LightType.Spot;
        default:
          return defaultValue;
      }
    }

    /// <summary>
    /// Converts String to ProjectionType enum type.
    /// </summary>
    /// <param name="projectionType">Projection type as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Projection type as ProjectionType.</returns>
    public static ProjectionType GetProjectionType(string projectionType, ProjectionType defaultValue)
    {
      switch (projectionType.Trim())
      {
        case "Perspective":
          return ProjectionType.Perspective;
        case "Orthogonal":
          return ProjectionType.Orthogonal;
        case "User":
          return ProjectionType.User;
        default:
          return defaultValue;
      }
    }

    /// <summary>
    /// Converts String to Axis enum type.
    /// </summary>
    /// <param name="axis">Axis as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Axis as Axis.</returns>
    public static Axis GetAxis(string axis, Axis defaultValue)
    {
      switch (axis.Trim())
      {
        case "X":
          return Axis.X;
        case "Y":
          return Axis.Y;
        case "Z":
          return Axis.Z;
        default:
          return defaultValue;
      }
    }

    /// <summary>
    /// Converts String to AngleUnits enum type.
    /// </summary>
    /// <param name="angleUnits">Angle unit as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Angle unit as AngleUnits.</returns>
    public static AngleUnits GetAngleUnits(string angleUnits, AngleUnits defaultValue)
    {
      switch (angleUnits.Trim())
      {
        case "Degrees":
          return AngleUnits.Degrees;
        case "Radians":
          return AngleUnits.Radians;
        default:
          return defaultValue;
      }
    }

    /// <summary>
    /// Converts String to ColorRGBA.
    /// </summary>
    /// <param name="color">Color as String.</param>
    /// <returns>Color as ColorRGBA.</returns>
    public static ColorRGBA GetColorRGBA(string color)
    {
      bool result = true;
      byte red = 224, green = 224, blue = 224, alpha = 255;

      try
      {
        string[] values = color.Split(' ');

        switch (values.Length)
        {
          case 3:
            result &= byte.TryParse(values[0], NumberStyles.Integer, CultureInfo.InvariantCulture.NumberFormat, out red);
            result &= byte.TryParse(values[1], NumberStyles.Integer, CultureInfo.InvariantCulture.NumberFormat, out green);
            result &= byte.TryParse(values[2], NumberStyles.Integer, CultureInfo.InvariantCulture.NumberFormat, out blue);
            break;
          case 4:
            result &= byte.TryParse(values[0], NumberStyles.Integer, CultureInfo.InvariantCulture.NumberFormat, out red);
            result &= byte.TryParse(values[1], NumberStyles.Integer, CultureInfo.InvariantCulture.NumberFormat, out green);
            result &= byte.TryParse(values[2], NumberStyles.Integer, CultureInfo.InvariantCulture.NumberFormat, out blue);
            result &= byte.TryParse(values[3], NumberStyles.Integer, CultureInfo.InvariantCulture.NumberFormat, out alpha);
            break;
          default:
            result = false;
            break;
        }
      }
      catch
      {
        result = false;
      }

      if (result)
      {
        return new ColorRGBA(red, green, blue, alpha);
      }
      else
      {
        Console.WriteLine("There was an error parsing color value: "
                          + color + ". Using default (224, 224, 224, 255).");

        return new ColorRGBA(255, 224, 224, 224);
      }
    }

    /// <summary>
    /// Converts String to Vector3D.
    /// </summary>
    /// <param name="vector3D">Vector as String.</param>
    /// <returns>Vector as Vector3D.</returns>
    public static Vector3D GetVector3D(string vector3D)
    {
      bool result = true;
      float x = 0f, y = 0f, z = 0f;

      try
      {
        string[] values = vector3D.Split(' ');

        result &= float.TryParse(values[0], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat, out x);
        result &= float.TryParse(values[1], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat, out y);
        result &= float.TryParse(values[2], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat, out z);
      }
      catch
      {
        result = false;
      }

      if (result)
      {
        return new Vector3D(x, y, z);
      }
      else
      {
        Console.WriteLine("There was an error parsing vector3D value: "
                          + vector3D + ". Using default (0, 0, 0).");

        return new Vector3D(0, 0, 0);
      }
    }

    /// <summary>
    /// Converts String to Point3D.
    /// </summary>
    /// <param name="point3D">Point as String.</param>
    /// <returns>Point as Point3D.</returns>
    public static Point3D GetPoint3D(string point3D)
    {
      bool result = true;
      float x = 0f, y = 0f, z = 0f;

      try
      {
        string[] values = point3D.Split(' ');

        result &= float.TryParse(values[0], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat, out x);
        result &= float.TryParse(values[1], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat, out y);
        result &= float.TryParse(values[2], NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat, out z);
      }
      catch
      {
        result = false;
      }

      if (result)
      {
        return new Point3D(x, y, z);
      }
      else
      {
        Console.WriteLine("There was an error parsing point3D value: "
                          + point3D + ". Using default (0, 0, 0).");

        return new Point3D(0, 0, 0);
      }
    }

    /// <summary>
    /// Converts String to Int.
    /// </summary>
    /// <param name="strInt">Integer as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Integer as Integer.</returns>
    public static int GetInt(string strInt, int defaultValue)
    {
      int tmp;

      if (!int.TryParse(strInt.Trim(), NumberStyles.Integer, CultureInfo.InvariantCulture.NumberFormat, out tmp))
      {
        tmp = defaultValue;
      }

      return tmp;
    }

    /// <summary>
    /// Converts String to Float.
    /// </summary>
    /// <param name="strFloat">Float as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Float as Float.</returns>
    public static float GetFloat(string strFloat, float defaultValue)
    {
      float tmp;

      if (!float.TryParse(strFloat.Trim(), NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat, out tmp))
      {
        tmp = defaultValue;
      }

      return tmp;
    }

    /// <summary>
    /// Converts String to Bool.
    /// </summary>
    /// <param name="strBool">Bool as String.</param>
    /// <param name="defaultValue">Default value.</param>
    /// <returns>Bool as Bool.</returns>
    public static bool GetBool(string strBool, bool defaultValue)
    {
      bool tmp;

      if (!bool.TryParse(strBool.Trim(), out tmp))
      {
        tmp = defaultValue;
      }

      return tmp;
    }
  }
}
