using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Zcu.Mve.Visualization;
using System.Diagnostics;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// This is a helper class for SceneObject. Its purpose is to create and maintain IndexBuffer array.
  /// </summary>
  public class IndexBufferWrapper : IDisposable
  {
    private Device device;
    private UniformDataArray cells;
    private PrimitiveType primitiveType;
    private IndexBuffer[] indexBuffer;
    private VertexBufferWrapper vertexBuffer;
    private long lmtCells;

    /// <summary>
    /// Default constructor. Initializes the object.
    /// </summary>
    public IndexBufferWrapper()
    {
      lmtCells = 0;
    }

    /// <summary>
    /// Update index buffer.
    /// </summary>
    /// <param name="device">Rendering device.</param>
    /// <param name="cells">Cell (SceneObject-UniformDataArray-Cells).</param>
    /// <param name="vertexBuffer">Reference to Vertex buffer.</param>
    public void Update(Device device, UniformDataArray cells, VertexBufferWrapper vertexBuffer)
    {
      this.device = device;
      this.cells = cells;
      this.vertexBuffer = vertexBuffer;

      if (lmtCells != cells.ModificationTime)
      {
        createIndexBuffer();
        lmtCells = cells.ModificationTime;
        Debug.WriteLine(">> IB Changed..");
      }
    }

    /// <summary>
    /// This method should be called during rendering between device.BeginScene() and device.EndScene().
    /// It draws indexed primitives.
    /// </summary>
    public void SetIndexBuffer()
    {
      if (indexBuffer != null)
      {       
          switch (cells.ElemType.ToString())
          {
            case "Zcu.Mve.Visualization.Triangle":
              device.Indices = indexBuffer[0];
              device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, cells.Count);         
              break;
            case "Zcu.Mve.Visualization.TriangleStrip":
              for (int i = 0; i < indexBuffer.Length; i++)
              {
                device.Indices = indexBuffer[i];

                if (vertexBuffer.Count < ushort.MaxValue)
                {
                  device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, (indexBuffer[i].SizeInBytes / sizeof(ushort)) - 2);
                }
                else
                {
                  device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, (indexBuffer[i].SizeInBytes / sizeof(int)) - 2);
                }
              }
              break;
            case "Zcu.Mve.Visualization.TriangleFan":
              for (int i = 0; i < indexBuffer.Length; i++)
              {
                device.Indices = indexBuffer[i];

                if (vertexBuffer.Count < ushort.MaxValue)
                {
                  device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, (indexBuffer[i].SizeInBytes / sizeof(ushort)) - 2);
                }
                else
                {
                  device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, (indexBuffer[i].SizeInBytes / sizeof(int)) - 2);
                }
              }
              break;
            case "Zcu.Mve.Visualization.Tetrahedron":
              for (int i = 0; i < indexBuffer.Length; i++)
              {
                device.Indices = indexBuffer[i];
                device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, 4);
              }
              break;
            case "Zcu.Mve.Visualization.Line":
              device.Indices = indexBuffer[0];
              device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, cells.Count); 
              break;
            case "Zcu.Mve.Visualization.PolyLine":
              for (int i = 0; i < indexBuffer.Length; i++)
              {
                device.Indices = indexBuffer[i];

                if (vertexBuffer.Count < ushort.MaxValue)
                {
                  device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, (indexBuffer[i].SizeInBytes / sizeof(ushort)) - 1);
                }
                else
                {
                  device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, (indexBuffer[i].SizeInBytes / sizeof(int)) - 1);
                }
              }
              break;     
            case "Zcu.Mve.Visualization.Slice":
              for (int i = 0; i < indexBuffer.Length; i++)
              {
                device.Indices = indexBuffer[i];

                if (vertexBuffer.Count < ushort.MaxValue)
                {
                  device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, (indexBuffer[i].SizeInBytes / sizeof(ushort)) - 1);
                }
                else
                {
                  device.DrawIndexedPrimitives(primitiveType, 0, 0, vertexBuffer.Count, 0, (indexBuffer[i].SizeInBytes / sizeof(int)) - 1);
                }
              }              
              break;
            // Polygon and Pixel is not implemented...
            default:
              Console.WriteLine(cells.ElemType + " not implemented yet, sorry.");
              break;
          }
      }
    }

    private void createIndexBuffer()
    {
      switch (cells.ElemType.ToString())
      {
        case "Zcu.Mve.Visualization.Triangle":
          createTriangleListIB();
          break;
        case "Zcu.Mve.Visualization.TriangleStrip":
          createTriangleStripIB();
          break;
        case "Zcu.Mve.Visualization.TriangleFan":
          createTriangleFanIB();
          break;
        case "Zcu.Mve.Visualization.Tetrahedron":
          createTetrahedronIB();
          break;
        case "Zcu.Mve.Visualization.Line":
          createLineListIB();
          break;
        case "Zcu.Mve.Visualization.PolyLine":
          createLineStripIB();
          break;
        case "Zcu.Mve.Visualization.Slice":
          createSliceIB();
          break;
          // Polygon and Pixel is not implemented...
        default:
          Console.WriteLine(cells.ElemType + " not implemented yet, sorry.");
          break;
      }
    }

    private void createSliceIB()
    {
      primitiveType = PrimitiveType.LineStrip;

      List<IndexBuffer> ib = new List<IndexBuffer>();

      for (int i = 0; i < cells.Count; i++)
      {
        Slice sl = cells[i] as Slice;

        for (int j = 0; j < sl.PolyLineCount; j++)
        {
          PolyLine pl = (PolyLine) sl.GetPolyLine(j);       

          if (vertexBuffer.Count < ushort.MaxValue)
          {
            ushort[] indices = new ushort[pl.PointCount];
            for (int k = 0; k < pl.PointCount; k++)
            {
              indices[k] = (ushort) pl[k];
            }

            IndexBuffer tmpIB = new IndexBuffer(typeof(ushort), pl.PointCount, device, Usage.WriteOnly, Pool.Managed);
            tmpIB.SetData(indices, 0, LockFlags.None);
            ib.Add(tmpIB);
          }
          else
          {
            int[] indices = new int[pl.PointCount];
            for (int k = 0; k < pl.PointCount; k++)
            {
              indices[k] = (int) pl[k];
            }

            IndexBuffer tmpIB = new IndexBuffer(typeof(int), pl.PointCount, device, Usage.WriteOnly, Pool.Managed);
            tmpIB.SetData(indices, 0, LockFlags.None);
            ib.Add(tmpIB);
          }
        }
      }

      indexBuffer = new IndexBuffer[ib.Count];
      ib.CopyTo(indexBuffer);
    }

    private void createLineStripIB()
    {
      primitiveType = PrimitiveType.LineStrip;
      indexBuffer = new IndexBuffer[cells.Count];

      for (int i = 0; i < cells.Count; i++)
      {
        PolyLine pl = (PolyLine) cells[i];

        if (vertexBuffer.Count < ushort.MaxValue)
        {
          ushort[] indices = new ushort[pl.PointCount];
          for (int j = 0; j < pl.PointCount; j++)
          {
            indices[j] = (ushort) pl[j];
          }

          indexBuffer[i] = new IndexBuffer(typeof(ushort), pl.PointCount, device, Usage.WriteOnly, Pool.Managed);
          indexBuffer[i].SetData(indices, 0, LockFlags.None);
        }
        else
        {
          int[] indices = new int[pl.PointCount];
          for (int j = 0; j < pl.PointCount; j++)
          {
            indices[j] = (int) pl[j];
          }

          indexBuffer[i] = new IndexBuffer(typeof(int), pl.PointCount, device, Usage.WriteOnly, Pool.Managed);
          indexBuffer[i].SetData(indices, 0, LockFlags.None);
        }
      }
    }

    private void createLineListIB()
    {
      primitiveType = PrimitiveType.LineList;
      indexBuffer = new IndexBuffer[1];

      if (vertexBuffer.Count < ushort.MaxValue)
      {
        indexBuffer[0] = new IndexBuffer(typeof(ushort), 2 * cells.Count, device, Usage.WriteOnly, Pool.Managed);

        ushort[] indices = new ushort[2 * cells.Count];

        for (int i = 0; i < cells.Count; i++)
        {
          indices[2 * i] = (ushort) (( Zcu.Mve.Visualization.Line) cells[i]).V1;
          indices[(2 * i) + 1] = (ushort)((Zcu.Mve.Visualization.Line)cells[i]).V2;
        }

        indexBuffer[0].SetData(indices, 0, LockFlags.None);
      }
      else
      {
        indexBuffer[0] = new IndexBuffer(typeof(int), 2 * cells.Count, device, Usage.WriteOnly, Pool.Managed);

        int[] indices = new int[2 * cells.Count];

        for (int i = 0; i < cells.Count; i++)
        {
            indices[2 * i] = ((Zcu.Mve.Visualization.Line)cells[i]).V1;
            indices[(2 * i) + 1] = ((Zcu.Mve.Visualization.Line)cells[i]).V2;
        }

        indexBuffer[0].SetData(indices, 0, LockFlags.None);
      }
    }

    private void createTetrahedronIB()
    {
      primitiveType = PrimitiveType.TriangleStrip;
      indexBuffer = new IndexBuffer[cells.Count];

      for (int i = 0; i < cells.Count; i++)
      {
        Tetrahedron tetra = (Tetrahedron) cells[i];

        if (vertexBuffer.Count < ushort.MaxValue)
        {
          indexBuffer[i] = new IndexBuffer(typeof(ushort), 6, device, Usage.WriteOnly, Pool.Managed);

          ushort[] indices = new ushort[6];

          indices[0] = (ushort) tetra.V1;
          indices[1] = (ushort) tetra.V2;
          indices[2] = (ushort) tetra.V3;
          indices[3] = (ushort) tetra.V4;
          indices[4] = (ushort) tetra.V1;
          indices[5] = (ushort) tetra.V2;

          indexBuffer[i].SetData(indices, 0, LockFlags.None);
        }
        else
        {
          indexBuffer[i] = new IndexBuffer(typeof(int), 6, device, Usage.WriteOnly, Pool.Managed);

          int[] indices = new int[6];

          indices[0] = tetra.V1;
          indices[1] = tetra.V2;
          indices[2] = tetra.V3;
          indices[3] = tetra.V4;
          indices[4] = tetra.V1;
          indices[5] = tetra.V2;

          indexBuffer[i].SetData(indices, 0, LockFlags.None);
        }
      }
    }

    private void createTriangleFanIB()
    {
      primitiveType = PrimitiveType.TriangleFan;
      indexBuffer = new IndexBuffer[cells.Count];

      for (int i = 0; i < cells.Count; i++)
      {
        TriangleFan tf = (TriangleFan) cells[i];

        if (vertexBuffer.Count < ushort.MaxValue)
        {
          indexBuffer[i] = new IndexBuffer(typeof(ushort), tf.PointCount, device, Usage.WriteOnly, Pool.Managed);

          ushort[] tmpPoints = new ushort[tf.PointCount];
          for (int j = 0; j < tmpPoints.Length; j++)
          {
            tmpPoints[j] = (ushort) tf.Points[j];
          }

          indexBuffer[i].SetData(tmpPoints, 0, LockFlags.None);
        }
        else
        {
          indexBuffer[i] = new IndexBuffer(typeof(int), tf.PointCount, device, Usage.WriteOnly, Pool.Managed);
          indexBuffer[i].SetData(tf.Points, 0, LockFlags.None);
        }
      }
    }

    private void createTriangleStripIB()
    {
      primitiveType = PrimitiveType.TriangleStrip;
      indexBuffer = new IndexBuffer[cells.Count];

      for (int i = 0; i < cells.Count; i++)
      {
        TriangleStrip ts = (TriangleStrip) cells[i];

        if (vertexBuffer.Count < ushort.MaxValue)
        {
          indexBuffer[i] = new IndexBuffer(typeof(ushort), ts.PointCount, device, Usage.WriteOnly, Pool.Managed);

          ushort[] tmpPoints = new ushort[ts.PointCount];
          for (int j = 0; j < tmpPoints.Length; j++)
          {
            tmpPoints[j] = (ushort) ts.Points[j];
          }

          indexBuffer[i].SetData(tmpPoints, 0, LockFlags.None);
        }
        else
        {
          indexBuffer[i] = new IndexBuffer(typeof(int), ts.PointCount, device, Usage.WriteOnly, Pool.Managed);
          indexBuffer[i].SetData(ts.Points, 0, LockFlags.None);
        }
      }
    }

    private void createTriangleListIB()
    {
      primitiveType = PrimitiveType.TriangleList;
      indexBuffer = new IndexBuffer[1];

      //Check caps...
      //Caps c = Manager.GetDeviceCaps(0, DeviceType.Hardware);
      //Console.WriteLine("MaxVertexIndex: " + c.MaxVertexIndex);
      //Console.WriteLine("MaxPrimitiveCount: " + c.MaxPrimitiveCount);
      
      if (vertexBuffer.Count < ushort.MaxValue)
      {
        indexBuffer[0] = new IndexBuffer(typeof(ushort), 3 * cells.Count, device, Usage.WriteOnly, Pool.Managed);
        ushort[] indices = new ushort[3 * cells.Count];

        for (int i = 0; i < cells.Count; i++)
        {
          indices[(3 * i)] = (ushort) ((Triangle) cells[i]).V1;
          indices[(3 * i) + 1] = (ushort) ((Triangle) cells[i]).V2;
          indices[(3 * i) + 2] = (ushort) ((Triangle) cells[i]).V3;
        }

        indexBuffer[0].SetData(indices, 0, LockFlags.None);
      }
      else
      {
        indexBuffer[0] = new IndexBuffer(typeof(int), 3 * cells.Count, device, Usage.WriteOnly, Pool.Managed);
        int[] indices = new int[3 * cells.Count];

        for (int i = 0; i < cells.Count; i++)
        {
          indices[(3 * i)] = ((Triangle) cells[i]).V1;
          indices[(3 * i) + 1] = ((Triangle) cells[i]).V2;
          indices[(3 * i) + 2] = ((Triangle) cells[i]).V3;
        }

        indexBuffer[0].SetData(indices, 0, LockFlags.None);
      }
    }

    #region IDisposable Members

    /// <summary>
    /// Dispose index buffer.
    /// </summary>
    public void Dispose()
    {
      foreach (IndexBuffer ib in indexBuffer)
      {
        ib.Dispose();
      }
    }

    #endregion
  }
}
