using System;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.Collections.Generic;

using Zcu.Mve.Visualization;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Zcu.Mve.UnstrGridRendering.Legend
{
    /// <summary>
    /// Class used for storing legend objects.
    /// </summary>
    public class LegendComposer : IDisposable
    {
        class LegendWrapper
        {
            public bool dataUpdated;

            public BaseLegend legend;
            public Texture defMemTexture;

            public LegendWrapper(BaseLegend l)
            {
                legend = l;
                dataUpdated = true;
                defMemTexture = null;
            }

            public LegendWrapper(BaseLegend l, Texture t)
            {
                legend = l;
                dataUpdated = true;
                defMemTexture = null;
            }

            /// <summary>
            /// Release default memory texture
            /// </summary>
            public void OnDeviceLost()
            {
                if ((defMemTexture != null) && (!defMemTexture.Disposed))
                {
                    defMemTexture.Dispose();
                }
            }

            /// <summary>
            /// Release all allocated resources
            /// </summary>
            public void Dispose()
            {
                legend = null;

                if (!defMemTexture.Disposed) defMemTexture.Dispose();

                defMemTexture = null;
            }
        };

        #region Fields

        private List<LegendWrapper> dataList;

        private bool anyChange;
        private Device device;
        private Size currWndSize;
        private PointF mult; // diference between init wnd size and currect size
        private static Color COLOR = Color.FromArgb(255, 255, 255, 255);

        #endregion

        public LegendComposer(Device device)
        {
            this.device = device;

            // window size
            Surface sTmp = device.GetBackBuffer(0, 0, BackBufferType.Mono);

            currWndSize = new Size(sTmp.Description.Width, sTmp.Description.Height);

            sTmp.Dispose();

            mult = new PointF(1.0f, 1.0f);

            dataList = new List<LegendWrapper>();

            anyChange = false;
        }

        /// <summary>
        /// Addition of single legend.
        /// </summary>
        /// <param name="bLegend"></param>
        public void Add(BaseLegend bLegend)
        {
            Insert(bLegend);
        }

        /// <summary>
        /// Addition of collection of legends.
        /// </summary>
        /// <param name="uda"></param>
        public void Add(UniformDataArray uda)
        {
            if (uda != null)
            {
                System.Collections.IEnumerator ie = uda.GetEnumerator();

                while (ie.MoveNext())
                {
                    Insert((BaseLegend)ie.Current);
                }
            }
        }

        /// <summary>
        /// Update legends if necessarily.
        /// </summary>
        /// <param name="device"></param>
        public void Update(Device device)
        {
            if (anyChange)
            {
                List<LegendWrapper>.Enumerator ie = dataList.GetEnumerator();

                // find who need update
                while (ie.MoveNext())
                {
                    if (ie.Current.dataUpdated)
                    {
                        ReCreateDefTexture(ie.Current);
                    }
                }

                ie.Dispose();
            }
        }

        /// <summary>
        /// Rednering legends.
        /// </summary>
        /// <param name="sprite"></param>
        public void Render(Sprite sprite)
        {
            if (dataList.Count == 0) return;

            List<LegendWrapper>.Enumerator ie = dataList.GetEnumerator();

            while (ie.MoveNext())
            {
                System.Drawing.Rectangle r = ie.Current.legend.Rectangle;

                System.Drawing.Point tmpLocation = new Point(currWndSize.Width - r.Left - r.Width, r.Top);//new Point((int)(r.Location.X * mult.X), (int)(r.Location.Y * mult.Y));

                sprite.Draw2D(ie.Current.defMemTexture, Rectangle.Empty, r.Size, tmpLocation, COLOR);
            }

            ie.Dispose();
        }

        private bool Insert(BaseLegend bLegend)
        {
            if (bLegend.Bitmap == null) return false;

            LegendWrapper lw = new LegendWrapper(bLegend);

            if (!CreateTexture(lw)) return false;

            dataList.Add(lw);

            return true;
        }

        private bool CreateTexture(LegendWrapper lw)
        {
            if (device != null)
            {
                Size s = lw.legend.Rectangle.Size;

                try
                {
                    lw.defMemTexture = new Texture(device, lw.legend.Bitmap, Usage.Dynamic, Pool.Default);

                    return true;
                }

                catch (OutOfVideoMemoryException eofvme)
                {
                    Console.WriteLine(eofvme);

                    return false;
                }

                catch (OutOfMemoryException eoome)
                {
                    Console.WriteLine(eoome);

                    return false;
                }

                catch (InvalidCallException oice)
                {
                    Console.WriteLine(oice);

                    return false;
                }
            }

            return false;
        }

        private void ReCreateDefTexture(LegendWrapper lw)
        {
            if (device != null)
            {
                Size s = lw.legend.Rectangle.Size;

                lw.defMemTexture = new Texture(device, lw.legend.Bitmap, Usage.Dynamic, Pool.Default);
            }
        }

        /// <summary>
        /// Release resources allocated in default memory pool.
        /// </summary>
        /// <param name="device"></param>
        public void OnDeviceLost(Device device)
        {
            // Release all resources on GPU - textures dealocation
            List<LegendWrapper>.Enumerator ie = dataList.GetEnumerator();

            while (ie.MoveNext())
            {
                ie.Current.OnDeviceLost();
            }

            ie.Dispose();
        }

        /// <summary>
        /// Resources recreation.
        /// </summary>
        /// <param name="device"></param>
        public void OnDeviceReset(Device device)
        {
            // window size
            Surface sTmp = device.GetBackBuffer(0, 0, BackBufferType.Mono);

            currWndSize.Width = sTmp.Description.Width;
            currWndSize.Height = sTmp.Description.Height;

            mult = new PointF((float)sTmp.Description.Width / (float)currWndSize.Width, (float)sTmp.Description.Height / (float)currWndSize.Height);

            sTmp.Dispose();

            List<LegendWrapper>.Enumerator ie = dataList.GetEnumerator();

            while (ie.MoveNext())
            {
                ReCreateDefTexture(ie.Current);
            }

            ie.Dispose();
        }

        #region IDisposable Members

        public void Dispose()
        {
            if (dataList != null)
            {
                List<LegendWrapper>.Enumerator ie = dataList.GetEnumerator();

                while (ie.MoveNext())
                {
                    ie.Current.Dispose();
                }

                ie.Dispose();
            }
        }

        void IDisposable.Dispose()
        {
            if (dataList != null)
            {
                List<LegendWrapper>.Enumerator ie = dataList.GetEnumerator();

                while (ie.MoveNext())
                {
                    ie.Current.Dispose();
                }

                ie.Dispose();
            }
        }

        #endregion
    }
}
