using System;
using System.Globalization;
using System.Text;
using System.Xml;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using System.Diagnostics;

namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// Defines light properties.
    /// </summary>
    [DataObject("Defines a set of lighting properties.")]
    public class Light : DataObject
    {
        #region Fields
        /// <summary>
        /// Type of the light source.
        /// </summary>
        private LightType type;

        /// <summary>
        /// Color of ambient part of the light.
        /// </summary>
        private ColorRGBA ambient;

        /// <summary>
        /// Color of diffuse part of the light.
        /// </summary>
        private ColorRGBA diffuse;

        /// <summary>
        /// Color of specular part of the light.
        /// </summary>
        private ColorRGBA specular;

        /// <summary>
        /// Position of the light source.
        /// This member has no effect on directional lights.
        /// </summary>
        private Vector3D position;

        /// <summary>
        /// Direction in which the light is pointing.
        /// This member has meaning only for directional light and spotlights.
        /// The vector need not be normalized, but it should have a nonzero length.
        /// </summary>
        private Vector3D direction;

        /// <summary>
        /// Value that indicates the decrease in illumination.
        /// The effect of Falloff, the light attenuation between a spotlight's
        /// inner and outer cones, on the lighting is subtle.
        /// Furthermore, a small performance penalty is incurred by
        /// shaping the Falloff curve. For these reasons, most developers
        /// set this value to 1.0.
        /// </summary>
        private float falloff;

        /// <summary>
        /// Distance beyond which the light has no effect.
        /// This member does not affect directional lights.
        /// </summary>
        private float range;

        /// <summary>
        /// Value that indicates the angle of the spotlight's inner cone.
        /// This value must be in the range of 0 through the value specified
        /// by OuterConeAngle.
        /// </summary>
        private float innerConeAngle;

        /// <summary>
        /// Value that indicates the angle of the outer edge of the spotlight's
        /// outer cone. Points outside of the cone are not lit by the spotlight.
        /// This value must fall between 0 and pi.
        /// </summary>
        private float outerConeAngle;

        /// <summary>
        /// Specifies how light intensity changes over distance. 
        /// Attenuation = AT0 + AT1 * distance + AT2 * distance^2.
        /// Result Light Intensity = Source Intensity / Attenuation. 
        /// Attenuation values are ignored for directional lights. 
        /// Its valid values range from 0.0 to infinity.
        /// For non-directional lights,
        /// all three attenuation values should not be set to 0.0 simultaneously.
        /// </summary>
        private float attenuation0;

        /// <summary>
        /// Specifies how light intensity changes over distance.
        /// Attenuation = AT0 + AT1 * distance + AT2 * distance^2.
        /// Result Light Intensity = Source Intensity / Attenuation. 
        /// Attenuation values are ignored for directional lights. 
        /// Its valid values range from 0.0 to infinity.
        /// For non-directional lights,
        /// all three attenuation values should not be set to 0.0 simultaneously.
        /// </summary>
        private float attenuation1;

        /// <summary>
        /// Specifies how light intensity changes over distance.
        /// Attenuation = AT0 + AT1 * distance + AT2 * distance^2.
        /// Result Light Intensity = Source Intensity / Attenuation. 
        /// Attenuation values are ignored for directional lights. 
        /// Its valid values range from 0.0 to infinity.
        /// For non-directional lights,
        /// all three attenuation values should not be set to 0.0 simultaneously.
        /// </summary>
        private float attenuation2;

        /// <summary>
        /// Turns the light on or off.
        /// </summary>
        private bool enabled;

        /// <summary>
        /// Show position and direction of the light source.
        /// </summary>
        private bool showLight;
        #endregion

        #region Properties
        /// <summary>
        /// Retrieves or sets the type of the light source.
        /// </summary>
        public LightType Type
        {
            get
            {
                return type;
            }
            set
            {
                type = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets color for ambient part of the light.
        /// </summary>
        public ColorRGBA Ambient
        {
            get
            {
                return this.ambient;
            }
            set
            {
                this.ambient = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets color for diffuse part of the light.
        /// </summary>
        public ColorRGBA Diffuse
        {
            get
            {
                return this.diffuse;
            }
            set
            {
                this.diffuse = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets color for specular part of the light.
        /// </summary>
        public ColorRGBA Specular
        {
            get
            {
                return this.specular;
            }
            set
            {
                this.specular = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets position of the light.
        /// This member has no effect on directional lights.
        /// </summary>
        public Vector3D Position
        {
            get
            {
                return this.position;
            }
            set
            {
                this.position = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets direction of the light.
        /// This member has meaning only for directional light and spotlights.
        /// The vector need not be normalized, but it should have a nonzero length.
        /// </summary>
        public Vector3D Direction
        {
            get
            {
                return direction;
            }
            set
            {
                direction = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets fallof of the light.
        /// The effect of Falloff, the light attenuation between a spotlight's
        /// inner and outer cones, on the lighting is subtle.
        /// Furthermore, a small performance penalty is incurred by
        /// shaping the Falloff curve. For these reasons, most developers
        /// set this value to 1.0.
        /// </summary>
        public float Falloff
        {
            get
            {
                return falloff;
            }
            set
            {
                falloff = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets distance beyond which the light has no effect. Distance must be greater
        /// then zero. This member does not affect directional lights.
        /// </summary>
        public float Range
        {
            get
            {
                return this.range;
            }
            set
            {
                if (value > 0.0f)
                    this.range = value;
                else
                    throw new MveException(this, "Range of light must be greater then zero.");

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets the angle of the spotlight's inner cone.
        /// This value must be in the range of 0 through the value specified
        /// by OuterConeAngle.  
        /// </summary>
        public float InnerConeAngle
        {
            get
            {
                return innerConeAngle;
            }
            set
            {
                if ((value >= 0.0f) && (value <= OuterConeAngle))
                    this.innerConeAngle = value;
                else
                    throw new MveException(this, "This value must be in the range of 0" +
                                           " through the value specified by OuterConeAngle.");

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets the angle of the outer edge of the spotlight's
        /// outer cone. Points outside of the cone are not lit by the spotlight.
        /// This value must fall between 0 and pi.
        /// </summary>
        public float OuterConeAngle
        {
            get
            {
                return outerConeAngle;
            }
            set
            {
                if ((value >= 0.0f) && (value <= Math.PI))
                    this.outerConeAngle = value;
                else
                    throw new MveException(this, "This value must fall between 0 and pi.");

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets attenuation constant 0. 
        /// Specifies how light intensity changes over distance. 
        /// Attenuation = AT0 + AT1 * distance + AT2 * distance^2.
        /// Result Light Intensity = Source Intensity / Attenuation. 
        /// Attenuation values are ignored for directional lights. 
        /// Its valid values range from 0.0 to infinity.
        /// For non-directional lights,
        /// all three attenuation values should not be set to 0.0 simultaneously.
        /// </summary>
        public float Attenuation0
        {
            get
            {
                return this.attenuation0;
            }
            set
            {
                if (value >= 0.0f)
                    this.attenuation0 = value;
                else
                    throw new MveException(this, "Attenuation constant 0 can't be less then zero.");

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets attenuation constant 1. 
        /// Specifies how light intensity changes over distance. 
        /// Attenuation = AT0 + AT1 * distance + AT2 * distance^2.
        /// Result Light Intensity = Source Intensity / Attenuation. 
        /// Attenuation values are ignored for directional lights. 
        /// Its valid values range from 0.0 to infinity.
        /// For non-directional lights,
        /// all three attenuation values should not be set to 0.0 simultaneously. 
        /// </summary>
        public float Attenuation1
        {
            get
            {
                return this.attenuation1;
            }
            set
            {
                if (value >= 0.0f)
                    this.attenuation1 = value;
                else
                    throw new MveException(this, "Attenuation constant 1 can't be less then zero.");

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets attenuation constant 2.
        /// Specifies how light intensity changes over distance. 
        /// Attenuation = AT0 + AT1 * distance + AT2 * distance^2.
        /// Result Light Intensity = Source Intensity / Attenuation. 
        /// Attenuation values are ignored for directional lights. 
        /// Its valid values range from 0.0 to infinity.
        /// For non-directional lights,
        /// all three attenuation values should not be set to 0.0 simultaneously.
        /// </summary>
        public float Attenuation2
        {
            get
            {
                return this.attenuation2;
            }
            set
            {
                if (value >= 0.0f)
                    this.attenuation2 = value;
                else
                    throw new MveException(this, "Attenuation constant 2 can't be less then zero.");

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Gets or sets value for turning on/off this light.
        /// </summary>
        public bool Enabled
        {
            get
            {
                return this.enabled;
            }
            set
            {
                this.enabled = value;

                UpdateModificationTime();
            }
        }

        /// <summary>
        /// Show position and direction of the light source.
        /// </summary>
        public bool ShowLight
        {
            get
            {
                return showLight;
            }
            set
            {
                showLight = value;
            }
        }
        #endregion

        /// <summary>
        /// Constructor initializes all fields to default values.
        /// </summary>
        public Light()
        {
            this.type = LightType.Directional;
            this.ambient = new ColorRGBA(0f, 0f, 0f);
            this.diffuse = new ColorRGBA(1f, 1f, 1f);
            this.specular = new ColorRGBA(1f, 1f, 1f);
            this.position = new Vector3D(0.0, 0.0, 0.0);
            this.direction = new Vector3D(0.0, -1f, 0.0);
            this.falloff = 1.0f;
            this.range = 100f;
            this.innerConeAngle = (float)Math.PI / 8f;
            this.outerConeAngle = (float)Math.PI / 4f;
            this.attenuation0 = 1.0f;
            this.attenuation1 = 0.0f;
            this.attenuation2 = 0.0f;
            this.enabled = true;
            this.showLight = false;

            UpdateModificationTime();
        }

        /// <summary>
        /// Check data consistence of current object.
        /// </summary>
        /// <returns>Returns true if succeed.</returns>
        public override bool CheckConsistence()
        {
            bool returnValue = true;

            // light type...
            returnValue &= this.ambient.CheckConsistence();
            returnValue &= this.diffuse.CheckConsistence();
            returnValue &= this.specular.CheckConsistence();
            returnValue &= this.position.CheckConsistence();
            returnValue &= this.direction.CheckConsistence();
            // returnValue &= this.falloff ???
            returnValue &= (this.range <= 0.0f);
            returnValue &= ((this.innerConeAngle >= 0.0f) && (this.innerConeAngle <= OuterConeAngle));
            returnValue &= ((this.outerConeAngle >= 0.0f) && (this.outerConeAngle <= Math.PI));
            returnValue &= (this.attenuation0 < 0.0f);
            returnValue &= (this.attenuation1 < 0.0f);
            returnValue &= (this.attenuation2 < 0.0f);
            returnValue &= (this.attenuation0 > 0.0f || this.attenuation1 > 0.0f || this.attenuation2 > 0.0f);

            return returnValue;
        }

        /// <summary>
        /// Creates deep copy of current object.
        /// </summary>
        /// <returns>Returns new instance of this class with deep copied data from current object.</returns>
        public override IDataObject DeepCopy()
        {
            Light l = new Light();

            l.type = this.type;
            l.ambient = (ColorRGBA)this.ambient.DeepCopy();
            l.diffuse = (ColorRGBA)this.diffuse.DeepCopy();
            l.specular = (ColorRGBA)this.specular.DeepCopy();
            l.position = (Vector3D)this.position.DeepCopy();
            l.direction = (Vector3D)this.direction.DeepCopy();
            l.falloff = this.falloff;
            l.range = this.range;
            l.innerConeAngle = this.innerConeAngle;
            l.outerConeAngle = this.outerConeAngle;
            l.attenuation0 = this.attenuation0;
            l.attenuation1 = this.attenuation1;
            l.attenuation2 = this.attenuation2;
            l.enabled = this.enabled;
            l.showLight = this.showLight;

            return l;
        }

        /// <summary>
        /// Gets names of all assemblies used by this class.
        /// </summary>
        public override string[] LibNames
        {
            get
            {
                AddLibName(this.GetType());
                AddLibName(typeof(ColorRGBA));
                AddLibName(typeof(Vector3D));
                AddLibName(typeof(LightType));

                return base.LibNames;
            }
        }

        /// <summary>
        /// Reads data from XML.
        /// </summary>
        /// <param name="xmlTextReader">Representation of XML stream.</param>
        public override void ReadData(XmlTextReader xmlTextReader)
        {
            Debug.WriteLine("Light reading...");
            while (xmlTextReader.Read())
            {
                if (xmlTextReader.NodeType == XmlNodeType.Element)
                {
                    switch (xmlTextReader.Name)
                    {
                        case "type":
                            this.type = FromString.GetLightType(xmlTextReader.ReadString(), LightType.Directional);
                            break;
                        case "ambient":
                            this.ambient = (ColorRGBA)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
                            this.ambient.ReadData(xmlTextReader);
                            break;
                        case "diffuse":
                            this.diffuse = (ColorRGBA)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
                            this.diffuse.ReadData(xmlTextReader);
                            break;
                        case "specular":
                            this.specular = (ColorRGBA)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
                            this.specular.ReadData(xmlTextReader);
                            break;
                        case "position":
                            this.position = (Vector3D)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
                            this.position.ReadData(xmlTextReader);
                            break;
                        case "direction":
                            this.direction = (Vector3D)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
                            this.direction.ReadData(xmlTextReader);
                            break;
                        case "falloff":
                            this.falloff = FromString.GetFloat(xmlTextReader.ReadString(), 1f);
                            break;
                        case "range":
                            this.range = FromString.GetFloat(xmlTextReader.ReadString(), 100f);
                            break;
                        case "innerconeangle":
                            this.innerConeAngle = FromString.GetFloat(xmlTextReader.ReadString(), (float)Math.PI / 8f);
                            break;
                        case "outerconeangle":
                            this.outerConeAngle = FromString.GetFloat(xmlTextReader.ReadString(), (float)Math.PI / 4f);
                            break;
                        case "attenuation0":
                            this.attenuation0 = FromString.GetFloat(xmlTextReader.ReadString(), 1f);
                            break;
                        case "attenuation1":
                            this.attenuation1 = FromString.GetFloat(xmlTextReader.ReadString(), 0f);
                            break;
                        case "attenuation2":
                            this.attenuation2 = FromString.GetFloat(xmlTextReader.ReadString(), 0f);
                            break;
                        case "enabled":
                            this.enabled = FromString.GetBool(xmlTextReader.ReadString(), true);
                            break;
                        case "showlight":
                            this.showLight = FromString.GetBool(xmlTextReader.ReadString(), true);
                            break;
                        default:
                            break;
                    } // switch
                } //if

                if (xmlTextReader.NodeType == XmlNodeType.EndElement)
                {
                    if (xmlTextReader.Name == "light")
                        break;
                }
            } // while			

            UpdateModificationTime();
            Debug.WriteLine(" OK");

            Console.WriteLine(this.ToString());
        }

        /// <summary>
        /// Writes current object's data to XML.
        /// </summary>
        /// <param name="xmlTextWriter">Representation of XML stream.</param>
        public override void WriteData(XmlTextWriter xmlTextWriter)
        {
            xmlTextWriter.WriteStartElement("light");

            xmlTextWriter.WriteStartElement("type");
            xmlTextWriter.WriteAttributeString("type", this.type.GetType().FullName);
            xmlTextWriter.WriteString(this.type.ToString());
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("ambient");
            xmlTextWriter.WriteAttributeString("type", this.ambient.GetType().FullName);
            this.ambient.WriteData(xmlTextWriter);
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("diffuse");
            xmlTextWriter.WriteAttributeString("type", this.diffuse.GetType().FullName);
            this.diffuse.WriteData(xmlTextWriter);
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("specular");
            xmlTextWriter.WriteAttributeString("type", this.specular.GetType().FullName);
            this.specular.WriteData(xmlTextWriter);
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("position");
            xmlTextWriter.WriteAttributeString("type", this.position.GetType().FullName);
            this.position.WriteData(xmlTextWriter);
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("direction");
            xmlTextWriter.WriteAttributeString("type", this.direction.GetType().FullName);
            this.direction.WriteData(xmlTextWriter);
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("falloff");
            xmlTextWriter.WriteAttributeString("type", this.falloff.GetType().FullName);
            xmlTextWriter.WriteString(this.falloff.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("range");
            xmlTextWriter.WriteAttributeString("type", this.range.GetType().FullName);
            xmlTextWriter.WriteString(this.range.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("innerconeangle");
            xmlTextWriter.WriteAttributeString("type", this.innerConeAngle.GetType().FullName);
            xmlTextWriter.WriteString(this.innerConeAngle.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("outerconeangle");
            xmlTextWriter.WriteAttributeString("type", this.outerConeAngle.GetType().FullName);
            xmlTextWriter.WriteString(this.outerConeAngle.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("attenuation0");
            xmlTextWriter.WriteAttributeString("type", this.attenuation0.GetType().FullName);
            xmlTextWriter.WriteString(this.attenuation0.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("attenuation1");
            xmlTextWriter.WriteAttributeString("type", this.attenuation1.GetType().FullName);
            xmlTextWriter.WriteString(this.attenuation1.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("attenuation2");
            xmlTextWriter.WriteAttributeString("type", this.attenuation2.GetType().FullName);
            xmlTextWriter.WriteString(this.attenuation2.ToString(CultureInfo.InvariantCulture.NumberFormat));
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("enabled");
            xmlTextWriter.WriteAttributeString("type", this.enabled.GetType().FullName);
            xmlTextWriter.WriteString(this.enabled.ToString());
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteStartElement("showlight");
            xmlTextWriter.WriteAttributeString("type", this.showLight.GetType().FullName);
            xmlTextWriter.WriteString(this.showLight.ToString());
            xmlTextWriter.WriteEndElement();

            xmlTextWriter.WriteEndElement();
        }

        /// <summary>
        /// Converts current object to its string description.
        /// </summary>
        /// <returns>Resulting string description.</returns>
        public override string ToString()
        {
            StringBuilder str = new StringBuilder(this.GetType().FullName + ":\r\n");

            str.Append("Type: ");
            str.Append(this.type.ToString() + "\r\n");

            str.Append("Ambient: ");
            str.Append(this.ambient.ToString() + "\r\n");

            str.Append("Diffuse: ");
            str.Append(this.diffuse.ToString() + "\r\n");

            str.Append("Specular: ");
            str.Append(this.specular.ToString() + "\r\n");

            str.Append("Position : ");
            str.Append(this.position.ToString() + "\r\n");

            str.Append("Direction : ");
            str.Append(this.direction.ToString() + "\r\n");

            str.Append("Falloff : ");
            str.Append(this.falloff.ToString() + "\r\n");

            str.Append("Range : ");
            str.Append(this.range.ToString() + "\r\n");

            str.Append("InnerConeAngle : ");
            str.Append(this.innerConeAngle.ToString() + "\r\n");

            str.Append("OuterConeAngle : ");
            str.Append(this.outerConeAngle.ToString() + "\r\n");

            str.Append("Attuentation0 : ");
            str.Append(this.attenuation0.ToString() + "\r\n");

            str.Append("Attuentation1 : ");
            str.Append(this.attenuation1.ToString() + "\r\n");

            str.Append("Attuentation2 : ");
            str.Append(this.attenuation2.ToString() + "\r\n");

            str.Append("Enabled : ");
            str.Append(this.enabled.ToString() + "\r\n");

            str.Append("ShowLight : ");
            str.Append(this.showLight.ToString() + "\r\n");

            str.Append("\r\n");

            return str.ToString();
        }
    }
}
