using System;
using System.Collections.Generic;
using System.Text;

using Zcu.Mve.Core;

using System.ComponentModel;
using System.Globalization;
using System.Xml;
using System.Diagnostics;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// Source of Light. Light is supposed to be placed in Scene.
    /// </summary>
    [ModuleInfo("Jan Kaiser", "Source of Light. Light is supposed to be placed in Scene.", IconName = "LightModule.ico")]
    [PortInfo("outLight", "Generated light.")]
    public class LightModule : Module
    {
        private bool dataChanged = true;
        private Light light;

        #region Properties

        /// <summary>
        /// Type of the light source. Directional lights are slightly 
        /// faster than point lights, but point lights look a little better. 
        /// Spotlights offer interesting visual effects but are slow.
        /// </summary>
        [Category("Light")]
        [DefaultValue(LightType.Directional)]
        [Description("Type of the light source. Directional lights are slightly "
                    + "faster than point lights, but point lights look a little better. "
                    + "Spotlights offer interesting visual effects but are slow.")]
        public LightType Type
        {
            get
            {
                return light.Type;
            }
            set
            {
                light.Type = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Ambient color.
        /// </summary>
        [Category("Light color")]
        [Description("Ambient color.")]
        public ColorRGBA Ambient
        {
            get
            {
                return light.Ambient;
            }
            set
            {
                light.Ambient = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Diffuse color.
        /// </summary>
        [Category("Light color")]
        [Description("Diffuse color.")]
        public ColorRGBA Diffuse
        {
            get
            {
                return light.Diffuse;
            }
            set
            {
                light.Diffuse = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Specular color.
        /// </summary>
        [Category("Light color")]
        [Description("Specular color.")]
        public ColorRGBA Specular
        {
            get
            {
                return light.Specular;
            }
            set
            {
                light.Specular = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Position of the light. Has no effect on directional lights.
        /// </summary>
        [Category("Light")]
        [Description("Position of the light. Has no effect on directional lights.")]
        public new Vector3D Position
        {
            get
            {
                return light.Position;
            }
            set
            {
                light.Position = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Direction of the light. Has meaning only for directional 
        /// light and spotlights. The vector need not be normalized,
        /// but it should have a nonzero length.
        /// </summary>
        [Description("Direction of the light. Has meaning only for directional "
                    + "light and spotlights. The vector need not be normalized, "
                    + "but it should have a nonzero length.")]
        [Category("Light")]
        public Vector3D Direction
        {
            get
            {
                return light.Direction;
            }
            set
            {
                light.Direction = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// The effect of Falloff, the light attenuation between a spotlight's 
        /// inner and outer cones, on the lighting is subtle.
        /// Furthermore, a small performance penalty is incurred by 
        /// shaping the Falloff curve. For these reasons, most developers 
        /// set this value to 1.0.
        /// </summary>
        [Description("The effect of Falloff, the light attenuation between a spotlight's "
                    + "inner and outer cones, on the lighting is subtle. "
                    + "Furthermore, a small performance penalty is incurred by "
                    + "shaping the Falloff curve. For these reasons, most developers "
                    + "set this value to 1.0.")]
        [DefaultValue(1.0f)]
        [Category("Light attenuation")]
        public float Falloff
        {
            get
            {
                return light.Falloff;
            }
            set
            {
                light.Falloff = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Distance beyond which the light has no effect. 
        /// Distance must be greater then zero. Does 
        /// not affect directional lights.
        /// </summary>
        [Description("Distance beyond which the light has no effect. "
                    + "Distance must be greater then zero. Does "
                    + "not affect directional lights.")]
        [Category("Light attenuation")]
        public float Range
        {
            get
            {
                return light.Range;
            }
            set
            {
                light.Range = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Angle of the spotlight's inner cone. This value must be 
        /// in the range of 0 through the value 
        /// specified by OuterConeAngle. 
        /// </summary>
        [Description("Angle of the spotlight's inner cone. This value must be "
                    + "in the range of 0 through the value "
                    + "specified by OuterConeAngle. ")]
        [Category("Light attenuation")]
        public float InnerConeAngle
        {
            get
            {
                return light.InnerConeAngle;
            }
            set
            {
                light.InnerConeAngle = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Angle of the outer edge of the spotlight's outer cone. 
        /// Points outside of the cone are not lit by the spotlight. 
        /// This value must fall between 0 and pi.
        /// </summary>
        [Description("Angle of the outer edge of the spotlight's outer cone. "
                    + "Points outside of the cone are not lit by the spotlight. "
                    + "This value must fall between 0 and pi.")]
        [Category("Light attenuation")]
        public float OuterConeAngle
        {
            get
            {
                return light.OuterConeAngle;
            }
            set
            {
                light.OuterConeAngle = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Attenuation constant 0. Specifies how light intensity
        /// changes over distance. Attenuation = AT0 + AT1 * distance + AT2 * distance^2.
        /// Result Light Intensity = Source Intensity / Attenuation.
        /// Attenuation values are ignored for directional lights.
        /// Its valid values range from 0.0 to infinity.
        /// For non-directional lights, all three attenuation values
        /// should not be set to 0.0 simultaneously.
        /// </summary>
        [Description("Attenuation constant 0. Specifies how light intensity "
                    + "changes over distance. Attenuation = AT0 + AT1 * distance + AT2 * distance^2."
                    + "Result Light Intensity = Source Intensity / Attenuation. "
                    + "Attenuation values are ignored for directional lights. "
                    + "Its valid values range from 0.0 to infinity. "
                    + "For non-directional lights, all three attenuation values "
                    + "should not be set to 0.0 simultaneously.")]
        [Category("Light attenuation")]
        [DefaultValue(1f)]
        public float Attenuation0
        {
            get
            {
                return light.Attenuation0;
            }
            set
            {
                light.Attenuation0 = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Attenuation constant 1. Specifies how light intensity
        /// changes over distance. Attenuation = AT0 + AT1 * distance + AT2 * distance^2.
        /// Result Light Intensity = Source Intensity / Attenuation.
        /// Attenuation values are ignored for directional lights.
        /// Its valid values range from 0.0 to infinity.
        /// For non-directional lights, all three attenuation values
        /// should not be set to 0.0 simultaneously.
        /// </summary>
        [Description("Attenuation constant 1. Specifies how light intensity "
                + "changes over distance. Attenuation = AT0 + AT1 * distance + AT2 * distance^2."
                + "Result Light Intensity = Source Intensity / Attenuation. "
                + "Attenuation values are ignored for directional lights. "
                + "Its valid values range from 0.0 to infinity. "
                + "For non-directional lights, all three attenuation values "
                + "should not be set to 0.0 simultaneously.")]
        [Category("Light attenuation")]
        [DefaultValue(0f)]
        public float Attenuation1
        {
            get
            {
                return light.Attenuation1;
            }
            set
            {
                light.Attenuation1 = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Attenuation constant 2. Specifies how light intensity
        /// changes over distance. Attenuation = AT0 + AT1 * distance + AT2 * distance^2.
        /// Result Light Intensity = Source Intensity / Attenuation.
        /// Attenuation values are ignored for directional lights.
        /// Its valid values range from 0.0 to infinity.
        /// For non-directional lights, all three attenuation values
        /// should not be set to 0.0 simultaneously.
        /// </summary>
        [Description("Attenuation constant 2. Specifies how light intensity "
                + "changes over distance. Attenuation = AT0 + AT1 * distance + AT2 * distance^2."
                + "Result Light Intensity = Source Intensity / Attenuation. "
                + "Attenuation values are ignored for directional lights. "
                + "Its valid values range from 0.0 to infinity. "
                + "For non-directional lights, all three attenuation values "
                + "should not be set to 0.0 simultaneously.")]
        [Category("Light attenuation")]
        [DefaultValue(0f)]
        public float Attenuation2
        {
            get
            {
                return light.Attenuation2;
            }
            set
            {
                light.Attenuation2 = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Enable/disable this light.
        /// </summary>
        [Category("Light")]
        [Description("Enable/disable this light.")]
        [DefaultValue(true)]
        public bool Enabled
        {
            get
            {
                return light.Enabled;
            }
            set
            {
                light.Enabled = value;
                dataChanged = true;
            }
        }

        /// <summary>
        /// Show position and direction of the light source.
        /// </summary>
        [Category("Light")]
        [Description("Show position and direction of the light source.")]
        [DefaultValue(false)]
        public bool ShowLight
        {
            get
            {
                return light.ShowLight;
            }
            set
            {
                light.ShowLight = value;
                dataChanged = true;
            }
        }
        #endregion

        /// <summary>
        /// Constructor, initializes object and adds output port.
        /// </summary>
        public LightModule()
        {
            AddOutPort("outLight", typeof(Light));

            light = new Light();
        }

        /// <summary>
        /// Generates/updates light and exposes it on output port.
        /// </summary>
        public override void Execute()
        {
            if (dataChanged == true)
            {
                SetOutput("outLight", light, false);
                dataChanged = false;
            }
            else
            {
                SetOutput("outLight", light, true);
            }
        }

        /// <summary>
        /// Writes Xml configuration of current module.
        /// </summary>
        /// <param name="config">XmlElement from map config file which belongs to current module.</param>
        public override void WriteConfig(XmlElement config)
        {
            XmlElement xmlProperties = config.OwnerDocument.CreateElement("Properties");

            XmlElement xmlType = config.OwnerDocument.CreateElement("Type");
            xmlType.SetAttribute("type", light.Type.GetType().FullName);
            xmlType.InnerText = light.Type.ToString();
            xmlProperties.AppendChild(xmlType);

            XmlElement xmlAmbient = config.OwnerDocument.CreateElement("Ambient");
            xmlAmbient.SetAttribute("type", light.Ambient.GetType().FullName);
            xmlAmbient.InnerText = light.Ambient.Red.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                 + light.Ambient.Green.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                 + light.Ambient.Blue.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                 + light.Ambient.Alpha.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlAmbient);

            XmlElement xmlDiffuse = config.OwnerDocument.CreateElement("Diffuse");
            xmlDiffuse.SetAttribute("type", light.Diffuse.GetType().FullName);
            xmlDiffuse.InnerText = light.Diffuse.Red.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                 + light.Diffuse.Green.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                 + light.Diffuse.Blue.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                 + light.Diffuse.Alpha.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlDiffuse);

            XmlElement xmlSpecular = config.OwnerDocument.CreateElement("Specular");
            xmlSpecular.SetAttribute("type", light.Specular.GetType().FullName);
            xmlSpecular.InnerText = light.Specular.Red.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                  + light.Specular.Green.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                  + light.Specular.Blue.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                  + light.Specular.Alpha.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlSpecular);

            XmlElement xmlPosition = config.OwnerDocument.CreateElement("Position");
            xmlPosition.SetAttribute("type", light.Position.GetType().FullName);
            xmlPosition.InnerText = light.Position.X.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                  + light.Position.Y.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                  + light.Position.Z.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlPosition);

            XmlElement xmlDirection = config.OwnerDocument.CreateElement("Direction");
            xmlDirection.SetAttribute("type", light.Direction.GetType().FullName);
            xmlDirection.InnerText = light.Direction.X.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                  + light.Direction.Y.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                  + light.Direction.Z.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlDirection);

            XmlElement xmlFalloff = config.OwnerDocument.CreateElement("Falloff");
            xmlFalloff.SetAttribute("type", light.Falloff.GetType().FullName);
            xmlFalloff.InnerText = light.Falloff.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlFalloff);

            XmlElement xmlRange = config.OwnerDocument.CreateElement("Range");
            xmlRange.SetAttribute("type", light.Range.GetType().FullName);
            xmlRange.InnerText = light.Range.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlRange);

            XmlElement xmlInnerConeAngle = config.OwnerDocument.CreateElement("InnerConeAngle");
            xmlInnerConeAngle.SetAttribute("type", light.InnerConeAngle.GetType().FullName);
            xmlInnerConeAngle.InnerText = light.InnerConeAngle.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlInnerConeAngle);

            XmlElement xmlOuterConeAngle = config.OwnerDocument.CreateElement("OuterConeAngle");
            xmlOuterConeAngle.SetAttribute("type", light.OuterConeAngle.GetType().FullName);
            xmlOuterConeAngle.InnerText = light.OuterConeAngle.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlOuterConeAngle);

            XmlElement xmlAttenuation0 = config.OwnerDocument.CreateElement("Attenuation0");
            xmlAttenuation0.SetAttribute("type", light.Attenuation0.GetType().FullName);
            xmlAttenuation0.InnerText = light.Attenuation0.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlAttenuation0);

            XmlElement xmlAttenuation1 = config.OwnerDocument.CreateElement("Attenuation1");
            xmlAttenuation1.SetAttribute("type", light.Attenuation1.GetType().FullName);
            xmlAttenuation1.InnerText = light.Attenuation1.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlAttenuation1);

            XmlElement xmlAttenuation2 = config.OwnerDocument.CreateElement("Attenuation2");
            xmlAttenuation2.SetAttribute("type", light.Attenuation2.GetType().FullName);
            xmlAttenuation2.InnerText = light.Attenuation2.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlAttenuation2);

            XmlElement xmlEnabled = config.OwnerDocument.CreateElement("Enabled");
            xmlEnabled.SetAttribute("type", light.Enabled.GetType().FullName);
            xmlEnabled.InnerText = light.Enabled.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlEnabled);

            XmlElement xmlShowLight = config.OwnerDocument.CreateElement("ShowLight");
            xmlShowLight.SetAttribute("type", light.ShowLight.GetType().FullName);
            xmlShowLight.InnerText = light.ShowLight.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlShowLight);

            config.AppendChild(xmlProperties);
        }

        /// <summary>
        /// Reads Xml configuration of current module.
        /// </summary>
        /// <param name="config">XmlElement from map config file which belongs to current module.</param>
        public override void ReadConfig(XmlElement config)
        {
            foreach (XmlElement xeConfig in config.ChildNodes)
            {
                if (xeConfig.Name == "Properties")
                {
                    foreach (XmlElement xeProperties in xeConfig.ChildNodes)
                    {
                        switch (xeProperties.Name)
                        {
                            case "Type":
                                light.Type = FromString.GetLightType(xeProperties.InnerText, LightType.Directional);
                                break;
                            case "Ambient":
                                light.Ambient = FromString.GetColorRGBA(xeProperties.InnerText);
                                break;
                            case "Diffuse":
                                light.Diffuse = FromString.GetColorRGBA(xeProperties.InnerText);
                                break;
                            case "Specular":
                                light.Specular = FromString.GetColorRGBA(xeProperties.InnerText);
                                break;
                            case "Position":
                                light.Position = FromString.GetVector3D(xeProperties.InnerText);
                                break;
                            case "Direction":
                                light.Direction = FromString.GetVector3D(xeProperties.InnerText);
                                break;
                            case "Falloff":
                                light.Falloff = FromString.GetFloat(xeProperties.InnerText, 1f);
                                break;
                            case "Range":
                                light.Range = FromString.GetFloat(xeProperties.InnerText, 100f);
                                break;
                            case "InnerConeAngle":
                                light.InnerConeAngle = FromString.GetFloat(xeProperties.InnerText, (float)Math.PI / 8f);
                                break;
                            case "OuterConeAngle":
                                light.OuterConeAngle = FromString.GetFloat(xeProperties.InnerText, (float)Math.PI / 4f);
                                break;
                            case "Attenuation0":
                                light.Attenuation0 = FromString.GetFloat(xeProperties.InnerText, 1f);
                                break;
                            case "Attenuation1":
                                light.Attenuation1 = FromString.GetFloat(xeProperties.InnerText, 0f);
                                break;
                            case "Attenuation2":
                                light.Attenuation2 = FromString.GetFloat(xeProperties.InnerText, 0f);
                                break;
                            case "Enabled":
                                light.Enabled = FromString.GetBool(xeProperties.InnerText, true);
                                break;
                            case "ShowLight":
                                light.ShowLight = FromString.GetBool(xeProperties.InnerText, true);
                                break;
                            default:
                                Debug.WriteLine("Unreaded value:" + xeProperties.Name + " - " + xeProperties.InnerText);
                                break;
                        } // switch
                    } // foreach
                } // if
            } // foreach
        }

        /// <summary>
        /// Creates new instance of module and copy all necessary member variables.
        /// This method is used by MapEditor in Copy/Paste process.
        /// Without overriding it returns implicit instance of current module.
        /// </summary>
        public override Module DeepCopy()
        {
            LightModule lightSource = (LightModule)base.DeepCopy();

            lightSource.light = (Light)light.DeepCopy();

            return lightSource;
        }
    }
}
