using System;
using System.Collections.Generic;
using dx3d = Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Text;
using System.Drawing;

namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// Class holds Light properties for D3D object representation.
    /// It also performs transition from Light to D3DLight. 
    /// </summary>
    public class LightWrapper
    {
        #region Fields

        private Color ambient;
        private Color diffuse;
        private Color specular;

        private dx3d.Vector3 position;
        private dx3d.Vector3 direction;

        private Light light;

        private long lmtLight;

        private VertexBuffer vertexBuffer;
        private Device device;

        #endregion

        /// <summary>
        /// Constructor, initializes light.
        /// </summary>
        public LightWrapper()
        {
            this.lmtLight = 0;
            this.device = null;
        }

        /// <summary>
        /// Update light properties.
        /// </summary>
        /// <param name="device">Rendering device.</param>
        /// <param name="light">Light.</param>
        public void Update(Device device, Light light)
        {
            if (lmtLight != light.ModificationTime)
            {
                this.light = light;

                ambient = Color.FromArgb((int)light.Ambient.Alpha,
                                         (int)light.Ambient.Red,
                                         (int)light.Ambient.Green,
                                         (int)light.Ambient.Blue);

                diffuse = Color.FromArgb((int)light.Diffuse.Alpha,
                                         (int)light.Diffuse.Red,
                                         (int)light.Diffuse.Green,
                                         (int)light.Diffuse.Blue);

                specular = Color.FromArgb((int)light.Specular.Alpha,
                                          (int)light.Specular.Red,
                                          (int)light.Specular.Green,
                                          (int)light.Specular.Blue);

                position = new dx3d.Vector3((float)light.Position.X, (float)light.Position.Y, (float)light.Position.Z);
                direction = new dx3d.Vector3((float)light.Direction.X, (float)light.Direction.Y, (float)light.Direction.Z);
            }

            if ((lmtLight != light.ModificationTime)
            || (this.device != device))
            {
                this.device = device;

                createVertexBuffer();
            }
        }

        /// <summary>
        /// This method should be called during rendering between device.BeginScene() and device.EndScene().
        /// It sets lights in rendering device.
        /// </summary>
        /// <param name="device">Rendering device.</param>
        /// <param name="id">Light id.</param>
        public void SetLight(Device device, int id)
        {
            device.Lights[id].Type = light.Type;

            device.Lights[id].Ambient = ambient;
            device.Lights[id].Diffuse = diffuse;
            device.Lights[id].Specular = specular;

            device.Lights[id].Position = position;
            device.Lights[id].Direction = direction;

            device.Lights[id].Falloff = light.Falloff;
            device.Lights[id].Range = light.Range;
            device.Lights[id].InnerConeAngle = light.InnerConeAngle;
            device.Lights[id].OuterConeAngle = light.OuterConeAngle;
            device.Lights[id].Attenuation0 = light.Attenuation0;
            device.Lights[id].Attenuation1 = light.Attenuation1;
            device.Lights[id].Attenuation2 = light.Attenuation2;
            device.Lights[id].Enabled = light.Enabled;

            if (light.ShowLight)
            {
                setRenderState();
                device.VertexFormat = CustomVertex.PositionColored.Format;
                device.SetStreamSource(0, vertexBuffer, 0);
                device.DrawPrimitives(PrimitiveType.PointList, 0, 1);
                if (light.Type != LightType.Point)
                    device.DrawPrimitives(PrimitiveType.LineList, 0, 1);
            }
        }

        /// <summary>
        /// Set device render state for drawing bounding box.
        /// </summary>
        private void setRenderState()
        {
            device.RenderState.Lighting = false;
            device.RenderState.ColorVertex = true;
            device.RenderState.PointSize = 5f;
        }

        /// <summary>
        /// Create vertex buffer of bounding box vertices.
        /// </summary>
        private void createVertexBuffer()
        {
            vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionColored), 2, device, Usage.WriteOnly, CustomVertex.PositionColored.Format, Pool.Managed);

            CustomVertex.PositionColored[] vertices = new CustomVertex.PositionColored[2];
            int i = 0;

            vertices[i++] = new CustomVertex.PositionColored((float)position.X, (float)position.Y, (float)position.Z, diffuse.ToArgb());
            vertices[i++] = new CustomVertex.PositionColored((float)position.X + direction.X, (float)position.Y + direction.Y, (float)position.Z + direction.Z, ambient.ToArgb());

            vertexBuffer.SetData(vertices, 0, LockFlags.None);
        }
    }
}
