using System;
using System.Globalization;
using System.Text;
using System.Xml;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.Diagnostics;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// Defines set of D3D Material properties.
  /// </summary>
  [DataObject("Defines set of d3dutMaterial properties.")]
  public class Material : DataObject
  {
    /// <summary>
    /// Ambient color of current Material.
    /// </summary>
    private ColorRGBA ambient;
    /// <summary>
    /// Diffuse color of current Material.
    /// </summary>
    private ColorRGBA diffuse;
    /// <summary>
    /// Specular color of current Material.
    /// </summary>
    private ColorRGBA specular; 
    /// <summary>
    /// Emissive color of current Material.
    /// </summary>
    private ColorRGBA emissive; 
    /// <summary>
    /// Specular sharpness of current Material.
    /// </summary>
    private float specularSharpness;
    /// <summary>
    /// Indicates if this Material has specular reflection or not.
    /// </summary>
    private bool specularEnable; 

    /// <summary>
    /// Gets or sets value for ambient color of current Material.
    /// </summary>
    public ColorRGBA Ambient
    {
      get
      {
        return this.ambient;
      }
      set
      {
        this.ambient = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Gets or sets value for diffuse color of current Material.
    /// </summary>
    public ColorRGBA Diffuse
    {
      get
      {
        return this.diffuse;
      }
      set
      {
        this.diffuse = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Gets or sets value for specular color of current Material.
    /// </summary>
    public ColorRGBA Specular
    {
      get
      {
        return this.specular;
      }
      set
      {
        this.specular = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Gets or sets value for emissive color of current Material.
    /// </summary>
    public ColorRGBA Emissive
    {
      get
      {
        return this.emissive;
      }
      set
      {
        this.emissive = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Gets or sets value for specular sharpness of current Material. Value must be equal or greater then zero.
    /// </summary>
    public float SpecularSharpness
    {
      get
      {
        return this.specularSharpness;
      }
      set
      {
        if (value >= 0.0f)
          this.specularSharpness = value;
        else
          throw new Exception("Specular sharpness of Material can't be less then zero.");

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Gets or sets value which indicate if this Material causes specular reflection.
    /// </summary>
    public bool SpecularEnable
    {
      get
      {
        return this.specularEnable;
      }
      set
      {
        this.specularEnable = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Creates new Material with default values. Ambient Color end Emissive Color are black, Diffuse Color and Specular 
    /// Color are white and Specular Sharpness is 1.0.
    /// </summary>
    public Material()
      : this(new ColorRGBA(0, 0, 0, 255), new ColorRGBA(255, 255, 255, 255),
      new ColorRGBA(255, 255, 255, 255), new ColorRGBA(0, 0, 0, 255), 20.0f, true)
    {
      UpdateModificationTime();
    }

    /// <summary>
    /// Creates new Material with these values.
    /// </summary>
    /// <param name="ambient">Ambient color</param>
    /// <param name="diffuse">Diffuse color</param>
    /// <param name="specular">Specular color</param>
    /// <param name="emissive">Emissive color</param>
    /// <param name="specularSharpness">Specular sharpness</param>
    /// <param name="specularEnable">Cause specular reflection or not</param>
    public Material(ColorRGBA ambient, ColorRGBA diffuse, ColorRGBA specular, ColorRGBA emissive, float specularSharpness, bool specularEnable)
    {
      this.ambient = ambient;
      this.diffuse = diffuse;
      this.specular = specular;
      this.emissive = emissive;
      this.specularSharpness = specularSharpness;
      this.specularEnable = specularEnable;

      UpdateModificationTime();
    }

    /// <summary>
    /// Creates new Material with these values and Emissive Color sets to black.
    /// </summary>
    /// <param name="ambient">Ambient color</param>
    /// <param name="diffuse">Diffuse color</param>
    /// <param name="specular">Specular color</param>
    /// <param name="specularSharpness">Specular sharpness</param>
    public Material(ColorRGBA ambient, ColorRGBA diffuse, ColorRGBA specular, float specularSharpness)
      : this(ambient, diffuse, specular, new ColorRGBA(0, 0, 0, 255), specularSharpness, true)
    {
      UpdateModificationTime();
    }

    /// <summary>
    /// Check data consistence of current object.
    /// </summary>
    /// <returns>Returns true if succeed.</returns>
    public override bool CheckConsistence()
    {
      bool returnValue = true;

      returnValue &= this.ambient.CheckConsistence();
      returnValue &= this.diffuse.CheckConsistence();
      returnValue &= this.specular.CheckConsistence();
      returnValue &= this.emissive.CheckConsistence();

      return returnValue;
    }

    /// <summary>
    /// Creates deep copy of current object.
    /// </summary>
    /// <returns>Returns new instance of this class with deep copied data from current object.</returns>
    public override IDataObject DeepCopy()
    {
      Material m = new Material();

      m.Ambient = (ColorRGBA) this.ambient.DeepCopy();
      m.Diffuse = (ColorRGBA) this.diffuse.DeepCopy();
      m.Specular = (ColorRGBA) this.specular.DeepCopy();
      m.Emissive = (ColorRGBA) this.emissive.DeepCopy();
      m.SpecularSharpness = this.specularSharpness;
      m.specularEnable = this.specularEnable;

      return m;
    }

    /// <summary>
    /// Gets names of all assemblies used by this class.
    /// </summary>
    public override string[] LibNames
    {
			get
			{
        AddLibName(this.GetType());
        AddLibName(typeof(ColorRGBA));
        
        return base.LibNames;
      }
    }

    /// <summary>
    /// Reads data from XML.
    /// </summary>
    /// <param name="xmlTextReader">Representation of XML stream.</param>
    public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
    {
      Debug.Write("Material reading...");

      while (xmlTextReader.Read())
      {
        if (xmlTextReader.NodeType == XmlNodeType.Element)
        {
          switch (xmlTextReader.Name)
          {
            case "ambient":
              this.ambient = (ColorRGBA) Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
              this.ambient.ReadData(xmlTextReader);
              break;
            case "diffuse":
              this.diffuse = (ColorRGBA) Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
              this.diffuse.ReadData(xmlTextReader);
              break;
            case "specular":
              this.specular = (ColorRGBA) Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
              this.specular.ReadData(xmlTextReader);
              break;
            case "emissive":
              this.emissive = (ColorRGBA) Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
              this.emissive.ReadData(xmlTextReader);
              break;
            case "specularsharpness":
              this.specularSharpness = FromString.GetFloat(xmlTextReader.ReadString(), 1f);
              break;
            case "specularenable":
              this.specularEnable = FromString.GetBool(xmlTextReader.ReadString(), true);
              break;
            default:
              break;
          } // switch
        } // if

        if (xmlTextReader.NodeType == XmlNodeType.EndElement)
        {
          if (xmlTextReader.Name == "d3dutmaterial")
            break;
        }
      } // while

      UpdateModificationTime();
      Debug.WriteLine(" OK");
    }

    /// <summary>
    /// Writes current object's data to XML.
    /// </summary>
    /// <param name="xmlTextWriter">Representation of XML stream.</param>
    public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
    {
      xmlTextWriter.WriteStartElement("d3dutmaterial");

      xmlTextWriter.WriteStartElement("ambient");
      xmlTextWriter.WriteAttributeString("type", this.ambient.GetType().FullName);
      this.ambient.WriteData(xmlTextWriter);
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("diffuse");
      xmlTextWriter.WriteAttributeString("type", this.diffuse.GetType().FullName);
      this.diffuse.WriteData(xmlTextWriter);
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("specular");
      xmlTextWriter.WriteAttributeString("type", this.specular.GetType().FullName);
      this.specular.WriteData(xmlTextWriter);
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("emissive");
      xmlTextWriter.WriteAttributeString("type", this.emissive.GetType().FullName);
      this.emissive.WriteData(xmlTextWriter);
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("specularsharpness");
      xmlTextWriter.WriteAttributeString("type", this.specularSharpness.GetType().FullName);
      xmlTextWriter.WriteString(this.specularSharpness.ToString(CultureInfo.InvariantCulture.NumberFormat));
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("specularenable");
      xmlTextWriter.WriteAttributeString("type", this.specularEnable.GetType().FullName);
      xmlTextWriter.WriteString(this.specularEnable.ToString(CultureInfo.InvariantCulture.NumberFormat));
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteEndElement();
    }

    /// <summary>
    /// Converts current object to its string description.
    /// </summary>
    /// <returns>Resulting string description.</returns>
    public override string ToString()
    {
      StringBuilder str = new StringBuilder(this.GetType().FullName + ":\r\n");

      str.Append("Ambient: ");
      str.Append(this.ambient.ToString() + "\r\n");

      str.Append("Diffuse: ");
      str.Append(this.diffuse.ToString() + "\r\n");

      str.Append("Specular: ");
      str.Append(this.specular.ToString() + "\r\n");

      str.Append("Emissive: ");
      str.Append(this.emissive.ToString() + "\r\n");

      str.Append("SpecularSharpness : ");
      str.Append(this.specularSharpness.ToString() + "\r\n");

      str.Append("SpecularEnable : ");
      str.Append(this.specularEnable.ToString() + "\r\n");

      str.Append("\r\n");

      return str.ToString();
    }
  }
}
