using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using Zcu.Mve.Core;
using System.ComponentModel;
using System.Drawing.Design;
using Zcu.Mve.Visualization;
using System.Globalization;
using System.Diagnostics;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// Source of Material. Material can be applied on scene object.
  /// </summary>
  [ModuleInfo("Jan Kaiser", "Source of Material. Material can be applied on scene object.", IconName = "MaterialModule.ico")]
  [PortInfo("outMaterial", "Generated Material.")]
  public class MaterialModule : Module
  {
    private bool dataChanged = true;
    private Material material;

    #region Properties
    /// <summary>
    /// Ambient color.
    /// </summary>
    [Category("Material color")]
    [Description("Ambient color.")]
    public ColorRGBA Ambient
    {
      get
      {
        return material.Ambient;
      }
      set
      {
        material.Ambient = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Diffuse color.
    /// </summary>
    [Category("Material color")]
    [Description("Diffuse color.")]
    public ColorRGBA Diffuse
    {
      get
      {
        return material.Diffuse;
      }
      set
      {
        material.Diffuse = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Emissive color.
    /// </summary>
    [Category("Material color")]
    [Description("Emissive color.")]
    public ColorRGBA Emissive
    {
      get
      {
        return material.Emissive;
      }
      set
      {
        material.Emissive = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Specular color.
    /// </summary>
    [Category("Material color")]
    [Description("Specular color.")]
    public ColorRGBA Specular
    {
      get
      {
        return material.Specular;
      }
      set
      {
        material.Specular = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Value that specifies the sharpness of specular highlights.
    /// </summary>
    [Description("Value that specifies the sharpness of specular highlights."), DefaultValue(1.0f)]
    public float SpecularSharpness
    {
      get
      {
        return material.SpecularSharpness;
      }
      set
      {
        material.SpecularSharpness = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Enable specular highlights.
    /// </summary>
    [Description("Enable specular highlights."), DefaultValue(true)]
    public bool SpecularEnable
    {
      get
      {
        return material.SpecularEnable;
      }
      set
      {
        material.SpecularEnable = value;
        dataChanged = true;
      }
    }
    #endregion

    /// <summary>
    /// Constructor, initializes object and adds output port.
    /// </summary>
    public MaterialModule()
    {
      AddOutPort("outMaterial", typeof(Material));

      material = new Material();
    }

    /// <summary>
    /// Generates/updates d3dutMaterial and exposes it on output port.
    /// </summary>
    public override void Execute()
    {
      if (dataChanged == true)
      {
        SetOutput("outMaterial", material, false);
        dataChanged = false;
      }
      else
      {
        SetOutput("outMaterial", material, true);
      }
    }

    /// <summary>
    /// Writes Xml configuration of current module.
    /// </summary>
    /// <param name="config">XmlElement from map config file which belongs to current module.</param>
    public override void WriteConfig(XmlElement config)
    {
      XmlElement xmlProperties = config.OwnerDocument.CreateElement("Properties");
     
      XmlElement xmlAmbient = config.OwnerDocument.CreateElement("Ambient");
      xmlAmbient.SetAttribute("type", material.Ambient.GetType().FullName);
      xmlAmbient.InnerText = material.Ambient.Red.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                           + material.Ambient.Green.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                           + material.Ambient.Blue.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                           + material.Ambient.Alpha.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlAmbient);

      XmlElement xmlDiffuse = config.OwnerDocument.CreateElement("Diffuse");
      xmlDiffuse.SetAttribute("type", material.Diffuse.GetType().FullName);
      xmlDiffuse.InnerText = material.Diffuse.Red.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                           + material.Diffuse.Green.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                           + material.Diffuse.Blue.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                           + material.Diffuse.Alpha.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlDiffuse);

      XmlElement xmlEmissive = config.OwnerDocument.CreateElement("Emissive");
      xmlEmissive.SetAttribute("type", material.Emissive.GetType().FullName);
      xmlEmissive.InnerText = material.Emissive.Red.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                            + material.Emissive.Green.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                            + material.Emissive.Blue.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                            + material.Emissive.Alpha.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlEmissive);

      XmlElement xmlSpecular = config.OwnerDocument.CreateElement("Specular");
      xmlSpecular.SetAttribute("type", material.Specular.GetType().FullName);
      xmlSpecular.InnerText = material.Specular.Red.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                            + material.Specular.Green.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                            + material.Specular.Blue.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                            + material.Specular.Alpha.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlSpecular);

      XmlElement xmlSpecularSharpness = config.OwnerDocument.CreateElement("SpecularSharpness");
      xmlSpecularSharpness.SetAttribute("type", material.SpecularSharpness.GetType().FullName);
      xmlSpecularSharpness.InnerText = material.SpecularSharpness.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlSpecularSharpness);

      XmlElement xmlSpecularEnable = config.OwnerDocument.CreateElement("SpecularEnable");
      xmlSpecularEnable.SetAttribute("type", material.SpecularEnable.GetType().FullName);
      xmlSpecularEnable.InnerText = material.SpecularEnable.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlSpecularEnable);
      
      config.AppendChild(xmlProperties);
    }

    /// <summary>
    /// Reads Xml configuration of current module.
    /// </summary>
    /// <param name="config">XmlElement from map config file which belongs to current module.</param>
    public override void ReadConfig(XmlElement config)
    {
      foreach (XmlElement xeConfig in config.ChildNodes)
      {
        if (xeConfig.Name == "Properties")
        {
          foreach (XmlElement xeProperties in xeConfig.ChildNodes)
          {
            switch (xeProperties.Name)
            {
              case "Ambient":
                material.Ambient = FromString.GetColorRGBA(xeProperties.InnerText);
                break;
              case "Diffuse":
                material.Diffuse = FromString.GetColorRGBA(xeProperties.InnerText);
                break;
              case "Emissive":
                material.Emissive = FromString.GetColorRGBA(xeProperties.InnerText);
                break;
              case "Specular":
                material.Specular = FromString.GetColorRGBA(xeProperties.InnerText);
                break;
              case "SpecularSharpness":
                material.SpecularSharpness = FromString.GetFloat(xeProperties.InnerText, 1f);
                break;              
              case "SpecularEnable":
                material.SpecularEnable = FromString.GetBool(xeProperties.InnerText, true);
                break;
              default:
                Debug.WriteLine("Unreaded value:" + xeProperties.Name + " - " + xeProperties.InnerText);
                break;
            }
          }
        }
      }
    }

    /// <summary>
    /// Creates new instance of module and copy all necessary member variables.
    /// This method is used by MapEditor in Copy/Paste process.
    /// Without overriding it returns implicit instance of current module.
    /// </summary>
    public override Module DeepCopy()
    {
      MaterialModule materialSource = (MaterialModule) base.DeepCopy();

      materialSource.material = (Material) material.DeepCopy();

      return materialSource;
    }
  }
}
