using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.Windows.Forms;
using Microsoft.DirectX.DirectInput;

using System.ComponentModel;
using System.Globalization;
using System.Xml;
using System.Diagnostics;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// Mouse interactor captures mouse device and provides information 
  /// about x, y position change and pressed mouse buttons.
  /// </summary>
  [ModuleInfo("Jan Kaiser", "Mouse interactor captures mouse device and provides information "
             +"about x, y position change and pressed mouse buttons.", IconName="MouseInteractor.ico")]
  [PortInfo("outX", "Change of position in horizontal direction.")]
  [PortInfo("outY", "Change of position in vertical direction.")]
  [PortInfo("outButtons", "Pressed buttons. Value is coded binary (left button = 1, right button = 2, left+right button = 3, etc.).")]
  [PortInfo("outVector", "Vector of mouse position change and pressed buttons. [OutX, OutY, OutButtons] is the format of outVector.")]
  public class MouseInteractor : Module
  {
    private Device mouseDevice;
    private Vector2D initialXY;

    private Scalar x = new Scalar(0);
    private Scalar y = new Scalar(0);
    private Integer buttons = new Integer(0);
    private Vector3D vector = new Vector3D();
    private float divisor;

    #region Properties
    private bool showOutX;
    /// <summary>
    /// Show outX port.
    /// </summary>
    [Category("Show output port")]
    [Description("Show outX port."), DefaultValue(true)]
    public bool outX
    {
      get
      {
        return showOutX;
      }
      set
      {
        if (value)
        {
          if (!showOutX)
            AddOutPort("outX", typeof(Scalar));
        }
        else
        {
          if (showOutX)
            RemoveOutPort("outX");
        }

        showOutX = value;
      }
    }

    private bool showOutY;
    /// <summary>
    /// Show outY port.
    /// </summary>
    [Category("Show output port")]
    [Description("Show outY port."), DefaultValue(true)]
    public bool outY
    {
      get
      {
        return showOutY;
      }
      set
      {
        if (value)
        {
          if (!showOutY)
            AddOutPort("outY", typeof(Scalar));
        }
        else
        {
          if (showOutY) 
            RemoveOutPort("outY");
        }

        showOutY = value;
      }
    }

    private bool showOutButtons;
    /// <summary>
    /// Show outButtons port.
    /// </summary>
    [Category("Show output port")]
    [Description("Show outButtons port."), DefaultValue(true)]
    public bool outButtons
    {
      get
      {
        return showOutButtons;
      }
      set
      {
        if (value)
        {
          if (!showOutButtons)
            AddOutPort("outButtons", typeof(Integer));
        }
        else
        {
          if (showOutButtons) 
            RemoveOutPort("outButtons");
        }

        showOutButtons = value;
      }
    }

    private bool showOutVector;
    /// <summary>
    /// Show outVector port. [OutX, OutY, OutButtons] is the format of outVector.
    /// </summary>
    [Category("Show output port")]
    [Description("Show outVector port. [OutX, OutY, OutButtons] is the format of outVector."), DefaultValue(true)]
    public bool outVector
    {
      get
      {
        return showOutVector;
      }
      set
      {
        if (value)
        {
          if (!showOutVector)
            AddOutPort("outVector", typeof(Vector3D));
        }
        else
        {
          if (showOutVector)
            RemoveOutPort("outVector");
        }

        showOutVector = value;
      }
    }

    /// <summary>
    /// Divide output value by this value. It allows to control sensitivity.
    /// </summary>
    [Description("Divide output value by this value. It allows to control sensitivity."), DefaultValue(100)]
    public float divideOutputBy
    {
      get
      {
        return divisor;
      }
      set
      {
        divisor = value;
      }
    }
    #endregion

    /// <summary>
    /// Constructor, initializes object.
    /// </summary>
    public MouseInteractor()
    {
      outX = false;
      outY = false;
      outButtons = false;
      outVector = false;

      x = new Scalar(0);
      y = new Scalar(0);
      divideOutputBy = 100;
      buttons = new Integer(0);

      this.SimulationEnding += new SimulationEndingEventHandler(MouseInteractor_SimulationEnding);
      this.SimulationStarted += new SimulationStartedEventHandler(MouseInteractor_SimulationStarted);
    }

    /// <summary>
    /// Acquire mouse device.
    /// </summary>
    private void MouseInteractor_SimulationStarted(object source)
    {
      mouseDevice = new Device(SystemGuid.Mouse);
      mouseDevice.Properties.AxisModeAbsolute = true;
      mouseDevice.SetCooperativeLevel(new Form(), CooperativeLevelFlags.Background | CooperativeLevelFlags.NonExclusive);
      mouseDevice.Acquire();
      
      initialXY.X = mouseDevice.CurrentMouseState.X;
      initialXY.Y = mouseDevice.CurrentMouseState.Y;
    }

    /// <summary>
    /// Unacquire mouse device.
    /// </summary>
    private void MouseInteractor_SimulationEnding(object source)
    {
      mouseDevice.Unacquire();
      mouseDevice.Dispose();
    }

    /// <summary>
    /// Exposes information about x, y position change and pressed mouse buttons.
    /// </summary>
    public override void Execute()
    {
      MouseState ms = mouseDevice.CurrentMouseState;

      int pressedButtons = 0;

      for (int i = 0; i < 8; i++)
      {
        if (ms.GetMouseButtons()[i] != 0)
          pressedButtons += (1 << i); 
      }
      
      x = (ms.X - initialXY.X) / divideOutputBy;
      y = (ms.Y - initialXY.Y) / divideOutputBy;

      initialXY.X = ms.X;
      initialXY.Y = ms.Y;

      if (showOutX)
      {
        if (x == 0)
        {
          //Debug.WriteLine("Mouse NO x: " + x.Value);
          this.SetOutput("outX", x, true);
        }
        else
        {
          //Debug.WriteLine("Mouse YO x: " + x.Value);
          this.SetOutput("outX", x, false);
        }
      }

      if (showOutY)
      {
        if (y == 0)
        {
          //Debug.WriteLine("Mouse NO y: " + y.Value);
          this.SetOutput("outY", y, true);
        }
        else
        {
          //Debug.WriteLine("Mouse YO y: " + y.Value);
          this.SetOutput("outY", y, false);
        }
      }

      if (showOutButtons)
      {
        if (pressedButtons == buttons)
        {
          //Debug.WriteLine("Mouse NO B: " + buttons.Value);
          this.SetOutput("outButtons", buttons, true);
        }
        else
        {
          buttons = pressedButtons;
          //Debug.WriteLine("Mouse YO B: " + buttons.Value);
          this.SetOutput("outButtons", buttons, false);
        }
      }

      if (showOutVector)
      {
        Vector3D tmpVector = new Vector3D(x, y, buttons);

        if (tmpVector == vector)
        {
          //Debug.WriteLine("Mouse NO V: " + vector.ToString());
          this.SetOutput("outVector", vector, true);
        }
        else
        {
          vector = tmpVector;
          //Debug.WriteLine("Mouse YO V: " + vector.ToString());
          this.SetOutput("outVector", vector, false);
        }
      }
    }

    /// <summary>
    /// Creates new instance of module and copy all necessary member variables.
    /// This method is used by MapEditor in Copy/Paste process.
    /// Without overriding it returns implicit instance of current module.
    /// </summary>
    public override Module DeepCopy()
    {
      MouseInteractor mouseInteractor = (MouseInteractor) base.DeepCopy();

      mouseInteractor.divideOutputBy = this.divideOutputBy;

      return mouseInteractor;
    }
  }
}