using System;
using System.Text;
using System.Threading;
using System.Diagnostics;
using System.Windows.Forms;
using System.Runtime.Serialization;
using System.Runtime.Remoting.Messaging;

using Zcu.Mve.Core;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// Class responsible for UnstrctureGrid class visualization. Class adds interactive view.
    /// </summary>
    public class Renderer : AdvancedRenderer
    {
        #region Fields

        protected SceneWrapper scene;
        protected CameraWrapper camera;
        protected float cameraDistance;

        protected FontWrapper fo;
        protected string infoStr;
        protected FPSCounter fpsCounter;

        private int mouseX;
        private int mouseY;
        private int startX;
        private int startY;
        private bool mouseCapture;

        const string INFO_BASE = "\n[Esc] - close\n[F1] - show/hide this help\n[F] - fullscreen/windowed mode\n[R] - render to texture\n[S] - take snapshot\n";
        const string INFO_ADVANCE = "[mouse wheel] - zoom in/out\n[leftMouseButton] & mouseMove - rotate the scene\n";

        #endregion

        // TODO:
        float angle;
        int index;

        /// <summary>
        /// Default constructor. Initializes renderer.
        /// </summary>
        /// <param name="scene">Scene.</param>
        /// <param name="camera">Camera.</param>
        /// <param name="rendererParameters">Renderer parameters.</param>
        public Renderer(Scene scene, Visualization.UniformDataArray uda, Camera camera, RendererParameters rendererParameters)
            : base(camera, uda, rendererParameters)
        {
            mouseCapture = false;
            cameraDistance = 1.0f;

            infoStr = INFO_BASE + ((_rParams.UseInternalVirtualTrackball == true) ? INFO_ADVANCE : string.Empty);

            this.scene = new SceneWrapper(scene);
            this.camera = new CameraWrapper(camera, this.scene, rendererParameters);
            fpsCounter = new FPSCounter();

            base.WindowCreated += new WindowCreatedEventHandler(Renderer_WindowCreated);
            base.RenderingLoopStart += new RenderingStartEventHandler(Renderer_RenderingLoopStart);
            base.RenderingLoopEnd += new RenderingEndEventHandler(Renderer_RenderingLoopEnd);

            angle = 0.0f;
            index = 0;
        }

        private void Renderer_WindowCreated()
        {
            if (_rParams.UseInternalVirtualTrackball)
            {
                _window.MouseMove   += new MouseEventHandler(renderer_MouseMove);
                _window.MouseDown   += new MouseEventHandler(renderer_MouseDown);
                _window.MouseWheel  += new MouseEventHandler(_window_MouseWheel);
                _window.MouseUp     += new MouseEventHandler(renderer_MouseUp);
            }

            _window.KeyDown     += new KeyEventHandler(renderer_KeyDown);
        }

        /// <summary>
        /// Device independent classes
        /// </summary>
        private void Renderer_RenderingLoopStart()
        {
            fo = new FontWrapper(_runtime.Device);

            fo.Enable = false;            
        }

        private void Renderer_RenderingLoopEnd()
        {
            if (fo != null) fo.Dispose();

            fo = null;
        }

        /// <summary>
        /// Update objects in scene by calling Update() foreach object.
        /// </summary>
        /// <param name="device"></param>
        protected override void UpdateScene(Device device)
        {
            fpsCounter.Update();

            scene.Update(device);

            camera.Update(_runtime.Size);            
        }

        /// <summary>
        /// Scene rendering.
        /// </summary>
        /// <param name="device"></param>
        protected override void DrawScene(Device device)
        {
            camera.SetCamera(device);            

            device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, _rParams.BackgroundColor, 1.0f, 0);

            device.BeginScene();

            fo.Draw("FPS: " + fpsCounter.GetFPS().ToString());
            fo.Draw(infoStr);

            scene.Draw(device);

            device.EndScene();
        }

        #region Key events

        protected void renderer_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.F1:
                    // draw font
                    if (fo != null) fo.Enable = !fo.Enable;
                    break;

                case Keys.Escape:
                    StopRendering();

                    break;

                case Keys.F:
                    ChangeWindowMode();

                    break;

                default:
                    break;
            }
        }

        #endregion

        #region Mouse events

        protected void _window_MouseWheel(object sender, MouseEventArgs e)
        {
            cameraDistance += (e.Delta / 120.0f) * 0.02f * cameraDistance;
                
                //(float)(((scene.MaxPoint.Z - scene.MinPoint.Z) * 0.01f * (e.Delta / 120.0f)));

            if (cameraDistance < 0.0f) cameraDistance = 0;

            camera.UpdatePosition(mouseX, mouseY, cameraDistance);
        }

        protected void renderer_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                mouseCapture = false;
            }
        }

        protected void renderer_MouseMove(object sender, MouseEventArgs e)
        {
            if (mouseCapture)
            {
                int dx = e.X - startX;
                int dy = e.Y - startY;

                while (mouseY > 360) mouseY -= 360;
                while (mouseY < 0) mouseY += 360;

                if (mouseY > 90 && mouseY < 270) mouseX += dx;
                else mouseX -= dx;

                mouseY += dy;

                startX = e.X;
                startY = e.Y;

                camera.UpdatePosition(mouseX, mouseY, cameraDistance);
            }
        }

        protected void renderer_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                mouseCapture = true;

                startX = e.X;
                startY = e.Y;
            }
        }

        #endregion
    }
}