using System;
using System.Drawing;
using System.Threading;
using System.Diagnostics;
using System.Windows.Forms;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Zcu.Mve.UnstrGridRendering
{
    public delegate void WindowModeChangedEventHandler(WindowMode wMode);

    /// <summary>
    /// Class responsible for Device management. Runtime class constructor create Device with optimal parameteres
    /// for current graphics card. Class has three main states: rendering, device_lost, device_reset. Switching 
    /// between states is done on the flags basis, which are indirectly set by calling suitable methods. When 
    /// device lost occure all allocated resources must be RELEASED, event DeviceLost is present by class Device.
    /// </summary>
    public class Runtime : IDisposable
    {
        private Device _device;
        private bool _deviceLost;           
        private bool _deviceNotReset;
        private WindowMode _wndMode;
        private bool _wndModeChanged;

        private Size size;
        private PresentParameters[] presentParams;

        public event WindowModeChangedEventHandler WindowModeChange;

        #region Properties

        public Size Size
        {
            get { return size; }
        }

        public Device Device
        {
            get { return _device; }
        }

        #endregion

        public Runtime(IntPtr wndHandle, ScreenTarget sTarget, int width, int height)
        {
            size.Width = width;
            size.Height = height;

            _deviceLost = false;
            _deviceNotReset = false;
            _wndModeChanged = false;
            _wndMode = WindowMode.Windowed;

            presentParams = new PresentParameters[1];

            IntPtr[] wndArray = new IntPtr[1];
            wndArray[0] = wndHandle;

            // device initialization
            if (!InitializeGraphics(wndArray, (int) sTarget))
            {
                Debug.WriteLine("Device initialization failed - Direct3D Device not created");
            }
        }

        public Runtime(IntPtr[] wndHandles, int width, int height)
        {
            size.Width = width;
            size.Height = height;

            _deviceLost = false;
            _deviceNotReset = false;
            _wndModeChanged = false;
            _wndMode = WindowMode.Windowed;

            presentParams = new PresentParameters[wndHandles.Length];

            // device initialization
            if (!InitializeGraphics(wndHandles, 0))
            {
                Debug.WriteLine("Device initialization failed - Direct3D Device not created");
            }
        }

        public SwapChain CreateAditionalSwapChain(int index)
        {
            SwapChain sch = new SwapChain(_device, presentParams[index]);

            return sch;
        }

        /// <summary>
        /// Create device.
        /// </summary>
        /// <param name="wndHandles">Window contexts</param>
        /// <param name="adapter">Required monitor</param>
        /// <returns>TRUE if device successfully created, else false</returns>
        private bool InitializeGraphics(IntPtr[] wndHandles, int adapter)
        {
            // selected adapter test, default = primary monitor
            AdapterInformation ai = Manager.Adapters.Default;

            if (Manager.Adapters.Count >= 2) { ai = Manager.Adapters[adapter]; }

            DisplayMode currentDm = ai.CurrentDisplayMode;

            // base hardware tests
            DeviceType dType = DeviceType.Reference;

            if (Manager.CheckDeviceType(ai.Adapter, DeviceType.Hardware, currentDm.Format, currentDm.Format, false))
            {
                dType = DeviceType.Hardware;
            }

            DepthFormat dFormat = DepthFormat.D16;
            DepthFormat[] dFormats = { DepthFormat.D32, DepthFormat.D24S8, DepthFormat.D24X4S4, DepthFormat.D24X8 };

            for (int i = 0; i < dFormats.Length; i++)
            {
                if (Manager.CheckDepthStencilMatch(ai.Adapter, dType, currentDm.Format, currentDm.Format, dFormats[i]))
                {
                    dFormat = dFormats[i];
                    break;
                }
            }

            // creation of present parameters
            for (int i = 0; i < wndHandles.Length; i++)
            {
                InitPresentParameters(wndHandles[i], dFormat, out presentParams[i]);
            }

            // create flags selection
            Caps caps = Manager.GetDeviceCaps(ai.Adapter, DeviceType.Hardware);

            CreateFlags cf = CreateFlags.SoftwareVertexProcessing;

            if ((caps.DeviceCaps.SupportsHardwareTransformAndLight) && (caps.VertexShaderVersion >= new Version(2, 0)))
            {
                cf = CreateFlags.HardwareVertexProcessing;
            }

            // create the Direct3D Device
            _device = new Device(ai.Adapter, dType, wndHandles[0], cf, presentParams[0]);

            Device.IsUsingEventHandlers = false;
            // basic events handling
            _device.DeviceLost += new EventHandler(OnDeviceLost);
            _device.DeviceReset += new EventHandler(OnDeviceReset);

            return true;
        }

        private void InitPresentParameters(IntPtr rendererWindowHandle, DepthFormat dFormat, out PresentParameters pParams)
        {
            DisplayMode dm = Manager.Adapters.Default.CurrentDisplayMode;

            pParams = new PresentParameters();

            // fill present params
            pParams.BackBufferCount = 1;
            pParams.BackBufferWidth = size.Width;
            pParams.BackBufferHeight = size.Height;

            pParams.FullScreenRefreshRateInHz = 0;
            pParams.BackBufferFormat = dm.Format;
            pParams.MultiSample = MultiSampleType.None;
            pParams.PresentationInterval = PresentInterval.Default;

            pParams.Windowed = true;
            pParams.SwapEffect = SwapEffect.Discard;

            pParams.EnableAutoDepthStencil = true;
            pParams.AutoDepthStencilFormat = dFormat;

            pParams.DeviceWindowHandle = rendererWindowHandle;
        }

        /// <summary>
        /// Set device lost flag.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void OnDeviceReset(object sender, EventArgs e)
        {
            _deviceNotReset = false;
        }

        /// <summary>
        /// Set device reset flag.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void OnDeviceLost(object sender, EventArgs e)
        {
            _deviceLost = true;
        }

        /// <summary>
        /// Device reset processing, call on the basis of deviceReset flag
        /// </summary>
        private void DeviceResetHandling()
        {
            if (_deviceNotReset)
            {
                try
                {
                    _device.Reset(presentParams[0]);

                    if (_wndModeChanged && WindowModeChange != null) WindowModeChange(_wndMode);

                    _deviceLost = false;
                    _deviceNotReset = false;
                }

                catch (Microsoft.DirectX.Direct3D.DeviceLostException) 
                {
                    _deviceLost = true;

                    return;
                }
            }
        }

        /// <summary>
        /// Device lost processing, call on the basis of deviceLost flag
        /// </summary>
        private void DeviceLostHandling()
        {
            int result;

            _device.CheckCooperativeLevel(out result);

            // device still lost 
            if (result == (int)ResultCode.DeviceLost)
            {
                _deviceLost = true;

                Thread.Sleep(100);

                return;
            }

            if (result == (int)ResultCode.DeviceNotReset)
            {
                _deviceLost = false;

                _deviceNotReset = true;
            }
        }

        /// <summary>
        /// Update present parameters, called when user resize form.
        /// </summary>
        /// <param name="nSize">Viewport size</param>
        public void DeviceResize(System.Drawing.Size nSize)
        {
            Debug.WriteLine("Resizing window to: " + nSize);

            if (size.Width == nSize.Width && size.Height == nSize.Height)
                return;

            size = nSize;

            for (int i = 0; i < presentParams.Length; i++)
            {
                presentParams[i].BackBufferWidth = size.Width;
                presentParams[i].BackBufferHeight = size.Height;
            }

            // set device reset flag
            _deviceNotReset = true;
        }

        /// <summary>
        ///  Check device falgs.
        /// </summary>
        /// <returns>TRUE if any flag is set, else FALSE</returns>
        public bool CheckDeviceState()
        {
            if (_deviceLost) DeviceLostHandling();

            if (_deviceNotReset) DeviceResetHandling();

            return !(_deviceLost || _deviceNotReset);
        }

        /// <summary>
        /// Switch to fullscreen mode.
        /// </summary>
        public void SwitchToFullscreen(int width, int height)
        {
            AdapterInformation ai = Manager.Adapters.Default;
            DisplayMode dm = ai.CurrentDisplayMode;

            size.Width = width;
            size.Height = height;

            for (int i = 0; i < presentParams.Length; i++)
            {
                presentParams[i].BackBufferWidth = size.Width;
                presentParams[i].BackBufferHeight = size.Height;
                presentParams[i].BackBufferFormat = dm.Format;
                presentParams[i].FullScreenRefreshRateInHz = dm.RefreshRate;

                presentParams[i].Windowed = false;
            }

            _deviceNotReset = true;
            _wndModeChanged = true;
            _wndMode = WindowMode.FullScreeen;
        }

        /// <summary>
        /// Switch to windowed mode.
        /// </summary>
        public void SwitchToWindowed(int width, int height)
        {
            AdapterInformation ai = Manager.Adapters.Default;
            DisplayMode dm = ai.CurrentDisplayMode;

            size.Width = width;
            size.Height = height;

            for (int i = 0; i < presentParams.Length; i++)

            {
                presentParams[i].BackBufferWidth = size.Width;
                presentParams[i].BackBufferHeight = size.Height;

                presentParams[i].Windowed = true;
                presentParams[i].FullScreenRefreshRateInHz = 0;
            }

            _deviceNotReset = true;
            _wndModeChanged = true;
            _wndMode = WindowMode.Windowed;
        }

        #region IDisposable Members

        /// <summary>
        /// Device release.
        /// </summary>
        public void Dispose()
        {
            if (_device != null) _device.Dispose();

            _device = null;

            presentParams = null;
        }

        #endregion
    }
}
