using System;
using Zcu.Mve.Visualization;
using Zcu.Mve.Core;
using System.Xml;
using System.Text;
using System.Globalization;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// Representation of the Scene. Scene is a composition of SceneObjects and Lights.
  /// </summary>
  [DataObject("Representation of the Scene. Scene is a composition of SceneObjects and Lights.")]
  public class Scene : DataObject
  {
    #region Fields

    private UniformDataArray sceneObjects;

    private UniformDataArray lights;

    private int frame;

    private bool showSceneOrigin;

    private float sceneOriginSize;

    #endregion

    #region Properties
    /// <summary>
    /// Array of SceneObjects.
    /// </summary>
    public UniformDataArray SceneObjects
    {
      get { return sceneObjects; }
      set 
      { 
        sceneObjects = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// animation frame to play
    /// </summary>
    public int Frame
    {
        get { return frame; }
        set { frame = value; }
    }

    /// <summary>
    /// Array of Lights
    /// </summary>
    public UniformDataArray Lights
    {
      get
      {
        return lights;
      }
      set
      {
        lights = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Show scene origin. 
    /// </summary>
    public bool ShowSceneOrigin
    {
      get
      {
        return showSceneOrigin;
      }
      set
      {
        showSceneOrigin = value;
      }
    }

    /// <summary>
    /// Scene origin gridSize.
    /// </summary>
    public float SceneOriginSize
    {
      get
      {
        return sceneOriginSize;
      }
      set
      {
        sceneOriginSize = value;
      }
    }
    #endregion

    /// <summary>
    /// Constructor, initializes Scene object.
    /// </summary>
    public Scene()
    {
      SceneObjects = null;

      Lights = null;

      ShowSceneOrigin = false;

      sceneOriginSize = 1f;

      UpdateModificationTime();
    }


    #region IDataObject Members

    /// <summary>
    /// Converts current object to its string description.
    /// </summary>
    /// <returns>Resulting string description.</returns>
    public override string ToString()
    {
      StringBuilder str = new StringBuilder(this.GetType().FullName + ":\r\n");

      str.Append("sceneObjects: \r\n" + SceneObjects.ToString() + "\r\n");

      foreach (SceneObject so in SceneObjects)
      {
        str.Append(so.ToString() + "\r\n");
      }

      if (Lights != null)
      {
        str.Append("Lights: \r\n" + Lights.ToString() + "\r\n");

        foreach (Light l in Lights)
        {
          str.Append(l.ToString() + "\r\n");
        }
      }

      str.Append("ShowSceneOrigin: " + ShowSceneOrigin + "\r\n");
      str.Append("SceneOriginSize: " + SceneOriginSize + "\r\n");
      
      str.Append("\r\n");

      return str.ToString();
    }

    /// <summary>
    /// Writes data of this object.
    /// </summary>
    /// <param name="xmlTextWriter">Writer.</param>
    public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
    {
      xmlTextWriter.WriteStartElement("scene");

      xmlTextWriter.WriteStartElement("sceneobjects");
      xmlTextWriter.WriteAttributeString("type", SceneObjects.GetType().FullName);
      SceneObjects.WriteData(xmlTextWriter);
      xmlTextWriter.WriteEndElement();

      if (Lights != null)
      {
        xmlTextWriter.WriteStartElement("lights");
        xmlTextWriter.WriteAttributeString("type", Lights.GetType().FullName);
        Lights.WriteData(xmlTextWriter);
        xmlTextWriter.WriteEndElement();
      }

      xmlTextWriter.WriteStartElement("showsceneorigin");
      xmlTextWriter.WriteAttributeString("type", ShowSceneOrigin.GetType().FullName);
      xmlTextWriter.WriteString(ShowSceneOrigin.ToString(CultureInfo.InvariantCulture.NumberFormat));
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("sceneoriginsize");
      xmlTextWriter.WriteAttributeString("type", SceneOriginSize.GetType().FullName);
      xmlTextWriter.WriteString(SceneOriginSize.ToString(CultureInfo.InvariantCulture.NumberFormat));
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteEndElement();
    }

    /// <summary>
    /// Reads data of this object.
    /// </summary>
    /// <param name="xmlTextReader">Reader.</param>
    public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
    {
      while (xmlTextReader.Read())
      {
        if (xmlTextReader.NodeType == XmlNodeType.Element)
        {
          switch (xmlTextReader.Name)
          {
            case "sceneobjects":
              SceneObjects = (UniformDataArray) Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
              SceneObjects.ReadData(xmlTextReader);
              break;
            case "lights":
              Lights = (UniformDataArray) Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
              Lights.ReadData(xmlTextReader);
              break;
            case "showsceneorigin":
              showSceneOrigin = FromString.GetBool(xmlTextReader.ReadString(), false);
              break;
            case "sceneoriginsize":
              sceneOriginSize = FromString.GetFloat(xmlTextReader.ReadString(), 1f);
              break;
            default:
              break;
          }
        }

        if (xmlTextReader.NodeType == XmlNodeType.EndElement)
        {
          if (xmlTextReader.Name == "scene")
            return;
        }
      }

      UpdateModificationTime();
    }

    /// <summary>
    /// Performs deep copy of current object.
    /// </summary>
    /// <returns>Deep copy of Camera object.</returns>
    public override IDataObject DeepCopy()
    {
      Scene s = new Scene();

      s.SceneObjects = (UniformDataArray) SceneObjects.DeepCopy();
      s.Lights = (UniformDataArray) Lights.DeepCopy();
      s.ShowSceneOrigin = ShowSceneOrigin;
      s.SceneOriginSize = SceneOriginSize;

      return s;
    }

    /// <summary>
    /// Check if data of this object are consistent.
    /// </summary>
    /// <returns>Returns true if data are consistent.</returns>
    public override bool CheckConsistence()
    {
      bool returnValue = true;

      returnValue &= SceneObjects.CheckConsistence();

      if (Lights != null) returnValue &= Lights.CheckConsistence();

      return returnValue;
    }

    /// <summary>
    /// Gets names of all assemblies used by this class.
    /// </summary>
    public override string[] LibNames
    {
      get
      {
        foreach (string s in SceneObjects.LibNames) AddLibName(s);
        if (Lights != null) foreach (string s in Lights.LibNames) AddLibName(s);

        return libNames.ToArray();
      }
    }
    #endregion
  }
}
