using System;
using System.Collections.Generic;
using System.Text;

using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.Diagnostics;

using System.ComponentModel;
using System.Xml;
using System.Globalization;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// Creates Scene by composition of SceneObjects and Lights.
  /// </summary>
  [ModuleInfo("Jan Kaiser", "Creates Scene by composition of SceneObjects and Lights.", IconName = "SceneComposer.ico")]
  [PortInfo("inSceneObjects", "Uniform data array with SceneObject items.")]
  [PortInfo("inLights", "Uniform data array with Light sources.")]
  [PortInfo("outScene", "Scene created by composition of SceneObjects and Lights.\n" 
  + "It is supposed to be rendered by Scene renderer.")]
  public class SceneComposer : Module
  {
    #region Fields
    private bool lightsConnected;
    private bool objectsConnected;
    private bool animatedConnected;

    private Scene scene = null;
    #endregion

    #region Properties
    /// <summary>
    /// Show scene origin.
    /// </summary>
    [Category("Scene")]
    [DefaultValue(false)]
    [Description("Show scene origin.")]
    public bool ShowSceneOrigin
    {
      get
      {
        return scene.ShowSceneOrigin;
      }
      set
      {
        scene.ShowSceneOrigin = value;
      }
    }

    /// <summary>
    /// GridSize of Scene origin displayed in viewprt.
    /// </summary>
    [Category("Scene")]
    [DefaultValue(1f)]
    [Description("GridSize of Scene origin displayed in viewprt.")]
    public float SceneOriginSize
    {
      get
      {
        return scene.SceneOriginSize;
      }
      set
      {
        scene.SceneOriginSize = value;
      }
    }
    #endregion

    /// <summary>
    /// Default constructor. Adds input and output ports and performs initialization.
    /// </summary>
    public SceneComposer()
    {
      AddInPort("inSceneObjects",    typeof(UniformDataArray),false);
      AddInPort("inAnimatedObjects", typeof(UniformDataArray),false);
      AddInPort("inLights", typeof(UniformDataArray), false);

      this.PortConnected += new PortConnectingEventHandler(SceneComposer_PortConnecting);
      this.PortDisconnected += new PortDisconnectingEventHandler(SceneComposer_PortDisconnecting);

      AddOutPort("outScene", typeof(Scene));
      scene = new Scene();
      lightsConnected = false;
      objectsConnected = false;
      animatedConnected = false;

      ShowSceneOrigin = false;
      SceneOriginSize = 1f;
    }

    private void SceneComposer_PortDisconnecting(object source, InPort portTo, out bool allowed)
    {
      allowed = true;

      if (portTo.Name == "inLights")
      {
        //Debug.WriteLine("SceneComposer: Lights disconnected");
        lightsConnected = false;
        scene.Lights = null;
      }

      if (portTo.Name == "inSceneObjects")
      {
          //Debug.WriteLine("SceneComposer: Lights disconnected");
          objectsConnected = false;
      }

      if (portTo.Name == "inAnimatedObjects")
      {
          //Debug.WriteLine("SceneComposer: Lights disconnected");
          animatedConnected = false;
      }

      if (!animatedConnected && !objectsConnected) scene.SceneObjects = null;

    }

    private void SceneComposer_PortConnecting(object source, OutPort portFrom, InPort portTo, out bool allowed)
    {
      allowed = true;

      if (portTo.Name == "inLights")
      {
        //Debug.WriteLine("SceneComposer: Lights connected");
        lightsConnected = true;
        scene.Lights = null;
      }

      if (portTo.Name == "inSceneObjects")
      {
          //Debug.WriteLine("SceneComposer: Lights disconnected");
          objectsConnected = true;
      }

      if (portTo.Name == "inAnimatedObjects")
      {
          //Debug.WriteLine("SceneComposer: Lights disconnected");
          animatedConnected = true;
      }
    }    

    /// <summary>
    /// Exposes composed/updated Scene to output port.
    /// </summary>
    public override void Execute()
    {
      UniformDataArray udaTemp;

      if ((objectsConnected) && (!IsDataSame("inSceneObjects") || (scene.SceneObjects == null)))
      {
        udaTemp = (UniformDataArray) GetInput("inSceneObjects");

        if (udaTemp.ElemType.FullName != "Zcu.Mve.UnstrGridRendering.SceneObject")
        {
          throw new MveException(this, "\nUniform data array connected to inSceneObjects port has wrong element type.\n"
                                     + "Element type is: " + udaTemp.ElemType.ToString() + ", required type: " + typeof(SceneObject));
        }
        else
        {
          scene.SceneObjects = udaTemp;
        }
      }

      if ( (animatedConnected) && (!IsDataSame("inAnimatedObjects") || (scene.SceneObjects == null)))
      {
          udaTemp = (UniformDataArray)GetInput("inAnimatedObjects");

          if (udaTemp.ElemType.FullName == "Zcu.Mve.UnstrGridRendering.SceneObject")
          {
              if (scene.SceneObjects == null) scene.SceneObjects = udaTemp;
              else
              {
                  UniformDataArray newuda = new UniformDataArray(typeof(SceneObject), scene.SceneObjects.Count + udaTemp.Count);

                  for (int i = 0; i < scene.SceneObjects.Count; i++) newuda[i] = scene.SceneObjects[i];
                  for (int i = 0; i < udaTemp.Count; i++) newuda[i + scene.SceneObjects.Count] = udaTemp[i];

                  scene.SceneObjects = newuda;
              }
          }
          else if (udaTemp.ElemType.FullName == "Zcu.Mve.Visualization.UniformDataArray")
          {
                int count = 0;
                UniformDataArray anim;
              
              for (int i = 0; i < udaTemp.Count; i++)
              {
                  anim = (UniformDataArray)udaTemp[i];
                  if (anim.ElemType.FullName == "Zcu.Mve.UnstrGridRendering.SceneObject") count += anim.Count;
              }
              if (scene.SceneObjects != null) count += scene.SceneObjects.Count;

              UniformDataArray newArray = new UniformDataArray(typeof(SceneObject), count);
              int j=0;

              if (scene.SceneObjects != null)
                  for (int i = 0; i < scene.SceneObjects.Count; i++)
                  {
                      newArray[j] = scene.SceneObjects[i];
                      j++;
                  }

              for (int i = 0; i < udaTemp.Count; i++)
              {
                  anim = (UniformDataArray)udaTemp[i];
                  if (anim.ElemType.FullName == "Zcu.Mve.UnstrGridRendering.SceneObject")
                  {
                      for (int k = 0; k < anim.Count; k++)
                      {
                          newArray[j] = anim[k];
                          j++;
                      }
                  }
              }

              scene.SceneObjects = newArray;
          }
      }

      if ((lightsConnected) && (!IsDataSame("inLights") || (scene.Lights == null)))
      {
        udaTemp = (UniformDataArray) GetInput("inLights");

        if (udaTemp.ElemType.FullName != "Zcu.Mve.UnstrGridRendering.Light")
        {
          throw new MveException(this, "\nUniform data array connected to inLights port has wrong element type.\n"
                                     + "Element type is: " + udaTemp.ElemType.ToString() + ", required type: " + typeof(Light));
        }
        else
        {
          scene.Lights = udaTemp;
        }
      }

      SetOutput("outScene", scene, this.IsAllDataSame());
    }

    /// <summary>
    /// Writes module settings.
    /// </summary>
    /// <param name="config">XmlElement from map config file which belongs to current module.</param>
    public override void WriteConfig(System.Xml.XmlElement config)
    {
      XmlElement xmlProperties = config.OwnerDocument.CreateElement("Properties");

      XmlElement xmlShowSceneOrigin = config.OwnerDocument.CreateElement("ShowSceneOrigin");
      xmlShowSceneOrigin.SetAttribute("type", ShowSceneOrigin.GetType().FullName);
      xmlShowSceneOrigin.InnerText = ShowSceneOrigin.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlShowSceneOrigin);

      XmlElement xmlSceneOriginSize = config.OwnerDocument.CreateElement("SceneOriginSize");
      xmlSceneOriginSize.SetAttribute("type", SceneOriginSize.GetType().FullName);
      xmlSceneOriginSize.InnerText = SceneOriginSize.ToString(CultureInfo.InvariantCulture.NumberFormat);
      xmlProperties.AppendChild(xmlSceneOriginSize);

      config.AppendChild(xmlProperties);
    }

    /// <summary>
    /// Reads module settings.
    /// </summary>
    /// <param name="config">XmlElement from map config file which belongs to current module.</param>
    public override void ReadConfig(System.Xml.XmlElement config)
    {
      foreach (XmlElement xeConfig in config.ChildNodes)
      {
        if (xeConfig.Name == "Properties")
        {
          foreach (XmlElement xeProperties in xeConfig.ChildNodes)
          {
            switch (xeProperties.Name)
            {
              case "ShowSceneOrigin":
                ShowSceneOrigin = FromString.GetBool(xeProperties.InnerText, false);
                break;
              case "SceneOriginSize":
                SceneOriginSize = FromString.GetFloat(xeProperties.InnerText, 1f);
                break;
              default:
                Debug.WriteLine("Unreaded value:" + xeProperties.Name + " - " + xeProperties.InnerText);
                break;
            } // switch
          } // foreach
        } // if
      } // foreach    
    }

    /// <summary>
    /// Creates new instance of module and copy all necessary member variables.
    /// This method is used by MapEditor in Copy/Paste process.
    /// Without overriding it returns implicit instance of current module.
    /// </summary>
    public override Module DeepCopy()
    {
      SceneComposer sceneComposer = (SceneComposer) base.DeepCopy();

      sceneComposer.ShowSceneOrigin = this.ShowSceneOrigin;
      sceneComposer.SceneOriginSize = this.SceneOriginSize;

      return sceneComposer;
    }
  }
}
