using System;
using Zcu.Mve.Visualization;
using Zcu.Mve.Core;
using System.Xml;
using System.Collections;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using System.Text;
using System.Diagnostics;
using System.Globalization;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// This enumeration is primarilly used in SceneObject class. 
  /// It defines possible combinations of vertex attributes. 
  /// User can choose what combination will be used while rendering.
  /// </summary>
  public enum VertexAttributes
  {
    /// <summary>
    /// None of attributes will be used.
    /// </summary>
    None = 0,
    /// <summary>
    /// Texture attributes will be used.
    /// </summary>
    TextureCoord = 1,
    /// <summary>
    /// Color attributes will be used.
    /// </summary>
    ColorRGBA = 2,
    /// <summary>
    /// Color and texture attributes will be used together.
    /// </summary>
    ColorRGBATextureCoord = 3,
    /// <summary>
    /// Normal attribute will be used.
    /// </summary>
    Normal3D = 4,
    /// <summary>
    /// Normal and color attribute will be used.
    /// </summary>
    Normal3DColorRGBA = 5,
    /// <summary>
    /// Normal and texture attribute will be used.
    /// </summary>
    Normal3DTextureCoord = 6
  };

  /// <summary>
  /// This enumeration is primarilly used in SceneObject class.
  /// User can choose to render cells or only points.  
  /// </summary>
  public enum RenderElements
  {
    /// <summary>
    /// Only points of DataSet will be rendered.
    /// </summary>
    Points = 0,
    /// <summary>
    /// Whole dataset will be rendered.
    /// </summary>
    Cells = 1
  };

  /// <summary>
  /// Representation of the object in the Scene. It is a composition of UnstructuredGrid and Material.
  /// </summary>
  [DataObject("Representation of the object in the Scene. It is a composition of UnstructuredGrid and Material.")]
  public class SceneObject : DataObject
  {
    #region Fields
    private String name;
    private VertexAttributes prefferedVertexAttributes;
    private bool showSceneObject;
    private bool lighting;
    private FillMode fillMode;
    private Cull cull;
    private ShadeMode shadeMode;
    private bool normalizeNormals;
    private bool colorVertex;

      private UnstrGrid unstructuredGrid;
      public int frame;                    //frame number; -1 if not animated; value>=0 set by animation composer

    private Material material;
    private TextureObject texture;
    private bool materialEnable;
    private string colorAttributeName;
    private string normalAttributeName;
    private string textureAttributeName;
    private bool showBoundingBox;
    private float pointSize;
    private RenderElements render;
    #endregion

    #region Properties
    /// <summary>
    /// Name of the object.
    /// </summary>
    public String Name
    {
      get
      {
        return name;
      }
      set
      {
        name = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Preffered vertex attributes that will be used for current object rendering.
    /// </summary>
    public VertexAttributes PrefferedVertexAttributes
    {
      get
      {
        return prefferedVertexAttributes;
      }
      set
      {
        prefferedVertexAttributes = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Show/hide scene object.
    /// </summary>
    public bool ShowSceneObject
    {
      get
      {
        return showSceneObject;
      }
      set
      {
        showSceneObject = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Enable/disable lighting
    /// </summary>
    public bool Lighting
    {
      get
      {
        return lighting;
      }
      set
      {
        lighting = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Contains a member of the FillMode enumeration that represents the fill mode. Default value is Solid.
    /// </summary>
    public FillMode FillMode
    {
      get
      {
        return fillMode;
      }
      set
      {
        fillMode = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Specifies how back-facing triangles are culled, if at all.
    /// </summary>
    public Cull Cull
    {
      get
      {
        return cull;
      }
      set
      {
        cull = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Specifies what shading method wil be used.
    /// </summary>
    public ShadeMode ShadeMode
    {
      get
      {
        return shadeMode;
      }
      set
      {
        shadeMode = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Enables or disables automatic normalization of vertex normals.
    /// </summary>
    public bool NormalizeNormals
    {
      get
      {
        return normalizeNormals;
      }
      set
      {
        normalizeNormals = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Retrieves or sets a value to enable or disable per-vertex color.
    /// </summary>
    public bool ColorVertex
    {
      get
      {
        return colorVertex;
      }
      set
      {
        colorVertex = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Geometric and topological information about scene object.
    /// </summary>
    public UnstrGrid UnstructuredGrid
    {
      get
      {
        return unstructuredGrid;
      }
      set
      {
        unstructuredGrid = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Material.
    /// </summary>
    public Material Material
    {
      get
      {
        return material;
      }
      set
      {
        material = value;

        UpdateModificationTime();
      }
    }

      /// <summary>
      /// texture
      /// </summary>
      internal TextureObject TextureObj
      {
          get { return texture; }
          set { texture = value; }
      }  

    /// <summary>
    /// Enables or disables use of d3dutMaterial during rendering.
    /// </summary>
    public bool MaterialEnable
    {
      get
      {
        return materialEnable;
      }
      set
      {
        materialEnable = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Name of attribute that will be used as color attribute.
    /// </summary>
    public string ColorAttributeName
    {
      get
      {
        return colorAttributeName;
      }
      set
      {
        colorAttributeName = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Name of attribute that will be used as normal attribute.
    /// </summary>
    public string NormalAttributeName
    {
      get
      {
        return normalAttributeName;
      }
      set
      {
        normalAttributeName = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Name of attribute that will be used as texture attribute.
    /// </summary>
    public string TextureAttributeName
    {
      get
      {
        return textureAttributeName;
      }
      set
      {
        textureAttributeName = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Enable/disable bounding box rendering.
    /// </summary>
    public bool ShowBoundingBox
    {
      get
      {
        return showBoundingBox;
      }
      set
      {
        showBoundingBox = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Specifies the gridSize to use for point gridSize computation (if FillMode == Point)
    /// in cases in which point gridSize is not specified for each vertex. 
    /// </summary>
    public float PointSize
    {
      get
      {
        return pointSize;
      }
      set
      {
        pointSize = value;

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Specifies if cells of dataset will be rendered or if only points will be rendered.
    /// </summary>
    public RenderElements Render
    {
      get
      {
        return render;
      }
      set
      {
        render = value;

        UpdateModificationTime();
      }
    }
    #endregion

    /// <summary>
    /// Constructor, initializes current object.
    /// </summary>
    public SceneObject()
    {
      PrefferedVertexAttributes = VertexAttributes.None;
      ShowSceneObject = true;
      ShowBoundingBox = false;
      Lighting = false;
      Render = RenderElements.Cells;
      FillMode = FillMode.Solid;
      PointSize = 1.0f;
      Cull = Cull.None;
      ShadeMode = ShadeMode.Gouraud;
      NormalizeNormals = true;
      ColorVertex = true;
      Name = "Scene object";
      ColorAttributeName = "ColorRGBA";
      NormalAttributeName = "Normal3D";
      TextureAttributeName = "TextureCoord";
      Material = new Material();
      MaterialEnable = false;

      unstructuredGrid = null;
      frame = -1;

      UpdateModificationTime();
    }
    #region DataObject Members

    /// <summary>
    /// Converts current object to its string description.
    /// </summary>
    /// <returns>Resulting string description.</returns>
    public override string ToString()
    {
      StringBuilder str = new StringBuilder(this.GetType().FullName + " [" + Name + "]: \r\n");

      str.Append("PrefferedVertexAttributes: ");
      str.Append(PrefferedVertexAttributes.ToString() + "\r\n");

      str.Append("ShowSceneObject: ");
      str.Append(ShowSceneObject.ToString() + "\r\n");

      str.Append("ShowBoundingBox: ");
      str.Append(ShowBoundingBox.ToString() + "\r\n");

      str.Append("Lighting: ");
      str.Append(Lighting.ToString() + "\r\n");

      str.Append("Render: ");
      str.Append(Render.ToString() + "\r\n");

      str.Append("FillMode: ");
      str.Append(FillMode.ToString() + "\r\n");

      str.Append("PointSize: ");
      str.Append(PointSize.ToString() + "\r\n");

      str.Append("Cull: ");
      str.Append(Cull.ToString() + "\r\n");

      str.Append("ShadeMode: ");
      str.Append(ShadeMode.ToString() + "\r\n");

      str.Append("NormalizeNormals: ");
      str.Append(NormalizeNormals.ToString() + "\r\n");

      str.Append("ColorVertex: ");
      str.Append(ColorVertex.ToString() + "\r\n");

      str.Append("ColorAttributeName: ");
      str.Append(ColorAttributeName.ToString() + "\r\n");

      str.Append("NormalAttributeName: ");
      str.Append(NormalAttributeName.ToString() + "\r\n");

      str.Append("TextureAttributeName: ");
      str.Append(TextureAttributeName.ToString() + "\r\n");

      str.Append("MaterialEnable: ");
      str.Append(MaterialEnable.ToString() + "\r\n");

      str.Append("Material: \r\n");
      str.Append(Material.ToString());

      str.Append("UnstructuredGrid: \r\n");
      str.Append(UnstructuredGrid.ToString());

      str.Append("\r\n");

      return str.ToString();
    }

    /// <summary>
    /// Writes data of this object.
    /// </summary>
    /// <param name="xmlTextWriter">Writer.</param>
    public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
    {
      xmlTextWriter.WriteStartElement("sceneobject");

      xmlTextWriter.WriteStartElement("name");
      xmlTextWriter.WriteAttributeString("type", this.name.GetType().FullName);
      xmlTextWriter.WriteString(this.name);
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("prefferedvertexattributes");
      xmlTextWriter.WriteAttributeString("type", this.prefferedVertexAttributes.GetType().FullName);
      xmlTextWriter.WriteString(this.prefferedVertexAttributes.ToString());
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("showsceneobject");
      xmlTextWriter.WriteAttributeString("type", this.showSceneObject.GetType().FullName);
      xmlTextWriter.WriteString(this.showSceneObject.ToString(CultureInfo.InvariantCulture.NumberFormat));
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("showboundingbox");
      xmlTextWriter.WriteAttributeString("type", this.showBoundingBox.GetType().FullName);
      xmlTextWriter.WriteString(this.showBoundingBox.ToString(CultureInfo.InvariantCulture.NumberFormat));
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("lighting");
      xmlTextWriter.WriteAttributeString("type", this.lighting.GetType().FullName);
      xmlTextWriter.WriteString(this.lighting.ToString());
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("render");
      xmlTextWriter.WriteAttributeString("type", this.render.GetType().FullName);
      xmlTextWriter.WriteString(this.render.ToString());
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("fillmode");
      xmlTextWriter.WriteAttributeString("type", this.fillMode.GetType().FullName);
      xmlTextWriter.WriteString(this.fillMode.ToString());
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("pointsize");
      xmlTextWriter.WriteAttributeString("type", this.pointSize.GetType().FullName);
      xmlTextWriter.WriteString(this.pointSize.ToString(CultureInfo.InvariantCulture.NumberFormat));
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("cull");
      xmlTextWriter.WriteAttributeString("type", this.cull.GetType().FullName);
      xmlTextWriter.WriteString(this.cull.ToString());
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("shademode");
      xmlTextWriter.WriteAttributeString("type", this.shadeMode.GetType().FullName);
      xmlTextWriter.WriteString(ShadeMode.ToString());
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("normalizenormals");
      xmlTextWriter.WriteAttributeString("type", this.normalizeNormals.GetType().FullName);
      xmlTextWriter.WriteString(NormalizeNormals.ToString());
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("colorvertex");
      xmlTextWriter.WriteAttributeString("type", this.colorVertex.GetType().FullName);
      xmlTextWriter.WriteString(ColorVertex.ToString());
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("materialenable");
      xmlTextWriter.WriteAttributeString("type", this.materialEnable.GetType().FullName);
      xmlTextWriter.WriteString(MaterialEnable.ToString());
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("d3dutMaterial");
      xmlTextWriter.WriteAttributeString("type", this.material.GetType().FullName);
      Material.WriteData(xmlTextWriter);
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("unstructuredgrid");
      xmlTextWriter.WriteAttributeString("type", this.unstructuredGrid.GetType().FullName);
      UnstructuredGrid.WriteData(xmlTextWriter);
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("colorattributename");
      xmlTextWriter.WriteAttributeString("type", this.colorAttributeName.GetType().FullName);
      xmlTextWriter.WriteString(ColorAttributeName);
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("normalattributename");
      xmlTextWriter.WriteAttributeString("type", this.normalAttributeName.GetType().FullName);
      xmlTextWriter.WriteString(NormalAttributeName);
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteStartElement("textureattributename");
      xmlTextWriter.WriteAttributeString("type", this.textureAttributeName.GetType().FullName);
      xmlTextWriter.WriteString(TextureAttributeName);
      xmlTextWriter.WriteEndElement();

      xmlTextWriter.WriteEndElement();
    }

    /// <summary>
    /// Reads data of this object.
    /// </summary>
    /// <param name="xmlTextReader">Reader.</param>
    public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
    {
      Debug.WriteLine("Scene object reading... ");

      while (xmlTextReader.Read())
      {
        if (xmlTextReader.NodeType == XmlNodeType.Element)
        {
          Debug.WriteLine("M >>" + xmlTextReader.Name);
          switch (xmlTextReader.Name)
          {
            case "name":
              Name = xmlTextReader.ReadString();
              Debug.WriteLine("M : Name>>" + Name);
              break;
            case "prefferedvertexattributes":
              prefferedVertexAttributes = FromString.GetVertexAttribute(xmlTextReader.ReadString(), VertexAttributes.None);
              break;
            case "showsceneobject":
              showSceneObject = FromString.GetBool(xmlTextReader.ReadString(), true);
              break;
            case "showboundingbox":
              showBoundingBox = FromString.GetBool(xmlTextReader.ReadString(), false);
              break;
            case "lighting":
              lighting = FromString.GetBool(xmlTextReader.ReadString(), false);
              break;
            case "render":
              render = FromString.GetRenderElements(xmlTextReader.ReadString(), RenderElements.Cells);
              break;
            case "fillmode":
              fillMode = FromString.GetFillMode(xmlTextReader.ReadString(), FillMode.Solid);
              break;
            case "pointsize":
              pointSize = FromString.GetFloat(xmlTextReader.ReadString(), 1.0f);
              break;
            case "cull":
              cull = FromString.GetCull(xmlTextReader.ReadString(), Cull.None);
              break;
            case "shademode":
              shadeMode = FromString.GetShadeMode(xmlTextReader.ReadString(), ShadeMode.Gouraud);
              break;
            case "normalizenormals":
              normalizeNormals = FromString.GetBool(xmlTextReader.ReadString(), true);
              break;
            case "colorvertex":
              colorVertex = FromString.GetBool(xmlTextReader.ReadString(), true);
              break;
            case "colorattributename":
              ColorAttributeName = xmlTextReader.ReadString();
              break;
            case "normalattributename":
              NormalAttributeName = xmlTextReader.ReadString();
              break;
            case "textureattributename":
              TextureAttributeName = xmlTextReader.ReadString();
              break;
            case "materialenable":
              materialEnable = FromString.GetBool(xmlTextReader.ReadString(), false);
              break;
            case "d3dutMaterial":
              Material = (Material) Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
              Material.ReadData(xmlTextReader);
              break;
            case "unstructuredgrid":
              UnstructuredGrid = (UnstrGrid) Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
              UnstructuredGrid.ReadData(xmlTextReader);
              break;
            default:
              break;
          }
        }
        if (xmlTextReader.NodeType == XmlNodeType.EndElement)
        {
          //Debug.WriteLine("soEnd>>" + xmlTextReader.Name);
          if (xmlTextReader.Name == "sceneobject")
            break;
        }
      }

      UpdateModificationTime();

      Debug.WriteLine(" OK");
    }

    /// <summary>
    /// Performs deep copy of current object.
    /// </summary>
    /// <returns>Deep copy of SceneObject.</returns>
    public override IDataObject DeepCopy()
    {
      SceneObject so = new SceneObject();

      if (UnstructuredGrid != null)
        so.UnstructuredGrid = (UnstrGrid) UnstructuredGrid.DeepCopy();

      so.PrefferedVertexAttributes = prefferedVertexAttributes;
      so.ShowSceneObject = showSceneObject;
      so.showBoundingBox = showBoundingBox;
      so.Lighting = lighting;
      so.render = render;
      so.FillMode = fillMode;
      so.pointSize = pointSize;
      so.Cull = cull;
      so.ShadeMode = shadeMode;
      so.NormalizeNormals = normalizeNormals;
      so.ColorVertex = colorVertex;
      so.Name = name;
      so.frame = frame;

      so.colorAttributeName = colorAttributeName;
      so.normalAttributeName = normalAttributeName;
      so.textureAttributeName = textureAttributeName;

      if (material != null)
        so.Material = (Material) material.DeepCopy();
      
      so.MaterialEnable = materialEnable;

      return so;
    }

    /// <summary>
    /// Check if data of this object are consistent.
    /// </summary>
    /// <returns>Returns true if data are consistent.</returns>
    public override bool CheckConsistence()
    {
      bool returnValue = true;

      returnValue &= UnstructuredGrid.CheckConsistence();
      returnValue &= Material.CheckConsistence();

      return returnValue;
    }
    
    public void SetFrame (Mve.Numerics.Integer i)
    {
        frame = i;
    }


    /// <summary>
    /// Gets names of all assemblies used by this class.
    /// </summary>
    public override string[] LibNames
    {
      get
      {
        foreach (string s in UnstructuredGrid.LibNames)
          AddLibName(s);

        foreach (string s in Material.LibNames)
          AddLibName(s);

        return libNames.ToArray();
      }
    }
    #endregion
  }
}
