using System;
using System.Collections.Generic;
using System.Text;

using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.ComponentModel;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using System.Xml;
using System.Diagnostics;

namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// Creates Scene object by composition of unstructured grid and Material.
    /// </summary>
    [ModuleInfo("Jan Kaiser", "Creates Scene object by composition of unstructured grid and Material.",
                IconName = "SceneObjectComposer.ico")]
    [PortInfo("inUnstrGrid", "Geometry and structure of the object.")]
    [PortInfo("inMaterial", "Material that will be applied on scene object.")]
    [PortInfo("inTexture", "Texture that will be applied on scene object.")]
    [PortInfo("outSceneObject", "SceneObject that is supposed to be placed in Scene.")]
    public class SceneObjectComposer : Module
    {
        #region Fields
        private SceneObject sceneObject = null;
        private Material defaultMaterial;
        private TextureObject texture;
        private bool textureConnected;
        #endregion

        #region Properties
        /// <summary>
        /// Preffered vertex attributes that will be used for current object rendering.
        /// </summary>
        [Category("Rendering")]
        [Description("Preffered vertex attributes that will be used for current object rendering.")]
        [DefaultValue(VertexAttributes.None)]
        public VertexAttributes PrefferedVertexAttributes
        {
            get
            {
                return sceneObject.PrefferedVertexAttributes;
            }
            set
            {
                sceneObject.PrefferedVertexAttributes = value;
            }
        }

        /// <summary>
        /// Show/hide scene object in scene being rendered.
        /// </summary>
        [Category("Scene object")]
        [Description("Show/hide scene object in scene being rendered.")]
        [DefaultValue(true)]
        public bool ShowSceneObject
        {
            get
            {
                return sceneObject.ShowSceneObject;
            }
            set
            {
                sceneObject.ShowSceneObject = value;
            }
        }

        /// <summary>
        /// Enable/disable bounding box rendering.
        /// </summary>
        [Category("Scene object")]
        [Description("Enable/disable bounding box rendering.")]
        [DefaultValue(false)]
        public bool ShowBoundingBox
        {
            get
            {
                return sceneObject.ShowBoundingBox;
            }
            set
            {
                sceneObject.ShowBoundingBox = value;
            }
        }

        /// <summary>
        /// Enable/disable lighting.
        /// </summary>
        [Category("Rendering")]
        [Description("Enable/disable lighting.")]
        [DefaultValue(false)]
        public bool Lighting
        {
            get
            {
                return sceneObject.Lighting;
            }
            set
            {
                sceneObject.Lighting = value;
            }
        }

        /// <summary>
        /// Specifies if cells of dataset will be rendered or if only points will be rendered.
        /// If dataset contains no cells, only points will be rendered.
        /// </summary>
        [Category("Rendering")]
        [Description("Specifies if cells of dataset will be rendered or if only points will be rendered. If dataset contains no cells, only points will be rendered.")]
        [DefaultValue(RenderElements.Cells)]
        public RenderElements Render
        {
            get
            {
                return sceneObject.Render;
            }
            set
            {
                sceneObject.Render = value;
            }
        }

        /// <summary>
        /// Contains a member of the FillMode enumeration that represents the fill mode.
        /// Default value is Solid.
        /// </summary>
        [Category("Rendering")]
        [Description("Contains a member of the FillMode enumeration that represents the fill mode. Default value is Solid.")]
        [DefaultValue(FillMode.Solid)]
        public FillMode FillMode
        {
            get
            {
                return sceneObject.FillMode;
            }
            set
            {
                sceneObject.FillMode = value;
            }
        }

        /// <summary>
        /// Specifies the gridSize to use for point gridSize computation in cases in which point gridSize is not specified for each vertex.
        /// </summary>
        [Category("Rendering")]
        [Description("Specifies the gridSize to use for point gridSize computation in cases in which point gridSize is not specified for each vertex.")]
        [DefaultValue(1.0f)]
        public float PointSize
        {
            get
            {
                return sceneObject.PointSize;
            }
            set
            {
                sceneObject.PointSize = value;
            }
        }

        /// <summary>
        /// Specifies how back-facing triangles are culled, if at all.
        /// </summary>
        [Category("Rendering")]
        [Description("Specifies how back-facing triangles are culled, if at all.")]
        [DefaultValue(Cull.None)]
        public Cull Cull
        {
            get
            {
                return sceneObject.Cull;
            }
            set
            {
                sceneObject.Cull = value;
            }
        }

        /// <summary>
        /// Specifies what shading method wil be used.
        /// </summary>
        [Category("Rendering")]
        [Description("Specifies what shading method wil be used.")]
        [DefaultValue(ShadeMode.Gouraud)]
        public ShadeMode ShadeMode
        {
            get
            {
                return sceneObject.ShadeMode;
            }
            set
            {
                sceneObject.ShadeMode = value;
            }
        }

        /// <summary>
        /// Enables or disables automatic normalization of vertex normals.
        /// </summary>
        [Category("Rendering")]
        [Description("Enables or disables automatic normalization of vertex normals.")]
        [DefaultValue(true)]
        public bool NormalizeNormals
        {
            get
            {
                return sceneObject.NormalizeNormals;
            }
            set
            {
                sceneObject.NormalizeNormals = value;
            }
        }

        /// <summary>
        /// Retrieves or sets a value to enable or disable per-vertex color.
        /// </summary>
        [Category("Rendering")]
        [Description("Retrieves or sets a value to enable or disable per-vertex color.")]
        [DefaultValue(true)]
        public bool ColorVertex
        {
            get
            {
                return sceneObject.ColorVertex;
            }
            set
            {
                sceneObject.ColorVertex = value;
            }
        }

        /// <summary>
        /// Enables or disables use of d3dutMaterial during rendering.
        /// </summary>
        [Category("Rendering")]
        [Description("Enables or disables use of d3dutMaterial during rendering.")]
        [DefaultValue(false)]
        public bool MaterialEnable
        {
            get
            {
                return sceneObject.MaterialEnable;
            }
            set
            {
                sceneObject.MaterialEnable = value;
            }
        }

        /// <summary>
        /// Name of the object.
        /// </summary>
        [Category("Scene object")]
        [Description("Name of the object.")]
        public string SceneObjectName
        {
            get
            {
                return sceneObject.Name;
            }
            set
            {
                sceneObject.Name = value;
            }
        }

        /// <summary>
        /// Name of attribute that will be used as color attribute.
        /// </summary>
        [Category("Attribute names")]
        [Description("Name of attribute that will be used as color attribute.")]
        [DefaultValue("ColorRGBA")]
        public string ColorAttributeName
        {
            get
            {
                return sceneObject.ColorAttributeName;
            }
            set
            {
                sceneObject.ColorAttributeName = value;
            }
        }

        /// <summary>
        /// Name of attribute that will be used as normal attribute.
        /// </summary>
        [Category("Attribute names")]
        [Description("Name of attribute that will be used as normal attribute.")]
        [DefaultValue("Normal3D")]
        public string NormalAttributeName
        {
            get
            {
                return sceneObject.NormalAttributeName;
            }
            set
            {
                sceneObject.NormalAttributeName = value;
            }
        }

        /// <summary>
        /// Name of attribute that will be used as texture attribute.
        /// </summary>
        [Category("Attribute names")]
        [Description("Name of attribute that will be used as texture attribute.")]
        [DefaultValue("TextureCoord")]
        public string TextureAttributeName
        {
            get
            {
                return sceneObject.TextureAttributeName;
            }
            set
            {
                sceneObject.TextureAttributeName = value;
            }
        }
        #endregion

        /// <summary>
        /// Default constructor. Adds input and output ports and performs initialization.
        /// </summary>
        public SceneObjectComposer()
        {
            AddInPort("inTexture", typeof(TextureObject), false, true);
            AddInPort("inMaterial", typeof(Material), false, true);
            AddInPort("inUnstrGrid", typeof(UnstrGrid), true, true);

            AddOutPort("outSceneObject", typeof(SceneObject));

            textureConnected = false;
            texture = new TextureObject("");

            this.PortDisconnected += new PortDisconnectingEventHandler(SceneObjectComposer_PortDisconnecting);
            this.PortConnected += new PortConnectingEventHandler(SceneObjectComposer_PortConnecting);

            defaultMaterial = new Material();

            sceneObject = new SceneObject();
        }

        void SceneObjectComposer_PortConnecting(object source, OutPort portFrom, InPort portTo, out bool allowed)
        {
            allowed = true;

            if (portTo.Name == "inTexture")
            {

                //Debug.WriteLine("SceneComposer: Lights connected");
                textureConnected = true;
                texture.Texture = null;
            }
        }

        private void SceneObjectComposer_PortDisconnecting(object source, InPort portTo, out bool allowed)
        {
            allowed = true;

            if (portTo.Name == "inMaterial")
            {
                sceneObject.Material = new Material();
            }

            if (portTo.Name == "inTexture")
            {
                //Debug.WriteLine("SceneComposer: Lights disconnected");
                textureConnected = false;
                texture.Texture = null;
            }
        }

        /// <summary>
        /// Exposes composed/updated SceneObject to output port.
        /// </summary>
        public override void Execute()
        {
            if (!IsDataSame("inUnstrGrid"))
            {
                sceneObject.UnstructuredGrid = (UnstrGrid)GetInput("inUnstrGrid");

                // Debug.WriteLine("SceneObject YO: ");
                //  + sceneObject.UnstructuredGrid.Transform.ModificationTime
                //  + " : " + sceneObject.UnstructuredGrid.Points.ModificationTime);
            }

            Material m = (Material)GetInput("inMaterial");
            if (m == null)
            {
                if (sceneObject.Material != defaultMaterial)
                {
                    sceneObject.Material = defaultMaterial;
                }
            }
            else
            {
                if (sceneObject.Material != m)
                {
                    sceneObject.Material = m;
                }

                if (!IsDataSame("inMaterial"))
                {
                    sceneObject.Material = m;
                }
            }


            if (textureConnected && !IsDataSame("inTexture"))
            {
                texture = (TextureObject)GetInput("inTexture");
                sceneObject.TextureObj = texture;
            }

            SetOutput("outSceneObject", sceneObject, this.IsAllDataSame());
        }

        /// <summary>
        /// Reads module settings.
        /// </summary>
        /// <param name="config">XmlElement from map config file which belongs to current module.</param>
        public override void ReadConfig(System.Xml.XmlElement config)
        {
            foreach (XmlElement xeConfig in config.ChildNodes)
            {
                if (xeConfig.Name == "Properties")
                {
                    foreach (XmlElement xeProperties in xeConfig.ChildNodes)
                    {
                        switch (xeProperties.Name)
                        {
                            case "PrefferedVertexAttributes":
                                sceneObject.PrefferedVertexAttributes = FromString.GetVertexAttribute(xeProperties.InnerText.Trim(), VertexAttributes.None);
                                break;
                            case "ShowSceneObject":
                                sceneObject.ShowSceneObject = FromString.GetBool(xeProperties.InnerText, true);
                                break;
                            case "ShowBoundingBox":
                                sceneObject.ShowBoundingBox = FromString.GetBool(xeProperties.InnerText, false);
                                break;
                            case "Lighting":
                                sceneObject.Lighting = FromString.GetBool(xeProperties.InnerText, false);
                                break;
                            case "Render":
                                sceneObject.Render = FromString.GetRenderElements(xeProperties.InnerText, RenderElements.Cells);
                                break;
                            case "FillMode":
                                sceneObject.FillMode = FromString.GetFillMode(xeProperties.InnerText, FillMode.Solid);
                                break;
                            case "PointSize":
                                sceneObject.PointSize = FromString.GetFloat(xeProperties.InnerText, 1.0f);
                                break;
                            case "Cull":
                                sceneObject.Cull = FromString.GetCull(xeProperties.InnerText, Cull.None);
                                break;
                            case "ShadeMode":
                                sceneObject.ShadeMode = FromString.GetShadeMode(xeProperties.InnerText, ShadeMode.Gouraud);
                                break;
                            case "NormalizeNormals":
                                sceneObject.NormalizeNormals = FromString.GetBool(xeProperties.InnerText, true);
                                break;
                            case "ColorVertex":
                                sceneObject.ColorVertex = FromString.GetBool(xeProperties.InnerText, true);
                                break;
                            case "MaterialEnable":
                                sceneObject.MaterialEnable = FromString.GetBool(xeProperties.InnerText, false);
                                break;
                            case "SceneObjectName":
                                sceneObject.Name = xeProperties.InnerText.Trim();
                                break;
                            case "ColorAttributeName":
                                sceneObject.ColorAttributeName = xeProperties.InnerText.Trim();
                                break;
                            case "NormalAttributeName":
                                sceneObject.NormalAttributeName = xeProperties.InnerText.Trim();
                                break;
                            case "TextureAttributeName":
                                sceneObject.TextureAttributeName = xeProperties.InnerText.Trim();
                                break;
                            default:
                                Debug.WriteLine("Unreaded value:" + xeProperties.Name + " - " + xeProperties.InnerText);
                                break;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Creates new instance of module and copy all necessary member variables.
        /// This method is used by MapEditor in Copy/Paste process.
        /// Without overriding it returns implicit instance of current module.
        /// </summary>
        public override Module DeepCopy()
        {
            SceneObjectComposer sceneObjectComposer = (SceneObjectComposer)base.DeepCopy();

            sceneObjectComposer.sceneObject = (SceneObject)this.sceneObject.DeepCopy();

            return sceneObjectComposer;
        }
    }
}
