using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Zcu.Mve.Visualization;
using System.Diagnostics;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// D3DUT representation of SceneObject. It consists of transformation matrix, vertex and index buffers.
  /// </summary>
  public class SceneObjectWrapper : IDisposable
  {
    #region Properties
    private VertexBufferWrapper vertexBuffer;
    private IndexBufferWrapper[] indexBuffer;
    private UGridBoundingBox boundingBox;
    private Matrix transformMatrix;
    private SceneObject sceneObject;
    private MaterialWrapper material;
    private long lmtSceneObject;
    #endregion

    /// <summary>
    /// Constructor. Initializes D3DUTSceneObject.
    /// </summary>
    public SceneObjectWrapper()
    {
      this.lmtSceneObject = 0;
 
      this.vertexBuffer = new VertexBufferWrapper();
      this.material = new MaterialWrapper();
      this.boundingBox = new UGridBoundingBox();

      this.transformMatrix = new Matrix();
    }

    /// <summary>
    /// Updates buffers and d3dutMaterial.
    /// </summary>
    /// <param name="sceneObject">Scene object.</param>
    /// <param name="device">Rendering device.</param>
    public void Update(Device device, SceneObject sceneObject)
    {
      if (lmtSceneObject != sceneObject.ModificationTime)
      {
        this.sceneObject = sceneObject;

        updateBuffers(device);
        material.Update(sceneObject.Material, sceneObject.MaterialEnable);
        if (sceneObject.TextureObj!=null) sceneObject.TextureObj.Update(device);


        if (sceneObject.ShowBoundingBox) boundingBox.Update(device, sceneObject.UnstructuredGrid);
  
        lmtSceneObject = sceneObject.ModificationTime;
      }
    }

    /// <summary>
    /// Draws scene object. This method should be used between device.BeginScene() and device.EndScene()
    /// </summary>
    /// <param name="device">Rendering device.</param>
    public void Draw(Device device, int frame)
    {
        if (sceneObject.frame == -1 || sceneObject.frame == frame)
        {
            if (sceneObject.UnstructuredGrid.Transform is Transform3D)
            {
                transform3DToMatrix((Transform3D)sceneObject.UnstructuredGrid.Transform);
                device.Transform.World = transformMatrix;
            }

            if (sceneObject.ShowSceneObject)
            {
                setRenderState(device);

                material.SetMaterial(device);

                if (sceneObject.TextureObj != null) 
                    sceneObject.TextureObj.SetTexture(device, sceneObject);
                else 
                    device.SetTexture(0, null);

                vertexBuffer.SetVertexBuffer();

                if ((indexBuffer.Length == 0) || (sceneObject.Render == RenderElements.Points))
                {
                    device.DrawPrimitives(PrimitiveType.PointList, 0, vertexBuffer.Count);
                }
                else
                {
                    foreach (IndexBufferWrapper ib in indexBuffer)
                    {
                        ib.SetIndexBuffer();
                    }
                }
            }

            if (sceneObject.ShowBoundingBox) boundingBox.Draw();
        }
    }
 
    /// <summary>
    /// Performs conversion of Transform3D to D3DUT Matrix. 
    /// </summary>
    /// <param name="transform">Transform3D matrix.</param>
    private void transform3DToMatrix(Transform3D transform)
    {
      transformMatrix.M11 = transform[0, 0];
      transformMatrix.M12 = transform[1, 0];
      transformMatrix.M13 = transform[2, 0];
      transformMatrix.M14 = transform[3, 0];

      transformMatrix.M21 = transform[0, 1];
      transformMatrix.M22 = transform[1, 1];
      transformMatrix.M23 = transform[2, 1];
      transformMatrix.M24 = transform[3, 1];

      transformMatrix.M31 = transform[0, 2];
      transformMatrix.M32 = transform[1, 2];
      transformMatrix.M33 = transform[2, 2];
      transformMatrix.M34 = transform[3, 2];

      transformMatrix.M41 = transform[0, 3];
      transformMatrix.M42 = transform[1, 3];
      transformMatrix.M43 = transform[2, 3];
      transformMatrix.M44 = transform[3, 3];
    }

    /// <summary>
    /// Sets render state for current object.
    /// </summary>
    /// <param name="device">Rendering device.</param>
    private void setRenderState(Device device)
    {
        device.RenderState.Lighting = sceneObject.Lighting;
        device.RenderState.FillMode = sceneObject.FillMode;
        device.RenderState.CullMode = sceneObject.Cull;
        device.RenderState.ShadeMode = sceneObject.ShadeMode;
        device.RenderState.NormalizeNormals = sceneObject.NormalizeNormals;
        device.RenderState.ColorVertex = sceneObject.ColorVertex;
        device.RenderState.PointSize = sceneObject.PointSize;
    }

    /// <summary>
    /// Updates vertex and index buffers.
    /// </summary>
    /// <param name="device">Rendering device.</param>
    private void updateBuffers(Device device)
    {
      vertexBuffer.Update(device, sceneObject);

      Debug.WriteLine("Cells count> " + sceneObject.UnstructuredGrid.GetCellIds().Count);

      if ((indexBuffer == null)
      || ((int) indexBuffer.Length != sceneObject.UnstructuredGrid.GetCellIds().Count))
      {
        indexBuffer = new IndexBufferWrapper[sceneObject.UnstructuredGrid.GetCellIds().Count];
      }

      int i = 0;
      foreach (String s in sceneObject.UnstructuredGrid.GetCellIds())
      {
        Debug.WriteLine(">>" + s);

        if (indexBuffer[i] == null)
          indexBuffer[i] = new IndexBufferWrapper();

        indexBuffer[i].Update(device, sceneObject.UnstructuredGrid.GetCells(s), vertexBuffer);
        i++;
      }
    }

    #region IDisposable Members
    /// <summary>
    /// Dispose D3DUTSceneObject together with its vertex and index buffers.
    /// </summary>
    public void Dispose()
    {
      if (vertexBuffer != null)
        vertexBuffer.Dispose();

      foreach (IndexBufferWrapper ib in indexBuffer)
        if (ib != null)
          ib.Dispose();

      if (boundingBox != null)
        boundingBox.Dispose();
    }

    #endregion
  }
}

