using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Drawing;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// Helper class for D3DUTScene. It provides functionality of drawing scene origin.
  /// </summary>
  public class SceneOrigin : IDisposable
  {
    private VertexBuffer vertexBuffer;
    private Device device;
    private float sceneOriginSize;

    /// <summary>
    /// Constructor, initializes object.
    /// </summary>
    public SceneOrigin()
    {
      this.device = null;
      this.sceneOriginSize = 0;
    }

    /// <summary>
    /// Update scene origin when device or gridSize has changed.
    /// </summary>
    /// <param name="device">Rendering device.</param>
    /// <param name="sceneOriginSize">Scene origin gridSize.</param>
    public void Update(Device device, float sceneOriginSize)
    {
      if ((this.device != device)
      || (this.sceneOriginSize != sceneOriginSize))
      {
        this.device = device;
        this.sceneOriginSize = sceneOriginSize;

        createVertexBuffer();
      }
    }

    /// <summary>
    /// Draws scene origin. This method should be used between device.BeginScene() and device.EndScene()
    /// </summary>
    public void Draw()
    {
      setRenderState();

      device.Transform.World = Matrix.Identity;

      device.VertexFormat = CustomVertex.PositionColored.Format;
      device.SetStreamSource(0, vertexBuffer, 0);

      device.DrawPrimitives(PrimitiveType.LineList, 0, 3);
    }

    /// <summary>
    /// Set render state for scene origin.
    /// </summary>
    private void setRenderState()
    {
      device.RenderState.Lighting = false;
      device.RenderState.ColorVertex = true;
    }

    /// <summary>
    /// Create vertex buffer for scene origin
    /// </summary>
    private void createVertexBuffer()
    {
      vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionColored), 6, device, Usage.WriteOnly, CustomVertex.PositionColored.Format, Pool.Managed);

      CustomVertex.PositionColored[] vertices = new CustomVertex.PositionColored[6];
      int i = 0;

      vertices[i++] = new CustomVertex.PositionColored(0f, 0f, 0f, Color.Red.ToArgb());
      vertices[i++] = new CustomVertex.PositionColored(sceneOriginSize, 0f, 0f, Color.Red.ToArgb());
      
      vertices[i++] = new CustomVertex.PositionColored(0f, 0f, 0f, Color.Green.ToArgb());
      vertices[i++] = new CustomVertex.PositionColored(0f, sceneOriginSize, 0f, Color.Green.ToArgb());
      
      vertices[i++] = new CustomVertex.PositionColored(0f, 0f, 0f, Color.Blue.ToArgb());
      vertices[i++] = new CustomVertex.PositionColored(0f, 0f, sceneOriginSize, Color.Blue.ToArgb());

      vertexBuffer.SetData(vertices, 0, LockFlags.None);
    }

    #region IDisposable Members
    /// <summary>
    /// Dispose scene origin and its vertex buffer.
    /// </summary>
    public void Dispose()
    {
      if (vertexBuffer != null)
        vertexBuffer.Dispose();
    }

    #endregion
  }
}
