using System;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.ComponentModel;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Drawing;

using System.Xml;
using System.Diagnostics;
using System.Drawing.Design;
using System.Globalization;
using System.Threading;

namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// Renders scene using D3DUT interface.
    /// </summary>
    [ModuleInfo("*", "Renders Scene using D3DUT interface.", IconName = "SceneRenderer.ico")]
    [PortInfo("inScene", "Representation of Scene. Scene is a composition of SceneObjects and Lights.")]
    [PortInfo("inCamera", "Camera settings.")]
    public class SceneRendererModule : Module
    {
        #region Fields

        public const int InitialWindowWidth = 400;
        public const int InitialWindowHeight = 300;
        public const int SaverFramerate = 25;

        protected Scene scene;
        protected Camera camera;
        protected UniformDataArray uda;
        protected AdvancedRenderer renderer;
        protected RendererParameters rendererParameters;
        protected bool frameIn;
        protected bool showInCamera;
        protected bool showLegend;

        private bool pickerSetToOutput;

        #endregion
                
        #region Properties

        [Category("Output window"),
        Description("Where to save recorded animation")]
        public String Path
        {
            get { return rendererParameters.PathSaveto; }
            set { rendererParameters.PathSaveto = value; }
        }

        [Category("Monitor"),
        Description("Define which monitor used for fullscreen scene"),
        DefaultValue(Zcu.Mve.UnstrGridRendering.ScreenTarget.Primary)]
        public ScreenTarget Screen
        {
            get { return rendererParameters.Screen; }
            set { rendererParameters.Screen = value; }
        }

        /// <summary>
        /// Caption of renderer window.
        /// </summary>
        [Category("Output window"),
        Description("Caption of renderer window."),
        DefaultValue("Renderer window")]
        public string Text
        {
            get { return this.rendererParameters.Text; }
            set { this.rendererParameters.Text = value; }
        }

        /// <summary>
        /// Output window size. Width of renderer window.
        /// </summary>
        [Category("Output window size"),
        Description("Width of renderer window."), DefaultValue(SceneRendererModule.InitialWindowWidth)]
        public int Width
        {
            get { return this.rendererParameters.Width; }
            set { this.rendererParameters.Width = value; }
        }

        /// <summary>
        /// Output window size. Height of renderer window.
        /// </summary>
        [Category("Output window size"),
        Description("Height of renderer window."), DefaultValue(SceneRendererModule.InitialWindowHeight)]
        public int Height
        {
            get { return this.rendererParameters.Height; }
            set { this.rendererParameters.Height = value; }
        }

        /// <summary>
        /// Color used as a background color.
        /// </summary>
        [Category("Output window")]
        [Description("Color used as a background color.")]
        public ColorRGBA BackgroundColor
        {
            get
            {
                return new ColorRGBA(rendererParameters.BackgroundColor.R,
                                     rendererParameters.BackgroundColor.G,
                                     rendererParameters.BackgroundColor.B,
                                     rendererParameters.BackgroundColor.A);
            }
            set
            {
                rendererParameters.BackgroundColor = Color.FromArgb(value.Alpha, value.Red, value.Green, value.Blue);
            }
        }

        /// <summary>
        /// "Use internal camera settings (position, fov, projection, etc.)
        /// or get these values from external CameraSource module.
        /// </summary>
        [Category("Camera")]
        [DefaultValue(true)]
        [Description("Use internal camera settings (position, fov, projection, etc.) "
                    + "or get these values from external CameraSource module.")]
        public bool UseInternalCamera
        {
            get
            {
                return rendererParameters.UseInternalCamera;
            }
            set
            {
                if (!value) inCamera = true;
                rendererParameters.UseInternalCamera = value;
            }
        }

        /// <summary>
        /// Use internal navigation thorough the scene
        /// or manage movement of objects and scene by yourself using some kind of interactor.
        /// </summary>
        [Category("Camera")]
        [DefaultValue(true)]
        [Description("Use internal navigation thorough the scene or manage movement of objects and scene by yourself using some kind of interactor.")]
        public bool UseInternalVirtualTrackball
        {
            get
            {
                return rendererParameters.UseInternalVirtualTrackball;
            }
            set
            {
                rendererParameters.UseInternalVirtualTrackball = value;
            }
        }

        [Category("Render to texture"),
        Description("SaveToTextureFPS")]
        public int SaveToTextureFPS
        {
            get { return rendererParameters.SavetextureFPS; }
            set
            {
                rendererParameters.SavetextureFPS = value;
            }
        }

        [Category("Show output port")]
        [DefaultValue(false)]
        [Description("Show output port for picker.")]
        public bool OutPicker
        {
            get { return rendererParameters.PickerEnabled; }
            set 
            {
                rendererParameters.PickerEnabled = value;

                if (rendererParameters.PickerEnabled)
                {
                    AddOutPort("outPicker", typeof(Mve.UnstrGridRendering.BasePicker), false);
                }
                else
                {
                    RemoveOutPort("outPicker");
                }
            }
        }

        /// <summary>
        /// Show input port for camera settings.
        /// </summary>
        [Category("Show input port")]
        [DefaultValue(false)]
        [Description("Show input port for camera settings.")]
        public bool inCamera
        {
            get
            {
                return showInCamera;
            }
            set
            {
                if (value)
                {
                    if (!showInCamera) AddInPort("inCamera", typeof(Camera), true);
                }
                else
                {
                    if (showInCamera) RemoveInPort("inCamera");

                    rendererParameters.UseInternalCamera = true;
                }

                showInCamera = value;
            }
        }
        #endregion

        /// <summary>
        /// Constructor. Initializes module and adds inScene port.
        /// </summary>
        public SceneRendererModule()
        {
            showInCamera = false;

            AddInPort("inScene", typeof(Scene), true, false);
            AddInPort("frameNumber", typeof(Mve.Numerics.Integer), false, false);
            AddInPort("inLegendArray", typeof(UniformDataArray), false, false);

            // render parameters initialization
            rendererParameters.PathSaveto = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + "\\MVE\\";
            rendererParameters.PickerEnabled = false;
            rendererParameters.Screen = ScreenTarget.Primary;
            rendererParameters.StereoMode = StereoSceneRendererModule.Mode.normal;
            rendererParameters.BackgroundColor = Color.Black;
            rendererParameters.UseInternalCamera = true;
            rendererParameters.UseInternalVirtualTrackball = true;
            rendererParameters.Text = "Renderer window";
            rendererParameters.Width = SceneRendererModule.InitialWindowWidth;
            rendererParameters.Height = SceneRendererModule.InitialWindowHeight;
            rendererParameters.SavetextureFPS = SceneRendererModule.SaverFramerate;

            pickerSetToOutput = false;
            frameIn = false;
            showLegend = false;

            base.ModuleExecutionAborted += new ModuleExecutionAbortedEventHandler(SceneRenderer_ModuleExecutionAborted);
            this.PortConnected         += new PortConnectingEventHandler(SceneRenderer_PortConnecting);
            this.PortDisconnected      += new PortDisconnectingEventHandler(SceneRenderer_PortDisconnecting);
        }

        void SceneRenderer_PortDisconnecting(object source, InPort portTo, out bool allowed)
        {
            allowed = true;
            if (portTo.Name == "frameNumber")
            {
                frameIn = false;
            }

            if (portTo.Name == "inLegendArray")
            {
                showLegend = false;
            }
        }

        void SceneRenderer_PortConnecting(object source, OutPort portFrom, InPort portTo, out bool allowed)
        {
            allowed = true;
            if (portTo.Name == "frameNumber")
            {
                frameIn = true;
            }

            if (portTo.Name == "inLegendArray")
            {
                showLegend = true;
            }
        }

        void SceneRenderer_ModuleExecutionAborted(object source, bool abortImmediately)
        {
            if (renderer != null)
            {
                renderer.StopRendering();

                Thread.Sleep(10);
            }
        }

        /// <summary>
        /// Updates the Scene and starts rendering loop. 
        /// </summary>
        public override void Execute()
        {
            pickerSetToOutput = false;

            InitializeInput();

            if (!UseInternalCamera)
                camera = (Camera)GetInput("inCamera");
            else
                camera = new Camera();

            renderer = new Renderer(scene, uda, camera, rendererParameters);

            // check for input data structures updates
            renderer.OnIdle += new OnIdleEventHandler(UpdateInput);

            renderer.Run(); // start rendering loop
        }

        protected void InitializeInput()
        {
            UpdateInput("inScene");
            scene = (Scene)GetInput("inScene");

            if (showLegend)
            {
                UpdateInput("inLegendArray");
                uda = (UniformDataArray)GetInput("inLegendArray");
            }
        }

        /// <summary>
        /// Check if input data changed, called in each step of rendering loop
        /// </summary>
        protected void UpdateInput()
        {
            UpdateInput("inScene");

            if (frameIn)
            {
                UpdateInput("frameNumber");
                scene.Frame = (Mve.Numerics.Integer)GetInput("frameNumber");
            }

            if (showLegend)
            {
                if (IsDataSame("inLegendArray"))
                {
                    UpdateInput("inLegendArray");
                    uda = (UniformDataArray)GetInput("inLegendArray");
                }
            }

            if (!pickerSetToOutput)
            {
                if ((renderer.Picker != null) && (rendererParameters.PickerEnabled))
                {
                    SetOutput("outPicker", renderer.Picker);
                }

                pickerSetToOutput = true;
            }
        }

        //protected void HookPickerToOutputPort()
        //{
        //    if (renderer.Picker != null)
        //    {
        //        renderer.PickerEnabled = rendererParameters.PickerEnabled;

        //        if (rendererParameters.PickerEnabled) SetOutput("outPicker", renderer.Picker);
        //    }
        //}

        /// <summary>
        /// Writes Xml configuration of current module.
        /// </summary>
        /// <param name="config">XmlElement from map config file which belongs to current module.</param>
        public override void WriteConfig(XmlElement config)
        {
            XmlElement xmlProperties = config.OwnerDocument.CreateElement("Properties");

            XmlElement xmlText = config.OwnerDocument.CreateElement("Text");
            xmlText.SetAttribute("type", Text.GetType().FullName);
            xmlText.InnerText = Text;
            xmlProperties.AppendChild(xmlText);

            XmlElement xmlWidth = config.OwnerDocument.CreateElement("Width");
            xmlWidth.SetAttribute("type", Width.GetType().FullName);
            xmlWidth.InnerText = Width.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlWidth);

            XmlElement xmlHeight = config.OwnerDocument.CreateElement("Height");
            xmlHeight.SetAttribute("type", Height.GetType().FullName);
            xmlHeight.InnerText = Height.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlHeight);

            XmlElement xmlBackgroundColor = config.OwnerDocument.CreateElement("BackgroundColor");
            xmlBackgroundColor.SetAttribute("type", BackgroundColor.GetType().FullName);
            xmlBackgroundColor.InnerText = BackgroundColor.Red.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                         + BackgroundColor.Green.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                         + BackgroundColor.Blue.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                         + BackgroundColor.Alpha.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlBackgroundColor);

            XmlElement xmlUseInternalCamera = config.OwnerDocument.CreateElement("UseInternalCamera");
            xmlUseInternalCamera.SetAttribute("type", UseInternalVirtualTrackball.GetType().FullName);
            xmlUseInternalCamera.InnerText = UseInternalCamera.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlUseInternalCamera);

            XmlElement xmlUseInternalVirtualTrackball = config.OwnerDocument.CreateElement("UseInternalVirtualTrackball");
            xmlUseInternalVirtualTrackball.SetAttribute("type", UseInternalVirtualTrackball.GetType().FullName);
            xmlUseInternalVirtualTrackball.InnerText = UseInternalVirtualTrackball.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlUseInternalVirtualTrackball);

            XmlElement xmlShowInCamera = config.OwnerDocument.CreateElement("inCamera");
            xmlShowInCamera.SetAttribute("type", inCamera.GetType().FullName);
            xmlShowInCamera.InnerText = inCamera.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlShowInCamera);

            XmlElement xmlScreen = config.OwnerDocument.CreateElement("Screen");
            xmlScreen.SetAttribute("type", rendererParameters.Screen.GetType().FullName);
            xmlScreen.InnerText = rendererParameters.Screen.ToString();
            xmlProperties.AppendChild(xmlScreen);

            XmlElement xmlPickerEnabled = config.OwnerDocument.CreateElement("PickerEnabled");
            xmlPickerEnabled.SetAttribute("type", rendererParameters.PickerEnabled.GetType().FullName);
            xmlPickerEnabled.InnerText = rendererParameters.PickerEnabled.ToString();
            xmlProperties.AppendChild(xmlPickerEnabled);

            XmlElement xmlFPS = config.OwnerDocument.CreateElement("SavetextureFPS");
            xmlFPS.SetAttribute("type", rendererParameters.SavetextureFPS.GetType().FullName);
            xmlFPS.InnerText = rendererParameters.SavetextureFPS.ToString();
            xmlProperties.AppendChild(xmlFPS);     

            config.AppendChild(xmlProperties);
        }

        /// <summary>
        /// Reads Xml configuration of current module.
        /// </summary>
        /// <param name="config">XmlElement from map config file which belongs to current module.</param>
        public override void ReadConfig(System.Xml.XmlElement config)
        {
            foreach (XmlElement xeConfig in config.ChildNodes)
            {
                if (xeConfig.Name == "Properties")
                {
                    foreach (XmlElement xeProperties in xeConfig.ChildNodes)
                    {
                        switch (xeProperties.Name)
                        {
                            case "Text":
                                rendererParameters.Text = xeProperties.InnerText;
                                break;
                            case "Width":
                                rendererParameters.Width = FromString.GetInt(xeProperties.InnerText, InitialWindowWidth);
                                break;
                            case "Height":
                                rendererParameters.Height = FromString.GetInt(xeProperties.InnerText, InitialWindowHeight);
                                break;
                            case "BackgroundColor":
                                BackgroundColor = FromString.GetColorRGBA(xeProperties.InnerText);
                                break;
                            case "UseInternalCamera":
                                rendererParameters.UseInternalCamera = FromString.GetBool(xeProperties.InnerText, true);
                                break;
                            case "UseInternalVirtualTrackball":
                                rendererParameters.UseInternalVirtualTrackball = FromString.GetBool(xeProperties.InnerText, true);
                                break;
                            case "inCamera":
                                inCamera = FromString.GetBool(xeProperties.InnerText, false);
                                break;
                            case "PathSaveto":
                                rendererParameters.PathSaveto = xeProperties.InnerText;
                                break;
                            case "PickerEnabled":
                                rendererParameters.PickerEnabled = FromString.GetBool(xeProperties.InnerText, true);
                                // if saved with enabled out port then restore need add out port too
                                if (rendererParameters.PickerEnabled) AddOutPort("outPicker", typeof(Mve.UnstrGridRendering.BasePicker), false);
                                break;
                            case "Screen":
                                rendererParameters.Screen = (ScreenTarget) Enum.Parse(typeof(ScreenTarget), xeProperties.InnerText);
                                break;
                            case "SavetextureFPS":
                                rendererParameters.SavetextureFPS = FromString.GetInt(xeProperties.InnerText, SceneRendererModule.SaverFramerate);
                                break;
                            default:
                                Debug.WriteLine("Unreaded value:" + xeProperties.Name + " - " + xeProperties.InnerText);
                                break;
                        }
                    }
                }
            }
        }
        
        /// <summary>
        /// Creates new instance of module and copy all necessary member variables.
        /// This method is used by MapEditor in Copy/Paste process.
        /// Without overriding it returns implicit instance of current module.
        /// </summary>
        public override Module DeepCopy()
        {
            SceneRendererModule sceneRenderer = (SceneRendererModule)base.DeepCopy();

            sceneRenderer.rendererParameters = (RendererParameters)this.rendererParameters.DeepCopy();

            return sceneRenderer;
        }
    }
}
