using System;
using System.Collections.Generic;
using System.Text;
using dx3d = Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// Scene composed of D3DUT objects.
  /// Trasformation of the Scene to D3DUTScene is performed in the constructor.
  /// Actualization of objects of the scene is done in Update method.
  /// </summary>
  public class SceneWrapper : IDisposable
  {
    private SceneObjectWrapper[] d3dutSceneObjects;
    private LightWrapper[] d3dutLights;
    private SceneOrigin d3dutSceneOrigin;
    private Scene scene;
    private long lmtScene;

    #region min, mid, max points

    Point3D minPoint;
    /// <summary>
    /// Minimum point of the scene (bottom, left, near).
    /// </summary>
    public Point3D MinPoint
    {
      get
      {
        return minPoint;
      }
      set
      {
        minPoint = value;
      }
    }

    Point3D midPoint;
    /// <summary>
    /// Middle point of the scene.
    /// </summary>
    public Point3D MidPoint
    {
      get
      {
        return midPoint;
      }
      set
      {
        midPoint = value;
      }
    }

    Point3D maxPoint;
    /// <summary>
    /// Maximum point of the scene (top, right, far).
    /// </summary>
    public Point3D MaxPoint
    {
      get
      {
        return maxPoint;
      }
      set
      {
        maxPoint = value;
      }
    }
    #endregion

    /// <summary>
    /// Constructor of D3DUTScene. Performs transformation
    /// of the Scene to the D3DUTScene and computes min, mid and max points.
    /// </summary>
    /// <param name="scene"></param>
    public SceneWrapper(Scene scene)
    {
      this.scene = scene;
      lmtScene = 0;
      d3dutSceneOrigin = new SceneOrigin();
      UpdateBoundingBox();
    }

    /// <summary>
    /// Updates the scene.
    /// </summary>
    /// <param name="device">Rendering device.</param>
    public void Update(Device device)
    {
      if (lmtScene == scene.ModificationTime)
      {
        return;
      }
    
      lmtScene = scene.ModificationTime;

      int realCount;

      if (scene.SceneObjects != null)
      {
        realCount = 0;
        foreach (SceneObject so in scene.SceneObjects)
          if (so != null)
            realCount++;

        // if number of sceneobjects in scene changed, recreate them
        if ((d3dutSceneObjects == null) || (realCount != d3dutSceneObjects.Length))
        {
          d3dutSceneObjects = new SceneObjectWrapper[realCount];

          for (int i = 0; i < realCount; i++)
          {
            d3dutSceneObjects[i] = new SceneObjectWrapper();
          }
        }

        // update scene objects
        for (int i = 0; i < scene.SceneObjects.Count; i++)
        {
          if (scene.SceneObjects[i] != null)
            d3dutSceneObjects[i].Update(device, (SceneObject) scene.SceneObjects[i]);
        }
      }

      if (scene.Lights != null)
      {
        realCount = 0;
        foreach (Light l in scene.Lights)
          if (l != null)
            realCount++;

        // if number of lights in scene changed, recreate them
        if ((d3dutLights == null) || (realCount != d3dutLights.Length))
        {
          d3dutLights = new LightWrapper[realCount];

          for (int i = 0; i < realCount; i++)
          {
            d3dutLights[i] = new LightWrapper();
          }
        }

        // update lights
        for (int i = 0; i < scene.Lights.Count; i++)
        {
          if (scene.Lights[i] != null)
            d3dutLights[i].Update(device, (Light) scene.Lights[i]);
        }
      }

      if (scene.ShowSceneOrigin)
        d3dutSceneOrigin.Update(device, scene.SceneOriginSize);
    }

    /// <summary>
    /// Draws scene. This method should be used between device.BeginScene() and device.EndScene()
    /// </summary>
    /// <param name="device">Rendering device.</param>
    public void Draw(Device device)
    {
      if (d3dutLights != null)
      {
        int i = 0;
        foreach (LightWrapper l in d3dutLights)
        {
          l.SetLight(device, i);
          i++;
        }
      }

      if (d3dutSceneObjects != null)
      {
        foreach (SceneObjectWrapper so in d3dutSceneObjects)
        {
          so.Draw(device, scene.Frame);
        }
      }

      if (scene.ShowSceneOrigin) 
        d3dutSceneOrigin.Draw();
    }

    /// <summary>
    /// Updates min, mid, max points. Primary use of these points is to setup camera target point.
    /// </summary>
    private void UpdateBoundingBox()
    {
      minPoint = new Point3D(float.MaxValue, float.MaxValue, float.MaxValue);
      midPoint = new Point3D(0, 0, 0);
      maxPoint = new Point3D(float.MinValue, float.MinValue, float.MinValue);

      Point3D tmpPoint;
      int visibleObjectsCount = 0; 

      foreach (SceneObject so in scene.SceneObjects)
      {
        if ((so.ShowSceneObject) && (so.UnstructuredGrid != null) && (so.UnstructuredGrid.Points != null))
        {
          tmpPoint = new Point3D((float) so.UnstructuredGrid.BoundingBox.MinPoint[0], (float) so.UnstructuredGrid.BoundingBox.MinPoint[1], (float) so.UnstructuredGrid.BoundingBox.MinPoint[2]);
          tmpPoint = ((Transform3D) so.UnstructuredGrid.Transform).Transform(tmpPoint);

          if (minPoint.X > tmpPoint.X)
            minPoint.X = tmpPoint.X;
          if (minPoint.Y > tmpPoint.Y)
            minPoint.Y = tmpPoint.Y;
          if (minPoint.Z > tmpPoint.Z)
            minPoint.Z = tmpPoint.Z;

          tmpPoint = new Point3D((float) so.UnstructuredGrid.BoundingBox.MaxPoint[0], (float) so.UnstructuredGrid.BoundingBox.MaxPoint[1], (float) so.UnstructuredGrid.BoundingBox.MaxPoint[2]);
          tmpPoint = ((Transform3D) so.UnstructuredGrid.Transform).Transform(tmpPoint);

          if (maxPoint.X < tmpPoint.X)
            maxPoint.X = tmpPoint.X;
          if (maxPoint.Y < tmpPoint.Y)
            maxPoint.Y = tmpPoint.Y;
          if (maxPoint.Z < tmpPoint.Z)
            maxPoint.Z = tmpPoint.Z;

          tmpPoint = new Point3D((float) so.UnstructuredGrid.BoundingBox.MidPoint[0], (float) so.UnstructuredGrid.BoundingBox.MidPoint[1], (float) so.UnstructuredGrid.BoundingBox.MidPoint[2]);
          midPoint += tmpPoint;

          visibleObjectsCount++;
        }

      }

      midPoint /= visibleObjectsCount;
    }

    #region IDisposable Members
    /// <summary>
    /// Dispose.
    /// </summary>
    public void Dispose()
    {
      foreach (SceneObjectWrapper so in d3dutSceneObjects)
      {
        so.Dispose();
      }
    }
    #endregion
  }
}
