using System;
using System.Text;
using System.Threading;
using System.Diagnostics;
using System.Windows.Forms;
using System.Runtime.Remoting.Messaging;

using System.IO;

using Zcu.Mve.Core;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace Zcu.Mve.UnstrGridRendering
{
    class StereoRenderer : Renderer
    {
        private Effect effect;
        private Texture renderLeft, renderRight;
        private Surface left, right;
        private VertexBuffer screenQuad;
        private float paralaxstep;

        /// <summary>
        /// Default constructor. Initializes renderer.
        /// </summary>
        /// <param name="scene">Scene.</param>
        /// <param name="camera">Camera.</param>
        /// <param name="rendererParameters">Renderer parameters.</param>
        public StereoRenderer(Scene scene,  Visualization.UniformDataArray uda, Camera camera, RendererParameters rendererParameters)
            : base(scene, uda, camera, rendererParameters)
        {
            effect = null;
            screenQuad = null;

            base.WindowCreated += new WindowCreatedEventHandler(D3DUTStereoRenderer_WindowCreated);
            base.RenderingLoopStart += new RenderingStartEventHandler(D3DUTStereoRenderer_RenderingLoopStart);

            paralaxstep = rendererParameters.EyeParalax * 0.1f;
            _rParams.EyeParalax = (0.01f * ((float)Math.Abs(this.scene.MaxPoint.X - this.scene.MinPoint.X)));
        }

        void D3DUTStereoRenderer_RenderingLoopStart()
        {
            RenderStereoShader();
        }

        void D3DUTStereoRenderer_WindowCreated()
        {
            _window.KeyDown += new KeyEventHandler(form_KeyDown);

            _window.ResizeEnd += new EventHandler(form_Resize); //change texture resolution
        }

        protected override void DrawScene(Device device)
        {
            if (_rParams.StereoMode == StereoSceneRendererModule.Mode.normal)
            {
                base.DrawScene(device);
            }
            else
            {
                RenderStereo(device);
            }
        }

        #region stereo_Hladik

        /// <summary>
        /// Scene objects rendering.
        /// sceond input describes distance beetween eyes
        /// </summary>
        protected void DrawStereoScene(Device device, float shift)
        {
            camera.SetCamera(device);
            //shift eye in x-direction
            device.Transform.View = camera.ViewMatrix * Matrix.Translation(shift, 0, 0);
            //draw scene
            scene.Draw(device);
        }

        /// <summary>
        /// draw scene with shader describing stereo effect
        /// </summary>
        private void RenderStereo(Device device)
        {
            device.BeginScene();

            Surface renderTarget = null;

            try
            {
               renderTarget = device.GetRenderTarget(0);

                device.SetRenderTarget(0,left);
                device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, _rParams.BackgroundColor, 1.0f, 0);
                DrawStereoScene(device, _rParams.EyeParalax);

                device.SetRenderTarget(0, right);
                device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, _rParams.BackgroundColor, 1.0f, 0);
                DrawStereoScene(device, -_rParams.EyeParalax);

                device.SetRenderTarget(0, renderTarget);

                renderTarget.Dispose();
            }

            catch (Microsoft.DirectX.Direct3D.DeviceLostException)
            {
                if (!renderTarget.Disposed) renderTarget.Dispose();
            }

            effect.Technique = "Stereo";

            effect.SetValue("left", renderLeft);
            effect.SetValue("right", renderRight);

            effect.SetValue("xPosition", _window.Location.X);
            effect.SetValue("yPosition", _window.Location.Y);

            effect.CommitChanges();
            device.Clear(ClearFlags.ZBuffer, _rParams.BackgroundColor, 1.0f, 0);

            int passes = effect.Begin(0);

            switch (_rParams.StereoMode)
            {
                case StereoSceneRendererModule.Mode.anaglyph:
                    effect.BeginPass(0);
                    break;
                case StereoSceneRendererModule.Mode.interlaced_horizontal:
                    effect.BeginPass(2);
                    break;
                case StereoSceneRendererModule.Mode.interlaced_vertical:
                    effect.BeginPass(1);
                    break;
            }

            device.SetStreamSource(0, screenQuad, 0);
            device.VertexFormat = CustomVertex.PositionTextured.Format;
            device.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 2);

            effect.EndPass();
            effect.End();

            // draw text
            fo.Draw("FPS: " + fpsCounter.GetFPS().ToString());
            fo.Draw(infoStr);

            device.EndScene();
        }

        private void RenderStereoShader()
        {
            InitTextures();             //init render targets for left and right eyes
            InitShader();               //init effect
            InitBuffer();               //init vertex buffer containig quad with -1,1 coordinates

            _runtime.Device.DeviceReset += new EventHandler(Device_DeviceReset);         //dispose textures
            _runtime.Device.DeviceLost += new EventHandler(Device_DeviceLost);           //new textures
        }

        void form_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.Left:
                    {
                        _rParams.EyeParalax -= (0.005f * ((float)Math.Abs(scene.MaxPoint.X - scene.MinPoint.X)));
                        break;
                    }
                case Keys.Right:
                    {
                        _rParams.EyeParalax += (0.005f * ((float)Math.Abs(scene.MaxPoint.X - scene.MinPoint.X)));
                        break;
                    }
            }
        }

        void form_Resize(object sender, EventArgs e)
        {
           // //dispose old textures and create new with correct size
           // InitTextures();
            //_runtime.DeviceResize(_window.Size);
        }

        void Device_DeviceLost(object sender, EventArgs e)
        {
            //quad is managed - so it wo don't have to dispose it
            if (renderLeft != null && !renderLeft.Disposed) 
                renderLeft.Dispose();

            if (renderRight != null && !renderRight.Disposed) 
                renderRight.Dispose();

            if (right != null && !right.Disposed)
                right.Dispose();

            if (left != null && !left.Disposed)
                left.Dispose();

            effect.OnLostDevice();
        }

        void Device_DeviceReset(object sender, EventArgs e)
        {
            InitTextures();
            effect.OnResetDevice();
        }

        #region init

        private void InitShader()
        {
            try
            {

                String effectString = StereoRendererResource.stereo;
                this.effect = Effect.FromString(_runtime.Device, effectString, null, null, ShaderFlags.None, null);

                //String path = Application.StartupPath + "\\Effects\\Stereo.fx";
                //this.effect = Effect.FromFile(_runtime.Device, path, null, null, ShaderFlags.NoPreShader, null);

                effect.Technique = "Stereo";

                effect.SetValue("xPosition", _window.Location.X);
                effect.SetValue("yPosition", _window.Location.Y);
            }
            catch (Exception)
            {
                MessageBox.Show("stereo effect failed to load");

                _rParams.StereoMode = StereoSceneRendererModule.Mode.normal;
            }
        }

        private bool InitTextures()
        {
            try
            {
                if (renderLeft != null && !renderLeft.Disposed)
                    renderLeft.Dispose();
                if (renderRight != null && !renderRight.Disposed)
                    renderRight.Dispose();

                renderLeft = null;
                renderRight = null;

                System.Drawing.Size s = _window.ClientSize;

                renderLeft = new Texture(_runtime.Device, s.Width, s.Height, 1, Usage.RenderTarget, Format.X8R8G8B8, Pool.Default);
                renderRight = new Texture(_runtime.Device, s.Width, s.Height, 1, Usage.RenderTarget, Format.X8R8G8B8, Pool.Default);

                left = renderLeft.GetSurfaceLevel(0);
                right = renderRight.GetSurfaceLevel(0);

                return true;
            }
            catch
            {
                _rParams.StereoMode = StereoSceneRendererModule.Mode.normal;

                return false;
            }
        }

        private void InitBuffer()
        {
            screenQuad = new VertexBuffer(typeof(CustomVertex.PositionTextured), 4, _runtime.Device, Usage.WriteOnly, CustomVertex.PositionTextured.Format, Pool.Managed);
            CustomVertex.PositionTextured[] verts = (CustomVertex.PositionTextured[])screenQuad.Lock(0, 0); // Lock the buffer (which will return our structs)


            verts[0].Position = new Vector3(-1, -1, 1);
            verts[0].Tu = 0; verts[0].Tv = 1;

            verts[1].Position = new Vector3(-1, 1, 1);
            verts[1].Tu = 0; verts[1].Tv = 0;

            verts[2].Position = new Vector3(1, -1, 1);
            verts[2].Tu = 1; verts[2].Tv = 1;

            verts[3].Position = new Vector3(1, 1, 1);
            verts[3].Tu = 1; verts[3].Tv = 0;

            // Unlock (and copy) the data
            screenQuad.Unlock();
        }

        #endregion

        #endregion
    }
}
