using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.ComponentModel;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Drawing;

using System.Xml;
using System.Diagnostics;
using System.Drawing.Design;
using System.Globalization;
using System.Threading;

namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// Renders Scene using D3DUT interface.
    /// </summary>
    [ModuleInfo("Vojtech Hladik", "Renders Scene in stereo mode using DirectX interface.", IconName = "SceneRenderer.ico")]
    [PortInfo("inScene", "Representation of Scene. Scene is a composition of SceneObjects and Lights.")]
    [PortInfo("inCamera", "Camera settings.")]
    public class StereoSceneRendererModule : SceneRendererModule
    {
        #region Fields

        //private D3DUTStereoRenderer renderer;
        public enum Mode { normal = 1, interlaced_vertical = 3, interlaced_horizontal = 4, anaglyph = 5 };

        #endregion

        #region Properties

        [Category("Output window"),
        Description("Caption of renderer window."),
        DefaultValue(Zcu.Mve.UnstrGridRendering.StereoSceneRendererModule.Mode.normal)]
        public StereoSceneRendererModule.Mode StereoMode
        {
            get { return rendererParameters.StereoMode; }
            set { rendererParameters.StereoMode = value; }
        }

        [Category("Output window"),
        Description("Caption of renderer window."), DefaultValue(0.015f)]
        public float Eyeparallax
        {
            get { return rendererParameters.EyeParalax; }
            set { rendererParameters.EyeParalax = value; }
        }


        #endregion

        /// <summary>
        /// Constructor. Initializes module and adds inScene port.
        /// </summary>
        public StereoSceneRendererModule() : base()
        {
            rendererParameters.StereoMode = Mode.normal;
            rendererParameters.EyeParalax = 0.015f;
            rendererParameters.PathSaveto = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + "\\MVE\\";  
        }

        /// <summary>
        /// Updates the Scene and starts rendering loop. 
        /// </summary>
        public override void Execute()
        {
            base.InitializeInput();

            if (!UseInternalCamera)
                camera = (Camera)GetInput("inCamera");
            else
                camera = new Camera();

            renderer = new StereoRenderer(scene, uda, camera, rendererParameters);

            renderer.OnIdle += new OnIdleEventHandler(base.UpdateInput);

            // must be here
            //base.HookPickerToOutputPort();

            renderer.Run(); // start rendering thread
        }

        /// <summary>
        /// Creates new instance of module and copy all necessary member variables.
        /// This method is used by MapEditor in Copy/Paste process.
        /// Without overriding it returns implicit instance of current module.
        /// </summary>
        public override Module DeepCopy()
        {
            StereoSceneRendererModule sceneRenderer = (StereoSceneRendererModule)base.DeepCopy();

            sceneRenderer.rendererParameters = (RendererParameters)this.rendererParameters.DeepCopy();

            return sceneRenderer;
        }

        /// <summary>
        /// Writes Xml configuration of current module.
        /// </summary>
        /// <param name="config">XmlElement from map config file which belongs to current module.</param>
        public override void WriteConfig(XmlElement config)
        {
            XmlElement xmlProperties = config.OwnerDocument.CreateElement("Properties");

            XmlElement xmlText = config.OwnerDocument.CreateElement("Text");
            xmlText.SetAttribute("type", Text.GetType().FullName);
            xmlText.InnerText = Text;
            xmlProperties.AppendChild(xmlText);

            XmlElement xmlWidth = config.OwnerDocument.CreateElement("Width");
            xmlWidth.SetAttribute("type", Width.GetType().FullName);
            xmlWidth.InnerText = Width.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlWidth);

            XmlElement xmlHeight = config.OwnerDocument.CreateElement("Height");
            xmlHeight.SetAttribute("type", Height.GetType().FullName);
            xmlHeight.InnerText = Height.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlHeight);

            XmlElement xmlBackgroundColor = config.OwnerDocument.CreateElement("BackgroundColor");
            xmlBackgroundColor.SetAttribute("type", BackgroundColor.GetType().FullName);
            xmlBackgroundColor.InnerText = BackgroundColor.Red.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                         + BackgroundColor.Green.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                         + BackgroundColor.Blue.ToString(CultureInfo.InvariantCulture.NumberFormat) + " "
                                         + BackgroundColor.Alpha.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlBackgroundColor);

            XmlElement xmlUseInternalCamera = config.OwnerDocument.CreateElement("UseInternalCamera");
            xmlUseInternalCamera.SetAttribute("type", UseInternalVirtualTrackball.GetType().FullName);
            xmlUseInternalCamera.InnerText = UseInternalCamera.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlUseInternalCamera);

            XmlElement xmlUseInternalVirtualTrackball = config.OwnerDocument.CreateElement("UseInternalVirtualTrackball");
            xmlUseInternalVirtualTrackball.SetAttribute("type", UseInternalVirtualTrackball.GetType().FullName);
            xmlUseInternalVirtualTrackball.InnerText = UseInternalVirtualTrackball.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlUseInternalVirtualTrackball);

            XmlElement xmlShowInCamera = config.OwnerDocument.CreateElement("inCamera");
            xmlShowInCamera.SetAttribute("type", inCamera.GetType().FullName);
            xmlShowInCamera.InnerText = inCamera.ToString(CultureInfo.InvariantCulture.NumberFormat);
            xmlProperties.AppendChild(xmlShowInCamera);

            XmlElement xmlStereo = config.OwnerDocument.CreateElement("Stereomode");
            xmlStereo.SetAttribute("type", rendererParameters.StereoMode.GetType().FullName);
            xmlStereo.InnerText = rendererParameters.StereoMode.ToString();
            xmlProperties.AppendChild(xmlStereo);

            XmlElement xmlParallax = config.OwnerDocument.CreateElement("Parallax");
            xmlParallax.SetAttribute("type", rendererParameters.EyeParalax.GetType().FullName);
            xmlParallax.InnerText = rendererParameters.EyeParalax.ToString();
            xmlProperties.AppendChild(xmlParallax);

            XmlElement xmlScreen = config.OwnerDocument.CreateElement("Screen");
            xmlScreen.SetAttribute("type", rendererParameters.Screen.GetType().FullName);
            xmlScreen.InnerText = rendererParameters.Screen.ToString();
            xmlProperties.AppendChild(xmlScreen);

            XmlElement xmlPickerEnabled = config.OwnerDocument.CreateElement("PickerEnabled");
            xmlPickerEnabled.SetAttribute("type", rendererParameters.PickerEnabled.GetType().FullName);
            xmlPickerEnabled.InnerText = rendererParameters.PickerEnabled.ToString();
            xmlProperties.AppendChild(xmlPickerEnabled);

            XmlElement xmlFPS = config.OwnerDocument.CreateElement("SavetextureFPS");
            xmlFPS.SetAttribute("type", rendererParameters.SavetextureFPS.GetType().FullName);
            xmlFPS.InnerText = rendererParameters.SavetextureFPS.ToString();
            xmlProperties.AppendChild(xmlFPS);          

            config.AppendChild(xmlProperties);
        }

        /// <summary>
        /// Reads Xml configuration of current module.
        /// </summary>
        /// <param name="config">XmlElement from map config file which belongs to current module.</param>
        public override void ReadConfig(System.Xml.XmlElement config)
        {
            base.ReadConfig(config);

            foreach (XmlElement xeConfig in config.ChildNodes)
            {
                if (xeConfig.Name == "Properties")
                {
                    foreach (XmlElement xeProperties in xeConfig.ChildNodes)
                    {
                        switch (xeProperties.Name)
                        {
                            case "Stereomode":
                                rendererParameters.StereoMode = (StereoSceneRendererModule.Mode)Enum.Parse(typeof(StereoSceneRendererModule.Mode), xeProperties.InnerText);
                                break;
                            case "Parallax":
                                rendererParameters.EyeParalax = FromString.GetFloat(xeProperties.InnerText, 0.015f);
                                break;
                        }
                    }
                }
            }
        }
    }
}
