using System;
using System.Threading;
using System.Diagnostics;
using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using Zcu.Mve.UnstrGridRendering.Legend;

namespace Zcu.Mve.UnstrGridRendering
{
    class StereoWallRenderer : AdvancedRenderer
    {
        #region Fileds

        protected SceneWrapper scene;
        protected CameraWrapper camera;
        protected FPSCounter fpsCounter;

        protected FontWrapper fo;
        protected string infoStr;
        
        private int mouseX;
        private int mouseY;
        private int startX;
        private int startY;
        private bool mouseCapture;

        private bool eye;
        private float eyeDistance;
        private float cameraDistance;
        private System.Drawing.Rectangle rect;

        private Form secWindow;
        private System.Drawing.Rectangle primaryRect;

        private SwapChain[] swapChain;
        private Surface[] backBufferSurface;

        const string INFO_BASE = "\n[Esc] - close\n[F1] - show/hide this help\n[F] - fullscreen/windowed mode\n";
        const string INFO_ADVANCE = "[mouse wheel] - zoom in/out\n[leftMouseButton] & mouseMove - rotate the scene\n";

        #endregion

        public StereoWallRenderer(Scene scene, Visualization.UniformDataArray uda, Camera camera, RendererParameters rendererParameters)
            : base(camera, uda, rendererParameters)
        {
            CheckScreens();

            infoStr = INFO_BASE + ((_rParams.UseInternalVirtualTrackball == true) ? INFO_ADVANCE : string.Empty);

            eyeDistance = 0.01f;
            cameraDistance = 1.0f;

            mouseCapture = false;
            mouseX = mouseY = 0;
            startX = startY = 0;

            this.scene = new SceneWrapper(scene);
            this.camera = new CameraWrapper(camera, this.scene, rendererParameters);

            fpsCounter = new FPSCounter();
            base.RenderingLoopStart += new RenderingStartEventHandler(D3DStereoWallRenderer_RenderingLoopStart);
            base.RenderingLoopEnd += new RenderingEndEventHandler(D3DStereoWallRenderer_RenderingLoopEnd);
        }

        public override void Run()
        {
            // left rendering context
            _window = new Form();
            // right rendering context
            secWindow = new Form();

            _window.Text = _rParams.Text + " right";
            _window.FormBorderStyle = FormBorderStyle.Sizable;
            _window.ClientSize = new System.Drawing.Size(_rParams.Width, _rParams.Height);
            _wndRestoreRect = new System.Drawing.Rectangle(_window.Location, _window.ClientSize);

            secWindow.Text = _rParams.Text + " left";
            secWindow.FormBorderStyle = FormBorderStyle.Sizable;
            secWindow.MaximizeBox = false;
            secWindow.ClientSize = _window.ClientSize;

            secWindow.Owner = _window;

            // window events
            _window.KeyDown += new KeyEventHandler(_window_KeyDown);
            _window.MouseDown += new MouseEventHandler(_window_MouseDown);
            _window.MouseMove += new MouseEventHandler(_window_MouseMove);
            _window.MouseUp += new MouseEventHandler(_window_MouseUp);
            _window.MouseWheel += new MouseEventHandler(_window_MouseWheel);

            _window.ResizeEnd += new EventHandler(_window_ResizeEnd);
            secWindow.FormClosing += new FormClosingEventHandler(secWindow_FormClosing);
            
            // hook events to picker
            HookEventsToPicker(_window);
            HookEventsToPicker(secWindow);

            IntPtr[] hndls = new IntPtr[2];
            hndls[0] = _window.Handle;
            hndls[1] = secWindow.Handle;

            // device creation
            _runtime = new Runtime(hndls, _rParams.Width, _rParams.Height);

            base.Render();

            _window.Close();
            secWindow.Close();

            if (_runtime != null) _runtime.Dispose();
        }

        protected override void Render(Device device)
        {
            UpdateScene(device);

            using (Surface origSurf = device.GetRenderTarget(0))
            {
                // primary / left window context
                device.SetRenderTarget(0, backBufferSurface[0]);

                device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, _rParams.BackgroundColor, 1.0f, 0);

                if (_lRenderer != null) _lRenderer.Render();

                device.BeginScene();

                eye = true;
                DrawScene(device);

                device.EndScene();

                //// save target to texture
                //if (_tSaver.toSave()) _tSaver.SaveToFile(backBufferSurface[0], @"C:\temp\out\L\", device);

                swapChain[0].Present();

                // secondary / right window context
                device.SetRenderTarget(0, backBufferSurface[1]);

                device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, _rParams.BackgroundColor, 1.0f, 0);

                if (_lRenderer != null) _lRenderer.Render();

                device.BeginScene();

                eye = false;
                DrawScene(device);

                device.EndScene();

                //// save target to texture
                //if (_tSaver.toSave()) _tSaver.SaveToFile(backBufferSurface[1], @"C:\temp\out\R\", device);

                swapChain[1].Present();

                // restore original render target
                device.SetRenderTarget(0, origSurf);

                // release allocated resources
                origSurf.Dispose();


            }            
        }

        protected override void UpdateScene(Device device)
        {
            fpsCounter.Update();

            scene.Update(device);

            camera.Update(_window.ClientSize);
        }

        protected override void DrawScene(Device device)
        {
            fo.Draw("FPS: " + fpsCounter.GetFPS().ToString());
            fo.Draw(infoStr);

            camera.SetCamera(device);

            if (eye)
            {
                device.Transform.View = camera.ViewMatrix * Matrix.Translation(eyeDistance, 0, 0);
            }
            else
            {
                device.Transform.View = camera.ViewMatrix * Matrix.Translation(-eyeDistance, 0, 0);
            }

            scene.Draw(device);
        }

        /// <summary>
        /// Check accessible screens.
        /// </summary>
        private void CheckScreens()
        {
            if (Screen.AllScreens.Length == 1)
            {
                primaryRect = new System.Drawing.Rectangle(
                    new System.Drawing.Point(),
                    Screen.AllScreens[0].WorkingArea.Size
                    );

                primaryRect.Width /= 2;
                primaryRect.Height /= 2;

                Console.WriteLine("Only ONE screen detected, stereo mode require two screens.");
            }

            for (int i = 0; i < Screen.AllScreens.Length; i++)
            {
                Screen s = Screen.AllScreens[i];

                if (s.Primary) { primaryRect = s.WorkingArea; break; }
            }
        }

        private void _window_ResizeEnd(object sender, EventArgs e)
        {
            secWindow.ClientSize = _window.ClientSize;

            // check secondary screen location
            UpdateSecondaryScreenPosition();

            secWindow.WindowState = _window.WindowState;

            if (_runtime != null) _runtime.DeviceResize(_window.ClientSize);            
        }

        private void secWindow_FormClosing(object sender, FormClosingEventArgs e)
        {
            StopRendering();
        }

        private void D3DStereoWallRenderer_RenderingLoopStart()
        {
            secWindow.Show();
            
            UpdateSecondaryScreenPosition();

            _window.Activate();

            swapChain = new SwapChain[2];
            backBufferSurface = new Surface[2];

            // font
            fo = new FontWrapper(_runtime.Device);

            if (_runtime != null) _runtime.Device.DeviceLost += new EventHandler(device_DeviceLost);
            if (_runtime != null) _runtime.Device.DeviceReset += new EventHandler(Device_DeviceReset);

            // get swap chaines for screens
            swapChain[0] = _runtime.CreateAditionalSwapChain(0);
            swapChain[1] = _runtime.CreateAditionalSwapChain(1);

            // get surfaces
            backBufferSurface[0] = swapChain[0].GetBackBuffer(0, BackBufferType.Mono);
            backBufferSurface[1] = swapChain[1].GetBackBuffer(0, BackBufferType.Mono);

            //_tSaver = new TextureSaver();
        }

        private void D3DStereoWallRenderer_RenderingLoopEnd()
        {
            device_DeviceLost(null, null);

            if (fo != null) fo.Dispose();
        }

        /// <summary>
        /// Called when device lost occure.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void device_DeviceLost(object sender, EventArgs e)
        {
            if ((backBufferSurface[0] != null) && (!backBufferSurface[0].Disposed)) backBufferSurface[0].Dispose();
            if ((backBufferSurface[1] != null) && (!backBufferSurface[1].Disposed)) backBufferSurface[1].Dispose();

            if ((swapChain[0] != null) && !swapChain[0].Disposed) swapChain[0].Dispose();
            if ((swapChain[1] != null) && !swapChain[1].Disposed) swapChain[1].Dispose();
        }

        private void Device_DeviceReset(object sender, EventArgs e)
        {
            swapChain = new SwapChain[2];
            backBufferSurface = new Surface[2];

            // get swap chaines for screens
            swapChain[0] = _runtime.CreateAditionalSwapChain(0);
            swapChain[1] = _runtime.CreateAditionalSwapChain(1);

            // get surfaces
            backBufferSurface[0] = swapChain[0].GetBackBuffer(0, BackBufferType.Mono);
            backBufferSurface[1] = swapChain[1].GetBackBuffer(0, BackBufferType.Mono);
        }

        private void _window_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.F1:
                    fo.Enable = !fo.Enable;
                    break;

                case Keys.Left:
                    eyeDistance += (0.01f * ((float)Math.Abs(scene.MaxPoint.X - scene.MinPoint.X)));
                    break;

                case Keys.Right:
                    eyeDistance -= (0.01f * ((float)Math.Abs(scene.MaxPoint.X - scene.MinPoint.X)));
                    break;

                case Keys.Escape:
                    base.StopRendering();
                    break;

                case Keys.F:
                    SwitchToFakeFullScreen();
                    break;

                //case Keys.R:
                //    _tSaver.RenderTotexture = !_tSaver.RenderTotexture;
                //    break;
            }
        }

        #region Mouse events

        private void _window_MouseWheel(object sender, MouseEventArgs e)
        {
            //float mult = (e.Delta / 120.0f) * 0.01f;
            //cameraDistance += ((float)Math.Abs(scene.MaxPoint.Z - scene.MinPoint.Z) * mult);

            cameraDistance += (e.Delta / 120.0f) * 0.02f * cameraDistance;

            if (cameraDistance < 0.0f) cameraDistance = 0;

            camera.UpdatePosition(mouseX, mouseY, cameraDistance);
        }

        private void _window_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                mouseCapture = false;
            }
        }

        private void _window_MouseMove(object sender, MouseEventArgs e)
        {
            UpdateSecondaryScreenPosition();

            if (mouseCapture)
            {
                int dx = e.X - startX;
                int dy = e.Y - startY;

                while (mouseY > 360) mouseY -= 360;
                while (mouseY < 0) mouseY += 360;

                if (mouseY > 90 && mouseY < 270) mouseX += dx;
                else mouseX -= dx;

                mouseY += dy;

                startX = e.X;
                startY = e.Y;

                camera.UpdatePosition(mouseX, mouseY, cameraDistance);
            }
        }

        private void _window_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                mouseCapture = true;

                startX = e.X;
                startY = e.Y;
            }
        }

        #endregion

        private void SwitchToFakeFullScreen()
        {
            if (_window.WindowState == FormWindowState.Normal)
            {
                _window.FormBorderStyle = FormBorderStyle.None;
                secWindow.FormBorderStyle = FormBorderStyle.None;

                rect = new System.Drawing.Rectangle(_window.Location, _window.ClientSize);

                _window.WindowState = FormWindowState.Maximized;
                secWindow.WindowState = FormWindowState.Maximized;

                base._runtime.DeviceResize(Screen.AllScreens[0].Bounds.Size);

                _window.Activate();

                return;
            }

            if (_window.WindowState == FormWindowState.Maximized)
            {
                _window.FormBorderStyle = FormBorderStyle.Sizable;
                secWindow.FormBorderStyle = FormBorderStyle.Sizable;

                _window.WindowState = FormWindowState.Normal;
                secWindow.WindowState = FormWindowState.Normal;

                base._runtime.DeviceResize(rect.Size);

                _window.Activate();
            }
        }

        /// <summary>
        /// Avoid screen overreach into other screen.
        /// </summary>
        private void UpdateSecondaryScreenPosition()
        {
            System.Drawing.Rectangle pRect;

            // if maximized then retur into normal size
            if (_window.WindowState == FormWindowState.Maximized)
                pRect = _window.RestoreBounds;
            else
                pRect = _window.Bounds;

            //no primary window overreach
            if (pRect.X + pRect.Width > primaryRect.Width)
            {
                pRect.X = Math.Max(0, primaryRect.Width - pRect.Width);
                pRect.Width = Math.Min(primaryRect.Width - 1, pRect.Width);
            }

            // no secondary window overreach
            if (pRect.X < 0) { pRect.X = 0; }

            // primary window size
            _window.Bounds = new System.Drawing.Rectangle(pRect.Location, pRect.Size);

            // secondary window size
            secWindow.Bounds = new System.Drawing.Rectangle(
                primaryRect.Width + _window.Bounds.X, _window.Bounds.Y, _window.Width, _window.Height);
        }
    }
}
