using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.ComponentModel;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// Test module. Generates triangle mesh that is changing in real time.
  /// </summary>
  [ModuleInfo("Jan Kaiser", "Test module. Generates triangle mesh that is changing in real time.")]
  [PortInfo("outTriangleMesh", "Resulting triangle mesh.")]
  public class TestModuleFunnyFunction : Module
  {
    private BackgroundWorker backgroundWorker;

    private bool swap;
    private TriangleMesh triangleMesh;
    private TriangleMesh tmpTriangleMesh;
    private TriangleMesh tmpTriangleMesh1;
    private TriangleMesh tmpTriangleMesh2;

    private Random random = new Random();
    private int lastTime;
    private float[,] lastValues;

    // For statistics...
    //private int st;
    //private int st2;

    private int updateDelay;
    /// <summary>
    /// Delay between updates.
    /// </summary>
    public int UpdateDelay
    {
      get
      {
        return updateDelay;
      }
      set
      {
        updateDelay = value;
      }
    }

    private int gridSize;
    /// <summary>
    /// Size of generated grid.
    /// </summary>
    public int GridSize
    {
      get
      {
        return gridSize;
      }
      set
      {
        gridSize = value;
        ResetLastValues();
      }
    }

    /// <summary>
    /// Constructor, initializes current module.
    /// </summary>
    public TestModuleFunnyFunction()
    {
      this.gridSize = 40;
      this.updateDelay = 150;

      AddOutPort("outTriangleMesh", typeof(TriangleMesh));

      this.SimulationStarted += new SimulationStartedEventHandler(TestModuleFunnyFunction_SimulationStarted);

      backgroundWorker = new BackgroundWorker();
      backgroundWorker.WorkerSupportsCancellation = false;
      backgroundWorker.DoWork += new DoWorkEventHandler(backgroundWorker_DoWork);

      lastTime = System.Environment.TickCount;
    }

    private void backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
    {
      //st = System.Environment.TickCount;
      if (swap)
        tmpTriangleMesh = UpdateFunnyFunction(tmpTriangleMesh2);
      else
        tmpTriangleMesh = UpdateFunnyFunction(tmpTriangleMesh1);

      swap = !swap;
      //Console.WriteLine("--UpdateTime>> " + (System.Environment.TickCount - st));
    }

    private void TestModuleFunnyFunction_SimulationStarted(object source)
    {
      ResetLastValues();
      tmpTriangleMesh = UpdateFunnyFunction(tmpTriangleMesh1);
      swap = true;
    }

    /// <summary>
    /// Exposes generated triangle mesh on out port.
    /// </summary>
    public override void Execute()
    {
      if ((System.Environment.TickCount - lastTime) > updateDelay)
      {
        lastTime = System.Environment.TickCount;
        //st2 = System.Environment.TickCount;
        
        while (backgroundWorker.IsBusy)
          System.Threading.Thread.Sleep(0);

        triangleMesh = tmpTriangleMesh;
        // Console.WriteLine("--BusyTime>> " + (System.Environment.TickCount - st2));

        SetOutput("outTriangleMesh", triangleMesh, false);

        //st2 = System.Environment.TickCount;
        backgroundWorker.RunWorkerAsync();
        //Console.WriteLine("-RWA>> " + (System.Environment.TickCount - st2));

      }
      else
        SetOutput("outTriangleMesh", triangleMesh);
    }

    private void ResetLastValues()
    {
      lastValues = new float[GridSize, GridSize];

      for (int i = 0; i < GridSize; i++)
        for (int j = 0; j < GridSize; j++)
          lastValues[i, j] = 0;

      tmpTriangleMesh1 = new TriangleMesh();
      tmpTriangleMesh1.Points = new UniformDataArray(typeof(Point3D), GridSize * GridSize);
      tmpTriangleMesh1.AddPointAttrs("ColorRGBA", new UniformDataArray(typeof(ColorRGBA), tmpTriangleMesh1.Points.Count));
      tmpTriangleMesh1.AddCells("Triangle", new UniformDataArray(typeof(Triangle), 2 * (GridSize - 1) * (GridSize - 1)));

      tmpTriangleMesh2 = new TriangleMesh();
      tmpTriangleMesh2.Points = new UniformDataArray(typeof(Point3D), GridSize * GridSize);
      tmpTriangleMesh2.AddPointAttrs("ColorRGBA", new UniformDataArray(typeof(ColorRGBA), tmpTriangleMesh2.Points.Count));
      tmpTriangleMesh2.AddCells("Triangle", new UniformDataArray(typeof(Triangle), 2 * (GridSize - 1) * (GridSize - 1)));
    }

    private TriangleMesh UpdateFunnyFunction(TriangleMesh triangleMesh)
    {
      //st = System.Environment.TickCount;

      for (int i = (GridSize - 1); i > 0; i--)
        for (int j = 0; j < GridSize; j++)
          lastValues[i, j] = lastValues[i - 1, j];

      //Console.WriteLine(">>lastValuesMove>>  \t" + (System.Environment.TickCount - st));
      //st = System.Environment.TickCount;
      System.Threading.Thread.Sleep(0);

      float rnd1 = (float) random.NextDouble();
      float rnd2 = (float) random.NextDouble();

      for (int j = 0; j < GridSize; j++)
      {

        lastValues[0, j] = 4 * (0.15f * lastValues[1, j] + (float) Math.Cos((rnd1 * Math.PI) + (rnd2 * 16 * Math.PI / (float) GridSize) * j) / 50f);

        if (lastValues[0, j] > 1)
          lastValues[0, j] -= 0.1f;

        if (lastValues[0, j] < -1)
          lastValues[0, j] += 0.1f;
      }

      //Console.WriteLine(">>lastValuesNewGen>>  \t" + (System.Environment.TickCount - st));
      //st = System.Environment.TickCount;
      System.Threading.Thread.Sleep(0);

      UniformDataArray points = triangleMesh.Points;
      for (int i = 0; i < GridSize; i++)
      {
        for (int j = 0; j < GridSize; j++)
          points[(i * GridSize) + j] = new Point3D((2 * ((float) i - (GridSize / 2f))) / GridSize, lastValues[i, j], (2 * (j - ((float) GridSize / 2f))) / GridSize);

        System.Threading.Thread.Sleep(0);
      }

      //Console.WriteLine(">>addPoints>>  \t\t" + (System.Environment.TickCount - st));
      //st = System.Environment.TickCount;

      UniformDataArray pointAttributes = triangleMesh.GetPointAttrs("ColorRGBA");
      for (int i = 0; i < points.Count; i++)
      {
        Point3D p = (Point3D) points[i];
        pointAttributes[i] = new ColorRGBA((float) (p.X + 1) / 2f, (float) (4 * p.Y + 1) / 2f, (float) (p.Z + 1) / 2f, 0.5f);
      }

      System.Threading.Thread.Sleep(0);
      //Console.WriteLine(">>addPointAttributes>>  \t" + (System.Environment.TickCount - st));
      //st = System.Environment.TickCount;

      UniformDataArray cells = triangleMesh.GetCells("Triangle");
      for (int i = 0; i < GridSize - 1; i++)
      {
        for (int j = 0; j < GridSize - 1; j++)
        {
          cells[(i * 2 * (GridSize - 1)) + (2 * j)] = new Triangle((i * GridSize) + j, ((i + 1) * GridSize) + j + 1, ((i + 1) * GridSize) + j);
          cells[(i * 2 * (GridSize - 1)) + (2 * j) + 1] = new Triangle((i * GridSize) + j, (i * GridSize) + j + 1, ((i + 1) * GridSize) + j + 1);
        }

        System.Threading.Thread.Sleep(0);
      }
      //Console.WriteLine(">>addCells>> \t\t" + (System.Environment.TickCount - st));

      return triangleMesh;
    }
  }
}
