using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;


using System.Xml;
using System.Diagnostics;
using System.Drawing.Design;


namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// structure used to read RGBA from texture memory
    /// </summary>
    public struct pix
    {
        public byte B;
        public byte G;
        public byte R;
        public byte A;
    }

    /// <summary>
    /// class derived from texture object, designed to be able to load data from given volume
    /// </summary>
    class TextureVolCut : TextureObject
    {
        //volume size
        private int _size;
        private int _sizeX, _sizeY, _sizeZ;
        //volume data
        private byte[] _data;
        //cut plane location
        private int _cutValue;
        //cut plane orientation
        private AxisEnum _direction;

        private bool rgbtable;
        //false if color table should be used
        public bool Rgbtable
        {
            get { return rgbtable; }
            set { rgbtable = value; }
        }

        private ColorMappingFunction mapFce;
        /// <summary>
        /// extern color mapping function 
        /// </summary>
        internal ColorMappingFunction MapFce
        {
            set
            {
                mapFce = value;
                Createtable();
            }
        }

        /// <summary>
        /// color table - RGB for 256 values
        /// </summary>
        private pix[] colortable;


        /// <summary>
        /// </summary>
        /// <param name="sizeX">volume size</param>
        /// <param name="sizeY">volume size</param>
        /// <param name="sizeZ">volume size</param>
        /// <param name="data">volume data</param>
        /// <param name="cutValue">plane postion</param>
        /// <param name="direction">plane orientation</param>
        public TextureVolCut(int sizeX, int sizeY, int sizeZ, byte[] data, int cutValue, AxisEnum direction)
        {
            path = null;
            _direction = direction;
            _size = sizeX;
            _data = data;
            _cutValue = cutValue;
            _device = null;
            mapFce = null;

            _sizeX = sizeX;
            _sizeY = sizeY;
            _sizeZ = sizeZ;

            Createtable();
        }

        /// <summary>
        /// compute color table
        /// </summary>
        private void Createtable()
        {
            colortable = new pix[256];
            bool tmp = rgbtable;
            rgbtable = false;

            if (mapFce == null)
            {
                for (int i = 0; i < 256; i++)
                {
                    colortable[i] = hsvtorgb(360-(60 + 300 * (i / 256.0f)), (float)(Math.Sqrt(i / 256.0f)), (float)(Math.Sqrt(i / 256.0f)));
                  // 
                }
            }
            else
            {
                ColorRGBA c;
                for (int i = 0; i < 256; i++)
                {
                    c = mapFce.GetColor(i/255.0f);
                    colortable[i].R = c.Red;
                    colortable[i].G = c.Green;
                    colortable[i].B = c.Blue;
                }
            }

            rgbtable = tmp;
        }

        /// <summary>
        /// set device events
        /// </summary>
        public override void Update(Device device)
        {
            if (path == null && _device ==null)
            { 
               // _device = device;
                _device_DeviceReset(device, null); 
                
                _device.DeviceLost += new EventHandler(_device_DeviceLost);
             _device.DeviceReset += new EventHandler(_device_DeviceReset);
            }
            Createtable();


        }

        /// <summary>
        /// reset texture
        /// </summary>
        void _device_DeviceReset(object sender, EventArgs e)
        {
            if (texture != null) texture.Dispose();

            _device = (Device)sender;

            switch (_direction)
            {
                case AxisEnum.X:
                    texture = new Texture(_device, _sizeY, _sizeZ, 1, Usage.Dynamic, Format.A8R8G8B8, Pool.Default);
                    break;
                case AxisEnum.Y:
                    texture = new Texture(_device, _sizeX, _sizeZ, 1, Usage.Dynamic, Format.A8R8G8B8, Pool.Default);
                    break;
                case AxisEnum.Z:
                    texture = new Texture(_device, _sizeX, _sizeY, 1, Usage.Dynamic, Format.A8R8G8B8, Pool.Default);
                    break;
            }
            UpdateData();

        }

        /// <summary>
        /// dispose texture
        /// </summary>
        void _device_DeviceLost(object sender, EventArgs e)
        {
            if(texture != null &&!texture.Disposed) texture.Dispose();
            texture = null;
            _device = null;
        }

        /// <summary>
        /// update texture according to cut value
        /// </summary>
        private void UpdateData()
        {
            UpdateData( _data, _cutValue, _direction);
        }

        /// <summary>
        /// update texture according to cut value
        /// </summary>
        public void UpdateData( byte[] data, int cutValue, AxisEnum direction)
        {
            _direction = direction;
            _cutValue = cutValue;




            unsafe
            {
                //lock texture
                GraphicsStream g = texture.LockRectangle(0, LockFlags.NoDirtyUpdate);
                pix* p = (pix*)g.InternalDataPointer;
                pix color;

                switch (_direction)
                {
                    case AxisEnum.X:
                        {
                            long step = _sizeY * _sizeZ;
                            long indexIn = cutValue;
                            
                            for (int i = 0; i < step; i++)
                            {
                                color = GetColor(data[indexIn]);

                                p[i].R = color.R;
                                p[i].G = color.G;
                                p[i].B = color.B;

                                indexIn += _sizeY;
                            }
                            break;
                        }
                    case AxisEnum.Y:
                        {
                            long step = _sizeX * _sizeY;
                            long indexIn =  (_sizeY - cutValue) * _sizeX;
                            long indexOut = 0;


                            for (int i = 0; i < _sizeZ; i++)
                            {
                                for (int j = 0; j < _sizeX; j++)
                                {
                                    color = GetColor(data[indexIn + j]);

                                    p[indexOut].R = color.R;
                                    p[indexOut].G = color.G;
                                    p[indexOut].B = color.B;

                                    indexOut++;
                                }
                            //    indexOut += 256 - _sizeZ;
                                indexIn += step;

                            }
                            break;
                        }
                    case AxisEnum.Z:
                        {
                            long step = _sizeX * _sizeY;
                            long sc = cutValue * step;
                            
                            for (int i = 0; i < step; i++)
                            {
                                color = GetColor(data[i + sc]);

                                p[i].R = color.R;
                                p[i].G = color.G;
                                p[i].B = color.B;

                            }
                            break;
                        }
                }
                //unlock texture 
                texture.UnlockRectangle(0);
            }
        }

        /// <summary>
        /// return value from color table (if it is enabled)
        /// </summary>
        private pix GetColor(byte value)
        {
            pix color;
            
            color.A = 1;

            if (rgbtable)
            {
                color.R = value;
                color.G = value;
                color.B = value;
            }
            else
            {
                color = colortable[value];
            }

            return color;
        }

        /// <summary>
        /// convert HSV to RGBA
        /// </summary>
        private pix hsvtorgb(float h, float s, float v)
        {
            pix color;

            color.A = 1;  
            color.R = 1;
            color.G = 1;
            color.B = 1;

            float hi = ((int)(h / 60)) % 60;
            float f = h / 60.0f - hi;
            float p = v * (1 - s);
            float q = v * (1 - f * s);
            float t = v * (1 - (1 - f) * s);

            if (hi == 0)
            {
                color.R = (byte)(255 * v);
                color.G = (byte)(255 * t);
                color.B = (byte)(255 * p);
            }
            else if (hi == 1)
            {
                color.R = (byte)(255 * q);
                color.G = (byte)(255 * v);
                color.B = (byte)(255 * p);
            }
            else if (hi == 2)
            {
                color.R = (byte)(255 * p);
                color.G = (byte)(255 * v);
                color.B = (byte)(255 * t);
            }
            else if (hi == 3)
            {
                color.R = (byte)(255 * p);
                color.G = (byte)(255 * q);
                color.B = (byte)(255 * v);
            }
            else if (hi == 4)
            {
                color.R = (byte)(255 * t);
                color.G = (byte)(255 * p);
                color.B = (byte)(255 * v);
            }
            else
            {
                color.R = (byte)(255 * v);
                color.G = (byte)(255 * p);
                color.B = (byte)(255 * q);
            }

            return color;

        }

    }
}
