using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Zcu.Mve.Visualization;
using System.Diagnostics;

namespace Zcu.Mve.UnstrGridRendering
{
  /// <summary>
  /// This is a helper class for D3DUTSceneObject. Its purpose is to create and maintain D3DUT VertexBuffer.
  /// </summary>
  public class VertexBufferWrapper : IDisposable  
  {
    private VertexBuffer vertexBuffer;
    private SceneObject sceneObject;
    private Device device;
    private VertexFormats vertexFormat;
    
    private int count;
    private UnstrGrid unstructuredGrid;

    private long lmtPoints;
    private long lmtTextureCoords;
    private long lmtColor;
    private long lmtNormal;

    /// <summary>
    /// Count of vertices stored in vertex buffer.
    /// </summary>
    public int Count
    {
      get { return count; }
    }

    /// <summary>
    /// Default constructor. Initializes the object.
    /// </summary>
    public VertexBufferWrapper()
    {
      this.lmtPoints = 0;
      this.lmtTextureCoords = 0;
      this.lmtColor = 0;
      this.lmtNormal = 0;
    }

    /// <summary>
    /// This method should be called during rendering between device.BeginScene() and device.EndScene().
    /// It assigns vertices to device. 
    /// </summary>
    public void SetVertexBuffer()
    {
      if (vertexBuffer != null)
      {
        device.VertexFormat = vertexFormat;
        device.SetStreamSource(0, vertexBuffer, 0);
      }
    }

    /// <summary>
    /// Update vertex buffer.
    /// </summary>
    /// <param name="sceneObject">Scene object.</param>
    /// <param name="device">Rendering device.</param>
    public void Update(Device device, SceneObject sceneObject)
    {
      this.device = device;
      this.sceneObject = sceneObject;

      if (sceneObject.UnstructuredGrid.Points == null)
        return;

      this.count = sceneObject.UnstructuredGrid.Points.Count;
      this.unstructuredGrid = sceneObject.UnstructuredGrid;

      if (this.count > ushort.MaxValue)
      {
        Debug.WriteLine("[!] Number of vertices (" + this.count + ") in VertexBuffer is greater than " + ushort.MaxValue + " what should cause problems with rendering on lowend and onboard graphics accelerators.");
      }

      switch (sceneObject.PrefferedVertexAttributes)
      {
        case VertexAttributes.None:
          updatePositionOnlyVB();
          break;

        case VertexAttributes.ColorRGBA:
          if (unstructuredGrid.ContainsPointAttrs(sceneObject.ColorAttributeName))
            updatePositionColoredVB();
          else
          {
            Console.WriteLine(sceneObject.ToString() + "[!] Point attributes ColorRGBA were not found, using Position only.");
            updatePositionOnlyVB();
          }
          break;

        case VertexAttributes.Normal3D:
          if (unstructuredGrid.ContainsPointAttrs(sceneObject.NormalAttributeName))
            updatePositionNormalVB();
          else
          {
            Console.WriteLine(sceneObject.ToString() + "[!] Point attributes Normal3D were not found, using Position only.");
            updatePositionOnlyVB();
          }
          break;

        case VertexAttributes.Normal3DColorRGBA:
          if (unstructuredGrid.ContainsPointAttrs(sceneObject.NormalAttributeName))
          {
            if (unstructuredGrid.ContainsPointAttrs(sceneObject.ColorAttributeName))
              updatePositionNormalColoredVB();
            else
            {
              Console.WriteLine(sceneObject.ToString() + "[!] Point attributes ColorRGBA were not found, using Position and Normal3D only.");
              updatePositionNormalVB();
            }
          }
          else
          {
            if (unstructuredGrid.ContainsPointAttrs(sceneObject.ColorAttributeName))
            {
              Console.WriteLine(sceneObject.ToString() + "[!] Point attributes Normal3D were not found, using Position and ColorRGBA only.");
              updatePositionColoredVB();
            }
            else
            {
              Console.WriteLine(sceneObject.ToString() + "[!] Point attributes Normal3D and ColorRGBA were not found, using Position only.");
              updatePositionOnlyVB();      
            }
          }
          break;

        case VertexAttributes.TextureCoord:
          if (unstructuredGrid.ContainsPointAttrs(sceneObject.TextureAttributeName))
            updatePositionTexturedVB();
          else
          {
            Console.WriteLine(sceneObject.ToString() + "[!] Point attributes TextureCoord were not found, using Position only.");
            updatePositionOnlyVB();
          }
          break;

        case VertexAttributes.Normal3DTextureCoord:
          if (unstructuredGrid.ContainsPointAttrs(sceneObject.NormalAttributeName))
          {
            if (unstructuredGrid.ContainsPointAttrs(sceneObject.TextureAttributeName))
              updatePositionNormalTexturedVB();
            else
            {
              Console.WriteLine(sceneObject.ToString() + "[!] Point attributes TextureCoord were not found, using Position and Normal3D only.");
              updatePositionNormalVB();
            }
          }
          else
          {
            if (unstructuredGrid.ContainsPointAttrs(sceneObject.TextureAttributeName))
            {
              Console.WriteLine(sceneObject.ToString() + "[!] Point attributes Normal3D were not found, using Position and TextureCoord only.");
              updatePositionTexturedVB();
            }
            else
            {
              Console.WriteLine(sceneObject.ToString() + "[!] Point attributes Normal3D and TexturedUV were not found, using Position only.");
              updatePositionOnlyVB();
            }
          }
          break; 

        case VertexAttributes.ColorRGBATextureCoord:
          if (unstructuredGrid.ContainsPointAttrs(sceneObject.ColorAttributeName))
          {
            if (unstructuredGrid.ContainsPointAttrs(sceneObject.TextureAttributeName))
              updatePositionColoredTexturedVB();
            else
            {
              Console.WriteLine(sceneObject.ToString() + "[!] Point attributes TextureCoord were not found, using Position and ColorRGBA only.");
              updatePositionColoredVB();
            }
          }
          else
          {
            if (unstructuredGrid.ContainsPointAttrs(sceneObject.TextureAttributeName))
            {
              Console.WriteLine(sceneObject.ToString() + "[!] Point attributes ColorRGBA were not found, using Position and TextureCoord only.");
              updatePositionTexturedVB();
            }
            else
            {
              Console.WriteLine(sceneObject.ToString() + "[!] Point attributes ColorRGBA and TexturedUV were not found, using Position only.");
              updatePositionOnlyVB();
            }
          }        
          break;

        default:
          Console.WriteLine(sceneObject.ToString() + "[!] " + sceneObject.PrefferedVertexAttributes + " not implemented yet, sorry.");
          break;
      }
    }

    private void updatePositionOnlyVB()
    {
      if (this.lmtPoints != this.unstructuredGrid.Points.ModificationTime)
      {
        createPositionOnlyVB();
        this.lmtPoints = unstructuredGrid.Points.ModificationTime;

        Debug.WriteLine(">VertexBuffer changed - position");
      }
    }

    private void createPositionOnlyVB()
    {
      vertexFormat = CustomVertex.PositionOnly.Format;

      vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionOnly), Count, device, Usage.WriteOnly, vertexFormat, Pool.Managed);

      CustomVertex.PositionOnly[] vertices = new CustomVertex.PositionOnly[Count];

      Point3D tmpPoint;

      for (int i = 0; i < Count; i++)
      {
        tmpPoint = (Point3D) unstructuredGrid.Points[i];
        vertices[i] = new CustomVertex.PositionOnly((float)tmpPoint.X, (float)tmpPoint.Y, (float)tmpPoint.Z);
      }

      vertexBuffer.SetData(vertices, 0, LockFlags.None);
    }

    private void updatePositionColoredVB()
    {
      UniformDataArray pointColors = unstructuredGrid.GetPointAttrs(sceneObject.ColorAttributeName);

      if ((this.lmtPoints != this.unstructuredGrid.Points.ModificationTime)
      || (this.lmtColor != pointColors.ModificationTime))
      {
        createPositionColoredVB();
        this.lmtPoints = unstructuredGrid.Points.ModificationTime;
        this.lmtColor = pointColors.ModificationTime;

        Debug.WriteLine(">VertexBuffer changed - position+colors");
      }
    }

    private void createPositionColoredVB()
    {
      vertexFormat = CustomVertex.PositionColored.Format;

      vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionColored), Count, device, Usage.WriteOnly, vertexFormat, Pool.Managed);

      CustomVertex.PositionColored[] vertices = new CustomVertex.PositionColored[Count];

      UniformDataArray pointColors = unstructuredGrid.GetPointAttrs(sceneObject.ColorAttributeName);

      Point3D tmpPoint;
      ColorRGBA tmpColor;

      for (int i = 0; i < Count; i++)
      {
        tmpPoint = (Point3D)unstructuredGrid.Points[i];
        tmpColor = (ColorRGBA)pointColors[i];

        vertices[i] = new CustomVertex.PositionColored((float)tmpPoint.X, (float)tmpPoint.Y, (float)tmpPoint.Z, tmpColor.GetInt());
      }

      vertexBuffer.SetData(vertices, 0, LockFlags.None);
    }

    private void updatePositionNormalVB()
    {
      UniformDataArray pointNormals = unstructuredGrid.GetPointAttrs(sceneObject.NormalAttributeName);

      if ((this.lmtPoints != this.unstructuredGrid.Points.ModificationTime)
      || (this.lmtNormal != pointNormals.ModificationTime))
      {
        createPositionNormalVB();
        this.lmtPoints = unstructuredGrid.Points.ModificationTime;
        this.lmtNormal = pointNormals.ModificationTime;

        Debug.WriteLine(">VertexBuffer changed - position+normals");
      }
    }

    private void createPositionNormalVB()
    {
      vertexFormat = CustomVertex.PositionNormal.Format;

      vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionNormal), Count, device, Usage.WriteOnly, vertexFormat, Pool.Managed);

      CustomVertex.PositionNormal[] vertices = new CustomVertex.PositionNormal[Count];

      UniformDataArray pointNormals = unstructuredGrid.GetPointAttrs(sceneObject.NormalAttributeName);

      Point3D tmpPoint;
      Normal3D tmpNormal;

      for (int i = 0; i < Count; i++)
      {
        tmpPoint = (Point3D) unstructuredGrid.Points[i];
        tmpNormal = (Normal3D) pointNormals[i];

        vertices[i] = new CustomVertex.PositionNormal((float) tmpPoint.X, (float) tmpPoint.Y, (float) tmpPoint.Z, 
                                                      (float) tmpNormal.Nx, (float) tmpNormal.Ny, (float) tmpNormal.Nz);
      }

      vertexBuffer.SetData(vertices, 0, LockFlags.None);    
    }

    private void updatePositionNormalColoredVB()
    {
      UniformDataArray pointNormals = unstructuredGrid.GetPointAttrs(sceneObject.NormalAttributeName);
      UniformDataArray pointColors = unstructuredGrid.GetPointAttrs(sceneObject.ColorAttributeName);

      if ((this.lmtPoints != this.unstructuredGrid.Points.ModificationTime)
      || (this.lmtNormal != pointNormals.ModificationTime)
      || (this.lmtColor != pointColors.ModificationTime))
      {
        createPositionNormalColoredVB();
        this.lmtPoints = unstructuredGrid.Points.ModificationTime;
        this.lmtNormal = pointNormals.ModificationTime;
        this.lmtColor = pointColors.ModificationTime;

        Debug.WriteLine(">VertexBuffer changed - position+normals+color");
      }
    }

    private void createPositionNormalColoredVB()
    {
      vertexFormat = CustomVertex.PositionNormalColored.Format;

      vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionNormalColored), Count, device, Usage.WriteOnly, vertexFormat, Pool.Managed);

      CustomVertex.PositionNormalColored[] vertices = new CustomVertex.PositionNormalColored[Count];

      UniformDataArray pointNormals = unstructuredGrid.GetPointAttrs(sceneObject.NormalAttributeName);
      UniformDataArray pointColors = unstructuredGrid.GetPointAttrs(sceneObject.ColorAttributeName);

      Point3D tmpPoint;
      Normal3D tmpNormal;
      ColorRGBA tmpColor;

      for (int i = 0; i < Count; i++)
      {
        tmpPoint = (Point3D) unstructuredGrid.Points[i];
        tmpNormal = (Normal3D) pointNormals[i];
        tmpColor = (ColorRGBA) pointColors[i];

        vertices[i] = new CustomVertex.PositionNormalColored((float) tmpPoint.X, (float) tmpPoint.Y, (float) tmpPoint.Z,
                                                             (float) tmpNormal.Nx, (float) tmpNormal.Ny, (float) tmpNormal.Nz, tmpColor.GetInt());
      }

      vertexBuffer.SetData(vertices, 0, LockFlags.None); 
    }

    private void updatePositionTexturedVB()
    {
      UniformDataArray pointTextureCoordinates = unstructuredGrid.GetPointAttrs(sceneObject.TextureAttributeName);

      if ((this.lmtPoints != this.unstructuredGrid.Points.ModificationTime)
      || (this.lmtTextureCoords != pointTextureCoordinates.ModificationTime))
      {
        createPositionTexturedVB();
        this.lmtPoints = unstructuredGrid.Points.ModificationTime;
        this.lmtTextureCoords = pointTextureCoordinates.ModificationTime;

        Debug.WriteLine(">VertexBuffer changed - position+texture");
      }
    }

    private void createPositionTexturedVB()
    {
      vertexFormat = CustomVertex.PositionTextured.Format;

      vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionTextured), Count, device, Usage.WriteOnly, vertexFormat, Pool.Managed);

      CustomVertex.PositionTextured[] vertices = new CustomVertex.PositionTextured[Count];
      UniformDataArray pointTextureCoordinates = unstructuredGrid.GetPointAttrs(sceneObject.TextureAttributeName);

      Point3D tmpPoint;
      TextureCoord tmpTextureCoord;

      for (int i = 0; i < Count; i++)
      {
        tmpPoint = (Point3D) unstructuredGrid.Points[i];
        tmpTextureCoord = (TextureCoord) pointTextureCoordinates[i];

        vertices[i] = new CustomVertex.PositionTextured((float) tmpPoint.X, (float) tmpPoint.Y, (float) tmpPoint.Z,
                                                        (float) tmpTextureCoord.U, (float) tmpTextureCoord.V);
      }

      vertexBuffer.SetData(vertices, 0, LockFlags.None);
    }

    private void updatePositionNormalTexturedVB()
    {
      UniformDataArray pointNormals = unstructuredGrid.GetPointAttrs(sceneObject.NormalAttributeName);
      UniformDataArray pointTextureCoordinates = unstructuredGrid.GetPointAttrs(sceneObject.TextureAttributeName);

      if ((this.lmtPoints != this.unstructuredGrid.Points.ModificationTime)
      || (this.lmtNormal != pointNormals.ModificationTime)
      || (this.lmtTextureCoords != pointTextureCoordinates.ModificationTime))
      {
        createPositionNormalTexturedVB();
        this.lmtPoints = unstructuredGrid.Points.ModificationTime;
        this.lmtNormal = pointNormals.ModificationTime;
        this.lmtTextureCoords = pointTextureCoordinates.ModificationTime;

        Debug.WriteLine(">VertexBuffer changed - position+normals+texture");
      }
    }

    private void createPositionNormalTexturedVB()
    {
      vertexFormat = CustomVertex.PositionNormalTextured.Format;

      vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionNormalTextured), Count, device, Usage.WriteOnly, vertexFormat, Pool.Managed);

      CustomVertex.PositionNormalTextured[] vertices = new CustomVertex.PositionNormalTextured[Count];
      UniformDataArray pointNormals = unstructuredGrid.GetPointAttrs(sceneObject.NormalAttributeName);
      UniformDataArray pointTextureCoordinates = unstructuredGrid.GetPointAttrs(sceneObject.TextureAttributeName);

      Point3D tmpPoint;
      Normal3D tmpNormal;
      TextureCoord tmpTextureCoord;

      for (int i = 0; i < Count; i++)
      {
        tmpPoint = (Point3D) unstructuredGrid.Points[i];
        tmpNormal = (Normal3D) pointNormals[i];
        tmpTextureCoord = (TextureCoord) pointTextureCoordinates[i];

        vertices[i] = new CustomVertex.PositionNormalTextured((float) tmpPoint.X, (float) tmpPoint.Y, (float) tmpPoint.Z,
                                                              (float) tmpNormal.Nx, (float) tmpNormal.Ny, (float) tmpNormal.Nz,
                                                              (float) tmpTextureCoord.U, (float) tmpTextureCoord.V);
      }

      vertexBuffer.SetData(vertices, 0, LockFlags.None);
    }

    private void updatePositionColoredTexturedVB()
    {
      UniformDataArray pointColors = unstructuredGrid.GetPointAttrs(sceneObject.ColorAttributeName);
      UniformDataArray pointTextureCoordinates = unstructuredGrid.GetPointAttrs(sceneObject.TextureAttributeName);

      if ((this.lmtPoints != this.unstructuredGrid.Points.ModificationTime)
      || (this.lmtColor != pointColors.ModificationTime)
      || (this.lmtTextureCoords != pointTextureCoordinates.ModificationTime))
      {
        createPositionColoredTexturedVB();
        this.lmtPoints = unstructuredGrid.Points.ModificationTime;
        this.lmtColor = pointColors.ModificationTime;
        this.lmtTextureCoords = pointTextureCoordinates.ModificationTime;

        Debug.WriteLine(">VertexBuffer changed - position+colors+texture");
      }
    }

    private void createPositionColoredTexturedVB()
    {
      vertexFormat = CustomVertex.PositionColoredTextured.Format;

      vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionColoredTextured), Count, device, Usage.WriteOnly, vertexFormat, Pool.Managed);

      CustomVertex.PositionColoredTextured[] vertices = new CustomVertex.PositionColoredTextured[Count];
      UniformDataArray pointColors = unstructuredGrid.GetPointAttrs(sceneObject.ColorAttributeName);
      UniformDataArray pointTextureCoordinates = unstructuredGrid.GetPointAttrs(sceneObject.TextureAttributeName);

      Point3D tmpPoint;
      ColorRGBA tmpColor;
      TextureCoord tmpTextureCoord;

      for (int i = 0; i < Count; i++)
      {
        tmpPoint = (Point3D) unstructuredGrid.Points[i];
        tmpColor = (ColorRGBA) pointColors[i];
        tmpTextureCoord = (TextureCoord) pointTextureCoordinates[i];

        vertices[i] = new CustomVertex.PositionColoredTextured((float) tmpPoint.X, (float) tmpPoint.Y, (float) tmpPoint.Z,
                                                               tmpColor.GetInt(),
                                                               (float) tmpTextureCoord.U, (float) tmpTextureCoord.V);
      }

      vertexBuffer.SetData(vertices, 0, LockFlags.None);
    }

    #region IDisposable Members

    /// <summary>
    /// Dispose vertex buffer.
    /// </summary>
    public void Dispose()
    {
      vertexBuffer.Dispose();
    }

    #endregion
  }
}
