using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.ComponentModel;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Windows.Forms;

using System.Xml;
using System.Diagnostics;
using System.Drawing.Design;
using System.Globalization;
using System.Threading;


namespace Zcu.Mve.UnstrGridRendering
{
    /// <summary>
    /// cutting plane orientation
    /// </summary>
    public enum AxisEnum { X = 1, Y = 2, Z = 3 };
    /// <summary>
    /// some color configuraions 
    /// </summary>
    public enum ColorEnum { GreyScale = 1, BlueToRed = 2, Extern = 3};


    /// <summary>
    /// cut volume in X,Y or Z direction and save output into texture
    /// </summary>
    [ModuleInfo("Vojtech Hladik", "cut volume in X,Y or Z direction and save output into texture")]
    [PortInfo("ColorFunction", "module which encapsulates color mapping function")]
    class VolumeCutter: Module
    {

        ///plane location
        private int _zLocation;
        private int _size, _sizeX, _sizeY, _sizeZ;

        ///class which contains mouse and keyboard events
        private BasePicker _picker;
        ///quad to which should output texture mapped
        private Zcu.Mve.Visualization.Mesh mesh;
        ///volume data
        private byte[] _data;
        ///class which encapsulate DX texture and methods to work with it
        private TextureVolCut _texture;
        private AxisEnum axis;
        private bool change, changeDone;
        private Keys moveUp;
        private Keys moveDown;
        //color function
        private ColorEnum colorMode;
        private bool exterColorFce;

        private RegGridND grid;

        [Category("color"),
        Description("color mapping function")]
        public ColorEnum ColorMode
        {
            get { return colorMode; }
            set
            {
                colorMode = value;
                if (colorMode == ColorEnum.Extern)
                {
                    AddInPort("ColorFunction", typeof(ColorMappingFunction));
                    exterColorFce = true;
                }
                else if (exterColorFce == true)
                {
                    RemoveInPort("ColorFunction");
                    exterColorFce = false;
                }
                
            }
        }

        [Category("control"),
        Description("key to move cutting plane up")]
        public Keys MoveUp
        {
            get { return moveUp; }
            set { moveUp = value; }
        }

        [Category("control"),
        Description("key to move cutting plane down")]
        public Keys MoveDown
        {
            get { return moveDown; }
            set { moveDown = value; }
        }

        [Category("control"),
        Description("cutting plane direction")]
        public AxisEnum Axis
        {
            get { return axis; }
            set { axis = value; }
        }


        public VolumeCutter()
        {
            AddOutPort("quad", typeof(Mve.Visualization.Mesh));
            AddOutPort("texture", typeof(TextureVolCut));

            AddInPort("picker", typeof(BasePicker));
            AddInPort("data", typeof(Mve.Visualization.RegGridND));

            DataReady += new DataReadyEventHandler(VolumeCutter_DataReady);

            change = false;
            changeDone = true;
            exterColorFce = false;
       
            //palce plane in the cente of volume
            _zLocation = 5;

            //default settings
            axis = AxisEnum.Z;
            moveUp = Keys.Up;
            moveDown = Keys.Down;
            colorMode = ColorEnum.BlueToRed;

            this.SimulationStarted += new SimulationStartedEventHandler(VolumeCutter_SimulationStarted);

        }

        void VolumeCutter_SimulationStarted(object source)
        {
            _texture = null;
        }

        public override void Execute()
        {
            if (_texture != null && changeDone == false)
            {

                SetOutput("quad", mesh, true);
                changeDone = true;
            }
            //first time runnig
            if (_texture == null)
            {
                grid = (RegGridND)GetInput("data");
                _data = VolumeLoad();
                InitUNSTRGrid();


                _texture = new TextureVolCut(_sizeX,_sizeY, _sizeZ, _data, _zLocation, axis);
                if (exterColorFce)
                    _texture.MapFce = (ColorMappingFunction)GetInput("ColorFunction");

                if (colorMode == ColorEnum.GreyScale) _texture.Rgbtable = true;
                else _texture.Rgbtable = false;

                SetOutput("texture", _texture, false);
                SetOutput("quad", mesh, false);
                changeDone = false;
            }
            else if (change)
            {

                SetOutput("quad", mesh, !change);
                change = false;
                changeDone = false;
            }
        }

        void VolumeCutter_DataReady(object source, InPort inPort)
        {
            if (inPort.Name == "picker")
            {
                _picker = (BasePicker)GetInput("picker");
                _picker.KeyDown += new KeyEventHandler(picker_KeyDown);
            }
        }

        /// <summary>
        /// move cutting plane
        /// </summary>
        void picker_KeyDown(object sender, System.Windows.Forms.KeyEventArgs e)
        {
            if (e.KeyCode == moveUp)
            {
                if (_zLocation < _size-2)
                {
                    _zLocation++; 
                    ReloadData();   //compute new texture and quad       
                    change = true;  //data has been changed
                }
            }
            if (e.KeyCode == moveDown)
            {
                if (_zLocation > 1)
                {
                    _zLocation--;
                    change = true;  //data has been changed
                    ReloadData();   //compute new texture and quad
                }
                
            }
        }

        /// <summary>
        /// init quad to which should be texture mapped
        /// </summary>
        private void InitUNSTRGrid()
        {
            //cells - two triagles
            Zcu.Mve.Visualization.UniformDataArray cellArray
                 = new Zcu.Mve.Visualization.UniformDataArray(typeof(Zcu.Mve.Visualization.Triangle), 2);

            //six vertices
            Zcu.Mve.Visualization.UniformDataArray verticesArray
                = new Zcu.Mve.Visualization.UniformDataArray(typeof(Zcu.Mve.Visualization.Point3D), 6);

            //six texture coordinates
            Zcu.Mve.Visualization.UniformDataArray verticesTexCoord
                = new Zcu.Mve.Visualization.UniformDataArray(typeof(Zcu.Mve.Visualization.TextureCoord), 6);

            //quad location
            float _height = 2*(_zLocation / (float)_size) -1;

            // guad vertices
            if (axis == AxisEnum.X)
            {
                verticesArray[0] = new Zcu.Mve.Visualization.Point3D(_height, -1, 1);
                verticesArray[1] = new Zcu.Mve.Visualization.Point3D(_height,  1, 1);
                verticesArray[2] = new Zcu.Mve.Visualization.Point3D(_height, -1, -1);

                verticesArray[3] = new Zcu.Mve.Visualization.Point3D(_height,  1, -1);
                verticesArray[4] = new Zcu.Mve.Visualization.Point3D(_height,  1,  1);
                verticesArray[5] = new Zcu.Mve.Visualization.Point3D(_height, -1, -1);
            }

            if (axis == AxisEnum.Y)
            {
                verticesArray[0] = new Zcu.Mve.Visualization.Point3D(1, _height, 1);
                verticesArray[1] = new Zcu.Mve.Visualization.Point3D(-1, _height, 1);
                verticesArray[2] = new Zcu.Mve.Visualization.Point3D(1, _height, -1);

                verticesArray[3] = new Zcu.Mve.Visualization.Point3D(-1, _height, -1);
                verticesArray[4] = new Zcu.Mve.Visualization.Point3D(-1, _height, 1);
                verticesArray[5] = new Zcu.Mve.Visualization.Point3D(1, _height, -1);
            }

            if (axis == AxisEnum.Z)
            {
                verticesArray[0] = new Zcu.Mve.Visualization.Point3D(1, -1, _height);
                verticesArray[1] = new Zcu.Mve.Visualization.Point3D(-1,-1, _height);
                verticesArray[2] = new Zcu.Mve.Visualization.Point3D(1, 1, _height);

                verticesArray[3] = new Zcu.Mve.Visualization.Point3D(-1, 1, _height);
                verticesArray[4] = new Zcu.Mve.Visualization.Point3D(-1, -1, _height);
                verticesArray[5] = new Zcu.Mve.Visualization.Point3D(1, 1, _height);
            }

            //texture coordinates
            verticesTexCoord[0] = new Zcu.Mve.Visualization.TextureCoord(1, 1, 0);
            verticesTexCoord[1] = new Zcu.Mve.Visualization.TextureCoord(0, 1, 0);
            verticesTexCoord[2] = new Zcu.Mve.Visualization.TextureCoord(1, 0, 0);

            verticesTexCoord[3] = new Zcu.Mve.Visualization.TextureCoord(0, 0, 0);
            verticesTexCoord[4] = new Zcu.Mve.Visualization.TextureCoord(0, 1, 0);
            verticesTexCoord[5] = new Zcu.Mve.Visualization.TextureCoord(1, 0, 0);

            //triangles
            cellArray[0] = new Zcu.Mve.Visualization.Triangle(0, 1, 2);
            cellArray[1] = new Zcu.Mve.Visualization.Triangle(3, 4, 5);

            //create mesh
            mesh = new Zcu.Mve.Visualization.Mesh();
            mesh.AddCells(typeof(Zcu.Mve.Visualization.Triangle).Name, cellArray);
            mesh.Points = verticesArray;
            mesh.AddPointAttrs(typeof(Zcu.Mve.Visualization.TextureCoord).Name, verticesTexCoord);
        }

        /// <summary>
        /// uptate quad and texture
        /// </summary>
        private void ReloadData()
        {
            _texture.UpdateData(_data, _zLocation,axis);
            InitUNSTRGrid();
            SetOutput("quad", mesh, false);
        }

        /// <summary>
        /// load volume data
        /// </summary>
        /// <returns></returns>
        private byte[] VolumeLoad()
        {
            Debug.WriteLine(grid.Dimensions.Length);

            System.Collections.ICollection l = grid.GetPointAttrIds();

            IEnumerator ie = l.GetEnumerator();
            ie.MoveNext();
            string ss = (string)ie.Current;

            MveArray points = grid.GetPointAttrs(ss);
            int count = 0;


            byte[] array = new byte[points.Lengths[0] * points.Lengths[1] * points.Lengths[2]];

            for (int i = 0; i < points.Lengths[0]; i++)
            {
                for (int j = 0; j < points.Lengths[1]; j++)
                {
                    for (int k = 0; k < points.Lengths[2]; k++)
                    {
                            Scalar s = (Scalar)points[i, j, k];
                            array[count] = (byte)s.Value;
                        count++;
                    }
                }
            }

            _sizeX = points.Lengths[0];
            _sizeY = points.Lengths[1];
            _sizeZ = points.Lengths[2];

            switch (axis)
            {
                case AxisEnum.X:
                    {
                        _size = _sizeX;  
                        break;
                    }
                case AxisEnum.Y:
                    {
                        _size = _sizeY;
                        break;
                    }
                case AxisEnum.Z:
                    {
                        _size = _sizeZ;
                        break;
                    }
            }
            _zLocation = _size / 2;

            return array;
        }

        /// <summary>
        /// writes Xml configuration of current module.
        /// </summary>
        public override void WriteConfig(XmlElement config)
        {
            XmlElement xmlProperties = config.OwnerDocument.CreateElement("Properties");

            XmlElement xmlAxis = config.OwnerDocument.CreateElement("axis");
            xmlAxis.SetAttribute("type", axis.GetType().FullName);
            xmlAxis.InnerText = axis.ToString();
            xmlProperties.AppendChild(xmlAxis);

            XmlElement xmlUp = config.OwnerDocument.CreateElement("up");
            xmlUp.SetAttribute("type", moveUp.GetType().FullName);
            xmlUp.InnerText = moveUp.ToString();
            xmlProperties.AppendChild(xmlUp);

            XmlElement xmlDown = config.OwnerDocument.CreateElement("down");
            xmlDown.SetAttribute("type", moveDown.GetType().FullName);
            xmlDown.InnerText = moveDown.ToString();
            xmlProperties.AppendChild(xmlDown);

            XmlElement xmlColor = config.OwnerDocument.CreateElement("color");
            xmlColor.SetAttribute("type", colorMode.GetType().FullName);
            xmlColor.InnerText = colorMode.ToString();
            xmlProperties.AppendChild(xmlColor);

            config.AppendChild(xmlProperties);
        }

        /// <summary>
        /// Reads Xml configuration of current module.
        /// </summary>
        /// <param name="config">XmlElement from map config file which belongs to current module.</param>
        public override void ReadConfig(System.Xml.XmlElement config)
        {
            foreach (XmlElement xeConfig in config.ChildNodes)
            {
                if (xeConfig.Name == "Properties")
                {
                    foreach (XmlElement xeProperties in xeConfig.ChildNodes)
                    {
                        switch (xeProperties.Name)
                        {
                            case "axis":
                                axis = (AxisEnum)Enum.Parse(typeof(AxisEnum), xeProperties.InnerText);
                                break;
                            case "up":
                                moveUp = (Keys)Enum.Parse(typeof(Keys), xeProperties.InnerText);
                                break;
                            case "down":
                                moveDown = (Keys)Enum.Parse(typeof(Keys), xeProperties.InnerText);
                                break;
                            case "color":
                                ColorMode = (ColorEnum)Enum.Parse(typeof(ColorEnum), xeProperties.InnerText);
                                break;
                            default:
                                Debug.WriteLine("Unreaded value:" + xeProperties.Name + " - " + xeProperties.InnerText);
                                break;
                        }
                    }
                }
            }
        }
    }
}
