using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;

using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization.Loaders
{
    /// <summary>
    /// Describes the main sections of the header file.
    /// </summary>
    enum VifSection : int
    {
        none = 0,
        main = 1,
        resolution = 2,
        position = 3,
        rotation = 4,
        distortion = 5
    }

    /// <summary>
    /// Class for reading the .vif header files.
    /// </summary>
    public class VifReader
    {
        private string vifString;
        private string[] vifLines;

        private NumberFormatInfo numberFormat1 = new NumberFormatInfo();
        private NumberFormatInfo numberFormat2 = new NumberFormatInfo();

        /// <summary>
        /// Reads the header file, prases it and copies the values to the module settings.
        /// </summary>
        /// <param name="vifFilePath">Path to the header file.</param>
        /// <param name="settings">Module settings, into which the values shall be copied.</param>
        public void ReadFromVif(string vifFilePath, BL_ModuleSettings settings)
        {
            numberFormat1.NumberDecimalSeparator = ".";
            numberFormat1.NumberDecimalSeparator = ",";

            FileStream fs;

            // Try to open the header file.
            try
            {
                fs = new FileStream(vifFilePath, FileMode.Open, FileAccess.Read, FileShare.Read);
                if (fs == null) throw new IOException();
            }
            catch (IOException)
            {
                throw new MveException(this, "Specified vif header file could not be opend.");
            }
            
            // Copy the file's characters to the 'vifString variable'.
            StreamReader reader = new StreamReader(fs);
            vifString = reader.ReadToEnd();

            // Get the files lines.
            vifLines = vifString.Split('\n');

            VifSection section = VifSection.none;
            string tempStr;
            float tempFl;
            float[] mins = new float[4];

            settings.DataDim = 3;
            string dataType = "";
            string signed = "";
            int bytesPerVoxel;

            // Try to read and parse the header file.
            // If the file is corrupted, an exception is chatched and propagated.
            try
            {
                for (int i = 0; i < vifLines.Length; i++)
                {
                    // Chceck if there is an section change on current line.
                    if (vifLines[i].ToLower().Contains("main"))
                    {
                        section = VifSection.main;
                    }
                    else if (vifLines[i].ToLower().Contains("resolution"))
                    {
                        section = VifSection.resolution;
                    }
                    else if (vifLines[i].ToLower().Contains("position"))
                    {
                        section = VifSection.position;
                    }
                    else if (vifLines[i].ToLower().Contains("rotation"))
                    {
                        section = VifSection.rotation;
                    }
                    else if (vifLines[i].ToLower().Contains("distortion"))
                    {
                        section = VifSection.distortion;
                    }

                    #region [MAIN]
                    if (section == VifSection.main)
                    {
                        // The 'header size' line check
                        if (vifLines[i].ToLower().Contains("header size"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            settings.HeaderSize = int.Parse(tempStr);
                        }
                        // The 'signed' line check
                        else if (vifLines[i].ToLower().Contains("signed"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            if (tempStr.ToLower().Trim().CompareTo("no") == 0)
                            {
                                signed = "u";
                            }
                            else
                            {
                                signed = "";
                            }
                        }
                        // The 'bytes per voxel' line check
                        else if (vifLines[i].ToLower().Contains("bytes per voxel"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            bytesPerVoxel = int.Parse(tempStr.Trim());
                            if (bytesPerVoxel == 1)
                            {
                                dataType = "int8";
                            }
                            else if (bytesPerVoxel == 2)
                            {
                                dataType = "int16";
                            }
                            else if (bytesPerVoxel == 4)
                            {
                                dataType = "int32";
                            }
                            else if (bytesPerVoxel == 8)
                            {
                                dataType = "int64";
                            }
                            else
                            {
                                throw new MveException(this, ".vif header describes uncompatible data type");
                            }
                        }
                    }
                    #endregion

                    #region  [RESOLUTION]
                    else if (section == VifSection.resolution)
                    {
                        // The 'x' line check
                        if (vifLines[i].ToLower().Contains("x"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            settings.Sizes = tempStr.Trim() + ";";
                        }
                        // The 'y' line check
                        else if (vifLines[i].ToLower().Contains("y"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            settings.Sizes = settings.Sizes + tempStr.Trim() + ";";
                        }
                        // The 'z' line check
                        else if (vifLines[i].ToLower().Contains("z"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            settings.Sizes = settings.Sizes + tempStr.Trim();
                        }
                    }
                    #endregion

                    #region [POSITION]
                    else if (section == VifSection.position)
                    {
                        if (vifLines[i].ToLower().Contains("x"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            settings.Mins = tempStr.Trim() + ";";
                            if (float.TryParse(tempStr.Trim(), NumberStyles.Float, numberFormat1, out mins[0]) == false)
                            {
                                mins[0] = float.Parse(tempStr.Trim(), numberFormat2);
                            }
                        }
                        else if (vifLines[i].ToLower().Contains("y"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            settings.Mins = settings.Mins + tempStr.Trim() + ";";
                            if (float.TryParse(tempStr.Trim(), NumberStyles.Float, numberFormat1, out mins[1]) == false)
                            {
                                mins[1] = float.Parse(tempStr.Trim(), numberFormat2);
                            }
                        }
                        else if (vifLines[i].ToLower().Contains("z"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            settings.Mins = settings.Mins + tempStr.Trim();
                            if (float.TryParse(tempStr.Trim(), NumberStyles.Float, numberFormat1, out mins[2]) == false)
                            {
                                mins[2] = float.Parse(tempStr.Trim(), numberFormat2);
                            }
                        }
                    }
                    #endregion

                    #region [DISTORTION]
                    else if (section == VifSection.distortion)
                    {
                        // The 'x' line check
                        if (vifLines[i].ToLower().Contains("x"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            if (float.TryParse(tempStr.Trim(), NumberStyles.Float, numberFormat1, out tempFl) == false)
                            {
                                tempFl = float.Parse(tempStr.Trim(), numberFormat2);
                            }
                            tempFl = tempFl + mins[0];
                            settings.Maxs = tempFl.ToString() + ";";
                        }

                        // The 'y' line check
                        else if (vifLines[i].ToLower().Contains("y"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            if (float.TryParse(tempStr.Trim(), NumberStyles.Float, numberFormat1, out tempFl) == false)
                            {
                                tempFl = float.Parse(tempStr.Trim(), numberFormat2);
                            }
                            tempFl = tempFl + mins[1];
                            settings.Maxs = settings.Maxs + tempFl.ToString() + ";";
                        }

                        // The 'z' line check
                        else if (vifLines[i].ToLower().Contains("z"))
                        {
                            tempStr = vifLines[i].Split('=')[1];
                            if (float.TryParse(tempStr.Trim(), NumberStyles.Float, numberFormat1, out tempFl) == false)
                            {
                                tempFl = float.Parse(tempStr.Trim(), numberFormat2);
                            }
                            tempFl = tempFl + mins[2];
                            settings.Maxs = settings.Maxs + tempFl.ToString();
                        }
                    }
                    #endregion
                }

                // Create a proper data description.
                settings.DataDesc = settings.DataDesc.Substring(0, settings.DataDesc.IndexOf(',') + 1) + signed + dataType + "[1]";
            }
            catch (Exception)
            {
                throw new MveException(this, ".vif header file corrupted");
            }
            finally
            {
                reader.Close();
                fs.Close();
            }
        }
    }
}
