﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using Zcu.Mve.Strings;
using System.IO;

namespace Zcu.Mve.Visualization.Savers
{
    [ModuleInfo("Libor Váša", "Module for saving volume data into a binary file. Processes one attribute of the Scalar or Integer type. Optionally can write a .vif file with file description")]
    [PortInfo("Data", "A 3D reggrid with input data")]
    [PortInfo("FileName", "Optional input, contains file name to be saved. If not connected, then the filename propert will be used instead")]
    class BinarySaver:Module
    {
        public BinarySaver()
        {            
            AddInPort("Data", typeof(RegGridND));
            AddInPort("FileName", typeof(MveString), false);
        }

        bool convertScalarsToIntegers = false;

        public bool ConvertScalarsToIntegers
        {
            get { return convertScalarsToIntegers; }
            set { convertScalarsToIntegers = value; }
        }        

        string attributeName = "";

        public string AttributeName
        {
            get { return attributeName; }
            set { attributeName = value; }
        }

        string fileName = "data";

        public string FileName
        {
            get { return fileName; }
            set { fileName = value; }
        }

        bool saveVif = true;

        public bool SaveVif
        {
            get { return saveVif; }
            set { saveVif = value; }
        }


        public override void Execute()
        {
            InPort fileNamePort = this.GetInPort("FileName");
            if (fileNamePort.Link != null)
            {
                MveString fileNameString = (MveString)GetInput("FileName");
                this.fileName = fileNameString.Value;
            }

            RegGridND input = (RegGridND)GetInput("Data");

            string usedAttributeName = attributeName;
            if (attributeName == "")
            {
                ICollection attrIds = input.GetPointAttrIds();
                foreach (string attrId in attrIds)
                {
                    if (usedAttributeName == "")
                    {
                        MveArray attr = input.GetPointAttrs(attrId);
                        if ((attr.ElemType == typeof(Integer)) || (attr.ElemType == typeof(Scalar)))
                        {
                            usedAttributeName = attrId;
                        }
                    }
                }
            }

            if (usedAttributeName == "")
            {
                throw new MveException("No scalar or integer point attribute found in the data.");
            }

            MveArray attribute = input.GetPointAttrs(usedAttributeName);
            
            BinaryWriter bw = new BinaryWriter(new FileStream(fileName+".vol", FileMode.Create, FileAccess.Write));
            for (int x = 0; x < attribute.Lengths[0]; x++)
            {
                for (int y = 0; y < attribute.Lengths[1]; y++)
                    for (int z = 0; z < attribute.Lengths[2]; z++)
                    {

                        if (attribute.ElemType == typeof(Scalar))
                        {
                            if (convertScalarsToIntegers)
                            {
                                int v = (int)((Scalar)attribute[x, y, z]);
                                bw.Write(v);
                            }
                            else
                            {
                                Scalar v = (Scalar)attribute[x, y, z];
                                bw.Write(v.Value);
                            }
                        }
                        if (attribute.ElemType == typeof(Integer))
                        {
                            Integer v = (Integer)attribute[x, y, z];
                            bw.Write(v.Value);
                        }
                    }
                ProgressInfo((double)x / attribute.Lengths[0]);
            }
            bw.Close();

            if (saveVif)
            {
                if ((attribute.ElemType == typeof(Scalar))&&(!(convertScalarsToIntegers)))
                    throw new MveException("VIF file cannot be generated for volume data with floating point values");

                StreamWriter sw = new StreamWriter(fileName + ".vif");
                sw.WriteLine("# Generated by MVE-2 for .vol file");

                sw.WriteLine("[Main]");
                
                sw.WriteLine("Caption ="); 
                sw.WriteLine("Source = vol");
                sw.WriteLine("Header Size = 0");
                sw.WriteLine("Bytes per voxel = 4");
                sw.WriteLine("Signed = yes");
                sw.WriteLine("Byte order file = file");
                 
                sw.WriteLine("[Resolution]");
                 
                sw.WriteLine("x = " + attribute.Lengths[0].ToString());
                sw.WriteLine("y = " + attribute.Lengths[1].ToString());
                sw.WriteLine("z = " + attribute.Lengths[2].ToString());

                sw.WriteLine("[Correction - position]");

                sw.WriteLine("x = 0");
                sw.WriteLine("y = 0");
                sw.WriteLine("z = 0");

                sw.WriteLine("[correction - rotation]");
                 
                sw.WriteLine("x = 0");
                sw.WriteLine("y = 0");
                sw.WriteLine("z = 0");

                sw.WriteLine("[Distortion]");

                sw.WriteLine("x = 1.0");
                double xSpan = input.Maxs[0] - input.Mins[0];
                double ySpan = input.Maxs[1] - input.Mins[1];
                double zSpan = input.Maxs[2] - input.Mins[2];
                double yDist = (ySpan / attribute.Lengths[1]) / (xSpan / attribute.Lengths[0]);
                double zDist = (zSpan / attribute.Lengths[2]) / (xSpan / attribute.Lengths[0]);
                sw.WriteLine("y = "+yDist.ToString(Globals.nfi));
                sw.WriteLine("z = "+zDist.ToString(Globals.nfi));

                sw.Close();
            }
        }
    }
}
