using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Syntax analyzer and expression evaluator class.
    /// </summary>
    class Analyzer
    {
        /// <summary>
        /// This enum contains types of symbols for use in the expression.
        /// </summary>
        private enum SymType
        {
            NUMBER = 0,
            VARIABLE,
            FUNCTION,
            ADD,
            SUB,
            MUL,
            DIV,
            POW,
            LOG,
            EXP,
            LN,
            SQRT,
            SIN,
            COS,
            TAN,
            SMUL,
            ASIN,
            ACOS,
            ATAN,
            VMUL,
            SINH,
            COSH,
            TANH,
            ABS,
            NORM,
            CEIL,
            FLOOR,
            LPAR1,
            PPAR1,
            LPAR2,
            PPAR2,
            SEMICLN,
            IVEC,
            JVEC,
            KVEC,
            UNKNOWN,
            NULL
        }

        /// <summary>
        /// Vector of the symbol types.
        /// </summary>
        private class SymTypes
        {
            private BitArray symSet;

            public SymTypes(params SymType[] types)
            {
                symSet = new BitArray((int)SymType.NULL + 1, false);

                for (int i = 0; i < types.Length; i++)
                {
                    symSet.Set((int)(types[i]), true);
                }
            }

            public SymTypes(SymTypes types)
            {
                if (types == null)
                {
                    symSet = new BitArray((int)SymType.NULL + 1, false);
                }
                else
                {
                    symSet = new BitArray(types.symSet);
                }
            }

            public static SymTypes operator +(SymTypes types, SymType type)
            {
                if (types == null)
                {
                    return null;
                }

                types.symSet.Set((int)type, true);

                return types;
            }

            public static SymTypes operator -(SymTypes types, SymType type)
            {
                if (types == null)
                {
                    return null;
                }

                types.symSet.Set((int)type, false);

                return types;
            }

            public static bool operator ==(SymTypes types, SymType type)
            {
                if (types == null)
                {
                    return false;
                }

                return types.symSet.Get((int)type);
            }

            public static bool operator !=(SymTypes types, SymType type)
            {
                if (types == null)
                {
                    return true;
                }

                return !types.symSet.Get((int)type);
            }
        }

        /// <summary>
        /// Single symbol with type, position and optional data.
        /// </summary>
        private class Symbol
        {
            private object data;
            private SymType type;
            private int position;

            public Symbol(object data, SymType type, int position)
            {
                this.data = data;
                this.type = type;
                this.position = position;
            }

            public object Data
            {
                get { return this.data; }

                set { this.data = value; }
            }

            public SymType Type
            {
                get { return this.type; }

                set { this.type = value; }
            }

            public int Position
            {
                get { return this.position; }

                set { this.position = value; }
            }

            public static bool operator ==(Symbol sym, SymType symType)
            {
                return (sym.Type == symType);
            }

            public static bool operator !=(Symbol sym, SymType symType)
            {
                return (sym.Type != symType);
            }
        }

        /// <summary>
        /// Evaluation tree node.
        /// </summary>
        private class Node
        {
            private Node left, right;
            private object data;
            private SymType type;
            private int position;

            private delegate double FunctionOneScalar(double value);
            private delegate double FunctionTwoScalars(double value1, double value2);

            public Node(object data, SymType type, int position)
            {
                this.data = data;
                this.type = type;
                this.position = position;
            }

            public Node Left
            {
                get { return this.left; }

                set { this.left = value; }
            }

            public Node Right
            {
                get { return this.right; }

                set { this.right = value; }
            }

            public double[] FirstResult(Dictionary<string, double[]> variables)
            {
                double[] result = new double[] { 0 };

                if (type == SymType.NUMBER)
                {
                    result = (double[])data;
                }
                else if (type == SymType.VARIABLE)
                {
                    if (left == null)
                    {
                        result = variables[data.ToString()];
                    }
                    else
                    {
                        if (variables[data.ToString()].Length == 1)
                        {
                            throw new Exception("Variable " + data.ToString() + " is not a vector! Position: " + position + ".");
                        }
                        
                        double[] lResult = left.FirstResult(variables);

                        if (lResult.Length != 1)
                        {
                            throw new Exception("Position in a vector must be a scalar! Position: " + position + ".");
                        }

                        if (Math.Floor(lResult[0]) != lResult[0])
                        {
                            throw new Exception("Position in a vector must be an integer! Position: " + position + ".");
                        }

                        int pos = (int)Math.Floor(lResult[0]);

                        if ((pos >= 0) && (pos < variables[data.ToString()].Length))
                        {
                            result = new double[] { variables[data.ToString()][pos] };
                        }
                        else
                        {
                            throw new Exception("Position in a vector is out of bounds! Position: " + position + ".");
                        }
                    }
                }
                else if (type == SymType.IVEC)
                {
                    result = new double[] { 1, 0, 0 };
                }
                else if (type == SymType.JVEC)
                {
                    result = new double[] { 0, 1, 0 };
                }
                else if (type == SymType.KVEC)
                {
                    result = new double[] { 0, 0, 1 };
                }
                else if (type == SymType.ADD)
                {
                    double[] lResult = left.FirstResult(variables);
                    double[] rResult = right.FirstResult(variables);
                    
                    if (lResult.Length == rResult.Length)
                    {
                        result = new double[lResult.Length];

                        for (int i = 0; i < result.Length; i++)
                        {
                            result[i] = lResult[i] + rResult[i];
                        }
                    }
                    else
                    {
                        throw new Exception("Addition of incompatible types! Position: " + position + ".");
                    }
                }
                else if (type == SymType.SUB)
                {
                    if (left != null)
                    {
                        double[] lResult = left.FirstResult(variables);
                        double[] rResult = right.FirstResult(variables);

                        if (lResult.Length == rResult.Length)
                        {
                            result = new double[lResult.Length];

                            for (int i = 0; i < result.Length; i++)
                            {
                                result[i] = lResult[i] - rResult[i];
                            }
                        }
                        else
                        {
                            throw new Exception("Substraction of incompatible types! Position: " + position + ".");
                        }
                    }
                    else
                    {
                        double[] rResult = right.FirstResult(variables);

                        result = new double[rResult.Length];

                        for (int i = 0; i < result.Length; i++)
                        {
                            result[i] = -rResult[i];
                        }
                    }
                }
                else if (type == SymType.MUL)
                {
                    double[] lResult = left.FirstResult(variables);
                    double[] rResult = right.FirstResult(variables);

                    if (lResult.Length == 1)
                    {
                        result = new double[rResult.Length];

                        for (int i = 0; i < result.Length; i++)
                        {
                            result[i] = lResult[0] * rResult[i];
                        }
                    }
                    else if (rResult.Length == 1)
                    {
                        result = new double[lResult.Length];

                        for (int i = 0; i < result.Length; i++)
                        {
                            result[i] = lResult[i] * rResult[0];
                        }
                    }
                    else
                    {
                        throw new Exception("Multiplication of vectors! Position: " + position + ".");
                    }
                }
                else if (type == SymType.DIV)
                {
                    double[] lResult = left.FirstResult(variables);
                    double[] rResult = right.FirstResult(variables);

                    if (rResult.Length == 1)
                    {
                        if (rResult[0] != 0)
                        {
                            result = new double[lResult.Length];

                            for (int i = 0; i < result.Length; i++)
                            {
                                result[i] = lResult[i] / rResult[0];
                            }
                        }
                        else
                        {
                            throw new Exception("Division by zero! Position: " + position + ".");
                        }
                    }
                    else
                    {
                        throw new Exception("Division of vectors! Position: " + position + ".");
                    }
                }
                else //if (type == SymType.FUNCTION)
                {
                    SymType function = (SymType)data;

                    if (function == SymType.POW)
                    {
                        double[] lResult = left.FirstResult(variables);
                        double[] rResult = right.FirstResult(variables);

                        FirstTwoScalarsCondition(lResult, rResult, ref result, !((lResult[0] == 0 && rResult[0] < 0) || (lResult[0] < 0 && rResult[0] > 0 && rResult[0] < 1) || (lResult[0] < 0 && rResult[0] > -1 && rResult[0] < 0)), "Power of  vectors! Position: " + position + ".", "Power has some invalid arguments! Position: " + position + ".", Math.Pow);
                    }
                    else if (function == SymType.LOG)
                    {
                        double[] lResult = left.FirstResult(variables);
                        double[] rResult = right.FirstResult(variables);

                        FirstTwoScalarsCondition(rResult, lResult, ref result, !(lResult[0] <= 0 || lResult[0] == 1 || rResult[0] <= 0), "Logarithm of  vectors! Position: " + position + ".", "Logarithm has some invalid arguments! Position: " + position + ".", Math.Log);
                    }
                    else if (function == SymType.EXP)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Exponent of a vector! Position: " + position + ".", Math.Exp);
                    }
                    else if (function == SymType.LN)
                    {
                        double[] lResult = left.FirstResult(variables);

                        FirstOneScalarCondition(lResult, ref result, lResult[0] > 0, "Natural logarithm of a vector! Position: " + position + ".", "Argument of a natural logarithm must be > 0! Position: " + position + ".", Math.Log);
                    }
                    else if (function == SymType.SQRT)
                    {
                        double[] lResult = left.FirstResult(variables);

                        FirstOneScalarCondition(lResult, ref result, lResult[0] >= 0, "Radix of a vector! Position: " + position + ".", "Argument of a radix must be >= 0! Position: " + position + ".", Math.Sqrt);
                    }
                    else if (function == SymType.SIN)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Sinus of a vector! Position: " + position + ".", Math.Sin);
                    }
                    else if (function == SymType.COS)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Cosinus of a vector! Position: " + position + ".", Math.Cos);
                    }
                    else if (function == SymType.TAN)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Tangenc of a vector! Position: " + position + ".", Math.Tan);
                    }
                    else if (function == SymType.SMUL)
                    {
                        double[] lResult = left.FirstResult(variables);
                        double[] rResult = right.FirstResult(variables);

                        if (lResult.Length == rResult.Length && lResult.Length != 1)
                        {
                            result = new double[] { 0 };

                            for (int i = 0; i < lResult.Length; i++)
                            {
                                result[0] += lResult[i] * rResult[i];
                            }
                        }
                        else
                        {
                            throw new Exception("Scalar multiplication of scalars or incompatible types! Position: " + position + ".");
                        }
                    }
                    else if (function == SymType.ASIN)
                    {
                        double[] lResult = left.FirstResult(variables);

                        FirstOneScalarCondition(lResult, ref result, lResult[0] >= -1 && lResult[0] <= 1, "Arcus sinus of a vector! Position: " + position + ".", "Argument of an arcus sinus must be >= -1 and <= 1! Position: " + position + ".", Math.Asin);
                    }
                    else if (function == SymType.ACOS)
                    {
                        double[] lResult = left.FirstResult(variables);

                        FirstOneScalarCondition(lResult, ref result, lResult[0] >= -1 && lResult[0] <= 1, "Arcus cosinus of a vector! Position: " + position + ".", "Argument of an arcus cosinus must be >= -1 and <= 1! Position: " + position + ".", Math.Acos);
                    }
                    else if (function == SymType.ATAN)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Arcus tangenc of a vector! Position: " + position + ".", Math.Atan);
                    }
                    else if (function == SymType.VMUL)
                    {
                        double[] lResult = left.FirstResult(variables);
                        double[] rResult = right.FirstResult(variables);

                        if (lResult.Length == rResult.Length && lResult.Length == 3)
                        {
                            result = new double[lResult.Length];

                            result[0] = lResult[1] * rResult[2] - lResult[2] * rResult[1];
                            result[1] = lResult[2] * rResult[0] - lResult[0] * rResult[2];
                            result[2] = lResult[0] * rResult[1] - lResult[1] * rResult[0];
                        }
                        else
                        {
                            throw new Exception("Vector multiplication of scalars or incompatible types! Position: " + position + ".");
                        }
                    }
                    else if (function == SymType.SINH)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Hyperbolic sinus of a vector! Position: " + position + ".", Math.Sinh);
                    }
                    else if (function == SymType.COSH)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Hyperbolic cosinus of a vector! Position: " + position + ".", Math.Cosh);
                    }
                    else if (function == SymType.TANH)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Hyperbolic tangenc of a vector! Position: " + position + ".", Math.Tanh);
                    }
                    else if (function == SymType.ABS)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Absolute value of a vector! Position: " + position + ".", Math.Abs);
                    }
                    else if (function == SymType.NORM)
                    {
                        double[] lResult = left.FirstResult(variables);

                        if (lResult.Length == 1)
                        {
                            throw new Exception("Norm of a scalar! Position: " + position + ".");
                        }

                        result = new double[] { 0 };

                        for (int i = 0; i < lResult.Length; i++)
                        {
                            result[0] += lResult[i] * lResult[i];
                        }

                        result[0] = Math.Sqrt(result[0]);
                    }
                    else if (function == SymType.CEIL)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Ceiling of a vector! Position: " + position + ".", Math.Ceiling);
                    }
                    else //if (function == SymType.FLOOR)
                    {
                        FirstOneScalar(left.FirstResult(variables), ref result, "Floor of a vector! Position: " + position + ".", Math.Floor);
                    }
                }

                return result;
            }

            public double[] SecondResult(Dictionary<string, double[]> variables)
            {
                double[] result = new double[] { 0 };

                if (type == SymType.NUMBER)
                {
                    result = (double[])data;
                }
                else if (type == SymType.VARIABLE)
                {
                    if (left == null)
                    {
                        result = variables[data.ToString()];
                    }
                    else
                    {
                        double[] lResult = left.SecondResult(variables);

                        int pos = (int)Math.Floor(lResult[0]);

                        result = new double[] { variables[data.ToString()][pos] };
                    }
                }
                else if (type == SymType.IVEC)
                {
                    result = new double[] { 1, 0, 0 };
                }
                else if (type == SymType.JVEC)
                {
                    result = new double[] { 0, 1, 0 };
                }
                else if (type == SymType.KVEC)
                {
                    result = new double[] { 0, 0, 1 };
                }
                else if (type == SymType.ADD)
                {
                    double[] lResult = left.SecondResult(variables);
                    double[] rResult = right.SecondResult(variables);

                    result = new double[lResult.Length];

                    for (int i = 0; i < result.Length; i++)
                    {
                        result[i] = lResult[i] + rResult[i];
                    }
                }
                else if (type == SymType.SUB)
                {
                    if (left != null)
                    {
                        double[] lResult = left.SecondResult(variables);
                        double[] rResult = right.SecondResult(variables);

                        result = new double[lResult.Length];

                        for (int i = 0; i < result.Length; i++)
                        {
                            result[i] = lResult[i] - rResult[i];
                        }
                    }
                    else
                    {
                        double[] rResult = right.SecondResult(variables);

                        result = new double[rResult.Length];

                        for (int i = 0; i < result.Length; i++)
                        {
                            result[i] = -rResult[i];
                        }
                    }
                }
                else if (type == SymType.MUL)
                {
                    double[] lResult = left.SecondResult(variables);
                    double[] rResult = right.SecondResult(variables);

                    if (lResult.Length == 1)
                    {
                        result = new double[rResult.Length];

                        for (int i = 0; i < result.Length; i++)
                        {
                            result[i] = lResult[0] * rResult[i];
                        }
                    }
                    else if (rResult.Length == 1)
                    {
                        result = new double[lResult.Length];

                        for (int i = 0; i < result.Length; i++)
                        {
                            result[i] = lResult[i] * rResult[0];
                        }
                    }
                }
                else if (type == SymType.DIV)
                {
                    double[] lResult = left.SecondResult(variables);
                    double[] rResult = right.SecondResult(variables);

                    if (rResult[0] != 0)
                    {
                        result = new double[lResult.Length];

                        for (int i = 0; i < result.Length; i++)
                        {
                            result[i] = lResult[i] / rResult[0];
                        }
                    }
                    else
                    {
                        throw new Exception("Division by zero! Position: " + position + ".");
                    }
                }
                else //if (type == SymType.FUNCTION)
                {
                    SymType function = (SymType)data;

                    if (function == SymType.POW)
                    {
                        double[] lResult = left.SecondResult(variables);
                        double[] rResult = right.SecondResult(variables);

                        SecondTwoScalarsCondition(lResult, rResult, ref result, !((lResult[0] == 0 && rResult[0] < 0) || (lResult[0] < 0 && rResult[0] > 0 && rResult[0] < 1) || (lResult[0] < 0 && rResult[0] > -1 && rResult[0] < 0)), "Power has some invalid arguments! Position: " + position + ".", Math.Pow);
                    }
                    else if (function == SymType.LOG)
                    {
                        double[] lResult = left.SecondResult(variables);
                        double[] rResult = right.SecondResult(variables);

                        SecondTwoScalarsCondition(rResult, lResult, ref result, !(lResult[0] <= 0 || lResult[0] == 1 || rResult[0] <= 0), "Logarithm has some invalid arguments! Position: " + position + ".", Math.Log);
                    }
                    else if (function == SymType.EXP)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Exp);
                    }
                    else if (function == SymType.LN)
                    {
                        double[] lResult = left.SecondResult(variables);

                        SecondOneScalarCondition(lResult, ref result, lResult[0] > 0, "Argument of a natural logarithm must be > 0! Position: " + position + ".", Math.Log);
                    }
                    else if (function == SymType.SQRT)
                    {
                        double[] lResult = left.SecondResult(variables);

                        SecondOneScalarCondition(lResult, ref result, lResult[0] >= 0, "Argument of a radix must be >= 0! Position: " + position + ".", Math.Sqrt);
                    }
                    else if (function == SymType.SIN)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Sin);
                    }
                    else if (function == SymType.COS)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Cos);
                    }
                    else if (function == SymType.TAN)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Tan);
                    }
                    else if (function == SymType.SMUL)
                    {
                        double[] lResult = left.SecondResult(variables);
                        double[] rResult = right.SecondResult(variables);

                        result = new double[] { 0 };

                        for (int i = 0; i < lResult.Length; i++)
                        {
                            result[0] += lResult[i] * rResult[i];
                        }
                    }
                    else if (function == SymType.ASIN)
                    {
                        double[] lResult = left.SecondResult(variables);

                        SecondOneScalarCondition(lResult, ref result, lResult[0] >= -1 && lResult[0] <= 1, "Argument of an arcus sinus must be >= -1 and <= 1! Position: " + position + ".", Math.Asin);
                    }
                    else if (function == SymType.ACOS)
                    {
                        double[] lResult = left.SecondResult(variables);

                        SecondOneScalarCondition(lResult, ref result, lResult[0] >= -1 && lResult[0] <= 1, "Argument of an arcus cosinus must be >= -1 and <= 1! Position: " + position + ".", Math.Acos);
                    }
                    else if (function == SymType.ATAN)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Atan);
                    }
                    else if (function == SymType.VMUL)
                    {
                        double[] lResult = left.SecondResult(variables);
                        double[] rResult = right.SecondResult(variables);

                        result = new double[lResult.Length];

                        result[0] = lResult[1] * rResult[2] - lResult[2] * rResult[1];
                        result[1] = lResult[2] * rResult[0] - lResult[0] * rResult[2];
                        result[2] = lResult[0] * rResult[1] - lResult[1] * rResult[0];
                    }
                    else if (function == SymType.SINH)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Sinh);
                    }
                    else if (function == SymType.COSH)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Cosh);
                    }
                    else if (function == SymType.TANH)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Tanh);
                    }
                    else if (function == SymType.ABS)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Abs);
                    }
                    else if (function == SymType.NORM)
                    {
                        double[] lResult = left.SecondResult(variables);

                        result = new double[] { 0 };

                        for (int i = 0; i < lResult.Length; i++)
                        {
                            result[0] += lResult[i] * lResult[i];
                        }

                        result[0] = Math.Sqrt(result[0]);
                    }
                    else if (function == SymType.CEIL)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Ceiling);
                    }
                    else //if (function == SymType.FLOOR)
                    {
                        SecondOneScalar(left.SecondResult(variables), ref result, Math.Floor);
                    }
                }

                return result;
            }

            private void FirstOneScalar(double[] lResult, ref double[] result, string errorMessage, FunctionOneScalar function)
            {
                if (lResult.Length == 1)
                {
                    result = new double[] { function(lResult[0]) };
                }
                else
                {
                    throw new Exception(errorMessage);
                }
            }

            private void FirstTwoScalars(double[] lResult, double[] rResult, ref double[] result, string errorMessage, FunctionTwoScalars function)
            {
                if (lResult.Length == 1 && rResult.Length == 1)
                {
                    result = new double[] { function(lResult[0], rResult[0]) };
                }
                else
                {
                    throw new Exception(errorMessage);
                }
            }

            private void FirstOneScalarCondition(double[] lResult, ref double[] result, bool condition, string errorMessage, string errorConditionMessage, FunctionOneScalar function)
            {
                if (lResult.Length == 1)
                {
                    if (!condition)
                    {
                        throw new Exception(errorConditionMessage);
                    }

                    result = new double[] { function(lResult[0]) };
                }
                else
                {
                    throw new Exception(errorMessage);
                }
            }

            private void FirstTwoScalarsCondition(double[] lResult, double[] rResult, ref double[] result, bool condition, string errorMessage, string errorConditionMessage, FunctionTwoScalars function)
            {
                if (lResult.Length == 1 && rResult.Length == 1)
                {
                    if (!condition)
                    {
                        throw new Exception(errorConditionMessage);
                    }

                    result = new double[] { function(lResult[0], rResult[0]) };
                }
                else
                {
                    throw new Exception(errorMessage);
                }
            }

            private void SecondOneScalar(double[] lResult, ref double[] result, FunctionOneScalar function)
            {
                result = new double[] { function(lResult[0]) };
            }

            private void SecondTwoScalars(double[] lResult, double[] rResult, ref double[] result, FunctionTwoScalars function)
            {
                result = new double[] { function(lResult[0], rResult[0]) };
            }

            private void SecondOneScalarCondition(double[] lResult, ref double[] result, bool condition, string errorConditionMessage, FunctionOneScalar function)
            {
                if (!condition)
                {
                    throw new Exception(errorConditionMessage);
                }

                result = new double[] { function(lResult[0]) };
            }

            private void SecondTwoScalarsCondition(double[] lResult, double[] rResult, ref double[] result, bool condition, string errorConditionMessage, FunctionTwoScalars function)
            {
                if (!condition)
                {
                    throw new Exception(errorConditionMessage);
                }

                result = new double[] { function(lResult[0], rResult[0]) };
            }
        }

        private Dictionary<string, SymType> resWord;
        private Dictionary<string, SymType> unitVec;
        private Dictionary<char, SymType> resChar;
        private Node result;
        private List<SymType> semiclnFunctions;
        private string line;
        private int position, symbolPosition;
        private char c;

        /// <summary>
        /// Class contructor.
        /// </summary>
        public Analyzer()
        {
            resWord = new Dictionary<string, SymType>();
            resWord.Add("pow", SymType.POW);
            resWord.Add("log", SymType.LOG);
            resWord.Add("exp", SymType.EXP);
            resWord.Add("ln", SymType.LN);
            resWord.Add("sqrt", SymType.SQRT);
            resWord.Add("sin", SymType.SIN);
            resWord.Add("cos", SymType.COS);
            resWord.Add("tan", SymType.TAN);
            resWord.Add("smul", SymType.SMUL);
            resWord.Add("asin", SymType.ASIN);
            resWord.Add("acos", SymType.ACOS);
            resWord.Add("atan", SymType.ATAN);
            resWord.Add("vmul", SymType.VMUL);
            resWord.Add("sinh", SymType.SINH);
            resWord.Add("cosh", SymType.COSH);
            resWord.Add("tanh", SymType.TANH);
            resWord.Add("abs", SymType.ABS);
            resWord.Add("norm", SymType.NORM);
            resWord.Add("ceil", SymType.CEIL);
            resWord.Add("floor", SymType.FLOOR);

            unitVec = new Dictionary<string, SymType>();
            unitVec.Add("ivec", SymType.IVEC);
            unitVec.Add("jvec", SymType.JVEC);
            unitVec.Add("kvec", SymType.KVEC);

            resChar = new Dictionary<char, SymType>();
            resChar.Add('+', SymType.ADD);
            resChar.Add('-', SymType.SUB);
            resChar.Add('*', SymType.MUL);
            resChar.Add('/', SymType.DIV);
            resChar.Add('(', SymType.LPAR1);
            resChar.Add(')', SymType.PPAR1);
            resChar.Add('[', SymType.LPAR2);
            resChar.Add(']', SymType.PPAR2);
            resChar.Add(';', SymType.SEMICLN);

            semiclnFunctions = new List<SymType>();
            semiclnFunctions.Add(SymType.POW);
            semiclnFunctions.Add(SymType.LOG);
            semiclnFunctions.Add(SymType.SMUL);
            semiclnFunctions.Add(SymType.VMUL);
        }

        /// <summary>
        /// Analyzes the expression string. Constructs an evaluation tree and returns list of variables used in the expression.
        /// </summary>
        /// <param name="line">expression to analyze</param>
        /// <returns>list of used variables' names</returns>
        public List<string> Analyze(string line)
        {
            this.line = line;
            position = 0;
            c = (char)1;

            Symbol help = null;
            result = null;

            List<string> variables = new List<string>();

            Expression(new SymTypes(SymType.NULL), ref help, ref result, true, true, variables);

            return variables;
        }

        /// <summary>
        /// Evaluates the tree constructed by the Analyze method. Checks for all possible errors and throws an exception while finds one.
        /// </summary>
        /// <param name="variables">list of values for variables found by Analyze</param>
        /// <returns>resulting value (scalar or vector) for the given values</returns>
        public double[] FirstResult(Dictionary<string, double[]> variables)
        {
            return result.FirstResult(variables);
        }

        /// <summary>
        /// Evaluates the tree constructed by the Analyze method. Does not check for errors that need not checking on every run.
        /// </summary>
        /// <param name="variables">list of values for variables found by Analyze</param>
        /// <returns>resulting value (scalar or vector) for the given values</returns>
        public double[] SecondResult(Dictionary<string, double[]> variables)
        {
            return result.SecondResult(variables);
        }

        /// <summary>
        /// Evaluates the tree with the specified node as root.
        /// </summary>
        /// <param name="endSym">vector of symbols to stop the evaluation on</param>
        /// <param name="op">reference to an instance of symbol for operator</param>
        /// <param name="node">pointer to the current tree root</param>
        /// <param name="unar">indicates the precedence of unar plus or minus</param>
        /// <param name="variable">indicates, whether this tree can contain variables</param>
        /// <param name="variables">pointer to a list where to add the newly found variables</param>
        private void Expression(SymTypes endSym, ref Symbol op, ref Node node, bool unar, bool variable, List<string> variables)
        {
            Symbol symbol = GetSymbol();

            if (symbol == SymType.NUMBER)
            {
                node = new Node(symbol.Data, SymType.NUMBER, symbol.Position);

                op = GetSymbol();
            }
            else if (symbol == SymType.VARIABLE)
            {
                if (!variable)
                {
                    throw new Exception("A variable can not be here! Position: " + symbol.Position + ".");
                }

                node = new Node(symbol.Data, SymType.VARIABLE, symbol.Position);

                if (!variables.Contains((string)symbol.Data))
                {
                    variables.Add((string)symbol.Data);
                }

                op = GetSymbol();

                if (op == SymType.LPAR2)
                {
                    Symbol help = null;
                    Node newNode = null;

                    Expression(new SymTypes(SymType.PPAR2), ref help, ref newNode, true, false, variables);

                    node.Left = newNode;

                    op = GetSymbol();
                }
            }
            else if (symbol == SymType.IVEC || symbol == SymType.JVEC || symbol == SymType.KVEC)
            {
                node = new Node(null, symbol.Type, symbol.Position);

                op = GetSymbol();
            }
            else if (symbol == SymType.ADD || symbol == SymType.SUB)
            {
                if (!unar)
                {
                    if (symbol == SymType.ADD)
                    {
                        throw new Exception("Unexpected operator +! Position: " + symbol.Position + ".");
                    }
                    else
                    {
                        throw new Exception("Unexpected operator -! Position: " + symbol.Position + ".");
                    }
                }

                Node newNode = null;

                Expression(new SymTypes(endSym) + SymType.ADD + SymType.SUB, ref op, ref newNode, false, variable, variables);

                if (symbol == SymType.SUB)
                {
                    node = new Node(null, SymType.SUB, symbol.Position);
                    node.Right = newNode;
                }
                else
                {
                    node = newNode;
                }
            }
            else if (symbol == SymType.LPAR1)
            {
                Symbol help = null;

                Expression(new SymTypes(SymType.PPAR1), ref help, ref node, true, variable, variables);

                op = GetSymbol();
            }
            else if (symbol == SymType.FUNCTION)
            {
                if (GetSymbol() != SymType.LPAR1)
                {
                    throw new Exception("Symbol ( expected! Position: " + symbol.Position + ".");
                }

                Symbol help = null;
                Node newNode = null;

                if (semiclnFunctions.Contains((SymType)symbol.Data))
                {
                    Expression(new SymTypes(SymType.SEMICLN), ref help, ref newNode, true, variable, variables);

                    Node lNode = newNode;

                    Expression(new SymTypes(SymType.PPAR1), ref help, ref newNode, true, variable, variables);

                    Node rNode = newNode;

                    node = new Node(symbol.Data, SymType.FUNCTION, symbol.Position);
                    node.Left = lNode;
                    node.Right = rNode;
                }
                else
                {
                    Expression(new SymTypes(SymType.PPAR1), ref help, ref newNode, true, variable, variables);

                    node = new Node(symbol.Data, SymType.FUNCTION, symbol.Position);
                    node.Left = newNode;
                }

                op = GetSymbol();
            }
            else
            {
                throw new Exception("Unexpected symbol! Position: " + symbol.Position + ".");
            }

            while (endSym != op.Type)
            {
                Node newNode = null, help = null;
                int opPosition = op.Position;

                switch (op.Type)
                {
                    case SymType.ADD:
                        Expression(new SymTypes(endSym) + SymType.ADD + SymType.SUB, ref op, ref newNode, false, variable, variables);

                        help = new Node(null, SymType.ADD, opPosition);
                        help.Left = node;
                        help.Right = newNode;
                        node = help;

                        break;

                    case SymType.SUB:
                        Expression(new SymTypes(endSym) + SymType.ADD + SymType.SUB, ref op, ref newNode, false, variable, variables);

                        help = new Node(null, SymType.SUB, opPosition);
                        help.Left = node;
                        help.Right = newNode;
                        node = help;

                        break;

                    case SymType.MUL:
                        Expression(new SymTypes(endSym) + SymType.ADD + SymType.SUB + SymType.MUL + SymType.DIV, ref op, ref newNode, true, variable, variables);

                        help = new Node(null, SymType.MUL, opPosition);
                        help.Left = node;
                        help.Right = newNode;
                        node = help;

                        break;

                    case SymType.DIV:
                        Expression(new SymTypes(endSym) + SymType.ADD + SymType.SUB + SymType.MUL + SymType.DIV, ref op, ref newNode, true, variable, variables);

                        help = new Node(null, SymType.DIV, opPosition);
                        help.Left = node;
                        help.Right = newNode;
                        node = help;

                        break;

                    default:
                        throw new Exception("Unexpected symbol! Position: " + opPosition + ".");
                }
            }
        }

        /// <summary>
        /// Reads a single symbol from the expression string.
        /// </summary>
        /// <returns>next symbol from the expression</returns>
        private Symbol GetSymbol()
        {
            Symbol sym = null;
            int newSymbolPosition = -1;

            while (c <= ' ' && c != (char)0)
            {
                symbolPosition = GetChar();
            }

            if (c == '.' || c == ',')
            {
                string textValue = "";

                textValue += ',';

                newSymbolPosition = GetChar();

                while (char.IsDigit(c))
                {
                    textValue += c;
                    newSymbolPosition = GetChar();
                }

                try
                {
                    double value = double.Parse(textValue,Globals.nfi);

                    sym = new Symbol(new double[] { value }, SymType.NUMBER, symbolPosition);
                }
                catch (Exception)
                {
                    sym = new Symbol(null, SymType.UNKNOWN, symbolPosition);
                }
            }
            else if (char.IsDigit(c))
            {
                string textValue = "";

                do
                {
                    textValue += c;
                    newSymbolPosition = GetChar();
                }
                while (char.IsDigit(c));

                if (c == '.' || c == ',')
                {
                    textValue += ',';

                    newSymbolPosition = GetChar();

                    while (char.IsDigit(c))
                    {
                        textValue += c;
                        newSymbolPosition = GetChar();
                    }
                }

                try
                {
                    double value = double.Parse(textValue, Globals.nfi);

                    sym = new Symbol(new double[] { value }, SymType.NUMBER, symbolPosition);
                }
                catch (Exception)
                {
                    sym = new Symbol(null, SymType.UNKNOWN, symbolPosition);
                }
            }
            else if (char.IsLetter(c))
            {
                string textValue = "";

                do
                {
                    textValue += c;
                    newSymbolPosition = GetChar();
                }
                while (char.IsLetterOrDigit(c));

                if (unitVec.ContainsKey(textValue.ToLower()))
                {
                    sym = new Symbol(null, unitVec[textValue.ToLower()], symbolPosition);
                }
                else if (resWord.ContainsKey(textValue.ToLower()))
                {
                    sym = new Symbol(resWord[textValue.ToLower()], SymType.FUNCTION, symbolPosition);
                }
                else
                {
                    sym = new Symbol(textValue, SymType.VARIABLE, symbolPosition);
                }
            }
            else if (resChar.ContainsKey(c))
            {
                sym = new Symbol(null, resChar[c], symbolPosition);
                newSymbolPosition = GetChar();
            }
            else if (c == (char)0)
            {
                sym = new Symbol(null, SymType.NULL, symbolPosition);
            }
            else
            {
                sym = new Symbol(null, SymType.UNKNOWN, symbolPosition);
                newSymbolPosition = GetChar();
            }

            symbolPosition = newSymbolPosition;

            return sym;
        }

        /// <summary>
        /// Reads and saves one character from the expression string.
        /// </summary>
        /// <returns>returns current position in the string</returns>
        private int GetChar()
        {
            if (position < line.Length)
            {
                c = line[position++];

                return position - 1;
            }

            c = (char)0;

            return position;
        }
    }
}
