using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Calculator module setup dialog class.
    /// </summary>
    public partial class CalculatorSetup : Mve.Core.ModuleSetup
    {
        private bool changed;
        private string expression = "";
        private int outputType = 0;
        private string cellsName = "name";
        private string outputName = "result";
        private bool outputCoords = false;
        private bool specialFields;
        private string helpStr;

        /// <summary>
        /// Indicates that some changes were made in the module setup.
        /// </summary>
        public bool Changed {
            get { return changed; }
            set { changed = value; }
        }

        /// <summary>
        /// Returns current expression string.
        /// </summary>
        public string Expression { get { return expression; } }

        /// <summary>
        /// Indicates the type of output.
        /// </summary>
        public bool CellData { get { return (outputType == 1); } }

        /// <summary>
        /// Returns name of the resulting attribute.
        /// </summary>
        public string OutputName { get { return outputName; } }

        /// <summary>
        /// Returns name of the cells array to work with.
        /// </summary>
        public string CellsName { get { return cellsName; } }

        /// <summary>
        /// Indicates whether to save results as a point attribute or point coordinates
        /// </summary>
        public bool CoordinateOutput { get { return outputCoords; } }

        /// <summary>
        /// Setup dialog constructor.
        /// </summary>
        public CalculatorSetup(bool showSpecialFields, string helpString)
        {
            InitializeComponent();
            specialFields = showSpecialFields;
            helpStr = helpString;
            UpdateConfig();
        }

        /// <summary>
        /// Sets new configuration values for this dialog.
        /// </summary>
        /// <param name="expression">expression string</param>
        /// <param name="cellData">output type</param>
        /// <param name="cellsName">cells array name</param>
        /// <param name="outputName">output attribute name</param>
        /// <param name="coordinateOutput">coordinate output option state</param>
        public void SetConfig(string expression, bool cellData, string cellsName, string outputName, bool coordinateOutput)
        {
            this.expression = expression;
            this.outputType = (cellData) ? 1 : 0;
            this.cellsName = cellsName;
            this.outputName = outputName;
            this.outputCoords = coordinateOutput;

            UpdateConfig();
        }

        /// <summary>
        /// Changes state of all the dialog controls according to the current configuration values.
        /// </summary>
        private void UpdateConfig()
        {
            if (!specialFields)
            {
                chbOutputCoords.Visible = false;
                tbxCellsName.Visible = false;
            }
            tbxExpression.Text = expression;
            cbxDataType.SelectedIndex = outputType;
            tbxCellsName.Text = cellsName;
            tbxOutputName.Text = outputName;
            chbOutputCoords.Checked = outputCoords;
            changed = true;
        }

        /// <summary>
        /// Updates configuration values from the dialog controls upon clicking the OK dialog button.
        /// </summary>
        /// <returns>true if the dialog is ready to close, false otherwise</returns>
        public override bool OnOkClick()
        {
            changed = (
                expression != tbxExpression.Text ||
                outputType != cbxDataType.SelectedIndex ||
                cellsName != tbxCellsName.Text ||
                outputName != tbxOutputName.Text ||
                outputCoords != chbOutputCoords.Checked
            );

            expression = tbxExpression.Text;
            outputType = cbxDataType.SelectedIndex;
            cellsName = tbxCellsName.Text;
            outputName = tbxOutputName.Text;
            outputCoords = chbOutputCoords.Checked;
            
            return base.OnOkClick();
        }

        /// <summary>
        /// Resets the controls' state from the configuration values upon clicking the Storno button.
        /// </summary>
        /// <returns>true if the dialog is ready to close, false otherwise</returns>
        public override bool OnStornoClick()
        {
            UpdateConfig();
            return base.OnStornoClick();
        }

        /// <summary>
        /// Displays a message box containing a brief module usage manual.
        /// </summary>
        private void ShowHelp()
        {
            MessageBox.Show(this, helpStr, "Calculator Help", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        /// <summary>
        /// Inserts one-parameter function to the current position in the expression editor.
        /// If some text is selected, it will be used as the parameter of the function.
        /// </summary>
        /// <param name="func">function name</param>
        private void InsertFunction(string func)
        {
            tbxExpression.Paste(func + "(" + tbxExpression.SelectedText + ")");
            tbxExpression.Focus();
        }

        /// <summary>
        /// Inserts two-parameter function to the current position in the expression editor.
        /// If some text is selected, it will be used as the first parameter of the function.
        /// </summary>
        /// <param name="func">function name</param>
        private void InsertFunction2(string func)
        {
            tbxExpression.Paste(func + "(" + tbxExpression.SelectedText + "; )");
            tbxExpression.Focus();
        }

        /// <summary>
        /// Inserts given text to the current location in the expression editor.
        /// If some text is selected, the new text will be placed before the selection.
        /// </summary>
        /// <param name="text"></param>
        private void InsertText(string text)
        {
            int start = tbxExpression.SelectionStart;
            int len = tbxExpression.SelectionLength;
            tbxExpression.SelectionLength = 0;
            tbxExpression.Paste(text);
            tbxExpression.SelectionStart = start + text.Length;
            tbxExpression.SelectionLength = len;
            tbxExpression.Focus();
        }

        private void btPow_Click(object sender, EventArgs e)
        {
            InsertFunction2("pow");
        }

        private void btLog_Click(object sender, EventArgs e)
        {
            InsertFunction2("log");
        }

        private void btExp_Click(object sender, EventArgs e)
        {
            InsertFunction("exp");
        }

        private void btLn_Click(object sender, EventArgs e)
        {
            InsertFunction("ln");
        }

        private void btSqrt_Click(object sender, EventArgs e)
        {
            InsertFunction("sqrt");
        }

        private void btSin_Click(object sender, EventArgs e)
        {
            InsertFunction("sin");
        }

        private void btCos_Click(object sender, EventArgs e)
        {
            InsertFunction("cos");
        }

        private void btTan_Click(object sender, EventArgs e)
        {
            InsertFunction("tan");
        }

        private void btAsin_Click(object sender, EventArgs e)
        {
            InsertFunction("asin");
        }

        private void btAcos_Click(object sender, EventArgs e)
        {
            InsertFunction("acos");
        }

        private void btAtan_Click(object sender, EventArgs e)
        {
            InsertFunction("atan");
        }

        private void btSinh_Click(object sender, EventArgs e)
        {
            InsertFunction("sinh");
        }

        private void btCosh_Click(object sender, EventArgs e)
        {
            InsertFunction("cosh");
        }

        private void btTanh_Click(object sender, EventArgs e)
        {
            InsertFunction("tanh");
        }

        private void btSmul_Click(object sender, EventArgs e)
        {
            InsertFunction2("smul");
        }

        private void btVmul_Click(object sender, EventArgs e)
        {
            InsertFunction2("vmul");
        }

        private void btNorm_Click(object sender, EventArgs e)
        {
            InsertFunction("norm");
        }

        private void btAbs_Click(object sender, EventArgs e)
        {
            InsertFunction("abs");
        }

        private void btCeil_Click(object sender, EventArgs e)
        {
            InsertFunction("ceil");
        }

        private void btFloor_Click(object sender, EventArgs e)
        {
            InsertFunction("floor");
        }

        private void btIvec_Click(object sender, EventArgs e)
        {
            InsertText("ivec");
        }

        private void btJvec_Click(object sender, EventArgs e)
        {
            InsertText("jvec");
        }

        private void btKvec_Click(object sender, EventArgs e)
        {
            InsertText("kvec");
        }

        private void btDiv_Click(object sender, EventArgs e)
        {
            InsertText("/");
        }

        private void btMul_Click(object sender, EventArgs e)
        {
            InsertText("*");
        }

        private void btSub_Click(object sender, EventArgs e)
        {
            InsertText("-");
        }

        private void btAdd_Click(object sender, EventArgs e)
        {
            InsertText("+");
        }

        private void btLpar_Click(object sender, EventArgs e)
        {
            InsertText("(");
        }

        private void btRpar_Click(object sender, EventArgs e)
        {
            InsertText(")");
        }

        private void btNum0_Click(object sender, EventArgs e)
        {
            InsertText("0");
        }

        private void btNum1_Click(object sender, EventArgs e)
        {
            InsertText("1");
        }

        private void btNum2_Click(object sender, EventArgs e)
        {
            InsertText("2");
        }

        private void btNum3_Click(object sender, EventArgs e)
        {
            InsertText("3");
        }

        private void btNum4_Click(object sender, EventArgs e)
        {
            InsertText("4");
        }

        private void btNum5_Click(object sender, EventArgs e)
        {
            InsertText("5");
        }

        private void btNum6_Click(object sender, EventArgs e)
        {
            InsertText("6");
        }

        private void btNum7_Click(object sender, EventArgs e)
        {
            InsertText("7");
        }

        private void btNum8_Click(object sender, EventArgs e)
        {
            InsertText("8");
        }

        private void btNum9_Click(object sender, EventArgs e)
        {
            InsertText("9");
        }

        private void cbxOutputType_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (!specialFields) return;

            if (cbxDataType.SelectedIndex == 1)
            {
                chbOutputCoords.Visible = false;
                tbxCellsName.Visible = true;
            }
            else
            {
                tbxCellsName.Visible = false;
                chbOutputCoords.Visible = true;
            }
        }

        private void btClear_Click(object sender, EventArgs e)
        {
            tbxExpression.Clear();
        }

        private void btHelp_Click(object sender, EventArgs e)
        {
            ShowHelp();
        }
    }
}