using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.Windows.Forms;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// This is an UnstrGrid version of a module that calculates new attribute values from existing attributes based on given expression.
    /// You can choose whether to work with point attributes or cell attributes.
    /// With point data, you can save the result as new coordinates or as a point attribute.
    /// For cell data, you have to specify the name of the cells array to be used.
    /// The results can be saved to a new attribute or an existing one, if an attribute with the specified name already exists.<br></br>
    /// <br></br>
    /// Expression syntax:<br></br>
    ///   <b>numbers</b> with '.' or ',' as decimal separator<br></br>
    ///   names of input <b>attributes</b><br></br>
    ///   basic operations '<b>+</b>', '<b>-</b>', '<b>*</b>', '<b>/</b>'<br></br>
    ///   <b>point</b> ... coordinates of the current point (only while working with points)<br></br>
    ///   <b>index</b> ... index of the current point or cell<br></br>
    ///   <b>pow(x; y)</b> ... x to the power of y<br></br>
    ///   <b>log(x; y)</b> ... logarithm of x with the root y<br></br>
    ///   <b>exp(x)</b> ... exponential function of x<br></br>
    ///   <b>ln(x)</b> ... natural logarithm of x<br></br>
    ///   <b>sqrt(x)</b> ... square root of x<br></br>
    ///   <b>sin(x)</b>, <b>cos(x)</b>, <b>tan(x)</b> ... goniometric functions of x<br></br>
    ///   <b>asin(x)</b>, <b>acos(x)</b>, <b>atan(x)</b> ... inverse goniometric functions of x<br></br>
    ///   <b>sinh(x)</b>, <b>cosh(x)</b>, <b>tanh(x)</b> ... hyperbolic functions of x<br></br>
    ///   <b>abs(x)</b> ... absolute value of x<br></br>
    ///   <b>ceil(x)</b> ... x rounded to the next higher integer<br></br>
    ///   <b>floor(x)</b> ... x rounded to the next lower integer<br></br>
    ///   <b>ivec</b>, <b>jvec</b>, <b>kvec</b> ... unit vectors of coordinate axes<br></br>
    ///   <b>norm(x)</b> ... magnitude of vector x<br></br>
    ///   <b>smul(x; y)</b> ... dot product of vectors x and y<br></br>
    ///   <b>vmul(x; y)</b> ... cross product of vectors x and y<br></br>
    ///   <b>x[n]</b> ... value of the nth element of vector x<br></br>
    /// <br></br>
    /// Functions norm, smul, vmul and [] require vector operand(s).
    /// Addition and subtraction can be performed on either two vectors or two scalars,
    /// multiplication and division on one vector and one scalar.
    /// All other functions work only with scalar operand(s).
    /// You can create a vector by multiplying the coordinates by ivec, jvec and kvec and adding them together.
    /// </summary>
    [ModuleInfo("Jan Krej, Oldich Petk", "A module that performs scientific calculations on point or cell attributes of an UnstrGrid.", IconName="calc.ico")]
    [PortInfo("Input", "Input UnstrGrid data structure. Must contain attributes used in the expression.")]
    [PortInfo("Ouptut", "Output UnstrGrid containing resulting attribute of the specified name. If attribute with such name already exists, it will be overwritten.")]
    public class UnstrGridCalculator : Mve.Core.Module
    {
        private const string helpStr =
            "This is an UnstrGrid version of a module that calculates new attribute values from existing attributes based on given expression. " +
            "You can choose whether to work with point attributes or cell attributes using the combo box labeled \"Work with\". " +
            "With point data selected, you can use the checkbox on the right to save the result as new coordinates or as a point attribute. " +
            "For cell data, you have to specify the name of the cells array to be used, in the text box on the right. " +
            "The results can be saved to a new attribute or an existing one, if an attribute with the specified name already exists.\n\n" +
            "Expression syntax:\n" +
            "  numbers with '.' or ',' as decimal separator\n" +
            "  names of input attributes\n" +
            "  basic operations '+', '-', '*', '/'\n" +
            "  point ... coordinates of the current point (only while working with points)\n" +
            "  index ... index of the current point or cell\n" +
            "  pow(x; y) ... x to the power of y\n" +
            "  log(x; y) ... logarithm of x with the root y\n" +
            "  exp(x) ... exponential function of x\n" +
            "  ln(x) ... natural logarithm of x\n" +
            "  sqrt(x) ... square root of x\n" +
            "  sin(x), cos(x), tan(x) ... goniometric functions of x\n" +
            "  asin(x), acos(x), atan(x) ... inverse goniometric functions of x\n" +
            "  sinh(x), cosh(x), tanh(x) ... hyperbolic functions of x\n" +
            "  abs(x) ... absolute value of x\n" +
            "  ceil(x) ... x rounded to the next higher integer\n" +
            "  floor(x) ... x rounded to the next lower integer\n" +
            "  ivec, jvec, kvec ... unit vectors of coordinate axes\n" +
            "  norm(x) ... magnitude of vector x\n" +
            "  smul(x; y) ... dot product of vectors x and y\n" +
            "  vmul(x; y) ... cross product of vectors x and y\n" +
            "  x[n] ... value of the nth element of vector x\n\n" +
            "Functions norm, smul, vmul and [] require vector operand(s). " +
            "Addition and subtraction can be performed on either two vectors or two scalars, " +
            "multiplication and division on one vector and one scalar. " +
            "All other functions work only with scalar operand(s). " +
            "You can create a vector by multiplying the coordinates by ivec, jvec and kvec and adding them together.";

        /// <summary>
        /// Expression analyzer and evaluator
        /// </summary>
        private Analyzer engine;

        /// <summary>
        /// Instance of a setup dialog
        /// </summary>
        private CalculatorSetup cs;

        /// <summary>
        /// List of attributes used in the expression along with their values for the current point or cell
        /// </summary>
        Dictionary<string, double[]> var;

        /// <summary>
        /// List of attributes used in the expression
        /// </summary>
        List<string> varNames;

        /// <summary>
        /// Number of elements in the input data array (points or cells)
        /// </summary>
        private int inputLength;

        /// <summary>
        /// Pointer to the input data structure
        /// </summary>
        UnstrGrid data;

        /// <summary>
        /// Module constructor.
        /// </summary>
        public UnstrGridCalculator()
        {
            AddInPort("Input", typeof(UnstrGrid));
            AddOutPort("Output", typeof(UnstrGrid));

            cs = new CalculatorSetup(true, helpStr);
        }

        /// <summary>
        /// Writes module configuration to the map file.
        /// </summary>
        /// <param name="config">XML element with the module configuration</param>
        public override void WriteConfig(System.Xml.XmlElement config)
        {
            config.SetAttribute("cellData", cs.CellData.ToString());
            config.SetAttribute("cellsName", cs.CellsName.ToString());
            config.SetAttribute("coordOutput", cs.CoordinateOutput.ToString());
            config.SetAttribute("outputName", cs.OutputName.ToString());
            config.InnerText = cs.Expression;
        }

        /// <summary>
        /// Reads saved module configuration.
        /// </summary>
        /// <param name="config">XML element with the module configuration</param>
        public override void ReadConfig(System.Xml.XmlElement config)
        {
            cs.SetConfig(
                config.InnerText,
                bool.Parse(config.GetAttribute("cellData")),
                config.GetAttribute("cellsName"),
                config.GetAttribute("outputName"),
                bool.Parse(config.GetAttribute("coordOutput"))
            );
        }

        /// <summary>
        /// Executes the calculation.
        /// </summary>
        public override void Execute()
        {
            Dictionary<string, UniformDataArray> attrs = new Dictionary<string,UniformDataArray>();
            double[] result;
            UniformDataArray results, current;
            int resLen;
            Vector2D v2d;
            Vector3D v3d;
            long ticks;

            if (cs.Expression.Trim().Length == 0)
            {
                SetOutput("Output", GetInput("Input"), IsAllDataSame());
                return;
            }

            // some of the settings have changed since last run, we need to analyze the expression
            if (cs.Changed)
            {
                engine = new Analyzer();
                varNames = engine.Analyze(cs.Expression);
            }

            // input or settings have changed, new calculation is needed
            if (!IsAllDataSame() || cs.Changed)
            {
                if (varNames == null) varNames = new List<string>();
                var = new Dictionary<string, double[]>(varNames.Count);

                ticks = DateTime.Now.Ticks;

                // make a shallow copy of the input
                data = ((UnstrGrid)GetInput("Input")).ShallowCopy();
                cs.Changed = false;

                // now we can extract the arrays of attributes needed for calculation
                if (cs.CellData)
                {
                    varNames.Remove("index");

                    if (!data.ContainsCells(cs.CellsName)) throw new Exception("Input data does not contain cells named " + cs.CellsName);

                    foreach (string attr in varNames)
                    {
                        if (data.ContainsCellAttrs(cs.CellsName, attr))
                            attrs.Add(attr, data.GetCellAttrs(cs.CellsName, attr));
                        else
                            throw new Exception("Input data does not contain cell attributes named " + attr);
                    }
                    inputLength = data.GetCells(cs.CellsName).Lengths[0];
                }
                else
                {
                    varNames.Remove("point");
                    varNames.Remove("index");

                    foreach (string attr in varNames)
                    {
                        if (data.ContainsPointAttrs(attr))
                            attrs.Add(attr, data.GetPointAttrs(attr));
                        else
                            throw new Exception("Input data does not contain point attributes named " + attr);
                    }
                    inputLength = data.Points.Lengths[0];
                }

                // we need to convert MVE types to double arrays for the calculator engine
                var["index"] = new double[] { 0 };
                if (!cs.CellData)
                {
                    if (data.Points.GetType() == typeof(Point2D))
                        var["point"] = ((PointND)(Point2D)(data.Points[0])).Values;
                    else if (data.Points.ElemType == typeof(Point3D))
                        var["point"] = ((PointND)(Point3D)(data.Points[0])).Values;
                    else
                        var["point"] = ((PointND)(data.Points[0])).Values;
                }
                foreach (string attr in varNames)
                {
                    current = attrs[attr];
                    if (current.ElemType == typeof(Scalar)) var[attr] = new double[] { ((Scalar)(attrs[attr][0])).Value };
                    else if (current.ElemType == typeof(Integer)) var[attr] = new double[] { ((Integer)attrs[attr][0]).Value };
                // this is what happens, when you have six incompatible types for the same thing
                    else if (current.ElemType == typeof(Vector2D))
                    {
                        v2d = (Vector2D)(attrs[attr][0]);
                        var[attr] = new double[] { v2d.X, v2d.Y };
                    }
                    else if (current.ElemType == typeof(Vector3D))
                    {
                        v3d = (Vector3D)(attrs[attr][0]);
                        var[attr] = new double[] { v3d.X, v3d.Y, v3d.Z };
                    }
                    else if (current.ElemType == typeof(VectorND))
                    {
                        var[attr] = ((VectorND)(attrs[attr][0])).Values;
                    }
                    else if (current.ElemType == typeof(Normal2D))
                    {
                        v2d = (Normal2D)(attrs[attr][0]);
                        var[attr] = new double[] { v2d.X, v2d.Y };
                    }
                    else if (current.ElemType == typeof(Normal3D))
                    {
                        v3d = (Normal3D)(attrs[attr][0]);
                        var[attr] = new double[] { v3d.X, v3d.Y, v3d.Z };
                    }
                    else if (current.ElemType == typeof(NormalND))
                    {
                        var[attr] = ((NormalND)(attrs[attr][0])).Values;
                    }
                }
                // run the first pass (checks for all errors - array overflows, invalid vector operations etc.)
                result = engine.FirstResult(var);

                // convert result back to MVE datatypes
                resLen = result.Length;
                switch (resLen)
                {
                    case 1:
                        results = new UniformDataArray(typeof(Scalar), inputLength);
                        results[0] = new Scalar(result[0]);
                        break;
                    case 2:
                        results = new UniformDataArray(typeof(Vector2D), inputLength);
                        results[0] = new Vector2D(result[0], result[1]);
                        break;
                    case 3:
                        results = new UniformDataArray(typeof(Vector3D), inputLength);
                        results[0] = new Vector3D(result[0], result[1], result[2]);
                        break;
                    default:
                        results = new UniformDataArray(typeof(VectorND), inputLength);
                        results[0] = new VectorND(result);
                        break;
                }

                // perform second pass calculation on the remaining elements (checks only for runtime errors - division by zero etc.)
                for (int i = 1; i < inputLength; i++)
                {
                    ProgressInfo((double)i / inputLength);

                    var["index"] = new double[] { i };
                    if (!cs.CellData)
                    {
                        if (data.Points.ElemType == typeof(Point2D))
                            var["point"] = ((PointND)(Point2D)(data.Points[i])).Values;
                        else if (data.Points.ElemType == typeof(Point3D))
                            var["point"] = ((PointND)(Point3D)(data.Points[i])).Values;
                        else
                            var["point"] = ((PointND)(data.Points[i])).Values;
                    }
                    foreach (string attr in varNames)
                    {
                        current = attrs[attr];
                        if (current.ElemType == typeof(Scalar)) var[attr] = new double[] { ((Scalar)attrs[attr][i]).Value };
                        else if (current.ElemType == typeof(Integer)) var[attr] = new double[] { ((Integer)attrs[attr][i]).Value };
                    // and again and again...
                        else if (current.ElemType == typeof(Vector2D))
                        {
                            v2d = (Vector2D)(attrs[attr][i]);
                            var[attr] = new double[] { v2d.X, v2d.Y };
                        }
                        else if (current.ElemType == typeof(Vector3D))
                        {
                            v3d = (Vector3D)(attrs[attr][i]);
                            var[attr] = new double[] { v3d.X, v3d.Y, v3d.Z };
                        }
                        else if (current.ElemType == typeof(VectorND))
                        {
                            var[attr] = ((VectorND)(attrs[attr][i])).Values;
                        }
                        else if (current.ElemType == typeof(Normal2D))
                        {
                            v2d = (Normal2D)(attrs[attr][i]);
                            var[attr] = new double[] { v2d.X, v2d.Y };
                        }
                        else if (current.ElemType == typeof(Normal3D))
                        {
                            v3d = (Normal3D)(attrs[attr][i]);
                            var[attr] = new double[] { v3d.X, v3d.Y, v3d.Z };
                        }
                        else if (current.ElemType == typeof(NormalND))
                        {
                            var[attr] = ((NormalND)(attrs[attr][i])).Values;
                        }
                    }

                    result = engine.SecondResult(var);

                    switch (resLen)
                    {
                        case 1:
                            results[i] = new Scalar(result[0]);
                            break;
                        case 2:
                            results[i] = new Vector2D(result[0], result[1]);
                            break;
                        case 3:
                            results[i] = new Vector3D(result[0], result[1], result[2]);
                            break;
                        default:
                            results[i] = new VectorND(result);
                            break;
                    }
                }

                // save resulting array
                if (cs.CellData)
                {
                    // add or replace attribute of the current cells
                    if (data.ContainsCellAttrs(cs.CellsName, cs.OutputName)) data.RemoveCellAttrs(cs.CellsName, cs.OutputName);
                    data.AddCellAttrs(cs.CellsName, cs.OutputName, results);
                }
                else
                {
                    // replace point coordinates with results (only valid for vector output)
                    if (cs.CoordinateOutput)
                    {
                        switch (resLen)
                        {
                            case 1:
                                throw new Exception("Scalar output can not be assigned into coordinates!");
                            case 2:
                                current = new UniformDataArray(typeof(Point2D), inputLength);
                                for (int i = 0; i < inputLength; i++) current[i] = (Point2D)(Vector2D)(results[i]);
                                break;
                            case 3:
                                current = new UniformDataArray(typeof(Point3D), inputLength);
                                for (int i = 0; i < inputLength; i++) current[i] = (Point3D)(Vector3D)(results[i]);
                                break;
                            default:
                                current = new UniformDataArray(typeof(PointND), inputLength);
                                for (int i = 0; i < inputLength; i++) current[i] = (PointND)(VectorND)(results[i]);
                                break;
                        }
                        data.Points = current;
                    }
                    // add or replace points' attribute
                    else
                    {
                        if (data.ContainsPointAttrs(cs.OutputName)) data.RemovePointAttrs(cs.OutputName);
                        data.AddPointAttrs(cs.OutputName, results);
                    }
                }

                ticks = DateTime.Now.Ticks - ticks;
                Console.WriteLine("Calculation ended: {0} values in {1} ms.", inputLength, (int)(ticks / 10000.0));

                // send resulting data structure to output
                SetOutput("Output", data, false);
            }
            // input data and module settings haven't changed since last run
            else
            {
                // we can send last data structure to output with dataSame indication
                SetOutput("Output", data, true);
            }
        }

        /// <summary>
        /// Shows the module setup dialog.
        /// </summary>
        /// <returns>instance of the setup dialog</returns>
        public override Zcu.Mve.Core.ModuleSetup InvokeSetup()
        {
            return cs;
        }
    }
}
