using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Visualization;
using Zcu.Mve.Core;
using System.ComponentModel;


namespace Zcu.Mve.Visualization
{
    /// <summary>
    /// Datastructure represents color in Grayscale. Internally it uses byte for the intensity componnent.
    /// </summary>
    public struct ColorIntensity : IAttribute, IColorVector
    {
        #region Public members
        /// <summary>
        /// Intensity.
        /// </summary>
        private byte intensity;
        #endregion

        #region Properties
        
        [Category("Color value"), Description("Gets/sets intensity.")]
        public byte Intensity
        {
            get
            {
                return intensity;
            }
            set
            {
                intensity = value;
            }
        }

        /// <summary>
        /// Gets or sets the intensity from float.
        /// </summary>
        [Category("Float color value"), Description("Gets/sets intensity as float (0 to 1).")]
        public float IntensityF
        {
            get
            {
                return (float)(Intensity / 256.0);
            }
            set
            {
                int fullint = (byte)(value * 256);
                Intensity = (Byte)Math.Max(0,Math.Min(255,fullint)); 
            }
        }

        /// <summary>
        /// Gets or sets the intensity from double.
        /// </summary>
        [Category("Double color value"), Description("Gets/sets intensity as float (0 to 1).")]
        public double IntensityD
        {
            get
            {
                return (double)(Intensity / 256.0);
            }
            set
            {
                IntensityF = (float)value;
            }
        }
        #endregion

        #region Construcotrs
        /// <summary>
        /// Constructor creates an instance of ColorGray.
        /// </summary>
        /// <param name="intensity">Byte intensity</param>
        public ColorIntensity(byte intensity)
        {
            this.intensity = intensity;
        }

        /// <summary>
        /// Constructor creates an instance of ColorGray.
        /// </summary>
        /// <param name="intensity">Float intensity clamped to (0.0,1.0), converted to byte.</param>
        public ColorIntensity(float intensity)
        {
            this.intensity = 0;
            IntensityF = intensity;
        }

        /// <summary>
        /// Constructor creates an instance of ColorGray.
        /// </summary>
        /// <param name="intensity">Double intensity clamped to (0.0,1.0), converted to byte.</param>
        public ColorIntensity(double intensity)
        {
            this.intensity = 0;
            this.IntensityD = intensity;
        }
        #endregion

        #region Operators
        #endregion

        #region Overrided methods of the class Object
        /// <summary>
        /// Method read data from Xml file to structures of this data object.
        /// </summary>
        /// <param name="xmlTextReader">Xml stream to read.</param>
        public void ReadData(System.Xml.XmlTextReader xmlTextReader)
        {
            string tmp = xmlTextReader.ReadString().Trim();
            this.Intensity = Byte.Parse(tmp);
        }

        /// <summary>
        /// Method write data from structures of this data object to Xml file.
        /// </summary>
        /// <param name="xmlTextWriter">Xml stream to write.</param>
        public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
        {
            xmlTextWriter.WriteString(this.Intensity.ToString());
        } 

        /// <summary>
        /// Deep copy of data object and its subelements.
        /// </summary>
        /// <returns>Same data, different memory space.</returns>
        public IDataObject DeepCopy()
        {
            return new ColorIntensity(this.Intensity);
        }

        /// <summary>
        /// Check the consistency of current data structure.
        /// </summary>
        /// <returns>True if structure is consistent.</returns>
        public bool CheckConsistence()
        {
            return true;
        } 

        /// <summary>
        /// Returns array of assembly name.
        /// </summary>
        [Browsable(false)]
        public string[] LibNames
        {
            get
            {
                return new string[] { Globals.GetLibraryName(this.GetType()) };
            }
        }

        /// <summary>
        /// Returns a String that represents the current Object.
        /// </summary>
        /// <returns>String that represents the current Object.</returns>
        public override string ToString()
        {
            return "ColorIntensity: Intensity = "+this.Intensity.ToString();
        }

        /// <summary>
        /// Determines whether the specified Objects are equal.
        /// </summary>
        /// <param name="a">First Object.</param>
        /// <param name="b">Second Object.</param>
        /// <returns>True if objects are equal. Otherwise, false.</returns>
        public static new bool Equals(object a, object b)
        {
            if (a.GetType() != typeof(ColorIntensity) || b.GetType() != typeof(ColorIntensity))
                return false;

            ColorIntensity c1 = (ColorIntensity)a;
            ColorIntensity c2 = (ColorIntensity)b;

            return c1.Intensity == c2.Intensity;
        }

        /// <summary>
        /// Determines whether the specified Object is equal to the current Object.
        /// </summary>
        /// <param name="obj">Specified Object to compare with the current Object.</param>
        /// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
        public override bool Equals(object obj)
        {
            return ColorIntensity.Equals(this, obj);
        }

        /// <summary>
        /// Returns a hash code for the current Object.
        /// </summary>
        /// <returns>A hash code for the current Object.</returns>
        public override int GetHashCode()
        {
            return this.Intensity.GetHashCode();
        }

        #endregion

        #region IColorVector Members

        /// <summary>
        /// Get/Set element of color vector.
        /// </summary>
        public byte this[int index]
        {
            get
            {
                if (index == 0) return Intensity;
                else throw new Exception("Color plane index is out of range.");
            }
            set
            {
                if (index == 0) Intensity = value;
                else throw new Exception("Color plane index is out of range.");
            }
        }

        /// <summary>
        /// Get number of elements of vector.
        /// </summary>
        [Browsable(false)]
        public int Length
        {
            get { return 1; }
        }

        #endregion
    }
}
