using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Globalization;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.Visualization.Renderers
{
	/// <summary>
	/// Panel for curve rendering.
	/// </summary>
	public class CurveRendererPanel : System.Windows.Forms.Form
	{
		private System.ComponentModel.IContainer components = null;
		private System.Windows.Forms.StatusBar statusBar1;
		private System.Windows.Forms.StatusBarPanel statusBarPanel1;
		private System.Windows.Forms.MainMenu mainMenu1;
		private System.Windows.Forms.MenuItem menuItem1;
		private System.Windows.Forms.MenuItem menuItem3;

		#region Constants and variables for rendering

		/// <summary> The renderer object. </summary>
		internal CurveRenderer renderer;
		/// <summary> Settings panel. </summary>
		private CurveRendererSettings settingsDialog;
		/// <summary> Data view panel. </summary>
		private CurveRendererDataView dataView;

		/// <summary> Auxiliary field for selected point highlighting. </summary>
		internal int highlightedPointIndex = 0;
		/// <summary> Auxiliary field for selected point highlighting. </summary>
		internal bool highlightProgress = false;

		/// <summary> 
		/// Extreme values - this values define bounding box of the graph. 
		/// </summary>
		private float minX;
		private float maxX;
		private float minY;
		private float maxY;

		/// <summary>
		/// Graph-top window margin spacing.
		/// </summary>
		private const float SPACING_TOP = 30;
		/// <summary>
		/// Graph-bottom window margin spacing.
		/// </summary>
		private const float SPACING_BOTTOM = 50;	 // including status bar panel
		/// <summary>
		/// Graph-left window margin spacing.
		/// </summary>
		private const float SPACING_LEFT = 50;
		/// <summary>
		/// Graph-right window margin spacing.
		/// </summary>
		private const float SPACING_RIGHT = 30;

		/// <summary>
		/// Graph width.
		/// </summary>
		private float width;
		/// <summary>
		/// Graph heihgt.
		/// </summary>
		private float height;

		/// <summary>
		/// Axis marker size.
		/// </summary>
		private const float AXIS_MARK_SIZE = 5;

		/// <summary>
		/// x-segment size.
		/// </summary>
		private float segmentX;
		/// <summary>
		/// y-segment size.
		/// </summary>
		private float segmentY;

		/// <summary>
		/// If true, rendering is in progress.
		/// </summary>
		private bool renderingInProgress = false;

        private System.Drawing.PointF[] bezierArray;

		#endregion

        #region Methods for spline computation

        private void prepareTangents()
        {
            // prepare points
            
            Point2D[] tangentArray = new Point2D[renderer.data.Count];
            Point2D[] pointArray = new Point2D[renderer.data.Count];
            for (int i = 0; i < renderer.data.Count; i++)
                pointArray[i] = (Point2D)renderer.data[i];
            
            Point2D[,] matrix = new Point2D[pointArray.Length, 4];

            // prepare matrix
            //oprava Jandys 7. 11. 2005
            /* 
            matrix[0, 0] = new Point2D(0, 0);
            matrix[0, 1] = pointArray[0] * 2;
            matrix[0, 2] = pointArray[1];
            matrix[0, 3] = 3 * (pointArray[1] - pointArray[0]);
            for (int i = 1; i < (pointArray.Length - 1); i++)
            {
                matrix[i, 0] = pointArray[i - 1];
                matrix[i, 1] = pointArray[i] * 4;
                matrix[i, 2] = pointArray[i + 1];
                matrix[i, 3] = 3 * (pointArray[i + 1] - pointArray[i - 1]);
            }
            int last = pointArray.Length - 1;
            matrix[last, 0] = pointArray[last-1];
            matrix[last, 1] = pointArray[last] * 2;
            matrix[last, 2] = new Point2D(0, 0);
            matrix[last, 3] = 3 * (pointArray[last] - pointArray[last-1]);
             */

            matrix[0, 0] = new Point2D(0, 0);
            matrix[0, 1] = new Point2D(2, 2);
            matrix[0, 2] = new Point2D(1, 1);
            matrix[0, 3] = 3 * (pointArray[1] - pointArray[0]);
            for (int i = 1; i < (pointArray.Length - 1); i++)
            {
                matrix[i, 0] = new Point2D(1, 1);
                matrix[i, 1] = new Point2D(4, 4);
                matrix[i, 2] = new Point2D(1, 1);
                matrix[i, 3] = 3 * (pointArray[i + 1] - pointArray[i - 1]);
            }
            int last = pointArray.Length - 1;
            matrix[last, 0] = new Point2D(1, 1);
            matrix[last, 1] = new Point2D(2, 2); 
            matrix[last, 2] = new Point2D(0, 0);
            matrix[last, 3] = 3 * (pointArray[last] - pointArray[last - 1]);

            // solve matrix

            // forward run
            Point2D c;
            for (int i = 1; i < pointArray.Length; i++)
            {
                c = matrix[i - 1, 1] / matrix[i, 0];
                
                // multiply by constant
                //matrix[i, 0] *= c; //
                matrix[i, 1] *= c;
                matrix[i, 2] *= c;
                matrix[i, 3] *= c;

                // subtract the row above
                //matrix[i, 0] -= matrix[i - 1, 1];//
                matrix[i, 1] -= matrix[i - 1, 2];                
                matrix[i, 3] -= matrix[i - 1, 3];
            }

            // backward run
            for (int i = pointArray.Length-2; i > -1; i--)
            {
                c = matrix[i + 1, 1] / matrix[i, 2];

                // multiply by constant
                //matrix[i, 0] *= c;//
                matrix[i, 1] *= c;
                //matrix[i, 2] *= c;//
                matrix[i, 3] *= c;

                // subtract the row below
                matrix[i, 3] -= matrix[i + 1, 3];
                //matrix[i, 1] -= matrix[i + 1, 0];//
                //matrix[i, 2] -= matrix[i + 1, 1];//
            }

            // copy values
            for (int i = 0; i < pointArray.Length; i++)
            {
                tangentArray[i] = (matrix[i, 3] / matrix[i, 1])/3;
            }

            // create array for DrawBeziers
            bezierArray = new PointF[pointArray.Length+2*(pointArray.Length-1)];
            bezierArray[0] = new PointF((float)pointArray[0].X, (float)pointArray[0].Y);
            for (int i = 1; i < pointArray.Length; i++)
            {
                Point2D p = tangentArray[i - 1] + pointArray[i - 1];
                bezierArray[i * 3 - 2] = new PointF((float)p.X, (float)p.Y);
                p = pointArray[i] - tangentArray[i];
                bezierArray[i * 3 - 1] = new PointF((float)p.X, (float)p.Y);
                p = pointArray[i];
                bezierArray[i * 3] = new PointF((float)p.X, (float)p.Y);
            }
        }

        #endregion

        /// <summary>
        /// Creates and initializes the panel.
        /// </summary>
        /// <param name="data"></param>
        public CurveRendererPanel(CurveRenderer renderer)
		{
			this.renderer = renderer;
			// Required for Windows Form Designer support
			InitializeComponent();		
			this.Text = this.renderer.Name + " panel";
		} // CurveRendererPanel()


		/// <summary>
		/// Starts rendering process.
		/// </summary>
		internal void Render()
		{
			this.renderingInProgress = true;

			// find extreme values
			this.minX = this.FindMinX();
			this.maxX = this.FindMaxX();
			this.minY = this.FindMinY();
			this.maxY = this.FindMaxY();

			width  = this.renderer.max_x - CurveRendererPanel.SPACING_LEFT - CurveRendererPanel.SPACING_RIGHT;
			height = this.renderer.max_y - CurveRendererPanel.SPACING_TOP - CurveRendererPanel.SPACING_BOTTOM;

			segmentX = width  / (float) this.renderer.x_segment_count;
			segmentY = height / (float) this.renderer.y_segment_count;

			// set renderer panel properties
			this.ClientSize = new Size(this.renderer.max_x, this.renderer.max_y);
			this.statusBarPanel1.Text = " Number of points: " + renderer.data.Count + " (for data view press F9)";
			
			this.ShowDialog();
		} // Render()


		#region Auxiliary methods for graph rendering


		#region Methods for extreme values finding

		private float FindMinX()
		{
			float min = Single.MaxValue;
			for (int i = 0; i < renderer.data.Count; i++)
			{
				Point2D currentPoint = (Point2D) renderer.data[i];
				if (currentPoint.X < min)
					min = (float) currentPoint.X;
			}
			return min;
		}

		private float FindMaxX()
		{
			float max = Single.MinValue;
			for (int i = 0; i < renderer.data.Count; i++)
			{
				Point2D currentPoint = (Point2D) renderer.data[i];
				if (currentPoint.X > max)
					max = (float) currentPoint.X;
			}
			return max;
		}

		private float FindMinY()
		{
			float min = Single.MaxValue;
			for (int i = 0; i < renderer.data.Count; i++)
			{
				Point2D currentPoint = (Point2D) renderer.data[i];
				if (currentPoint.Y < min)
					min = (float) currentPoint.Y;
			}
			return min;
		}

		private float FindMaxY()
		{
			float max = Single.MinValue;
			for (int i = 0; i < renderer.data.Count; i++)
			{
				Point2D currentPoint = (Point2D) renderer.data[i];
				if (currentPoint.Y > max)
					max = (float) currentPoint.Y;
			}
			return max;
		}

		#endregion


		/// <summary>
		/// Draw the graph.
		/// </summary>
		/// <param name="e">Event</param>
		protected override void OnPaint(PaintEventArgs e)
		{
			Graphics g = e.Graphics;
			this.RenderCurve(g);
			base.OnPaint (e);
		} // OnPaint()


		/// <summary>
		/// Refresh the panel after resize.
		/// </summary>
		/// <param name="e">Event</param>
		protected override void OnResize(EventArgs e)
		{
			if (this.renderingInProgress)
				this.Refresh();
			base.OnResize (e);
		} // OnResize()


		/// <summary>
		/// Refresh the panel.
		/// </summary>
		public override void Refresh()
		{
			if (this.renderingInProgress)
			{
				this.renderer.max_x = this.ClientSize.Width;
				this.renderer.max_y = this.ClientSize.Height;
	
				width  = this.renderer.max_x - CurveRendererPanel.SPACING_LEFT - CurveRendererPanel.SPACING_RIGHT;
				height = this.renderer.max_y - CurveRendererPanel.SPACING_TOP - CurveRendererPanel.SPACING_BOTTOM;

				segmentX = width  / (float) this.renderer.x_segment_count;
				segmentY = height / (float) this.renderer.y_segment_count;
			
				this.RenderCurve(this.CreateGraphics());
				base.Refresh ();
			}
		} // Refresh()


		/// <summary>
		/// Render all the graph including axis, grid and legend.
		/// </summary>
		/// <param name="g"></param>
		private void RenderCurve(Graphics g)
		{
			if (this.renderer.data.Count > 1)
			{
				/*if (this.renderer.interpolation == CurveRenderer.Interpolation.Bezier && this.CheckDataBezierConsistency(g) == false)
					return;*/

				// Inicializace pouzitych kreslicich nastroju.
				Pen axisPen = new Pen(new SolidBrush(this.renderer.axesLineColor), 2);
				Pen graphPen = new Pen(new SolidBrush(this.renderer.graphLineColor), renderer.graphLineWidth);
				// Vodorovna osa grafu
				DrawXAxis(g, axisPen);
				// Vodorovna osa grafu
				DrawYAxis(g, axisPen);
				// Ohraniceni grafu - obdelnik okolo.	
				if (this.renderer.showRectangle)
				{
					g.DrawRectangle(Pens.Black, 
						CurveRendererPanel.SPACING_LEFT, 
						CurveRendererPanel.SPACING_TOP, 
						this.renderer.max_x - CurveRendererPanel.SPACING_RIGHT - CurveRendererPanel.SPACING_LEFT, 
						this.renderer.max_y - CurveRendererPanel.SPACING_BOTTOM - CurveRendererPanel.SPACING_TOP);
				}
				// Vykresleni vlastniho grafu
				DrawMyGraph(g, graphPen);
			}
			else
			{
				Font warningFont = new Font(FontFamily.GenericMonospace, 10, FontStyle.Bold);
				g.DrawString("               -== ERROR: ==-\n"
					+ "Input array must contain two points minimally.\n"
					+ "           Cannot render the graph.", warningFont,
					new SolidBrush(Color.Red), 
					this.renderer.max_x / 2 - 185, 
					this.renderer.max_y / 2 - 50);
			}
		} // RenderCurve()


		/// <summary>
		/// Draw x-axis including legend.
		/// </summary>
		/// <param name="g">Graphics.</param>
		/// <param name="axisPen">Style of line.</param>
		/// <param name="axisFont">Font of legend.</param>
		private void DrawXAxis(Graphics g, Pen axisPen)
		{
			// computation of the y position of the x-axis
			float yPosition;
			if (this.minY > 0)
				yPosition = this.renderer.max_y - CurveRendererPanel.SPACING_BOTTOM;
			else if (this.maxY < 0)
				yPosition = CurveRendererPanel.SPACING_TOP;
			else
			{
				float ratio = maxY / (maxY - minY);
				yPosition = height * ratio + CurveRendererPanel.SPACING_TOP;
			}

			// draw axis line
			g.DrawLine(axisPen, 
				CurveRendererPanel.SPACING_LEFT, 
				yPosition,
				this.renderer.max_x - CurveRendererPanel.SPACING_RIGHT,
				yPosition);

			// axis scale step
			double step = (this.maxX - this.minX) / (float) this.renderer.x_segment_count;

			for (int i = 0; i <= this.renderer.x_segment_count; i++)
			{
				if (this.renderer.showXAxesLegend == true || (this.renderer.showXAxesLegend == false && (i == 0 || i == this.renderer.x_segment_count)))
				{
					g.DrawLine(axisPen, // svisla znacka na ose
						CurveRendererPanel.SPACING_LEFT + i * this.segmentX,
						yPosition - CurveRendererPanel.AXIS_MARK_SIZE,
						CurveRendererPanel.SPACING_LEFT + i * this.segmentX,
						yPosition + CurveRendererPanel.AXIS_MARK_SIZE);
					double currentXValue = this.minX + i * step;
					string textFormat = "f" + this.renderer.decPlaceCountX;
					g.DrawString(currentXValue.ToString(textFormat), this.renderer.axisFont, // popisky
						new SolidBrush(this.renderer.axesTextColor), 
						CurveRendererPanel.SPACING_LEFT + i * this.segmentX - 10,
						yPosition + 10);
				}
				if (this.renderer.showXGrid == true)
					g.DrawLine(Pens.Black,	// svisla mrizka grafu
						CurveRendererPanel.SPACING_LEFT + i * this.segmentX,
						CurveRendererPanel.SPACING_TOP,
						CurveRendererPanel.SPACING_LEFT + i * this.segmentX,
						this.renderer.max_y - CurveRendererPanel.SPACING_BOTTOM);
			}
		} // DrawXAxis()


		/// <summary>
		/// Draw y-axis including legend.
		/// </summary>
		/// <param name="g">Graphics.</param>
		/// <param name="axisPen">Style of line.</param>
		/// <param name="axisFont">Font of legend.</param>
		private void DrawYAxis(Graphics g, Pen axisPen)
		{
			// computation of the x position of the y-axis
			float xPosition;
			if (this.minX > 0)
				xPosition = CurveRendererPanel.SPACING_LEFT;
			else if (this.maxX < 0)
				xPosition = this.renderer.max_x - CurveRendererPanel.SPACING_RIGHT;
			else
			{
				float ratio = Math.Abs((minX / (minX - maxX)));
				xPosition = width * ratio + CurveRendererPanel.SPACING_LEFT;
			}

			// draw axis line
			g.DrawLine(axisPen,
				xPosition, 
				this.renderer.max_y - CurveRendererPanel.SPACING_BOTTOM,
				xPosition,
				CurveRendererPanel.SPACING_TOP);

			// axis scale step
			double step = (this.maxY - this.minY) / (float) this.renderer.y_segment_count;

			for (int i = 0; i <= this.renderer.y_segment_count; i++)
			{
				if (this.renderer.showYAxesLegend == true || (this.renderer.showYAxesLegend == false && (i == 0 || i == this.renderer.y_segment_count)))
				{
					g.DrawLine(axisPen,	// vodorovne znacky na ose
						xPosition - CurveRendererPanel.AXIS_MARK_SIZE,
						CurveRendererPanel.SPACING_TOP + height - i * this.segmentY,
						xPosition + CurveRendererPanel.AXIS_MARK_SIZE,
						CurveRendererPanel.SPACING_TOP + height - i * this.segmentY);
					double currentYValue = this.minY + i * step;
					string textFormat = "f" + this.renderer.decPlaceCountY;
					g.DrawString(currentYValue.ToString(textFormat), this.renderer.axisFont, // popisky
						new SolidBrush(this.renderer.axesTextColor), 
						xPosition - 40,
						CurveRendererPanel.SPACING_TOP + height - i * this.segmentY - 7);
				}
				if (this.renderer.showYGrid == true)
					g.DrawLine(Pens.Black,	// vodorovna mrizka grafu
						CurveRendererPanel.SPACING_LEFT,
						CurveRendererPanel.SPACING_TOP + height - i * this.segmentY,
						CurveRendererPanel.SPACING_LEFT + width,
						CurveRendererPanel.SPACING_TOP + height - i * this.segmentY);
			}		
		} // DrawYAxis()


		/// <summary>
		/// Draw the graph (marks and line).
		/// </summary>
		/// <param name="g">Graphics.</param>
		/// <param name="graphPen">Pen style.</param>
		private void DrawMyGraph(Graphics g, Pen graphPen)
		{
			double stepX = (this.maxX - this.minX) / (float) this.renderer.x_segment_count;
			double stepY = (this.maxY - this.minY) / (float) this.renderer.y_segment_count;

            // setup transform matrix
            RectangleF source = new RectangleF(minX, minY, maxX - minX, maxY - minY);
            PointF upperLeft = new PointF(CurveRendererPanel.SPACING_LEFT, CurveRendererPanel.SPACING_TOP + height);
            PointF upperRight = new PointF(CurveRendererPanel.SPACING_LEFT + width, CurveRendererPanel.SPACING_TOP + height);
            PointF lowerLeft = new PointF(CurveRendererPanel.SPACING_LEFT, CurveRendererPanel.SPACING_TOP);
            System.Drawing.Drawing2D.Matrix transform = new System.Drawing.Drawing2D.Matrix(source, new PointF[] { upperLeft, upperRight, lowerLeft });            

			System.Drawing.PointF[] pointArray = new System.Drawing.PointF[renderer.data.Count];
			for (int i = 0; i < pointArray.Length; i++)
			{
				Point2D p2D = (Point2D) renderer.data[i];
                System.Drawing.PointF point = new System.Drawing.PointF(
                    //(int) (CurveRendererPanel.SPACING_LEFT + ((p2D.X - minX) / (maxX - minX)) * width), 
                    //(int) (CurveRendererPanel.SPACING_TOP + ((maxY - p2D.Y) / (maxY - minY)) * height ));
                    (float)p2D.X, (float)p2D.Y);
				pointArray[i] = point;                				
			}

            transform.TransformPoints(pointArray);            

            for (int i = 0; i < pointArray.Length; i++)
            {
                if (this.renderer.showGraphMarks == true)
                {
                    if (this.renderer.allowHighlighting == false || this.highlightProgress == false)
                        DrawPoint(g, pointArray[i], i);
                    else
                    {
                        if (this.highlightedPointIndex != i)
                            DrawPoint(g, pointArray[i], i);
                    }
                }
            }

            if (this.renderer.allowHighlighting == true && this.highlightProgress == true)
            {
                DrawHighlightedPoint(g, pointArray[highlightedPointIndex]);
            }

			if (this.renderer.showGraphLine)
			{
                if (this.renderer.interpolation == CurveRenderer.Interpolation.PolyLine)
                {
                    //g.Transform = transform;
                    //transform.TransformPoints(pointArray);
                    g.DrawLines(graphPen, pointArray);
                    //g.Transform = new System.Drawing.Drawing2D.Matrix();
                }
                if (this.renderer.interpolation == CurveRenderer.Interpolation.Bezier)
                {
                    prepareTangents();
                    //g.Transform = transform;
                    g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
                    transform.TransformPoints(bezierArray);
                    g.DrawBeziers(graphPen, bezierArray);
                    //g.Transform = new System.Drawing.Drawing2D.Matrix();
                }
			}            
		} // DrawMyGraph()


		/// <summary>
		/// Checks consistency of input data for rendering using Bezier spline.
		/// If the data are inconsistent, writes message.
		/// </summary>
		/// <param name="g"></param>
		/// <returns></returns>
		private bool CheckDataBezierConsistency(Graphics g)
		{
			if ((this.renderer.data.Count - 4) % 3 == 0)
				return true;
			else
			{
				string message = "";
				message += "                            -== ERROR: ==- \n";
				message += "    Invalid number of input points for Bezier spline rendering.\n";
				message += "       Use polyline interpolation or correct the input data.\n\n";
				message += "Remarks: The first Bezier spline is drawn from the first point to the \n";
				message += "         fourth point in the point array. The second and third points \n";
				message += "         are control points that determine the shape of the curve. \n";
				message += "         Each subsequent curve needs exactly three more points: two \n";
				message += "         more control points and an ending point. The ending point \n";
				message += "         of the previous curve is used as the starting point for each \n";
				message += "         additional curve. \n";

				Font warningFont = new Font(FontFamily.GenericMonospace, 10, FontStyle.Bold);
				g.Clear(this.BackColor);			
				g.DrawString(message, warningFont,
					new SolidBrush(Color.Red), 
					this.renderer.max_x / 2 - 290, 
					this.renderer.max_y / 2 - 90);

				return false;
			}
		} // CheckDataBezierConsistency()


		/// <summary>
		/// Draw mark.
		/// </summary>
		/// <param name="g">Graphics.</param>
		/// <param name="point">Point.</param>
		/// <param name="index">Index of the point.</param>
		private void DrawPoint(Graphics g, System.Drawing.PointF point, int index)
		{
			Pen pen = new Pen(new SolidBrush(renderer.marksLineColor), renderer.graphMarkLineWidth);
			g.DrawLine(pen,
				point.X - renderer.graph_mark_size,
				point.Y,
				point.X + renderer.graph_mark_size,
				point.Y);
			g.DrawLine(pen,
				point.X,
				point.Y - renderer.graph_mark_size,
				point.X,
				point.Y + renderer.graph_mark_size);
			if (this.renderer.showIndexes)
			{
				g.DrawString(index.ToString(), this.renderer.indexFont, // popisky
					new SolidBrush(renderer.marksLineColor), 
					point.X + 3,
					point.Y - 15);
			}
		} // DrawPoint()


		/// <summary>
		/// Draw highlighted point.
		/// </summary>
		/// <param name="g">Graphics.</param>
		/// <param name="point">Point.</param>
		private void DrawHighlightedPoint(Graphics g, System.Drawing.PointF point)
		{
			Pen pen = new Pen(new SolidBrush(renderer.hMarksColor), renderer.hMarksLineWidth);
			g.DrawLine(pen,
				point.X - renderer.hMarksSize,
				point.Y,
				point.X + renderer.hMarksSize,
				point.Y);
			g.DrawLine(pen,
				point.X,
				point.Y - renderer.hMarksSize,
				point.X,
				point.Y + renderer.hMarksSize);
			/*g.DrawString(this.highlightedPointIndex.ToString(), this.renderer.hMarksFont, // popisky
				new SolidBrush(renderer.hMarksColor), 
				point.X + 3,
				point.Y - 15);*/
		} // DrawHighlightedPoint()


		#endregion


		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (this.settingsDialog != null)
				{
					this.settingsDialog.Close();
					this.settingsDialog.Dispose();
				}
				if (this.dataView != null)
				{
					this.dataView.Close();
					this.dataView.Dispose();
				}
				if(components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}


		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.statusBar1 = new System.Windows.Forms.StatusBar();
			this.statusBarPanel1 = new System.Windows.Forms.StatusBarPanel();
			this.mainMenu1 = new System.Windows.Forms.MainMenu();
			this.menuItem3 = new System.Windows.Forms.MenuItem();
			this.menuItem1 = new System.Windows.Forms.MenuItem();
			((System.ComponentModel.ISupportInitialize)(this.statusBarPanel1)).BeginInit();
			this.SuspendLayout();
			// 
			// statusBar1
			// 
			this.statusBar1.Font = new System.Drawing.Font("Tahoma", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((System.Byte)(238)));
			this.statusBar1.Location = new System.Drawing.Point(0, 395);
			this.statusBar1.Name = "statusBar1";
			this.statusBar1.Panels.AddRange(new System.Windows.Forms.StatusBarPanel[] {
																						  this.statusBarPanel1});
			this.statusBar1.ShowPanels = true;
			this.statusBar1.Size = new System.Drawing.Size(624, 22);
			this.statusBar1.TabIndex = 1;
			// 
			// statusBarPanel1
			// 
			this.statusBarPanel1.AutoSize = System.Windows.Forms.StatusBarPanelAutoSize.Spring;
			this.statusBarPanel1.Width = 608;
			// 
			// mainMenu1
			// 
			this.mainMenu1.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																					  this.menuItem3,
																					  this.menuItem1});
			this.mainMenu1.RightToLeft = System.Windows.Forms.RightToLeft.No;
			// 
			// menuItem3
			// 
			this.menuItem3.Index = 0;
			this.menuItem3.Shortcut = System.Windows.Forms.Shortcut.F9;
			this.menuItem3.Text = "Data View [F9]";
			this.menuItem3.Click += new System.EventHandler(this.menuItem3_Click);
			// 
			// menuItem1
			// 
			this.menuItem1.Index = 1;
			this.menuItem1.Shortcut = System.Windows.Forms.Shortcut.F10;
			this.menuItem1.Text = "Settings [F10]";
			this.menuItem1.Click += new System.EventHandler(this.menuItem1_Click);
			// 
			// CurveRendererPanel
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(624, 417);
			this.Controls.Add(this.statusBar1);
			this.Menu = this.mainMenu1;
			this.MinimumSize = new System.Drawing.Size(300, 200);
			this.Name = "CurveRendererPanel";
			this.Text = "CurveRenderer";
			((System.ComponentModel.ISupportInitialize)(this.statusBarPanel1)).EndInit();
			this.ResumeLayout(false);

		}
		#endregion


		#region Event handling


		/// <summary>
		/// Click on the menu 'Data view'
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void menuItem3_Click(object sender, System.EventArgs e)
		{
			dataView = new CurveRendererDataView(this);
			dataView.Text = this.renderer.Name + " data view";
			dataView.Show();	
			dataView.Owner = this;
		}


		/// <summary>
		/// Click on the menu 'Settings'
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void menuItem1_Click(object sender, System.EventArgs e)
		{
			settingsDialog = new CurveRendererSettings(this.renderer);
			settingsDialog.ShowDialog(this);
		}


		#endregion


	} // class CurveRendererPanel
} // namespace
