﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.ComponentModel;

using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.Visualization.Processing
{
  /// <summary>
  /// This modul provide cutting off geometric cells according to points scalar attribute. The name of
  /// points scalar attribute with evaluation for cut can be set by property. If attribut with given name
  /// not found, first points scalar attribute is use.
  /// </summary>
  [ModuleInfo("Tomáš Hofhans", 
    "This modul provide cutting off geometric cells according to points scalar attribute. The name of " +
    "points scalar attribute with evaluation for cut can be set by property. If property not set, " + 
    "first points scalar attribute is use.",
     IconName = "GeometryCut.ico")]
  [PortInfo("In", "Input port for UnstrGrid value with points evaluation to cut.")]
  [PortInfo("Out", "UnstrGrid with new cutted values.")]
  public class GeometryCut : Module
  {

    private static double EPSILON = Globals.EpsilonDoubleImplicit;


    #region ATTRS

    /// <summary>
    /// Cut function attrubute name.
    /// </summary>
    private String fAttrName;

    /// <summary>
    /// Treshold value for cutting.
    /// </summary>
    private double tresholdValue;

    /// <summary>
    /// True if cut geater values from treshold value, otherwise cut less values.
    /// </summary>
    private bool cutGreater;

    /// <summary>
    /// Flag indicate if parameters change.
    /// </summary>
    /// 
    private bool paramChange;

    /// <summary>
    /// Dictionary with maket cuts.
    /// </summary>
    private Dictionary<Line, Line> cuts;

    /// <summary>
    /// Array with new points. Last cutted index is (points.Count - 1).
    /// </summary>
    private ArrayList points;

    /// <summary>
    /// Maping table for indexes.
    /// </summary>
    private int[] mapTable;

    /// <summary>
    /// Output grid.
    /// </summary>
    private UnstrGrid outGrid;
    /// <summary>
    /// Input grid.
    /// </summary>
    private UnstrGrid inGrid;

    /// <summary>
    /// Cutting attributes.
    /// </summary>
    Scalar[] attrs;

    /// <summary>
    /// Total number of items to cut.
    /// </summary>
    private double itemsToCutDiv;
    /// <summary>
    /// Actual cutted item.
    /// </summary>
    private int actualItem;

    #endregion

    #region PROPERTIES

    /// <summary>
    /// Attribute name of cut function.
    /// </summary>
    [Description("Name of preffered points scalar attribute with evaluation for cut.")]
    public String CutFunctionAttrName
    {
      get { return fAttrName; }
      set {
        paramChange = true;
        fAttrName = value; 
      }
    }

    /// <summary>
    /// Set treshold value for cut.
    /// </summary>
    [Description("Treshold value for cut.")]
    public double TresholdValue
    {
      get { return tresholdValue; }
      set {
        paramChange = true;
        tresholdValue = value;
      }
    }

    /// <summary>
    /// True if cut geater values from treshold value, otherwise cut less values.
    /// </summary>
    [Description("True if you want cut greater values from treshold value, otherwise cut less values.")]
    public bool CutGreater
    {
      get { return cutGreater; }
      set {
        paramChange = true;
        cutGreater = value;
      }
    }

    #endregion

    /// <summary>
    /// Constructor. Add ports and set default values.
    /// </summary>
    public GeometryCut() : base()
    {
      //fAttrName = "";
      //tresholdValue = 0;
      cutGreater = false;

      AddInPort("In", typeof(UnstrGrid));
      AddOutPort("Out", typeof(UnstrGrid));
      cuts = new Dictionary<Line, Line>();
      paramChange = true;
    }

    /// <summary>
    /// 
    /// </summary>
    public override void Execute()
    {
      // Update when parameters or data was change ;-)
      if (IsDataSame("In") && paramChange == false)
      {
        SetOutput("Out", outGrid, true);
      } else
      {
        InitForExecute();
        Cut();
        outGrid.Points = UniformDataArray.FromArrayList(inGrid.Points.ElemType, points);
        SetOutput("Out", outGrid, false);
      }
    }

    /// <summary>
    /// Initialize values before cutting.
    /// </summary>
    private void InitForExecute()
    {
      cuts.Clear();
      outGrid = new UnstrGrid();
      inGrid = (UnstrGrid)GetInput("In");
      paramChange = false;
      // copy points which will not cutted and generate mapTable
      Array inPointData = inGrid.Points.Array;
      if (fAttrName != null && !fAttrName.Equals(""))
      {
        if (inGrid.ContainsPointAttrs(fAttrName))
        {
          attrs = (Scalar[])(inGrid.GetPointAttrs(fAttrName).Array);
        } else
        {
          throw new MveException(this, "No point attributes with ID " + fAttrName + " found.");
        }
      } else
      {
        if (inGrid.GetPointAttrsIds().Count > 0)
        {
          attrs = (Scalar[])(inGrid.GetPointAttrs(typeof(Scalar))[0].Array);
        } else
        {
          throw new MveException(this, "No scalar point attribut found. No point evaluation for cut.");
        }
      }

      // Set number of items. It is used for progress.
      int pointsCount = inGrid.Points.Count;
      itemsToCutDiv = pointsCount;
      ICollection<String> ids = inGrid.GetCellIds();
      foreach (String id in ids)
      {
        itemsToCutDiv = itemsToCutDiv + inGrid.GetCells(id).Count;
      }
      itemsToCutDiv = 1 / itemsToCutDiv;
      actualItem = 0;
      
      mapTable = new int[pointsCount];
      points = new ArrayList(pointsCount);
      int newIndex;

      for (int i = 0; i != pointsCount; i++)
      {
        if (CutPoint(i))
        {
          // Cut point - maping set to -1 value
          mapTable[i] = -1;
        } else
        {
          // Copy point
          newIndex = points.Add(inPointData.GetValue(i));
          mapTable[i] = newIndex;
        }
        if (i % 100 == 0)
        {
          actualItem = i;
          ProgressInfo(actualItem * itemsToCutDiv);
        }
      }
      actualItem = pointsCount;
      ProgressInfo(actualItem * itemsToCutDiv);
    }

    /// <summary>
    /// Cut geometric data in cells.
    /// </summary>
    private void Cut()
    {
      ICollection<String> ids = inGrid.GetCellIds();
      UniformDataArray array;
      Type type;
      int lastItems = actualItem;

      foreach (string id in ids)
      {
        array = inGrid.GetCells(id);
        type = array.ElemType;
        if (type.Equals(typeof(Line)))
        {
          CutLines(id, array);
        } else if (type.Equals(typeof(PolyLine)))
        {
          CutPolyLines(id, array);
        } else if (type.Equals(typeof(Polygon)))
        {
          CutPolygons(id, array);
        } else if (type.Equals(typeof(Triangle)))
        {
          CutTriangles(id, array);
        } else if (type.Equals(typeof(TriangleFan)))
        {
          CutTriangleFans(id, array);
        } else if (type.Equals(typeof(TriangleStrip)))
        {
          CutTriangleStrip(id, array);
        } else if (type.Equals(typeof(Tetrahedron)))
        {
          CutTetrahedron(id, array);
        }
        actualItem = lastItems + array.Count;
        lastItems = actualItem;
        ProgressInfo(actualItem * itemsToCutDiv);
      }
    }


    /// <summary>
    /// Cut tetrahedrons from given array and set cutted tetrahedrons in cells with getted id.
    /// </summary>
    /// <param name="id">Id of cutted cells array.</param>
    /// <param name="tetraArray">Array with tetrahedrons to cut.</param>
    private void CutTetrahedron(String id, UniformDataArray tetraArray)
    {
      ArrayList outTetrahedrons = new ArrayList();
      Tetrahedron[] tetrahedrons = (Tetrahedron[])tetraArray.Array;
      List<Int32> cut = new List<Int32>();
      List<Int32> keep = new List<Int32>();
      // Helper array to sort point - mamimum number of points after cut is 6
      // First is not cutted points, then si point which cross the cut and then is cutted point
      int[] helper = new int[6];
      Line cutLine;
      int keepPoints, crossPoints;

      foreach (Tetrahedron t in tetrahedrons)
      {
        for (int i = 0; i != 4; i++)
        {
          if (CutPoint(t.GetPoint(i)))
          {
            cut.Add(t.GetPoint(i));
          } else
          {
            keep.Add(t.GetPoint(i));
          }
        }
        // No vertex cutted
        switch (cut.Count) {
          case 0:
            outTetrahedrons.Add(new Tetrahedron(MapPoint(t.V1), MapPoint(t.V2), MapPoint(t.V3), MapPoint(t.V4)));
            break;
  
          // One vertex cutted
          case 1:
            keepPoints = 0;
            crossPoints = 0;
            // Try to cut all lines to cutted point
            for (int j = 0; j != 3; j++)
            {
              CutLine(keep[j], cut[0]);
              if (cuts.TryGetValue(new Line(keep[j], cut[0], false), out cutLine))
              {
                helper[keepPoints] = MapPoint(keep[j]);
                if (cutLine.V1 == helper[keepPoints])
                {
                  helper[keepPoints + 3] = cutLine.V2;
                } else
                {
                  helper[keepPoints + 3] = cutLine.V1;
                }
                keepPoints = keepPoints + 1;
              } else
              {
                helper[2 - crossPoints] = MapPoint(keep[j]);
                helper[5 - crossPoints] = -1;
                crossPoints = crossPoints + 1;
              }
            }
            switch (crossPoints) {
              case 0:
                outTetrahedrons.Add(new Tetrahedron(helper[0], helper[1], helper[2], helper[4]));
                outTetrahedrons.Add(new Tetrahedron(helper[0], helper[2], helper[3], helper[4]));
                outTetrahedrons.Add(new Tetrahedron(helper[2], helper[3], helper[4], helper[5]));
                break;
              case 1:
                outTetrahedrons.Add(new Tetrahedron(helper[0], helper[1], helper[2], helper[4]));
                outTetrahedrons.Add(new Tetrahedron(helper[0], helper[2], helper[3], helper[4]));
                break;
              case 2:
                outTetrahedrons.Add(new Tetrahedron(helper[0], helper[1], helper[2], helper[3]));
                break;
            }
            break;
          // 2 points cutted
          case 2:
            keepPoints = 0;
            crossPoints = 0;
            // Try to cut all lines to cutted point
            for (int j = 0; j != 2; j++)
            {
              CutLine(keep[j], cut[0]);
              if (cuts.TryGetValue(new Line(keep[j], cut[0], false), out cutLine))
              {
                helper[keepPoints] = MapPoint(keep[j]);
                if (cutLine.V1 == helper[keepPoints])
                {
                  helper[keepPoints + 2] = cutLine.V2;
                } else
                {
                  helper[keepPoints + 2] = cutLine.V1;
                }
                CutLine(keep[j], cut[1]);
                if (cuts.TryGetValue(new Line(keep[j], cut[1], false), out cutLine))
                {
                  if (cutLine.V1 == helper[keepPoints])
                  {
                    helper[keepPoints + 4] = cutLine.V2;
                  } else
                  {
                    helper[keepPoints + 4] = cutLine.V1;
                  }
                  keepPoints = keepPoints + 1;
                }
              } else
              {
                helper[1 - crossPoints] = MapPoint(keep[j]);
                helper[5 - crossPoints] = -1;
                helper[3 - crossPoints] = -1;
                crossPoints = crossPoints + 1;
              }
            }
            switch (crossPoints) {
              case 0:
                outTetrahedrons.Add(new Tetrahedron(helper[0], helper[1], helper[3], helper[5]));
                outTetrahedrons.Add(new Tetrahedron(helper[0], helper[3], helper[4], helper[5]));
                outTetrahedrons.Add(new Tetrahedron(helper[0], helper[2], helper[3], helper[4]));
                break;
              case 1:
                outTetrahedrons.Add(new Tetrahedron(helper[0], helper[1], helper[2], helper[4]));
                break;
            }
            break;
          // 3 points cutted
          case 3:
            crossPoints = 0;
            // Try to cut all lines to cutted point
            helper[0] = MapPoint(keep[0]);
            for (int j = 0; j != 3; j++)
            {
              CutLine(keep[0], cut[j]);
              if (cuts.TryGetValue(new Line(keep[0], cut[j], false), out cutLine))
              {
                if (cutLine.V1 == helper[0])
                {
                  helper[j + 1] = cutLine.V2;
                } else
                {
                  helper[j + 1] = cutLine.V1;
                }
              } else
              {
                crossPoints = crossPoints + 1;
                break;
              }
            }
            if (crossPoints == 0)
            {
              outTetrahedrons.Add(new Tetrahedron(helper[0], helper[1], helper[2], helper[3]));
            }
            break;
        }

        cut.Clear();
        keep.Clear();
        actualItem = actualItem + 1;
        ProgressInfo(actualItem * itemsToCutDiv);
      }

      // Set output tetrahedrons to outGrid.
      UniformDataArray uda = UniformDataArray.FromArrayList(typeof(Tetrahedron), outTetrahedrons);
      outGrid.AddCells(id, uda);
    }

    /// <summary>
    /// Cut triangle strips from given array and set cutted strips in cells with getted id.
    /// </summary>
    /// <param name="id">Id of cutted cells array.</param>
    /// <param name="stripArray">Array with fans to cut.</param>
    private void CutTriangleStrip(String id, UniformDataArray stripArray)
    {
      ArrayList outStrips = new ArrayList();
      TriangleStrip[] strips = (TriangleStrip[])stripArray.Array;
      List<Int32> actStrip = new List<int>();
      bool cut0, cut1, cut2;
      bool cont;
      Line cutLine;

      // for optimalize cutting tetragon to triangles, and for help
      int p0 = -1;
      int p1 = -1;
      int p2 = -1;
      int p3 = -1;
      int p4 = -1;
      bool optim = false;

      foreach (TriangleStrip strip in strips)
      {
        cut0 = CutPoint(strip.GetPoint(0));
        cut1 = CutPoint(strip.GetPoint(1));
        cont = false;
        for (int i = 2; i != strip.PointCount; i++)
        {
          cut2 = CutPoint(strip.GetPoint(i));
          if (cont && !cut2)
          { // Strip continuing
            actStrip.Add(MapPoint(strip.GetPoint(i)));
          } else if (!cut0 && !cut1 && !cut2)
          { // Start new continuing strip - this is first in line
            actStrip.Add(MapPoint(strip.GetPoint(i - 2)));
            actStrip.Add(MapPoint(strip.GetPoint(i - 1)));
            actStrip.Add(MapPoint(strip.GetPoint(i)));
            cont = true;
          } else
          { // Some cut.
            if (cont)
            { // End of continuing strip - both previous poinst not cutted and in actual strip list
              CutLine(strip.GetPoint(i - 2), strip.GetPoint(i));
              if (cuts.TryGetValue(new Line(strip.GetPoint(i - 2), strip.GetPoint(i), false), out cutLine))
              {
                if (cutLine.V1 == MapPoint(strip.GetPoint(i - 2)))
                {
                  actStrip.Add(cutLine.V2);
                } else
                {
                  actStrip.Add(cutLine.V1);
                }
              }
              CutLine(strip.GetPoint(i - 1), strip.GetPoint(i));
              if (cuts.TryGetValue(new Line(strip.GetPoint(i - 1), strip.GetPoint(i), false), out cutLine))
              {
                if (cutLine.V1 == MapPoint(strip.GetPoint(i - 1)))
                {
                  actStrip.Add(cutLine.V2);
                } else
                {
                  actStrip.Add(cutLine.V1);
                }
              }
              outStrips.Add(new TriangleStrip(actStrip.ToArray()));
              actStrip.Clear();
              cont = false;
            } else 
            {  // Noting in actual strip add only this cutted strip (1 or 2 triangles)
              optim = true;
              // Found line which is not cutted
              if (!cut0 && !cut1)
              {
                p0 = strip.GetPoint(i - 2);
                p1 = strip.GetPoint(i - 1);
                p2 = strip.GetPoint(i);
              } else if (!cut0 && !cut2)
              {
                p0 = strip.GetPoint(i - 2);
                p1 = strip.GetPoint(i);
                p2 = strip.GetPoint(i - 1);
              } else if (!cut1 && !cut2)
              {
                p0 = strip.GetPoint(i - 1);
                p1 = strip.GetPoint(i);
                p2 = strip.GetPoint(i - 2);
              } else
              {
                optim = false;
              }
              // Try to cut lines
              if (optim)
              {
                CutLine(p0, p2);
                if (cuts.TryGetValue(new Line(p0, p2, false), out cutLine))
                {
                  if (cutLine.V1 == MapPoint(p0))
                  {
                    p3 = cutLine.V2;
                  } else
                  {
                    p3 = cutLine.V1;
                  }
                } else
                {
                  optim = false;
                }
                CutLine(p1, p2);
                if (cuts.TryGetValue(new Line(p1, p2, false), out cutLine))
                {
                  if (cutLine.V1 == MapPoint(p1))
                  {
                    p4 = cutLine.V2;
                  } else
                  {
                    p4 = cutLine.V1;
                  }
                } else
                {
                  optim = false;
                }
                // Count optimal triangles
                if (optim)
                {
                  IVector d1 = null;
                  IVector d2 = null;
                  if (inGrid.Points.ElemType.Equals(typeof(Point2D)))
                  {
                    Point2D a = (Point2D)points[MapPoint(p1)];
                    Point2D b = (Point2D)points[p3];
                    d1 = (a - b);
                    a = (Point2D)points[MapPoint(p0)];
                    b = (Point2D)points[p4];
                    d2 = (a - b);
                  } else if (inGrid.Points.ElemType.Equals(typeof(Point3D)))
                  {
                    Point3D a = (Point3D)points[MapPoint(p1)];
                    Point3D b = (Point3D)points[p3];
                    d1 = (a - b);
                    a = (Point3D)points[MapPoint(p0)];
                    b = (Point3D)points[p4];
                    d2 = (a - b);
                  }
                  if (d1.Abs() < d2.Abs())
                  {
                    actStrip.Add(MapPoint(p0));
                    actStrip.Add(MapPoint(p1));
                    actStrip.Add(p3);
                    actStrip.Add(p4);
                    outStrips.Add(new TriangleStrip(actStrip.ToArray()));
                  } else
                  {
                    actStrip.Add(MapPoint(p1));
                    actStrip.Add(MapPoint(p0));
                    actStrip.Add(p4);
                    actStrip.Add(p3);
                    outStrips.Add(new TriangleStrip(actStrip.ToArray()));
                  }
                  actStrip.Clear();
                }
              }
              if (!optim)
              {
                // Add all vertices which is not cutted
                if (!cut0)
                {
                  actStrip.Add(MapPoint(strip.GetPoint(i - 2)));
                }
                if (!cut1)
                {
                  actStrip.Add(MapPoint(strip.GetPoint(i - 1)));
                }
                if (!cut2)
                {
                  actStrip.Add(MapPoint(strip.GetPoint(i)));
                }

                // Try to cut all 3 lines and add cutted points
                if (CutLine(strip.GetPoint(i - 2), strip.GetPoint(i - 1)))
                {
                  if (cut0)
                  {
                    p0 = MapPoint(strip.GetPoint(i - 1));
                  } else
                  {
                    p0 = MapPoint(strip.GetPoint(i - 2));
                  }
                  if (cuts.TryGetValue(new Line(strip.GetPoint(i - 2), strip.GetPoint(i - 1), false), out cutLine))
                  {
                    if (cutLine.V1 == p0)
                    {
                      actStrip.Add(cutLine.V2);
                    } else
                    {
                      actStrip.Add(cutLine.V1);
                    }
                  }
                }

                if (CutLine(strip.GetPoint(i - 2), strip.GetPoint(i)))
                {
                  if (cut0)
                  {
                    p0 = MapPoint(strip.GetPoint(i));
                  } else
                  {
                    p0 = MapPoint(strip.GetPoint(i - 2));
                  }
                  if (cuts.TryGetValue(new Line(strip.GetPoint(i - 2), strip.GetPoint(i), false), out cutLine))
                  {
                    if (cutLine.V1 == p0)
                    {
                      actStrip.Add(cutLine.V2);
                    } else
                    {
                      actStrip.Add(cutLine.V1);
                    }
                  }
                }

                if (CutLine(strip.GetPoint(i - 1), strip.GetPoint(i)))
                {
                  if (cut1)
                  {
                    p0 = MapPoint(strip.GetPoint(i));
                  } else
                  {
                    p0 = MapPoint(strip.GetPoint(i - 1));
                  }
                  if (cuts.TryGetValue(new Line(strip.GetPoint(i - 1), strip.GetPoint(i), false), out cutLine))
                  {
                    if (cutLine.V1 == p0)
                    {
                      actStrip.Add(cutLine.V2);
                    } else
                    {
                      actStrip.Add(cutLine.V1);
                    }
                  }
                }
                if (actStrip.Count > 2)
                {
                  outStrips.Add(new TriangleStrip(actStrip.ToArray()));
                }
                actStrip.Clear();
              }
            }
          }

          cut0 = cut1;
          cut1 = cut2;
        }
        if (actStrip.Count > 2)
        {
          outStrips.Add(new TriangleStrip(actStrip.ToArray()));
        }
        actStrip.Clear();
        actualItem = actualItem + 1;
        ProgressInfo(actualItem * itemsToCutDiv);
      }

      // Set output strips to outGrid.
      UniformDataArray uda = UniformDataArray.FromArrayList(typeof(TriangleStrip), outStrips);
      outGrid.AddCells(id, uda);
    }


    /// <summary>
    /// Cut triangle fans from given array and set cutted fans in cells with getted id.
    /// </summary>
    /// <param name="id">Id of cutted cells array.</param>
    /// <param name="fansArray">Array with fans to cut.</param>
    private void CutTriangleFans(String id, UniformDataArray fansArray)
    {
      ArrayList outFans = new ArrayList();
      TriangleFan[] fans = (TriangleFan[])fansArray.Array;
      List<Int32> actFan = new List<int>();
      List<Int32> startFan = new List<int>();
      Line cutLine;
      bool prev, act;
      bool start;
      int index;

      // Fro each fan in input grid
      foreach (TriangleFan fan in fans)
      {
        if (CutPoint(fan.GetPoint(0)))
        {
          // Center point is cutted
          // Start at second point
          index = 2;
          if (CutPoint(fan.GetPoint(2)))
          {
            if (!CutPoint(fan.GetPoint(1)))
            {
              // Cut must exist
              CutLine(fan.GetPoint(0), fan.GetPoint(1));
              if (cuts.TryGetValue(new Line(fan.GetPoint(0), fan.GetPoint(1), false), out cutLine))
              {
                if (cutLine.V1 == MapPoint(fan.GetPoint(1)))
                {
                  actFan.Add(cutLine.V2);
                } else
                {
                  actFan.Add(cutLine.V1);
                }
              }
              CutLine(fan.GetPoint(1), fan.GetPoint(2));
              if (cuts.TryGetValue(new Line(fan.GetPoint(1), fan.GetPoint(2), false), out cutLine))
              {
                actFan.Add(cutLine.V1);
                actFan.Add(cutLine.V2);
              }
              if (actFan.Count > 2)
              {
                outFans.Add(new TriangleFan(actFan.ToArray()));
              }
              actFan.Clear();
            }
          } else
          {
            CutLine(fan.GetPoint(1), fan.GetPoint(2));
            // cut crossing second point
            if (cuts.TryGetValue(new Line(fan.GetPoint(1), fan.GetPoint(2), false), out cutLine) == false)
            {
              CutLine(fan.GetPoint(0), fan.GetPoint(1));
              if (cuts.TryGetValue(new Line(fan.GetPoint(0), fan.GetPoint(1), false), out cutLine))
              {
                actFan.Add(cutLine.V1);
                actFan.Add(cutLine.V2);
                actFan.Add(MapPoint(fan.GetPoint(2)));
                outFans.Add(new TriangleFan(actFan.ToArray()));
                actFan.Clear();
              }

            }
          }
          // Go through all verticies
          while (index < fan.PointCount)
          {
            if (CutPoint(fan.GetPoint(index)))
            { // Go to next
              index++;
              continue;
            }
            CutLine(fan.GetPoint(0), fan.GetPoint(index));
            if (cuts.TryGetValue(new Line(fan.GetPoint(0), fan.GetPoint(index), false), out cutLine))
            {
              if (cutLine.V1 == MapPoint(fan.GetPoint(index)))
              {
                actFan.Add(cutLine.V2); // add centre of new fan
              } else
              {
                actFan.Add(cutLine.V1); // add centre of new fan
              }
              // Test previous point
              if (CutPoint(fan.GetPoint(index - 1)))
              {
                CutLine(fan.GetPoint(index), fan.GetPoint(index - 1));
                if (cuts.TryGetValue(new Line(fan.GetPoint(index), fan.GetPoint(index - 1), false), out cutLine))
                {
                  if (cutLine.V1 == MapPoint(fan.GetPoint(index)))
                  {
                    actFan.Add(cutLine.V2);  // add next left cutted point
                  } else
                  {
                    actFan.Add(cutLine.V1);  // add next left cutted point
                  }
                }
              } else
              {
                CutLine(fan.GetPoint(index - 1), fan.GetPoint(0));
                if (cuts.TryGetValue(new Line(fan.GetPoint(index - 1), fan.GetPoint(0), false), out cutLine))
                {
                  if (cutLine.V1 == MapPoint(fan.GetPoint(index - 1)))
                  {
                    actFan.Add(cutLine.V2); // add previous line cutted point
                  } else
                  {
                    actFan.Add(cutLine.V1); // add previous line cutted point
                  }
                }
                actFan.Add(MapPoint(fan.GetPoint(index - 1))); // add previous point
              }

              // Add actual point in fan
              actFan.Add(MapPoint(fan.GetPoint(index)));

              // Test next point - if exist
              if ((index + 1) < fan.PointCount)
              {
                if (CutPoint(fan.GetPoint(index + 1)))
                {
                  CutLine(fan.GetPoint(index + 1), fan.GetPoint(index));
                  if (cuts.TryGetValue(new Line(fan.GetPoint(index + 1), fan.GetPoint(index), false), out cutLine))
                  {
                    if (cutLine.V1 == MapPoint(fan.GetPoint(index)))
                    {
                      actFan.Add(cutLine.V2); // add point from next right cutted line
                    } else
                    {
                      actFan.Add(cutLine.V1); // add point from next right cutted line
                    }
                  }
                } else
                {
                  actFan.Add(MapPoint(fan.GetPoint(index + 1))); // add next point
                  CutLine(fan.GetPoint(index + 1), fan.GetPoint(0));
                  if (cuts.TryGetValue(new Line(fan.GetPoint(index + 1), fan.GetPoint(0), false), out cutLine))
                  {
                    if (cutLine.V1 == MapPoint(fan.GetPoint(index + 1)))
                    {
                      actFan.Add(cutLine.V2); // add next cutted point
                    } else
                    {
                      actFan.Add(cutLine.V1); // add next cutted point
                    }
                  }
                }
              }
              outFans.Add(new TriangleFan(actFan.ToArray()));
              actFan.Clear();
              // Skip next point and go on
              index = index + 2;
            } else
            { // Only one point cut - go to next
              index++;
              continue;
            }
          }

        } else // Center point is not cutted
        {
          // Copy center point
          actFan.Add(MapPoint(fan.GetPoint(0)));
          start = fan.GetPoint(1) == fan.GetPoint(fan.PointCount - 1);
          prev = !CutPoint(fan.GetPoint(1));
          if (prev)
          {
            actFan.Add(MapPoint(fan.GetPoint(1)));
          } else
          {
            if (CutLine(fan.GetPoint(0), fan.GetPoint(1)))
            {
              if (cuts.TryGetValue(new Line(fan.GetPoint(0), fan.GetPoint(1), false), out cutLine))
              {
                if (cutLine.V1 == MapPoint(fan.GetPoint(0)))
                {
                  actFan.Add(cutLine.V2);
                } else
                {
                  actFan.Add(cutLine.V1);
                }
                prev = true;
              }
            } // else is imposible
          }
          start = start && prev;

          for (int i = 2; i != fan.PointCount; i++)
          {
            act = !CutPoint(fan.GetPoint(i));
            // Cause II - fan continue
            if (act && prev)
            {
              if (CutPoint(fan.GetPoint(i - 1)))
              {
                // Test cut between this point and previous because previous point was cut.
                CutLine(fan.GetPoint(i), fan.GetPoint(i - 1));
                if (cuts.TryGetValue(new Line(fan.GetPoint(i), fan.GetPoint(i - 1), false), out cutLine))
                {
                  if (cutLine.V1 == MapPoint(fan.GetPoint(i)))
                  {
                    actFan.Add(cutLine.V2);
                  } else
                  {
                    actFan.Add(cutLine.V1);
                  }
                }
              }
              actFan.Add(MapPoint(fan.GetPoint(i)));
            // Cause 0I - cut fan line or end of fan
            } else if (!act && prev)
            {
              // Cut must always exist
              CutLine(fan.GetPoint(i), fan.GetPoint(i - 1));
              if (cuts.TryGetValue(new Line(fan.GetPoint(i), fan.GetPoint(i - 1), false), out cutLine))
              {
                if (cutLine.V1 == MapPoint(fan.GetPoint(i - 1)))
                {
                  actFan.Add(cutLine.V2);
                } else
                {
                  actFan.Add(cutLine.V1);
                }
              }
              // Cut must always exist
              CutLine(fan.GetPoint(0), fan.GetPoint(i));
              if (cuts.TryGetValue(new Line(fan.GetPoint(0), fan.GetPoint(i), false), out cutLine))
              {
                if (cutLine.V1 ==  MapPoint(fan.GetPoint(0)))
                {
                  actFan.Add(cutLine.V2);
                } else
                {
                  actFan.Add(cutLine.V1);
                }
                act = true;
              } else
              {
                if (start)
                {
                  startFan = actFan;
                  actFan = new List<int>();
                  actFan.Add(MapPoint(fan.GetPoint(0)));
                  start = false;
                } else
                {
                  if (actFan.Count > 2)
                  {
                    outFans.Add(new TriangleFan(actFan.ToArray()));
                  }
                  actFan.Clear();
                  actFan.Add(MapPoint(fan.GetPoint(0)));
                }
              }

            // Cause IO
            } else if (act && !prev)
            {
              CutLine(fan.GetPoint(i), fan.GetPoint(i - 1));
              if (cuts.TryGetValue(new Line(fan.GetPoint(i), fan.GetPoint(i - 1), false), out cutLine))
              {
                if (cutLine.V1 == MapPoint(fan.GetPoint(i)))
                {
                  actFan.Add(cutLine.V2);
                } else
                {
                  actFan.Add(cutLine.V1);
                }
              }
              actFan.Add(MapPoint(fan.GetPoint(i)));
            }// Cause OO - means thas cut cross the center point and no point is add fan.
            
            prev = act;
          }
          if (startFan.Count > 2)
          {
            // Remove center point
            startFan.RemoveAt(0);
            startFan.RemoveAt(0);
            actFan.AddRange(startFan);
            startFan.Clear();
          }
          if (actFan.Count > 2)
          {
            outFans.Add(new TriangleFan(actFan.ToArray()));
          }
          actFan.Clear();
        }
        actualItem = actualItem + 1;
        ProgressInfo(actualItem * itemsToCutDiv);
      }

      // Set output fans to outGrid.
      UniformDataArray uda = UniformDataArray.FromArrayList(typeof(TriangleFan), outFans);
      outGrid.AddCells(id, uda);
    }

    /// <summary>
    /// Cut triangles and from given array and set cutted triangles in cells with getted id.
    /// </summary>
    /// <param name="id">Id of cutted cells array.</param>
    /// <param name="trianglesArray">Array with triangles to cut.</param>
    private void CutTriangles(String id, UniformDataArray trianglesArray)
    {
      ArrayList outTriangles = new ArrayList();
      Triangle[] triangles = (Triangle[])trianglesArray.Array;
      // lines of triangle
      Line[] l = new Line[3];
      int center = 0;
      Line cutLine;
      int cutted, removes, news;

      // Fro each triangle in input grid
      foreach (Triangle tri in triangles)
      {
        // Cut all 3 lines
        cutted = 0;
        removes = 0;
        news = 0;
        for (int i = 0; i != 3; i++)
        {
          if (CutLine(tri[i], tri[(i + 1) % 3]))
          {
            if (cuts.TryGetValue(new Line(tri[i], tri[(i + 1) % 3], false), out cutLine))
            {
              l[news] = cutLine;
              news++;
              cutted++;
            } else
            {
              removes++;
            }
          } else
          {
            l[news] = new Line(MapPoint(tri[i]), MapPoint(tri[(i + 1) % 3]));
            center = news;
            news++;
          }
        }

        // No cut
        if (cutted == 0 && removes == 0)
        {
          outTriangles.Add(new Triangle(MapPoint(tri.V1), MapPoint(tri.V2), MapPoint(tri.V3)));
        // Cut one side of triangle
        } else if (removes == 1) {
          // Found shared point to both edges
          if (l[0].V1 == l[1].V1) {
            outTriangles.Add(new Triangle(l[0].V1, l[0].V2, l[1].V2));
          } else if (l[0].V1 == l[1].V2) {
            outTriangles.Add(new Triangle(l[0].V1, l[0].V2, l[1].V1));
          } else if (l[0].V2 == l[1].V1)
          {
            outTriangles.Add(new Triangle(l[0].V1, l[0].V2, l[1].V2));
          } else
          {
            outTriangles.Add(new Triangle(l[0].V1, l[0].V2, l[1].V1));
          }
        // Cut triangle vertex
        } else if (removes == 0)
        {
          int first, second;
          IVector d1 = null;
          IVector d2 = null;
          if (l[center].V1 == l[(center + 1) % 3].V1 || l[center].V1 == l[(center + 1) % 3].V2)
          {
            if (l[center].V1 == l[(center + 1) % 3].V1)
            {
              first = l[(center + 1) % 3].V2;
            } else
            {
              first = l[(center + 1) % 3].V1;
            }
            if (l[center].V2 == l[(center + 2) % 3].V1)
            {
              second = l[(center + 2) % 3].V2;
            } else
            {
              second = l[(center + 2) % 3].V1;
            }
          } else
          {
            if (l[center].V1 == l[(center + 2) % 3].V1)
            {
              first = l[(center + 2) % 3].V2;
            } else
            {
              first = l[(center + 2) % 3].V1;
            }
            if (l[center].V2 == l[(center + 1) % 3].V1)
            {
              second = l[(center + 1) % 3].V2;
            } else
            {
              second = l[(center + 1) % 3].V1;
            }
          }
          if (inGrid.Points.ElemType.Equals(typeof(Point2D)))
          {
            Point2D a = (Point2D) points[first];
            Point2D b = (Point2D) points[l[center].V2];
            d1 = (a - b);
            a = (Point2D)points[second];
            b = (Point2D) points[l[center].V1];
            d2 = (a - b);
          } else if (inGrid.Points.ElemType.Equals(typeof(Point3D)))
          {
            Point3D a = (Point3D) points[first];
            Point3D b = (Point3D) points[l[center].V2];
            d1 = (a - b);
            a = (Point3D) points[second];
            b = (Point3D) points[l[center].V1];
            d2 = (a - b);
          }
          if (d1.Abs() < d2.Abs())
          {
            outTriangles.Add(new Triangle(l[center].V1, l[center].V2, first));
            outTriangles.Add(new Triangle(l[center].V2, first, second));
          } else
          {
            outTriangles.Add(new Triangle(l[center].V1, l[center].V2, second));
            outTriangles.Add(new Triangle(l[center].V1, first, second));
          }
        }
        actualItem = actualItem + 1;
        ProgressInfo(actualItem * itemsToCutDiv);
      }
      // Set output triangles to outGrid.
      UniformDataArray uda = UniformDataArray.FromArrayList(typeof(Triangle), outTriangles);
      outGrid.AddCells(id, uda);
    }

    /// <summary>
    /// Cut polygons from given array and set it in cells with given id.
    /// </summary>
    /// <param name="id">Id of output cell.</param>
    /// <param name="polygonsArray">Array with polygons to cut.</param>
    private void CutPolygons(String id, UniformDataArray polygonsArray)
    {
      ArrayList outPolygons = new ArrayList();
      Polygon[] polygons = (Polygon[])polygonsArray.Array;
      List<Int32> actPolygon = new List<int>();
      List<Int32> startPart = new List<int>();
      bool start;
      // previous point is not cutted
      bool prev;
      // actual point is not cutted
      bool act;
      int prevIndex;
      Line cutLine;

      // Fro each polygon in input grid
      foreach (Polygon polygon in polygons)
      {
        prev = !CutPoint(polygon[polygon.PointCount - 1]);
        if (prev)
        {
          start = true;
        } else
        {
          start = false;
        }
        for (int i = 0; i != polygon.PointCount; i++)
        {
          if (i == 0)
          {
            prevIndex = polygon.PointCount - 1;
          } else
          {
            prevIndex = i - 1;
          }
          act = !CutPoint(polygon[i]);
          // Test posibilities
          // posibility II - no point cut
          if (prev && act)
          {
            actPolygon.Add(MapPoint(polygon[i]));
            // posibility IO - actual point cut
          } else if (prev && !act)
          {
            if (CutLine(polygon[prevIndex], polygon[i]))
            {
              if (cuts.TryGetValue(new Line(polygon[prevIndex], polygon[i], false), out cutLine))
              {
                // Add cutted point
                if (cutLine.V1 == MapPoint(polygon[prevIndex]))
                {
                  actPolygon.Add(cutLine.V2);
                } else
                {
                  actPolygon.Add(cutLine.V1);
                }
              } // In this if is else cause imposible (I hope)
            }
            if (start)
            {
              startPart.AddRange(actPolygon);
              start = false;
            } else
            {
              if (actPolygon.Count > 1)
              {
                outPolygons.Add(new Polygon(actPolygon.ToArray()));
              }
            }
            actPolygon.Clear();
            // posibility OI - cut previous point point
          } else if (!prev && act)
          {
            if (CutLine(polygon[prevIndex], polygon[i]))
            {
              if (cuts.TryGetValue(new Line(polygon[prevIndex], polygon[i], false), out cutLine))
              {
                // Add cutted point
                if (cutLine.V1 == MapPoint(polygon[i]))
                {
                  actPolygon.Add(cutLine.V2);
                } else
                {
                  actPolygon.Add(cutLine.V1);
                }
              }
              // Always add actual point
              actPolygon.Add(MapPoint(polygon[i]));
            }
          }
          prev = act;
        }

        if (actPolygon.Count != 0)
        {
          startPart.AddRange(actPolygon);
          actPolygon.Clear();
        }
        if (startPart.Count > 1)
        {
          outPolygons.Add(new Polygon(startPart.ToArray()));
        }
        startPart.Clear();
        actualItem = actualItem + 1;
        ProgressInfo(actualItem * itemsToCutDiv);
      }
      // Set output polygons to outputGrid.
      UniformDataArray uda = UniformDataArray.FromArrayList(typeof(Polygon), outPolygons);
      outGrid.AddCells(id, uda);
    }


    /// <summary>
    /// Cut poly lines from given array and set it in cells with given id.
    /// </summary>
    /// <param name="id">Id of output cell.</param>
    /// <param name="linesArray">Array with polylines to cut.</param>
    private void CutPolyLines(String id, UniformDataArray linesArray)
    {
      ArrayList outLines = new ArrayList();
      List<Int32> actLine = new List<int>();
      PolyLine[] lines = (PolyLine[])linesArray.Array;
      // previous point is not cutted
      bool prev;
      // actual point is not cutted
      bool act;
      Line cutLine;
      // Fro each line in input grid
      foreach (PolyLine line in lines)
      {
        prev = !CutPoint(line[0]);
        if (prev) 
        {
          actLine.Add(MapPoint(line[0]));
        }
        for (int i = 1; i != line.PointCount; i++)
        {
          act = !CutPoint(line[i]);
          // Test posibilities
          // posibility II - no point cut
          if (prev && act)
          {
            actLine.Add(MapPoint(line[i]));
          // posibility IO - actual point cut
          } else if (prev && !act)
          {
            if (CutLine(line[i - 1], line[i]))
            {
              if (cuts.TryGetValue(new Line(line[i - 1], line[i], false), out cutLine))
              {
                // Add cutted point
                if (cutLine.V1 == MapPoint(line[i - 1]))
                {
                  actLine.Add(cutLine.V2);
                } else
                {
                  actLine.Add(cutLine.V1);
                }
              } // In this if is else cause imposible (I hope)
            }
            if (actLine.Count > 1)
            {
              outLines.Add(new PolyLine(actLine.ToArray()));
            }
            actLine.Clear();
          // posibility OI - cut previous point point
          } else if (!prev && act)
          {
            if (CutLine(line[i - 1], line[i]))
            {
              if (cuts.TryGetValue(new Line(line[i - 1], line[i], false), out cutLine))
              {
                // Add cutted point
                if (cutLine.V1 == MapPoint(line[i]))
                {
                  actLine.Add(cutLine.V2);
                } else
                {
                  actLine.Add(cutLine.V1);
                }
              }
              // Always add actual point
              actLine.Add(MapPoint(line[i]));
            }
          }

          prev = act;
        }
        if (actLine.Count > 1)
        {
          outLines.Add(new PolyLine(actLine.ToArray()));
        }
        actLine.Clear();
        actualItem = actualItem + 1;
        ProgressInfo(actualItem * itemsToCutDiv);
      }
      // Set output lines to outputGrid.
      UniformDataArray linesUDA = UniformDataArray.FromArrayList(typeof(PolyLine), outLines);
      outGrid.AddCells(id, linesUDA);
    }


    /// <summary>
    /// Cut lines and from given array and set cutted lines in cells with getted id.
    /// </summary>
    /// <param name="id">Id of cutted cells array.</param>
    /// <param name="linesArray">Array with lines to cut.</param>
    private void CutLines(String id, UniformDataArray linesArray)
    {
      ArrayList outLines = new ArrayList();
      Line outLine;
      Line[] lines = (Line[]) linesArray.Array;
      // Fro each line in input grid
      foreach (Line line in lines)
      {
        if (CutLine(line.V1, line.V2))
        {
          if (cuts.TryGetValue(new Line(line.V1, line.V2, false), out outLine) == true)
          {
            outLines.Add(outLine);
          }
        } else
        { // Line no cutted - but indexes must be mapped to output indexes
          outLine = new Line(MapPoint(line.V1), MapPoint(line.V2), true);
          outLines.Add(outLine);
        }
        actualItem = actualItem + 1;
        ProgressInfo(actualItem * itemsToCutDiv);
      }
      // Set output lines to outputGrid.
      UniformDataArray linesUDA = UniformDataArray.FromArrayList(typeof(Line), outLines);
      outGrid.AddCells(id, linesUDA);
    }

    /// <summary>
    /// Try to cut line between point on indexes v1 and v2.
    /// If function cut the line, cut this line, add cutted line to dictionary and return true.
    /// Otherwise return false. If whole line cutted, return true, and nothing is add to dictionary.
    /// If this line olready cutted, return true.
    /// </summary>
    /// <param name="v1">Index of first point</param>
    /// <param name="v2">Index of second point</param>
    /// <returns>True if this line must be cutted.</returns>
    private bool CutLine(int v1, int v2)
    {
      if (cuts.ContainsKey(new Line(v1, v2, false)))
      {
        return true;
      }
      bool greaterV1 = attrs[v1].Value > tresholdValue;
      bool greaterV2 = attrs[v2].Value > tresholdValue;
      bool cut = false;

      bool lieOnThV1 = Math.Abs(attrs[v1].Value - tresholdValue) < EPSILON;
      bool lieOnThV2 = Math.Abs(attrs[v2].Value - tresholdValue) < EPSILON;
      if (lieOnThV1 && lieOnThV2)
      {
        return false;
      }

      if (lieOnThV1) 
      {
        if (!(cutGreater ^ greaterV2))
        {
          return true;
        } else
        {
          return false;
        }
      }

      if (lieOnThV2)
      {
        if (!(cutGreater ^ greaterV1))
        {
          return true;
        } else
        {
          return false;
        }
      }
        


      if (cutGreater)
      {
        if (greaterV1 || greaterV2)
        {
          cut = true;
          if (!(greaterV1 && greaterV2))
          {
            CutLine(v1, v2, greaterV1, greaterV2);
          }
        }
      } else // if less values cutting
      {
        if (!(greaterV1 && greaterV2))
        {
          cut = true;
          if (greaterV1 || greaterV2)
          {
            CutLine(v1, v2, greaterV1, greaterV2);
          }
        }
      }
      return cut;
    }

    /// <summary>
    /// Call this only if function cutting line, this method clip line and add it to dictionary.
    /// </summary>
    /// <param name="v1">Index of first line point.</param>
    /// <param name="v2">Index of second point.</param>
    /// <param name="greaterV1">True if first point have greater value that treshold.</param>
    /// <param name="greaterV2">True if second point have greater value that treshold.</param>
    private void CutLine(int v1, int v2, bool greaterV1, bool greaterV2)
    {
      IPoint newPoint = null;
      double p = 0;
      // Cut this line
      if (inGrid.Points.ElemType.Equals(typeof(Point2D)))
      {
        Point2D a = (Point2D)inGrid.Points.Array.GetValue(v1);
        Point2D b = (Point2D)inGrid.Points.Array.GetValue(v2);
        p = (attrs[v1].Value - tresholdValue) / (attrs[v1].Value - attrs[v2].Value);
        newPoint = a + (b - a) * p;
      } else if (inGrid.Points.ElemType.Equals(typeof(Point3D)))
      {
        Point3D a = (Point3D)inGrid.Points.Array.GetValue(v1);
        Point3D b = (Point3D)inGrid.Points.Array.GetValue(v2);
        p = (attrs[v1].Value - tresholdValue) / (attrs[v1].Value - attrs[v2].Value);
        newPoint = a + (b - a) * p;
      }
      int newIndex = points.Add(newPoint);

      // Add line in dictionary - to both orientations
      Line k = new Line(v1, v2, false);
      Line val;
      if (!(greaterV1 ^ cutGreater))
      {
        // Map index to output index
        val = new Line(MapPoint(v2), newIndex, false);
      } else
      {
        // Map index to output index
        val = new Line(MapPoint(v1), newIndex, false);
      }
      cuts.Add(k, val);
    }

    /// <summary>
    /// Map old point index to new point index.
    /// </summary>
    /// <param name="index">Old point index.</param>
    /// <returns>New point index.</returns>
    public int MapPoint(int index)
    {
      if (index < mapTable.Length)
      {
        return mapTable[index];
      } else
      {
        return index;
      }
    }

    /// <summary>
    /// Test if given point will be cut or not.
    /// </summary>
    /// <param name="index">Index of point</param>
    /// <returns>True when point will be cutted.</returns>
    public bool CutPoint(int index)
    {
      if ((!cutGreater && attrs[index].Value >= (tresholdValue - EPSILON)) ||
          (cutGreater && attrs[index].Value <= (tresholdValue + EPSILON)))
      {
        return false;
      } else
      {
        return true;
      }
    }
  }
}
