using System;
using System.Diagnostics;
using System.Collections;
using System.Reflection;
using System.Xml;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Zkladn tda pro vechny 2D mky.
	/// Mka se skld z pixel, kad pixel je tvoen 4-mi body.
	/// S pixely i s body me bt asociovn libovoln poet atribut.
	/// Pixely ani body nejsou v pamti fyzicky uloeny, pracuje se pouze s atributy.
	/// Proto je tento DataSet pamov sporn. Zkladn charakteristikou mky 
	/// je rozlien - poet pixel ve smrech x a y.
	/// </summary>
	public abstract class Grid2D_old : StructGrid
	{
		/// <summary>
		/// Poet pixel ve smru osy y (vka mky).
		/// </summary>
		protected int height;

		/// <summary>
		/// Poet pixel ve smru osy x (ka mky).
		/// </summary>
		protected int width;

		/// <summary>
		/// 2D pole (mka) seznam atribut bod. Kad bod me mt libovoln 
		/// poet atribut. Ty jsou uloeny v seznamu. K seznamu atribut bodu se 
		/// pistupuje pes indexy [i,j] do tohoto pole.
		/// </summary>
		protected ArrayList[,] pointAttributes;

		/// <summary>
		/// 2D pole (mka) seznam atribut pixel. Kad pixel me mt libovoln
		/// poet atribut. Ty jsou uloeny v seznamu. K seznamu atributm pixelu se 
		/// pistupuje pes indexy [i,j] do tohoto pole.
		/// </summary>
		protected ArrayList[,] pixelAttributes;

		/// <summary>
		/// Implicitn konstruktor - vytvo przdn objekt (pouv se typicky pi natn dat ze souboru).
		/// </summary>
		public Grid2D_old() : this(0, 0)
		{
		} // Grid2D()

		/// <summary>
		/// Konstruktor - vytvo przdnou mku.
		/// </summary>
		/// <param name="height">Poet pixel v y-ovm smru (vka mky).</param>
		/// <param name="width">Poet pixel v x-ovm smru (ka mky).</param>
		public Grid2D_old(int height, int width)
		{
			InitializeGrid2D(width, height);
		} // Grid2D()

		/// <summary>
		/// Vrac poet pixel ve smru osy y (vka mky).
		/// </summary>
		public int Height
		{
			get
			{
				return this.height;
			}
		} // Height

		/// <summary>
		/// Vrac poet pixel ve smru osy x (ka mky).
		/// </summary>
		public int Width
		{
			get
			{
				return this.width;
			}
		} // Width

// ===== Metody DataObject ========================================================

		/// <summary>
		/// Prov konzistenci datov struktury.
		/// </summary>
		/// <returns>Vrac true v ppad, e je datov struktura v podku, jinak vrac false.</returns>
		public override bool CheckConsistence()
		{
			bool isOK = true;
			if (this.width < 0 || this.height < 0)
			{
				Console.WriteLine("Error! " + this.GetType().Name + ": resolution has negative value.");
				isOK = false;
			}
			return isOK;
		} // CheckConsistence()
		
		/// <summary>
		/// Informace o Grid2D.
		/// </summary>
		/// <returns>etzec obsahujc informace o Grid2D - rozlien atd...</returns>
		public override string ToString()
		{
			return this.GetType().Name.ToString() + ": [" + this.height + " x " + this.width + "]"
				+ "\n\t- height                           : " + this.height
				+ "\n\t- width                            : " + this.width
				+ "\n\t- number of points                 : " + ((this.width + 1) * (this.height + 1))
				+ "\n\t- number of pixels                 : " + (this.width * this.height)
				+ "\n\t- number of points with attributes : " + this.GetCountPointsWithAttr()
				+ "\n\t- number of pixels with attributes : " + this.GetCountPixelsWithAttr()
				+ "\n\t- number of point's attributes     : " + this.GetPointsAttrsCount() 
				+ "\n\t- number of pixel's attributes     : " + this.GetPixelsAttrsCount();
		} // ToString()

// ===== Pomocn metody ===========================================================

		/// <summary>
		/// Vytvo przdnou mku - przdn pole atribut bod a pixel.
		/// </summary>
		/// <param name="width">Poet pixel v x-ovm smru.</param>
		/// <param name="height">Poet pixel v y-ovm smru.</param>
		protected void InitializeGrid2D(int width, int height)
		{
			this.width = width;
			this.height = height;
			this.pointAttributes = new ArrayList[height + 1, width + 1];
			this.pixelAttributes = new ArrayList[height, width];
			for(int i = 0; i < this.pointAttributes.GetLength(0); i++)
				for(int j = 0; j < this.pointAttributes.GetLength(1); j++)
					this.pointAttributes[i, j] = new ArrayList();
			for(int i = 0; i < this.pixelAttributes.GetLength(0); i++)
				for(int j = 0; j < this.pixelAttributes.GetLength(1); j++)
					this.pixelAttributes[i, j] = new ArrayList();
		} // InitializeGrid2D()

		/// <summary>
		/// Pomocn metoda pro naten atribut bod z xml souboru.
		/// </summary>
		/// <param name="xmlTextReader">Datov proud pro ten dat z xml souboru.</param>
		protected void ReadPointsAttrs(XmlTextReader xmlTextReader)
		{
			// poet vech atribut
			int aCount = Int32.Parse(xmlTextReader.GetAttribute("count"));
			// poet akutln natench atribut
			int i = 0;
			// natej dokud nejsou nateny vechny atributy nebo nen dosaeno konce xml souboru
			while(xmlTextReader.Read() && i < aCount)
			{
				if (xmlTextReader.NodeType == XmlNodeType.Element && xmlTextReader.Name == "a")
				{
					// indexy bodu ktermu nle naten atribut
					int indexI = Int32.Parse(xmlTextReader.GetAttribute("i"));
					int indexJ = Int32.Parse(xmlTextReader.GetAttribute("j"));
					Type type = Globals.FindType(this.libNames, xmlTextReader.GetAttribute("type"));
					DataObject attr = (DataObject) Activator.CreateInstance(type, true);
					attr.ReadData(xmlTextReader);
					this.AddPointAttr(indexI, indexJ, attr);
					i++;
				}
			} // while
		} // ReadPointsAttrs()

		/// <summary>
		/// Pomocn metoda pro naten atribut pixel z xml souboru.
		/// </summary>
		/// <param name="xmlTextReader">Datov proud pro ten dat z xml souboru.</param>
		protected void ReadPixelsAttrs(XmlTextReader xmlTextReader)
		{
			// poet vech atribut
			int aCount = Int32.Parse(xmlTextReader.GetAttribute("count"));
			// poet akutln natench atribut
			int i = 0;
			// natej dokud nejsou nateny vechny atributy nebo nen dosaeno konce xml souboru
			while(xmlTextReader.Read() && i < aCount)
			{
				if (xmlTextReader.NodeType == XmlNodeType.Element && xmlTextReader.Name == "a")
				{
					// indexy bodu ktermu nle naten atribut
					int indexI = Int32.Parse(xmlTextReader.GetAttribute("i"));
					int indexJ = Int32.Parse(xmlTextReader.GetAttribute("j"));
					Type type = Globals.FindType(this.libNames, xmlTextReader.GetAttribute("type"));
					DataObject attr = (DataObject) Activator.CreateInstance(type, true);
					attr.ReadData(xmlTextReader);
					this.AddPixelAttr(indexI, indexJ, attr);
					i++;
				}
			} // while
		} // ReadPixelsAttrs()

		/// <summary>
		/// Pomocn metoda - vrt datov typ, kter nalezne v assembly visualization.
		/// </summary>
		/// <param name="typeName">Jmno vyhledvanho datovho typu</param>
		/// <returns>Nalezen datov typ</returns>
		protected Type GetType(string typeName)
		{
			try  
			{
				Assembly assembly = Assembly.GetExecutingAssembly();
				foreach (Type t in assembly.GetTypes())
				{
					if (t.Name == typeName)
					{
						return t;
					}
				} // foreach
			} // try
			catch (Exception ex)
			{
				Debug.WriteLine(this.GetType().Name + ": Throw exception at searching type " + typeName);
				Debug.WriteLine(ex.Message);

			} // catch
			throw new MveException("Type " + typeName + " not found in Visualization library.");
		} // GetType()

		/// <summary>
		/// Pomocn metoda pro zpis atribut bod do xml souboru.
		/// </summary>
		/// <param name="xmlTextWriter">Datov proud pro zpis dat do xml souboru.</param>
		protected void WritePointsAttrs(XmlTextWriter xmlTextWriter)
		{
			xmlTextWriter.WriteStartElement("pointsattrs");
			xmlTextWriter.WriteAttributeString("count", this.GetPointsAttrsCount().ToString());
			for(int i = 0; i < this.pointAttributes.GetLength(0); i++)
				for(int j = 0; j < this.pointAttributes.GetLength(1); j++)
				{
					ArrayList pointAttrList = this.pointAttributes[i, j];
					for (int k = 0; k < pointAttrList.Count; k++)
					{
						DataObject attr = (DataObject) pointAttrList[k];
						xmlTextWriter.WriteStartElement("a");
						xmlTextWriter.WriteAttributeString("i", i.ToString());
						xmlTextWriter.WriteAttributeString("j", j.ToString());
						xmlTextWriter.WriteAttributeString("type", attr.GetType().Name);
						attr.WriteData(xmlTextWriter);
						xmlTextWriter.WriteEndElement();
					}
				}
			xmlTextWriter.WriteEndElement();
		} // WritePointsAttrs()

		/// <summary>
		/// Pomocn metoda pro zpis atribut pixel do xml souboru.
		/// </summary>
		/// <param name="xmlTextWriter">Datov proud pro zpis dat do xml souboru.</param>
		protected void WritePixelsAttrs(XmlTextWriter xmlTextWriter)
		{
			xmlTextWriter.WriteStartElement("pixelsattrs");
			xmlTextWriter.WriteAttributeString("count", this.GetPixelsAttrsCount().ToString());
			for(int i = 0; i < this.pixelAttributes.GetLength(0); i++)
				for(int j = 0; j < this.pixelAttributes.GetLength(1); j++)
				{
					ArrayList pixelAttrList = this.pixelAttributes[i, j];
					for (int k = 0; k < pixelAttrList.Count; k++)
					{
						DataObject attr = (DataObject) pixelAttrList[k];
						xmlTextWriter.WriteStartElement("a");
						xmlTextWriter.WriteAttributeString("i", i.ToString());
						xmlTextWriter.WriteAttributeString("j", j.ToString());
						xmlTextWriter.WriteAttributeString("type", attr.GetType().Name);
						attr.WriteData(xmlTextWriter);
						xmlTextWriter.WriteEndElement();
					}
				}
			xmlTextWriter.WriteEndElement();
		} // WritePixelsAttrs()

		/// <summary>
		/// Vrac poet vech bod se ktermi jsou asociovny njak atributy.
		/// </summary>
		/// <returns>Poet vech bod s atributy.</returns>
		protected int GetCountPointsWithAttr()
		{
			int count = 0;
			for(int i = 0; i < this.pointAttributes.GetLength(0); i++)
				for(int j = 0; j < this.pointAttributes.GetLength(1); j++)
					if (this.pointAttributes[i, j].Count != 0)
						count++;
			return count;
		} // GetCountPointsWithAttr()

		/// <summary>
		/// Vrac poet vech pixel se ktermi jsou asociovny njak atributy.
		/// </summary>
		/// <returns>Poet vech pixel s atributy.</returns>
		protected int GetCountPixelsWithAttr()
		{
			int count = 0;
			for(int i = 0; i < this.pixelAttributes.GetLength(0); i++)
				for(int j = 0; j < this.pixelAttributes.GetLength(1); j++)
					if (this.pixelAttributes[i, j].Count != 0)
						count++;
			return count;
		} // GetCountPixelsWithAttr()

		/// <summary>
		/// Vrac poet vech atribut bod.
		/// </summary>
		/// <returns>Poet vech atribut bod.</returns>
		protected int GetPointsAttrsCount()
		{
			int count = 0;
			for(int i = 0; i < this.pointAttributes.GetLength(0); i++)
				for(int j = 0; j < this.pointAttributes.GetLength(1); j++)
					count += this.pointAttributes[i, j].Count;
			return count;
		} // GetPointsAttrsCount()

		/// <summary>
		/// Vrac poet vech atribut pixel.
		/// </summary>
		/// <returns>Poet vech atribut pixel.</returns>
		protected int GetPixelsAttrsCount()
		{
			int count = 0;
			for(int i = 0; i < this.pixelAttributes.GetLength(0); i++)
				for(int j = 0; j < this.pixelAttributes.GetLength(1); j++)
					count += this.pixelAttributes[i, j].Count;
			return count;
		} // GetPixelsAttrsCount()

// ===== Metody DataSetu ==========================================================

		/// <summary>
		/// Vrac bod zadan pozic v mce. Vrcen bod m ji konkrtn souadnice - reln sla v intervalu 0.0 a 1.0
		/// </summary>
		/// <param name="i">Index bodu v mce ve smru osy y. Mus bt v intervalu 0 a height.</param>
		/// <param name="j">Index bodu v mce ve smru osy x. Mus bt v intervalu 0 a width.</param>
		/// <returns>Bod s konkrtnmi souadnicemi - reln sla v intervalu 0.0 a 1.0</returns>
		public abstract Point2D GetPoint(int i, int j);

		/// <summary>
		/// Vrac pixel zadan pozic v mce. Pixel se skld ze 4 bod.
		/// </summary>
		/// <param name="i">Index pixelu v mce ve smru osy y. Mus bt v intervalu 0 a height - 1.</param>
		/// <param name="j">Index pixelu v mce ve smru osy x. Mus bt v intervalu 0 a width - 1.</param>
		/// <returns>Pixel na danch pozicch.</returns>
		public Pixel GetPixel(int i, int j)
		{
			if (i < this.height && j < this.width)
				return new Pixel(i + 1, j);
			else
			{
				Debug.WriteLine(this.GetType().Name + ": pixel [" + i + ", " + j + "] is out of range \n" + this.ToString());
				return null;
			}
		} // GetPixel()

		/// <summary>
		/// Pid atribut bodu. Bod je dn pozic [i, j] v mce, kad bod me mt libovoln poet atribut.
		/// </summary>
		/// <param name="i">Index bodu v mce ve smru osy y. Mus bt v intervalu 0 a height.</param>
		/// <param name="j">Index bodu v mce ve smru osy x. Mus bt v intervalu 0 a width.</param>
		/// <param name="attr">Pidvan atribut.</param>
		public void AddPointAttr(int i, int j, DataObject attr)
		{
			if (this.GetPoint(i, j) == null)
				return;		// bod je mimo rozsah mrizky
			this.pointAttributes[i, j].Add(attr);
			this.AddLibName(attr);
		} // AddPointAttr()

		/// <summary>
		/// Pid atribut pixelu. Pixel je dn pozic [i, j] v mce, kad pixel me mt libovoln poet atribut.
		/// </summary>
		/// <param name="i">Index pixelu v mce ve smru osy y. Mus bt v intervalu 0 a height - 1.</param>
		/// <param name="j">Index pixelu v mce ve smru osy x. Mus bt v intervalu 0 a width - 1.</param>
		/// <param name="attr">Pidvan atribut.</param>
		public void AddPixelAttr(int i, int j, DataObject attr)
		{
			if (this.GetPixel(i, j) == null)
				return;		// pixel je mimo rozsah mrizky
			this.pixelAttributes[i, j].Add(attr);
			this.AddLibName(attr);
		} // AddPixelAttr()

		/// <summary>
		/// Vrac seznam vech atribut asociovanch s bodem danm pozic [i, j] v mce.
		/// </summary>
		/// <param name="i">Index bodu v mce ve smru osy y. Mus bt v intervalu 0 a height.</param>
		/// <param name="j">Index bodu v mce ve smru osy x. Mus bt v intervalu 0 a width.</param>
		/// <returns>Seznam vech atribut asociovanch s danm bodem.</returns>
		public ArrayList GetPointAttrs(int i, int j)
		{
			if (this.GetPoint(i, j) == null)
				return null;		// bod je mimo rozsah mrizky
			return this.pointAttributes[i, j];
		} // GetPointAttrs()

		/// <summary>
		/// Vrac seznam vech atribut asociovanch s pixelem danm pozic [i, j] v mce.
		/// </summary>
		/// <param name="i">Index pixelu v mce ve smru osy y. Mus bt v intervalu 0 a height - 1.</param>
		/// <param name="j">Index pixelu v mce ve smru osy x. Mus bt v intervalu 0 a width - 1.</param>
		/// <returns>Seznam vech atribut asociovanch s danm pixelem.</returns>
		public ArrayList GetPixelAttrs(int i, int j)
		{
			if (this.GetPixel(i, j) == null)
				return null;		// pixel je mimo rozsah mrizky
			return this.pixelAttributes[i, j];
		} // GetPixelAttrs()

		/// <summary>
		/// Vrac prvn atribut danho jmna asociovan s bodem.
		/// </summary>
		/// <param name="i">Index bodu v mce ve smru osy y. Mus bt v intervalu 0 a height.</param>
		/// <param name="j">Index bodu v mce ve smru osy x. Mus bt v intervalu 0 a width.</param>
		/// <param name="attrType">Typ atributu (zadava se jako typeof(atribut)).</param>
		/// <returns>Prvn atribut danho jmna asociovan s bodem.</returns>
		public DataObject GetFirstPointAttr(int i, int j, Type attrType)
		{
			ArrayList pointAttrs;
			if ((pointAttrs = this.GetPointAttrs(i, j)) == null)
				return null;	// bod nem atributy
			foreach(DataObject attr in pointAttrs)
			{
				if (attr.GetType() == attrType)
					return attr;
			}
			Debug.WriteLine(this.GetType().Name + ": point [" + i + ", " + j + "] has no attribute " + attrType.Name + ".");
			return null;		
		} // GetFirstPointAttr()

		/// <summary>
		/// Vrac prvn atribut danho jmna asociovan s pixelem.
		/// </summary>
		/// <param name="i">Index pixelu v mce ve smru osy y. Mus bt v intervalu 0 a height - 1.</param>
		/// <param name="j">Index pixelu v mce ve smru osy x. Mus bt v intervalu 0 a width - 1.</param>
		/// <param name="attrType">Typ atributu (zadava se jako typeof(atribut)).</param>
		/// <returns>Prvn atribut danho jmna asociovan s pixelem.</returns>
		public DataObject GetFirstPixelAttr(int i, int j, Type attrType)
		{
			ArrayList pixelAttrs;
			if ((pixelAttrs = this.GetPixelAttrs(i, j)) == null)
				return null;	// pixel nem atributy
			foreach(DataObject attr in pixelAttrs)
			{
				if (attr.GetType() == attrType)
					return attr;
			}
			Debug.WriteLine(this.GetType().Name + ": pixel [" + i + ", " + j + "] has no attribute " + attrType.Name + ".");
			return null;		
		} // GetFirstPixelAttr()

		/// <summary>
		/// Vrac seznam vech atribut danho jmna asociovanch s bodem.
		/// </summary>
		/// <param name="i">Index bodu v mce ve smru osy y. Mus bt v intervalu 0 a height.</param>
		/// <param name="j">Index bodu v mce ve smru osy x. Mus bt v intervalu 0 a width.</param>
		/// <param name="attrType">Typ atributu (zadava se jako typeof(atribut)).</param>
		/// <returns>Seznam vech atribut danho jmna asociovanch s bodem.</returns>
		public ArrayList GetPointAttrsType(int i, int j, Type attrType)
		{
			ArrayList pointAttrs;
			if ((pointAttrs = this.GetPointAttrs(i, j)) == null)
				return null;	// bod nem dn atributy
			ArrayList myPointAttrs = new ArrayList();
			foreach(DataObject attr in pointAttrs)
			{
				if (attr.GetType() == attrType)
					myPointAttrs.Add(attr);
			}
			if (myPointAttrs.Count == 0)
			{
				Debug.WriteLine(this.GetType().Name + ": point [" + i + ", " + j + "] has no attribute " + attrType.Name + ".");
				return null;
			}
			else
				return myPointAttrs;
		} // GetPointAttrsType()

		/// <summary>
		/// Vrac seznam vech atribut danho jmna asociovanch s pixelem.
		/// </summary>
		/// <param name="i">Index pixelu v mce ve smru osy y. Mus bt v intervalu 0 a height - 1.</param>
		/// <param name="j">Index pixelu v mce ve smru osy x. Mus bt v intervalu 0 a width - 1.</param>
		/// <param name="attrType">Typ atributu (zadava se jako typeof(atribut)).</param>
		/// <returns>Seznam vech atribut danho jmna asociovanch s pixelem.</returns>
		public ArrayList GetPixelAttrsType(int i, int j, Type attrType)
		{
			ArrayList pixelAttrs;
			if ((pixelAttrs = this.GetPixelAttrs(i, j)) == null)
				return null;	// pixel nem dn atributy
			ArrayList myPixelAttrs = new ArrayList();
			foreach(DataObject attr in pixelAttrs)
			{
				if (attr.GetType() == attrType)
					myPixelAttrs.Add(attr);
			}
			if (myPixelAttrs.Count == 0)
			{
				Debug.WriteLine(this.GetType().Name + ": pixel [" + i + ", " + j + "] has no attribute " + attrType.Name + ".");
				return null;
			}		
			else
				return myPixelAttrs;
		} // GetPixelAttrsType()

	} // class Grid2D
} // namespace
