using System;
using Zcu.Mve.Visualization;
using System.Collections.Generic;

namespace Zcu.Mve.Visualization.Loaders
{
    public struct PropertyInfo
    {
        public string name;
        public Type type;

        public PropertyInfo(string name, Type type)
        {
            this.name = name;
            this.type = type;
        }
    }

	/// <summary>
	/// Class that contains all data structures.
	/// </summary>
	public class DataSet
	{
		/// <summary>
		/// Number of cells (triangles, polygons) in input file.
		/// </summary>
		private int numberOfCells = 0;

		/// <summary>
		/// Number of vertices in input file.
		/// </summary>
		private int numberOfVertices  = 0;

		/// <summary>
		/// Complete mesh.
		/// </summary>
		private Zcu.Mve.Visualization.Mesh             mesh                 = null;
		/// <summary>
		/// Array of vertices.
		/// </summary>
		private Zcu.Mve.Visualization.UniformDataArray verticesArray        = null;
		/// <summary>
		/// Array of cells (triangles, polygons).
		/// </summary>
		private Zcu.Mve.Visualization.UniformDataArray cellArray            = null;
		/// <summary>
		/// Array of vertex normals.
		/// </summary>
		private Zcu.Mve.Visualization.UniformDataArray verticesNormalsArray = null;
		/// <summary>
		/// Array of cell normals.
		/// </summary>
		private Zcu.Mve.Visualization.UniformDataArray cellNormalsArray     = null;
		/// <summary>
		/// Array of cells neighbours. Used only for TRI file format (if available).
		/// </summary>
		private Zcu.Mve.Visualization.UniformDataArray cellNeighboursArray  = null;

		/// <summary>
		/// True if DataSet contains polygons.
		/// </summary>
		private bool containsPolygons = false;
		/// <summary>
		/// True if DataSet contains vertex normals.
		/// </summary>
		private bool containsVerticesNormals = false;
		/// <summary>
		/// True if DataSet contains cell normals.
		/// </summary>
		private bool containsCellsNormals = false;
		/// <summary>
		/// True if DataSet contains cell neighbours.
		/// </summary>
		private bool containsCellsNeighbours = false;

        public UniformDataArray[] vertexAttrs;        

        public List<PropertyInfo> vertexPropertyInfoList;

		/// <summary>
		/// Constructor of DataSet class.
		/// </summary>
		/// <param name="numberOfVertices">Number of vertices in input file.</param>
		/// <param name="numberOfCells">Number of cells in input file.</param>
		/// <param name="typeOfCell">Type of cells: 0 - triangles,
		///                                         1 - polygons.</param>
		public DataSet(int numberOfVertices, int numberOfCells, int typeOfCell)
		{
			this.numberOfVertices = numberOfVertices;
			this.numberOfCells = numberOfCells;

			if (typeOfCell == 0) 
			{
				this.cellArray = new Zcu.Mve.Visualization.UniformDataArray(typeof(Zcu.Mve.Visualization.Triangle), numberOfCells);
			}
			else if (typeOfCell == 1)
			{
				this.cellArray = new Zcu.Mve.Visualization.UniformDataArray(typeof(Zcu.Mve.Visualization.Polygon), numberOfCells);
			}

			this.verticesArray = new Zcu.Mve.Visualization.UniformDataArray(typeof(Zcu.Mve.Visualization.Point3D), numberOfVertices);
			this.verticesNormalsArray = new Zcu.Mve.Visualization.UniformDataArray(typeof(Zcu.Mve.Visualization.Normal3D), numberOfVertices);
			this.cellNormalsArray = new Zcu.Mve.Visualization.UniformDataArray(typeof(Zcu.Mve.Visualization.Normal3D), numberOfCells);
			this.cellNeighboursArray = new Zcu.Mve.Visualization.UniformDataArray(typeof(Zcu.Mve.Visualization.TriangleNeighbor), numberOfCells);
		}

		/// <summary>
		/// Fill mesh data structure.
		/// </summary>
		/// <returns>Filled mesh.</returns>
		public Zcu.Mve.Visualization.Mesh GetMesh()
		{
			if (this.mesh == null)
			{
				this.mesh = new Zcu.Mve.Visualization.Mesh();

				this.mesh.Points = this.verticesArray;

				if (this.containsPolygons == false)
				{
					this.mesh.AddCells(typeof(Zcu.Mve.Visualization.Triangle).Name, this.cellArray);
				}
				else
				{
					this.mesh.AddCells(typeof(Zcu.Mve.Visualization.Polygon).Name, this.cellArray);
				}
				
				if (this.containsVerticesNormals)
				{
					this.mesh.AddPointAttrs(typeof(Zcu.Mve.Visualization.Normal3D).Name, this.verticesNormalsArray);
				}

				if (this.containsCellsNormals)
				{
					this.mesh.AddCellAttrs(typeof(Zcu.Mve.Visualization.Triangle).Name, typeof(Zcu.Mve.Visualization.Normal3D).Name, this.cellNormalsArray);
				}

				if (this.containsCellsNeighbours)
				{
					this.mesh.AddCellAttrs(typeof(Zcu.Mve.Visualization.Triangle).Name, typeof(Zcu.Mve.Visualization.TriangleNeighbor).Name, this.cellNeighboursArray);
				}

                //if (vertexAttrs != null)
                //{
                //    for (int i = 0; i < vertexAttrs.Length; i++)
                //    {
                //        this.mesh.AddPointAttrs(vertexPropertyInfoList[i].name, vertexAttrs[i]);
                //    }
                //}
			}

			return this.mesh;
		}

		/// <summary>
		/// Adds vertex to vertices array.
		/// </summary>
		/// <param name="index">Index of vertex.</param>
		/// <param name="vertex">Structure of vertex (Zcu.Mve.Visualization.Point3D).</param>
		public void AddVertex(int index, Zcu.Mve.Visualization.Point3D vertex)
		{
			this.verticesArray[index] = vertex;
		}

		/// <summary>
		/// Adds triangle to cells array.
		/// </summary>
		/// <param name="index">Index of triangle.</param>
		/// <param name="triangle">Structure of triangle (Zcu.Mve.Visualization.Triangle).</param>
		public void AddTriangle(int index, Zcu.Mve.Visualization.Triangle triangle)
		{
			this.cellArray[index] = triangle;
		}

		/// <summary>
		/// Adds polygon to cells array.
		/// </summary>
		/// <param name="index">Index of polygon.</param>
		/// <param name="polygon">Structure of polygon (Zcu.Mve.Visualization.Polygon).</param>
		public void AddPolygon(int index, Zcu.Mve.Visualization.Polygon polygon)
		{
			this.containsPolygons = true;

			this.cellArray[index] = polygon;
		}

		/// <summary>
		/// Adds cell normal to cell normals array.
		/// </summary>
		/// <param name="index">Index of cell.</param>
		/// <param name="normal">Structure of normal (Zcu.Mve.Visualization.Normal3D).</param>
		public void AddCellNormal(int index, Zcu.Mve.Visualization.Normal3D normal)
		{
			this.containsCellsNormals = true;

			this.cellNormalsArray[index] = normal;
		}

		/// <summary>
		/// Adds vertex normal to vertex normals array.
		/// </summary>
		/// <param name="index">Index of vertex.</param>
		/// <param name="normal">Structure of normal (Zcu.Mve.Visualization.Normal3D).</param>
		public void AddVertexNormal(int index, Zcu.Mve.Visualization.Normal3D normal)
		{
			this.containsVerticesNormals = true;

			this.verticesNormalsArray[index] = normal;
		}

		/// <summary>
		/// Adds triangle neighbour to neighbours array.
		/// </summary>
		/// <param name="index">Index of triangle</param>
		/// <param name="neighbor">Structure of triangle neighbour (Zcu.Mve.Visualization.TriangleNeighbour.</param>
		public void AddTriangleNeighbours(int index, Zcu.Mve.Visualization.TriangleNeighbor neighbor)
		{
			this.containsCellsNeighbours = true;

			this.cellNeighboursArray[index] = neighbor;
		}
	}
}
