using System;
using System.Drawing;
using System.Windows.Forms;
using System.IO;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization.Loaders
{
	/// <summary>
	/// Dialog for selecting input file.
	/// </summary>
	public class MeshLoaderSetup : ModuleSetup
	{
		// Dialog components.
		private System.Windows.Forms.Label   label;
		private System.Windows.Forms.Button  browse;
		private System.Windows.Forms.TextBox filePath;
		
		private System.Windows.Forms.OpenFileDialog openFileDialog;

		/// <summary>
		/// Reference to MeshLoader object.
		/// </summary>
		private MeshLoader meshLoader;

		/// <summary>
		/// Name of previous selected file.
		/// </summary>
		private string oldFileName = null;

		/// <summary>
		/// Constructor of MeshLoaderSetup.
		/// </summary>
		/// <param name="meshLoader">MeshLoader object.</param>
		/// <param name="oldFileName">Name of previous selected file.</param>
		public MeshLoaderSetup(MeshLoader meshLoader, string oldFileName)
		{
			Initialize();

			this.meshLoader  = meshLoader;
			this.oldFileName = oldFileName;

			this.filePath.Text = oldFileName;
		}

		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void Initialize()
		{
			// creating components
			this.label = new System.Windows.Forms.Label();
			this.filePath = new System.Windows.Forms.TextBox();
			this.browse = new System.Windows.Forms.Button();
			// label
			this.label.Location = new System.Drawing.Point(8, 8);
			this.label.Size = new System.Drawing.Size(72, 16);
			this.label.Text = "Selected file:";
			// filePath
			this.filePath.Location = new System.Drawing.Point(8, 24);
			this.filePath.Size = new System.Drawing.Size(248, 20);
			this.filePath.TabIndex = 0;
			this.filePath.Text = "";
			// browse
			this.browse.Location = new System.Drawing.Point(256, 22);
			this.browse.Size = new System.Drawing.Size(24, 24);
			this.browse.TabIndex = 1;
			this.browse.Text = "...";
			this.browse.Click +=new EventHandler(browse_Click);

			// fileDialog
			this.ClientSize = new System.Drawing.Size(292, 53);
			this.Controls.Add(this.browse);
			this.Controls.Add(this.filePath);
			this.Controls.Add(this.label);
			this.Name = "FileDialog";
		}



		/// <summary>
		/// Handler for browse button click.
		/// </summary>
		private void browse_Click(object sender, EventArgs e)
		{
			this.openFileDialog = new OpenFileDialog();
			this.openFileDialog.Filter = "All supported files|*.tri; *.stl; *.ply; *.cmx|TRI files|*.tri|STL files|*.stl|PLY files|*.ply|CMX files|*.cmx";
			this.openFileDialog.ShowDialog();
			
			if (this.openFileDialog.FileName != "") 
			{
				this.filePath.Text = this.openFileDialog.FileName;
			}
			
			this.openFileDialog.Dispose();
		}

		/// <summary>
		/// Analyze type of selected file based on file extension.
		/// </summary>
		/// <param name="path">Full filename (full path and filename).</param>
		/// <returns>Return type of selected input file - 1 - TRI format,
		///                                               2 - STL format,
		///                                               3 - PLY format,
		///                                               4 - CMX format,
		///                                               0 - ??? format.</returns>
		private byte tellFileType(string path)
		{
			if (path.EndsWith(".tri"))
			{
				return 1;
			}
			else if (path.EndsWith(".stl"))
			{
				return 2;
			}
			else if (path.EndsWith(".ply"))
			{
				return 3;
			}
			else if (path.EndsWith(".cmx"))
			{
				return 4;
			}
			else
			{
				return 0;
			}
		}

		/// <summary>
		/// Handler for OK button click - store input file name and tests filename. 
		/// </summary>
		/// <returns>Returns True if selected file is supported.</returns>
		public override bool OnOkClick()
		{
			byte type;

			if (!File.Exists(this.filePath.Text))          // Test if the file exists.
			{
				MessageBox.Show(this, "Invalid file name!\n\nClick Cancel or choose existing file.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;                              // Doesn't close dialog.
			}

			if ((type = this.tellFileType(this.filePath.Text.ToLower())) == 0) 
			{                                              // Test the extension of file - 0 is code for unsupported type.
				MessageBox.Show(this, "Unknown file format!\n\nSee help for list of supported file types.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;                              // Doesn't close dialog.
			}			

			this.meshLoader.FileName = this.filePath.Text;
			                                               // Change meshLoader file name property.
			this.meshLoader.typeOfFile = type;             // Change meshLoader file type property.

			if (this.oldFileName != this.filePath.Text)    // If new file selected, ...
			{
				this.meshLoader.newFileSelected = true;    // ... set flag that cause generating new triangle mesh.
			}
			return true;                                   // Everything is allright, close dialog.
		}
	}
}
