using System;
using System.IO;

namespace Zcu.Mve.Visualization.Loaders
{
	/// <summary>
	/// Class for reading CMX file format.
	/// </summary>
	public class ReaderCmx
	{
		private int numberOfCells = 0;
		private int numberOfVertices  = 0;

		private DataSet dataSet;

		private int indexOfVertex = 0;
		private int indexOfTriangle = 0;

		/// <summary>
		/// Reads input file.
		/// </summary>
		/// <param name="fileName">Full filename.</param>
		/// <returns>Filled object of DataSet class.</returns>
		public DataSet ReadFile(string fileName)
		{
			FileStream fileStream = new FileStream(fileName, FileMode.Open, FileAccess.Read);

			BinaryReader binReader = new BinaryReader(fileStream);

			byte[] buffer = new byte[6];                   // Create buffer for head of file.

			int numberOfLightSchemes = 0;                  // Numbers of file elements.
			int numberOfTextures     = 0;                  //
			int numberOfObjects      = 0;                  //

			int firstObjectByte    = 0;

			binReader.Read(buffer, 0, 6);                  // Read head of file.

			if (this.parseString(buffer, 0, 6) != "LFSCMX")
			{                                              // Test head of file.
				fileStream.Close();
				throw new Zcu.Mve.Core.MveException("Input file isn't original Live for Speed .cmx file.");
			}

			binReader.BaseStream.Position = 9;             // In CMX file 9th byte indicates car model.
			if (binReader.ReadByte() != 1)
			{
				fileStream.Close();
				throw new Zcu.Mve.Core.MveException("Input file isn't car model.");
			}

			try
			{
				binReader.BaseStream.Position = 68;        // Trys to read header informations.
				numberOfLightSchemes = binReader.ReadInt32();

				binReader.BaseStream.Position = 76;
				numberOfTextures = binReader.ReadInt32();

				binReader.BaseStream.Position = 80;
				numberOfObjects = binReader.ReadInt32();

				binReader.BaseStream.Position += numberOfLightSchemes * 72
				                               + numberOfTextures     * 20;
				                                           // Sets position on first Object.

				firstObjectByte = (int)binReader.BaseStream.Position;

				ReadNumberOfVerticesAndCells(binReader, numberOfObjects);

				binReader.BaseStream.Position = firstObjectByte;

				this.dataSet = new DataSet(this.numberOfVertices, this.numberOfCells, 0);
				                                           // Creates DataSet object.

				for (int index = 0; index < numberOfObjects; index++)
				{                                          // Reads all objects data.
					this.AddObjectToDataSet(binReader);
				}
			}
			catch (System.Exception){
				long position = binReader.BaseStream.Position;
				fileStream.Close();
		
				throw new Zcu.Mve.Core.MveException("Error in input file on byte "+ position + ".");
			}
			
			fileStream.Close();

			return this.dataSet;
		}

		/// <summary>
		/// Reads string in binary file.
		/// </summary>
		/// <param name="buff">Buffer with readed bytes.</param>
		/// <param name="start">Starting byte.</param>
		/// <param name="maxLength">Maximal length of string.</param>
		/// <returns>Readed string.</returns>
		private string parseString(byte[] buff, int start, int maxLength)
		{
			string str = "";
			for (int index = start;  index < start + maxLength; index++)
			{
				if (buff[index] == 0) break;

				str += (char)buff[index];
			}

			return str;
		}

		/// <summary>
		/// Passes file and counts number of vertices and cells.
		/// </summary>
		/// <param name="binReader">Reference of input file.</param>
		/// <param name="numberOfObjects">Number of objects in file.</param>
		private void ReadNumberOfVerticesAndCells(BinaryReader binReader, int numberOfObjects)
		{
			int numberOfTextures;
			int numberOfPoints;
			int numberOfTriangles;

			for (int index = 0; index < numberOfObjects; index++)
			{
				binReader.BaseStream.Position += 20;       // Object header.
				
				numberOfTextures = binReader.ReadInt16();
				numberOfPoints = binReader.ReadInt16();
				numberOfTriangles = binReader.ReadInt16();

				binReader.BaseStream.Position += 2         // The rest of Object attributes.
					                          + numberOfTextures  *  8
				                              + numberOfPoints    * 36
				                              + numberOfTriangles *  8;

				this.numberOfVertices += numberOfPoints;
				this.numberOfCells += numberOfTriangles;
			}
		}

		/// <summary>
		/// Reads one object in input file and stores it to DataSet.
		/// </summary>
		/// <param name="binReader">Reference of input file.</param>
		private void AddObjectToDataSet(BinaryReader binReader)
		{
			int numberOfTextures;
			int numberOfPoints;
			int numberOfTriangles;

			Zcu.Mve.Visualization.Point3D  vertex   = new Zcu.Mve.Visualization.Point3D(0, 0, 0);
			Zcu.Mve.Visualization.Normal3D normal   = new Zcu.Mve.Visualization.Normal3D(0.0, 0.0, 0.0);
			Zcu.Mve.Visualization.Triangle triangle = new Zcu.Mve.Visualization.Triangle(0, 0, 0);

			int startIndexOfVertex = 0;

			binReader.BaseStream.Position += 20;
				
			numberOfTextures  = binReader.ReadInt16();
			numberOfPoints    = binReader.ReadInt16();
			numberOfTriangles = binReader.ReadInt16();

			binReader.BaseStream.Position += 2;            // Skip rest of Object head.

			binReader.BaseStream.Position += numberOfTextures * 8;
                                                           // Skip Texture info blocks.

			startIndexOfVertex = this.indexOfVertex;
			for (int index = 0; index < numberOfPoints; index++)
			{                                              // Reads objects vertices.
				vertex.X = (double)binReader.ReadInt32();
				vertex.Y = (double)binReader.ReadInt32();
				vertex.Z = (double)binReader.ReadInt32();

				this.dataSet.AddVertex(this.indexOfVertex, vertex);

				normal.Nx = binReader.ReadSingle();
				normal.Ny = binReader.ReadSingle();
				normal.Nz = binReader.ReadSingle();

				this.dataSet.AddVertexNormal(this.indexOfVertex, normal);

				binReader.BaseStream.Position += 12;
				                                           // Skip other point attributes.
				this.indexOfVertex++;
			}

			for (int index = 0; index < numberOfTriangles; index++)
			{                                              // Reads object triangles.
				triangle.V1 = startIndexOfVertex + binReader.ReadInt16();
				triangle.V2 = startIndexOfVertex + binReader.ReadInt16();
				triangle.V3 = startIndexOfVertex + binReader.ReadInt16();

				this.dataSet.AddTriangle(this.indexOfTriangle, triangle);

				binReader.BaseStream.Position += 2;
                                                           // Skip other triangle attributes.
				this.indexOfTriangle++;
			}
		}
	}
}