using System;
using System.IO;
using System.Text;
using Zcu.Mve.Visualization;
using System.Collections.Generic;
using Zcu.Mve.Visualization.Loaders;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.Visualization.Loaders
{
	/// <summary>
	/// Class for reading PLY file format.
	/// </summary>
	public class ReaderPly
	{
		private int numberOfCells = 0;
		private int numberOfVertices  = 0;

		private int actualLine = 0;                        // Actual line of readed file.

		private DataSet dataSet;

		private string exception = null;                   // Exception error string.

        private enum Format{ascii,binary_big_endian, binary_little_endian};
        private enum ReadingStage { none, vertex, face };                

		/// <summary>
		/// Reads input file.
		/// </summary>
		/// <param name="fileName">Full filename.</param>
		/// <returns>Filled object of DataSet class.</returns>
		public DataSet ReadFile(string fileName)
		{
			StreamReader streamReader = new StreamReader(fileName);

			string line = null;
			byte   numberOf = 2;                           //  Counting parts of file head.

            Format format = Format.ascii;
            ReadingStage stage = ReadingStage.none;

            int vertexDataLength = 0;
            int vertexCoordinateLength = 0;
            int faceIndexCountLength = 0;
            int faceIndexLength = 0;

            int headerLength = 0;            

			try
			{
                List<PropertyInfo> PropertyInfoList = new List<PropertyInfo>();
				while (((line = streamReader.ReadLine()) != null)&&(!line.StartsWith("end_header"))) 
				{                                          // File head processing.
					this.actualLine++;
                    headerLength += line.Length + 1;


					if (line.StartsWith("element ver"))
					{                                      // Reads number of vertices.
						this.numberOfVertices = this.IntNumber(line);
                        stage = ReadingStage.vertex;
						numberOf--;
					}
					else if (line.StartsWith("element fac"))
					{                                      // Reads number of cells.
						this.numberOfCells = this.IntNumber(line);
                        stage = ReadingStage.face;
						numberOf--;
					}
                    else if (line.StartsWith("format"))
                    {
                        string[] parts = line.Split(new char[] { ' ' });
                        if (parts[1] == "binary_big_endian")
                        {
                            format = Format.binary_big_endian;
                        }
                        else if (parts[1] == "binary_little_endian")
                        {
                            format = Format.binary_little_endian;
                        }
                        else
                        {
                            format = Format.ascii;
                        }
                    }
                    else if (line.StartsWith("element"))
                    {
                        stage = ReadingStage.none;
                    }
                    else if (line.StartsWith("property"))
                    {
                        switch (stage)
                        {
                            case ReadingStage.vertex:
                                string[] parts = line.Split(new char[] { ' ' });
                                if (vertexDataLength == 0)
                                {
                                    switch (parts[1])
                                    {
                                        case "float":
                                            vertexCoordinateLength = 4;
                                            break;
                                        case "double":
                                            vertexCoordinateLength = 8;
                                            break;
                                    }
                                }
                                switch (parts[1])
                                {
                                    case "float":
                                        vertexDataLength += 4;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2],typeof(float)));
                                        break;
                                    case "double":
                                        vertexDataLength += 8;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(double)));
                                        break;
                                    case "uint":
                                        vertexDataLength += 4;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(uint)));
                                        break;
                                    case "int":
                                        vertexDataLength += 4;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(int)));
                                        break;
                                    case "ushort":
                                        vertexDataLength += 2;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(ushort)));
                                        break;
                                    case "short":
                                        vertexDataLength += 2;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(short)));
                                        break;
                                    case "uchar":
                                        vertexDataLength += 1;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(byte)));
                                        break;
                                    case "char":
                                        vertexDataLength += 1;
                                        PropertyInfoList.Add(new PropertyInfo(parts[2], typeof(sbyte)));
                                        break;
                                }
                                break;
                            case ReadingStage.face:
                                parts = line.Split(new char[] { ' ' });                                
                                switch (parts[2])
                                {
                                    case "float":
                                        faceIndexCountLength += 4;
                                        break;
                                    case "double":
                                        faceIndexCountLength += 8;
                                        break;
                                    case "uint":
                                        faceIndexCountLength += 4;
                                        break;
                                    case "int":
                                        faceIndexCountLength += 4;
                                        break;
                                    case "ushort":
                                        faceIndexCountLength += 2;
                                        break;
                                    case "short":
                                        faceIndexCountLength += 2;
                                        break;
                                    case "uchar":
                                        faceIndexCountLength += 1;
                                        break;
                                    case "char":
                                        faceIndexCountLength += 1;
                                        break;
                                }
                                switch (parts[3])
                                {
                                    case "float":
                                        faceIndexLength += 4;
                                        break;
                                    case "double":
                                        faceIndexLength += 8;
                                        break;
                                    case "uint":
                                        faceIndexLength += 4;
                                        break;
                                    case "int":
                                        faceIndexLength += 4;
                                        break;
                                    case "ushort":
                                        faceIndexLength += 2;
                                        break;
                                    case "short":
                                        faceIndexLength += 2;
                                        break;
                                    case "uchar":
                                        faceIndexLength += 1;
                                        break;
                                    case "char":
                                        faceIndexLength += 1;
                                        break;
                                }
                                break;
                        }
                    }

					if (numberOf == 0) 
					{                                      // Head reading is done, creating DataSet object.
						this.dataSet = new DataSet(this.numberOfVertices, this.numberOfCells, 0);
                        this.dataSet.vertexPropertyInfoList = PropertyInfoList;
					}
				}

                headerLength += line.Length + 1;

				if (numberOf != 0)                             // If number of triangles or vertices wasn't found - throws exception.
				{
					this.exception = "Invalid file format - can't find number of triangles and/or number of vertices.";
					throw new Zcu.Mve.Core.MveException(null);
				}

                dataSet.vertexAttrs = new UniformDataArray[dataSet.vertexPropertyInfoList.Count - 3];
                for (int i = 0; i < (dataSet.vertexPropertyInfoList.Count - 3); i++)
                {
                    Type type = dataSet.vertexPropertyInfoList[i + 3].type;
                    if ((type == typeof(byte))||(type == typeof(sbyte))||(type == typeof(int))||(type == typeof(uint))||(type == typeof(short))||(type == typeof(ushort)))
                        dataSet.vertexAttrs[i] = new UniformDataArray(typeof(Integer), numberOfVertices);
                }

                if (format == Format.ascii)
                {
                    this.ReadVertices(ref streamReader);
                    this.ReadPolygons(ref streamReader);
                }
                else
                {                    
                    this.ReadVerticesBinary(ref streamReader, format, vertexDataLength, vertexCoordinateLength, numberOfVertices, headerLength, dataSet.vertexPropertyInfoList);
                    this.ReadFacesBinary(ref streamReader, format, numberOfCells, faceIndexCountLength, faceIndexLength);
                }
                
				
			}
			catch (System.Exception)
			{
				streamReader.Close();
				if (this.exception == null)
				{
					throw new Zcu.Mve.Core.MveException("Error in input file on line " + this.actualLine + ".");
				}
				else
				{
					throw new Zcu.Mve.Core.MveException(this.exception);
				}
			}

			streamReader.Close();

			return this.dataSet;
		}

		/// <summary>
		/// Reads integer number in input string.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Readed number.</returns>
		private int IntNumber(string line) 
		{
			string[] elements;

			elements = line.Split(" :\t".ToCharArray());

			return Int32.Parse(elements[elements.Length - 1]);
		}

		/// <summary>
		/// Reads one vertex in input string.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Readed vertex (3D point).</returns>
		private Zcu.Mve.Visualization.Point3D ReadPoint3D(string line) 
		{
			string[] elements;
			int elementsLength = 0;

			System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();

			nfi.CurrencyDecimalSeparator = ".";            // Setting of decimal separator.

			elements = line.Split(" :\t".ToCharArray());

			for (int i = 0; i < elements.Length; i++)      // Delete elements that contains spaces only.
			{
				if (elements[i].Trim() != "")
				{
					elements[elementsLength++] = elements[i];
				}
			}

			return new Zcu.Mve.Visualization.Point3D(Double.Parse(elements[0], nfi),
				                                     Double.Parse(elements[1], nfi),
				                                     Double.Parse(elements[2], nfi));
		}

		/// <summary>
		/// Reads one polygon in input string.
		/// </summary>
		/// <param name="line">Input string.</param>
		/// <returns>Readed polygon.</returns>
		private Zcu.Mve.Visualization.Polygon ReadPolygon(string line) 
		{
			string[] elements;
			int[] polygon;
			int elementsLength = 0;
			int polyDimension;

			elements = line.Split(" :\t".ToCharArray());

			for (int i = 0; i < elements.Length; i++)      // Deletes elements that contains spaces only.
				if (elements[i].Trim() != "")
					elements[elementsLength++] = elements[i];


			polyDimension = Int32.Parse(elements[0]);      // Reads number of polygon vertices.
			polygon = new int[polyDimension];              // Allocate memory for polygon.

			for (int i = 0; i < polyDimension; i++)        // Reads indexes of polygon vertices.
			{
				polygon[i] = Int32.Parse(elements[i + 1]);
			}

			return new Zcu.Mve.Visualization.Polygon(polygon);
		}

		/// <summary>
		/// Reads all vertices from input file.
		/// </summary>
		/// <param name="streamReader">Reference to input file.</param>
		private void ReadVertices(ref StreamReader streamReader)
		{
            
                Zcu.Mve.Visualization.Point3D vertex;
                string line;
                int i = 0;                                  // Index of actual vertex.

                while ((line = streamReader.ReadLine()) != null)
                {
                    this.actualLine++;

                    line = line.Trim();

                    if (line == "")                            // Skips free line.
                    {
                        continue;
                    }
                    else                                       // Trys to read one vertex.
                    {
                        vertex = this.ReadPoint3D(line);
                        this.dataSet.AddVertex(i, vertex);

                        if (++i == numberOfVertices)           // Number of vertices reached - exit.
                        {
                            break;
                        }
                    }
                }
                       
		}

        private byte[] reverse4(byte[] original, int pos)
        {
            byte[] result = new byte[4];
            for (int i = 0; i < 4; i++)
                result[i] = original[pos+3-i];
            return result;
        }

        private byte[] reverse2(byte[] original, int pos)
        {
            byte[] result = new byte[2];
            for (int i = 0; i < 2; i++)
                result[i] = original[pos + 1 - i];
            return result;
        }

        private byte[] reverse8(byte[] original, int pos)
        {
            byte[] result = new byte[8];
            for (int i = 0; i < 8; i++)
                result[i] = original[pos + 7 - i];
            return result;
        }

        private void ReadVerticesBinary(ref StreamReader streamReader, Format format, int vertexDataLength, int vertexCoordinateLength, int vertexCount, int headerLength, List<PropertyInfo> piList)
        {                        
            BinaryReader br = new BinaryReader(streamReader.BaseStream);
            br.BaseStream.Seek(headerLength, SeekOrigin.Begin);            
            if (format == Format.binary_big_endian)
                {                    
                    for (int i = 0;i<vertexCount;i++)
                    {                        
                        byte[] buffer = br.ReadBytes(vertexDataLength);
                        double[] vertex = new double[3];
                        if (vertexCoordinateLength == 4)
                        {
                            
                            vertex[0] = BitConverter.ToSingle(reverse4(buffer,0), 0);
                            vertex[1] = BitConverter.ToSingle(reverse4(buffer,4), 0);
                            vertex[2] = BitConverter.ToSingle(reverse4(buffer,8), 0);
                            int pos = 4 * 3;
                            //ReadProps(i, buffer, pos);
                        }
                        else if (vertexCoordinateLength == 8)
                        {
                            vertex[0] = BitConverter.ToDouble(reverse8(buffer, 0),0);
                            vertex[1] = BitConverter.ToDouble(reverse8(buffer, 8),0);
                            vertex[2] = BitConverter.ToDouble(reverse8(buffer, 16),0);
                            int pos = 8 * 3;
                            //ReadProps(i, buffer, pos);
                        }
                        this.dataSet.AddVertex(i,new Point3D(vertex[0], vertex[1], vertex[2]));                        
                    }
                }
            else // little endian
                {
                    for (int i = 0; i < vertexCount; i++)
                    {
                        byte[] buffer = br.ReadBytes(vertexDataLength);
                        double[] vertex = new double[3];
                        if (vertexCoordinateLength == 4)
                        {

                            vertex[0] = BitConverter.ToSingle(buffer, 0);
                            vertex[1] = BitConverter.ToSingle(buffer, 4);
                            vertex[2] = BitConverter.ToSingle(buffer, 8);
                            int pos = 4 * 3;
                            ReadProps(i, buffer, pos);
                        }
                        else if (vertexCoordinateLength == 8)
                        {
                            vertex[0] = BitConverter.ToDouble(buffer, 0);
                            vertex[1] = BitConverter.ToDouble(buffer, 8);
                            vertex[2] = BitConverter.ToDouble(buffer, 16);
                            int pos = 8 * 3;
                            ReadProps(i, buffer, pos);
                        }
                        this.dataSet.AddVertex(i, new Point3D(vertex[0], vertex[1], vertex[2]));
                    }
                }
        }

        private int ReadProps(int i, byte[] buffer, int pos)
        {
            for (int j = 0; j < dataSet.vertexAttrs.Length; j++)
            {
                UniformDataArray attr = dataSet.vertexAttrs[j];
                if (attr.ElemType == typeof(byte))
                {
                    attr[i] = new Integer(buffer[pos]);
                    pos++;
                }
            }
            return pos;
        }
        private void ReadFacesBinary(ref StreamReader streamReader, Format format, int faceCount, int listLengthLength, int ListItemLength)
        {
            BinaryReader br = new BinaryReader(streamReader.BaseStream);
            if (format == Format.binary_big_endian)
            {
                for (int i = 0; i < faceCount; i++)
                {                    
                    byte[] listLength = br.ReadBytes(listLengthLength);
                    int length = 0;
                    if (listLengthLength == 1)
                        length = listLength[0];
                    if (listLengthLength == 2)
                        length = BitConverter.ToUInt16(listLength, 0);
                    int[] indices = new int[length];
                    
                    byte[] buffer = br.ReadBytes(length * ListItemLength);
                    for (int j = 0; j < length; j++)
                    {
                        if (ListItemLength == 2)
                            indices[j] = BitConverter.ToUInt16(reverse2(buffer, j * ListItemLength),0);
                        else if (ListItemLength == 4)
                            indices[j] = (int)BitConverter.ToUInt32(reverse4(buffer,j * ListItemLength), 0);
                    }
                    if (length == 3)
                        this.dataSet.AddTriangle(i, new Triangle(indices[0], indices[1], indices[2]));
                    else
                        this.dataSet.AddPolygon(i, new Polygon(indices));
                }
            }
            else                
            {
                for (int i = 0; i < faceCount; i++)
                {
                    byte[] listLength = br.ReadBytes(listLengthLength);
                    int length = 0;
                    if (listLengthLength == 1)
                        length = listLength[0];
                    if (listLengthLength == 2)
                        length = BitConverter.ToUInt16(listLength, 0);
                    int[] indices = new int[length];

                    byte[] buffer = br.ReadBytes(length * ListItemLength);
                    for (int j = 0; j < length; j++)
                    {
                        if (ListItemLength == 2)
                            indices[j] = BitConverter.ToUInt16(buffer, j * ListItemLength);
                        else if (ListItemLength == 4)
                            indices[j] = (int)BitConverter.ToUInt32(buffer, j * ListItemLength);
                    }
                    if (length == 3)
                        this.dataSet.AddTriangle(i, new Triangle(indices[0], indices[1], indices[2]));
                    else
                        this.dataSet.AddPolygon(i, new Polygon(indices));
                }
            }
        }

        /// <summary>
        /// Reads all polygons from input file.
        /// </summary>
        /// <param name="streamReader">Reference to input file.</param>
        private void ReadPolygons(ref StreamReader streamReader)
		{
			Zcu.Mve.Visualization.Polygon polygon;
			string   line;
			int    i = 0;                                  // Index of actual triangle.

			while ((line = streamReader.ReadLine()) != null)
			{
				this.actualLine++;

				line = line.Trim();

				if (line == "")                            // Skips free line.
				{
					continue;
				}
				else                                       // Try to reads one triangle.
				{
					polygon = this.ReadPolygon(line);

					for (int index = 0; index < polygon.PointCount; index++)
					{
						if (polygon.GetPoint(index) > this.numberOfVertices)
						{
							this.exception = "Error in input file on line " + this.actualLine + " - index is greater then number of vertices.";
							throw new Zcu.Mve.Core.MveException(null);
						}
					}

					this.dataSet.AddPolygon(i, polygon);

					if (++i == numberOfCells)              // Number of triangles reached - exit.
					{
						break;
					}
				}
			}

			if (i != this.numberOfCells)
			{
				this.exception = "Error in input file - number of cells and/or vertices doesn't match.";
				throw new Zcu.Mve.Core.MveException(null);
			}
		}
	}
}
