using System;
using Zcu.Mve.Core;
using Zcu.Mve.Strings;

namespace Zcu.Mve.Visualization.Savers
{
	/// <summary>
	/// Main class of MeshSaver. Loads and saves configuration file and runs file savers.
	/// </summary>
	[ModuleInfo("Jan Brta", "Module for saving of mesh to .tri, .stl and .ply file formats.",
                IconName="MeshLS.ico", Assemblied = "2005-5-24", Category = "File Saver")]
	[PortInfo("Input", "Zcu.Mve.Visualization.Mesh - must contains vertices, triangles or polygons (for ply format), ...")]
	public class MeshSaver : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Stores type of output file.
		/// </summary>
		internal byte typeOfFile   = 0;
		/// <summary>
		/// Stores name of input file.
		/// </summary>
		internal string nameOfFile = null;
		/// <summary>
		/// Stores flags that indicates which part of TRI file will be saved.
		/// </summary>
		internal bool[] flags      = new bool[] {true, true, true, true};

		/// <summary>
		/// Mesh data that is on input port of module.
		/// </summary>
		private Zcu.Mve.Visualization.Mesh mesh = null;

		/// <summary>
		/// Constructor of MeshSaver. Creates input port.
		/// </summary>
		public MeshSaver()
		{
			AddInPort("Input", typeof (Zcu.Mve.Visualization.Mesh));
            AddInPort("FileName", typeof(MveString), false);
            AddOutPort("Output", typeof(Zcu.Mve.Visualization.Mesh));
		}

		/// <summary>
		/// Entry point of module. Saves data on input port to file.
		/// </summary>
		public override void Execute()
		{
			mesh = (Zcu.Mve.Visualization.Mesh)GetInput("Input");
			
//			if (mesh.CheckConsistence())
//			{
//				Console.WriteLine("OK");
//			}
//			else
//			{
//				Console.WriteLine("Not OK");
//				return;
//			}

			int startTics = Environment.TickCount;
			this.SaveFile();
			int endTics   = Environment.TickCount;

			//Console.WriteLine("File saved in {0} ms", (endTics - startTics));
            SetOutput("Output", mesh);
		}

		/// <summary>
		/// Creates MeshSaverSetup, where user can sets output file and choose blocks of TRI file to save.
		/// </summary>
		/// <returns>Module setup.</returns>
		public override ModuleSetup InvokeSetup()
		{
			return new MeshSaverSetup(this, this.nameOfFile);
		}

		/// <summary>
		/// Configuration is saved to XML map file.
		/// </summary>
		/// <param name="config"></param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			System.Xml.XmlElement xmlEl;

			xmlEl = config.OwnerDocument.CreateElement("Output_file");
			xmlEl.SetAttribute("path", this.nameOfFile);
			xmlEl.SetAttribute("type", this.typeOfFile.ToString());
			config.AppendChild(xmlEl);

			xmlEl = config.OwnerDocument.CreateElement("TRI_attributes");
			xmlEl.SetAttribute("Triangles",           this.flags[0].ToString());
			xmlEl.SetAttribute("Triangles_normals",   this.flags[1].ToString());
			xmlEl.SetAttribute("Triangles_neighbors", this.flags[2].ToString());
			xmlEl.SetAttribute("Vertices_normals",    this.flags[3].ToString());
			config.AppendChild(xmlEl);
		}

		/// <summary>
		/// Read of configuration from XML map file.
		/// </summary>
		/// <param name="config"></param>
		public override void ReadConfig(System.Xml.XmlElement config)
		{
			System.Xml.XmlNodeList childList = config.ChildNodes;

			try
			{
				if ((this.nameOfFile = childList[0].Attributes["path"].Value) == "")
				{                                          // No file name stored in map file.
					this.nameOfFile = null;
				}

				this.typeOfFile = Byte.Parse(childList[0].Attributes["type"].Value);

				flags[0] = bool.Parse(childList[1].Attributes["Triangles"].Value);
				flags[1] = bool.Parse(childList[1].Attributes["Triangles_normals"].Value);
				flags[2] = bool.Parse(childList[1].Attributes["Triangles_neighbors"].Value);
				flags[3] = bool.Parse(childList[1].Attributes["Vertices_normals"].Value);
			}
			catch
			{                                              // If some problem appear, default configuration will be set.
				this.nameOfFile = null;
				this.typeOfFile = 0;

				flags[0] = true;
				flags[1] = true;
				flags[2] = true;
				flags[3] = true;
			}
		}

		/// <summary>
		/// Call methods for save output file based on type of file.
		/// </summary>
		private void SaveFile() 
		{
            if (this.GetInPort("FileName").Link != null)
            {
                MveString fn = (MveString)GetInput("FileName");
                this.typeOfFile = MeshSaverSetup.tellFileType(fn.Value);
                this.nameOfFile = fn.Value;
            }

			switch (this.typeOfFile) 
			{
				case 1: this.SaveTriFile();
					break;
				case 2: this.SaveStlFile();
					break;
				case 3: this.SavePlyFile();
					break;
			}
		}

		/// <summary>
		/// Creates saver object of TRI format.
		/// </summary>
		private void SaveTriFile()
		{
			SaverTri saver = new SaverTri(this.flags);     // Creates file saver.

			saver.SaveFile(this.mesh, this.nameOfFile);    // Saves file.
		}

		/// <summary>
		/// Creates saver object of text STL format.
		/// </summary>
		private void SaveStlFile()
		{
			SaverStl saver = new SaverStl();               // Creates file saver.

			saver.SaveFile(this.mesh, this.nameOfFile);    // Saves file.
		}

		/// <summary>
		/// Creates saver object of PLY format.
		/// </summary>
		private void SavePlyFile()
		{
			SaverPly saver = new SaverPly();               // Creates file saver.

			saver.SaveFile(this.mesh, this.nameOfFile);    // Saves file.
		}
	}
}
