using System;
using System.IO;

namespace Zcu.Mve.Visualization.Savers
{
	/// <summary>
	/// Class for saving PLY file format.
	/// </summary>
	public class SaverPly
	{
		private int numberOfCells = 0;
		private int numberOfVertices  = 0;

		private Zcu.Mve.Visualization.UniformDataArray verticesArray = null;

		private Zcu.Mve.Visualization.UniformDataArray[] polygonsArrays;
		private Zcu.Mve.Visualization.UniformDataArray[] trianglesArrays;

		/// <summary>
		/// Saves mesh data to output file.
		/// </summary>
		/// <param name="mesh">Input mesh.</param>
		/// <param name="nameOfFile">Name of output file.</param>
		public void SaveFile(Zcu.Mve.Visualization.Mesh mesh, string nameOfFile)
		{
			this.verticesArray = mesh.Points;              // Gets vertices array.
			this.numberOfVertices = mesh.Points.Count;     // Gets number of vertices.

			this.polygonsArrays = mesh.GetCells(typeof(Zcu.Mve.Visualization.Polygon));
			this.trianglesArrays = mesh.GetCells(typeof(Zcu.Mve.Visualization.Triangle));
			                                               // Gets cells arrays.
			this.numberOfCells = this.SumNumberOfCells();
			                                               // Gets number of cells.

			if ((this.numberOfCells < 1) || (this.numberOfVertices < 3))
			{
				throw new Zcu.Mve.Core.MveException("No polygons/triangles or not enough vertices to store.");
			}

			try
			{
				this.WriteDataToFile(nameOfFile);
			}
			catch (System.Exception)
			{
				throw new Zcu.Mve.Core.MveException("Error writing output file.");
			}
		}

		/// <summary>
		/// Count number of cells.
		/// </summary>
		/// <returns>Number of polygons.</returns>
		private int SumNumberOfCells()
		{
			int sum = 0;

			for (int index = 0; index < this.polygonsArrays.Length; index++)
			{
				sum += this.polygonsArrays[index].Count;
			}

			for (int index = 0; index < this.trianglesArrays.Length; index++)
			{
				sum += this.trianglesArrays[index].Count;
			}

			return sum;
		}

		/// <summary>
		/// Writes complete data to output file.
		/// </summary>
		/// <param name="nameOfFile">Name of input file.</param>
		private void WriteDataToFile(string nameOfFile)
		{
			Zcu.Mve.Visualization.Point3D vertex;
			Zcu.Mve.Visualization.Triangle triangle;
			Zcu.Mve.Visualization.Polygon polygon;
			
			StreamWriter streamWriter = new StreamWriter(nameOfFile);

			System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();

			nfi.CurrencyDecimalSeparator = ".";

			streamWriter.WriteLine("ply");                 // Writes head of file.
			streamWriter.WriteLine("format ascii 1.0");
			streamWriter.WriteLine("element vertex {0}", this.numberOfVertices);
			streamWriter.WriteLine("property float32 x");
			streamWriter.WriteLine("property float32 y");
			streamWriter.WriteLine("property float32 z");
			streamWriter.WriteLine("element face {0}", this.numberOfCells);
			streamWriter.WriteLine("property list uint8 int32 vertex_indices");
			streamWriter.WriteLine("end_header");
			
			for (int index = 0; index < this.numberOfVertices; index++)
			{                                              // Writes all vertices.
				vertex = (Zcu.Mve.Visualization.Point3D)this.verticesArray[index];	
				streamWriter.WriteLine("{0} {1} {2}", vertex[0].ToString("0.######0.0", nfi), vertex[1].ToString("0.######0.0", nfi), vertex[2].ToString("0.######0.0", nfi));
			}

			if ((this.polygonsArrays != null) && (this.polygonsArrays.Length != 0))
			{                                              // Writes all polygons.
				for (int arrayIndex = 0; arrayIndex < this.polygonsArrays.Length; arrayIndex++)
				{
					for (int index = 0; index < this.polygonsArrays[arrayIndex].Count; index++)
					{
						polygon = (Zcu.Mve.Visualization.Polygon)this.polygonsArrays[arrayIndex][index];
						streamWriter.Write(polygon.PointCount);
						for (int point = 0; point < polygon.PointCount; point++)
						{
							streamWriter.Write(" {0}", polygon.GetPoint(point));
						}
						streamWriter.WriteLine();
					}
				}
			}

			if ((this.trianglesArrays != null) && (this.trianglesArrays.Length != 0))
			{                                              // Writes all triangles.
				for (int arrayIndex = 0; arrayIndex < this.trianglesArrays.Length; arrayIndex++)
				{
					for (int index = 0; index < this.trianglesArrays[arrayIndex].Count; index++)
					{
						triangle = (Zcu.Mve.Visualization.Triangle)this.trianglesArrays[arrayIndex][index];	
						streamWriter.WriteLine("3 {0} {1} {2}", triangle.V1, triangle.V2, triangle.V3);
					}
				}
			}
			
			streamWriter.Close();
		}
	}
}
