using System;
using System.IO;
using System.Collections;

namespace Zcu.Mve.Visualization.Savers
{
	/// <summary>
	/// Class for saving TRI file format.
	/// </summary>
	public class SaverTri
	{
		private int numberOfCells = 0;
		private int numberOfVertices  = 0;

		private Zcu.Mve.Visualization.UniformDataArray verticesArray = null;

		private Zcu.Mve.Visualization.UniformDataArray[] cellsArrays;

		private ArrayList verticesNormalsArrayList = new ArrayList();
		private ArrayList cellNormalsArrayList     = new ArrayList();
		private ArrayList cellNeighborsArrayList   = new ArrayList();

		private bool[] flags = new bool[4];

		/// <summary>
		/// Constructor of SaverTri class.
		/// </summary>
		/// <param name="_flags">Flags indicating which blocks of TRI format will be saved.</param>
		public SaverTri(bool[] _flags)
		{
			this.flags = _flags;
		}

		/// <summary>
		/// Saves mesh data to output file.
		/// </summary>
		/// <param name="mesh">Input mesh.</param>
		/// <param name="nameOfFile">Name of output file.</param>
		public void SaveFile(Zcu.Mve.Visualization.Mesh mesh, string nameOfFile)
		{
			this.verticesArray = mesh.Points;              // Gets vertices array.
			this.numberOfVertices = mesh.Points.Count;     // Gets number of vertices.

			if (this.numberOfVertices == 0)
			{
				throw new Zcu.Mve.Core.MveException("Mesh must contain one vertex at least.");
			}

			this.cellsArrays = mesh.GetCells(typeof(Zcu.Mve.Visualization.Triangle));
				                                           // Gets cells arrays.
			this.numberOfCells = this.SumNumberOfCells();
			                                               // Gets number of cells.

			this.ReadAttributsVertices(mesh);
			this.ReadAttributsCells(mesh);

			try
			{
				this.WriteDataToFile(nameOfFile);
			}
			catch (System.Exception)
			{
				throw new Zcu.Mve.Core.MveException("Error writing output file. Please check consistency of mesh or write rights to output file.");
			}
		}

		/// <summary>
		/// Counts number of cells.
		/// </summary>
		/// <returns>Number of triangles.</returns>
		private int SumNumberOfCells()
		{
			int sum = 0;

			for (int index = 0; index < this.cellsArrays.Length; index++)
			{
				sum += this.cellsArrays[index].Count;
			}
			return sum;
		}

		/// <summary>
		/// Reads attributs of vertices from mesh and store it to temporal array.
		/// </summary>
		/// <param name="mesh">Input mesh.</param>
		private void ReadAttributsVertices(Zcu.Mve.Visualization.Mesh mesh)
		{
			string[] attributsVertices;

			attributsVertices = new string[mesh.GetPointAttrsIds().Count];
			mesh.GetPointAttrsIds().CopyTo(attributsVertices, 0);

			for (int index = 0; index < attributsVertices.Length; index++)
			{
				if (this.flags[3] && (mesh.GetPointAttrs(attributsVertices[index]).ElemType == typeof(Zcu.Mve.Visualization.Normal3D)))
				{
					this.verticesNormalsArrayList.Add(mesh.GetPointAttrs(attributsVertices[index]));

					if (mesh.GetPointAttrs(attributsVertices[index]).Count > this.numberOfVertices)
					{
						throw new Zcu.Mve.Core.MveException("Number of vertex normals and number of vertices is not same.");
					}
				}
			}
		}

		/// <summary>
		/// Reads attributs of cells from mesh and store it to temporal array.
		/// </summary>
		/// <param name="mesh">Input mesh.</param>
		private void ReadAttributsCells(Zcu.Mve.Visualization.Mesh mesh)
		{
			string[] cellsIds;
			string[] attributesIds;

			cellsIds = new string[mesh.GetCellIds().Count];
			mesh.GetCellIds().CopyTo(cellsIds, 0);

			for (int cellIndex = 0; cellIndex < cellsIds.Length; cellIndex++)
			{
				attributesIds = new string[mesh.GetCellAttrsIds(cellsIds[cellIndex]).Count];
				mesh.GetCellAttrsIds(cellsIds[cellIndex]).CopyTo(attributesIds, 0);

				for (int index = 0; index < attributesIds.Length; index++)
				{
					if (this.flags[1] && (mesh.GetCellAttrs(cellsIds[cellIndex], attributesIds[index]).ElemType == typeof(Zcu.Mve.Visualization.Normal3D)))
					{
						this.cellNormalsArrayList.Add(mesh.GetCellAttrs(cellsIds[cellIndex], attributesIds[index]));
					}
					else if (this.flags[2] && (mesh.GetCellAttrs(cellsIds[cellIndex], attributesIds[index]).ElemType == typeof(Zcu.Mve.Visualization.TriangleNeighbor)))
					{
						this.cellNeighborsArrayList.Add(mesh.GetCellAttrs(cellsIds[cellIndex], attributesIds[index]));
					}
				}
			}
		}

		/// <summary>
		/// Writes complete data to output file.
		/// </summary>
		/// <param name="nameOfFile">Name of output file.</param>
		private void WriteDataToFile(string nameOfFile)
		{
			Zcu.Mve.Visualization.Point3D vertex;
			Zcu.Mve.Visualization.Triangle triangle;
			Zcu.Mve.Visualization.TriangleNeighbor neighbor;
			Zcu.Mve.Visualization.Normal3D normal3d;
			
			StreamWriter streamWriter = new StreamWriter(nameOfFile);

			System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();

			nfi.CurrencyDecimalSeparator = ".";

			streamWriter.WriteLine("# File generated by MVE2 module MeshSaver");
			streamWriter.WriteLine("# Number of triangles : " + this.numberOfCells);
			streamWriter.WriteLine("# Number of vertices : " + this.numberOfVertices);
			streamWriter.WriteLine();                      // Writing of head.


			streamWriter.WriteLine("\n[Vertices]\n");

			for (int index = 0; index < this.numberOfVertices; index++)
			{                                              // Writes vertices.
				vertex = (Zcu.Mve.Visualization.Point3D)this.verticesArray[index];	
				streamWriter.WriteLine("  {0} {1} {2}", vertex[0].ToString("0.######0.0", nfi), vertex[1].ToString("0.######0.0", nfi), vertex[2].ToString("0.######0.0", nfi));
			}

			if ((this.cellsArrays != null) && (this.cellsArrays.Length != 0) && this.flags[0])
			{                                              // Writes triangles.
				streamWriter.WriteLine("\n[Triangles]\n");

				for (int arrayIndex = 0; arrayIndex < this.cellsArrays.Length; arrayIndex++)
				{
					for (int index = 0; index < this.cellsArrays[arrayIndex].Count; index++)
					{
						triangle = (Zcu.Mve.Visualization.Triangle)this.cellsArrays[arrayIndex][index];	
						streamWriter.WriteLine("  {0} {1} {2}", triangle.V1, triangle.V2, triangle.V3);
					}
				}
			}

			if ((this.cellNormalsArrayList.Count != 0) && this.flags[1])
			{                                              // Writes triangle normals.
				streamWriter.WriteLine("\n[Triangles' Normals]\n");

				Zcu.Mve.Visualization.UniformDataArray normals;
				for (int arrayList = 0; arrayList < this.cellNormalsArrayList.Count; arrayList++)
				{
					normals = (Zcu.Mve.Visualization.UniformDataArray)this.cellNormalsArrayList[arrayList];
					for (int index = 0; index < normals.Count; index++)
					{
						normal3d = (Zcu.Mve.Visualization.Normal3D)normals[index];	
						streamWriter.WriteLine("  {0} {1} {2}", normal3d[0].ToString("0.######0.0", nfi), normal3d[1].ToString("0.######0.0", nfi), normal3d[2].ToString("0.######0.0", nfi));
					}
				}
			}

			if ((this.cellNeighborsArrayList.Count != 0) && this.flags[2])
			{                                              // Writes triangle neighbours.
				streamWriter.WriteLine("\n[Triangles' Neighbours]\n");

				Zcu.Mve.Visualization.UniformDataArray neighbors;
				for (int arrayList = 0; arrayList < this.cellNeighborsArrayList.Count; arrayList++)
				{
					neighbors = (Zcu.Mve.Visualization.UniformDataArray)this.cellNeighborsArrayList[arrayList];
					for (int index = 0; index < neighbors.Count; index++)
					{
						neighbor = (Zcu.Mve.Visualization.TriangleNeighbor)neighbors[index];	
						streamWriter.WriteLine("  {0} {1} {2}", neighbor.Neighbors[0], neighbor.Neighbors[1], neighbor.Neighbors[2]);
					}
				}
			}

			if ((this.verticesNormalsArrayList.Count != 0) && this.flags[3])
			{                                              // Writes vertex normals.
				streamWriter.WriteLine("\n[Vertices' Normals]\n");

				Zcu.Mve.Visualization.UniformDataArray normals;
				for (int arrayList = 0; arrayList < this.verticesNormalsArrayList.Count; arrayList++)
				{
					normals = (Zcu.Mve.Visualization.UniformDataArray)this.verticesNormalsArrayList[arrayList];
					for (int index = 0; index < normals.Count; index++)
					{
						normal3d = (Zcu.Mve.Visualization.Normal3D)normals[index];	
						streamWriter.WriteLine("  {0} {1} {2}", normal3d[0].ToString("0.######0.0", nfi), normal3d[1].ToString("0.######0.0", nfi), normal3d[2].ToString("0.######0.0", nfi));
					}
				}
			}

			streamWriter.Close();
		}
	}
}
