using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

using Zcu.Mve.Core;
using Zcu.Mve.Visualization;


namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Module makes TriangleMesh more complex using PN Triangles.
    /// </summary>
    [ModuleInfo("Lubo Kolk",
    "Module makes TriangleMesh more complex using PN Triangles.", IconName = "MeshSmooth.ico")]
    [PortInfo("InMesh", "Input triangle mesh")]
    [PortInfo("OutMesh", "Input triangle mesh")] 
   
    public class MeshSmooth : Module
    {
        /// <summary>
        /// Enum for select of normal interpolation method
        /// </summary>
        public enum ENormalInterpolation {
            /// <summary>
            /// Simply linear interpolation is performed
            /// </summary>
            Linear,
            /// <summary>
            /// Like linear interpolation, but in some cases mirrors result
            /// </summary>
            Quadratic,
            /// <summary>
            /// Theoretically more accurate and more complex method of normal interpolation
            /// </summary>
            Local
        };
        #region Variables
        bool paramsChanged = false;
        TriangleMesh output = null;
        int iterations = 1;
        private ENormalInterpolation normalInterpolation = ENormalInterpolation.Linear;
        private double strength = 1.0;
        private String normalAttributteName = "Normal3D";
        #endregion Variables
        #region Properties
        /// <summary>
        /// Number of iterations
        /// </summary>
        public int Iterations { get { return iterations; } set { iterations = value; paramsChanged = true; } }
        /// <summary>
        /// Strength of smoothing. Selecting of bigger strength will result in more rounded mesh
        /// </summary>
        public double Strength { get { return strength; } set { strength = value; paramsChanged = true; } }
        /// <summary>
        /// Normal interpolation method
        /// </summary>
        public ENormalInterpolation NormalInterpolation
        {
            get { return normalInterpolation; }
            set
            {
                normalInterpolation = value;
                paramsChanged = true;
            }
        }
        /// <summary>
        /// Name of normals attribute
        /// </summary>
        public String NormalAttributeName { get { return normalAttributteName; } set { normalAttributteName = value; paramsChanged = true; } }
        #endregion Properies
        #region Overriden methods

        /// <summary>
        /// Default constructor.
        /// </summary>
        public MeshSmooth()
        {
            AddInPort("InMesh", typeof(TriangleMesh));
            AddOutPort("OutMesh", typeof(TriangleMesh));
        }
        /// <summary>
        /// Adds new triangles to the mesh, so mesh looks smoother
        /// </summary>
        public override void Execute()
        {
            if ((paramsChanged) || (!IsAllDataSame()) || (output == null))
            {
                paramsChanged = false;
                //stazeni dat ze vstupu
                TriangleMesh mesh = (TriangleMesh)GetInput("InMesh");
                output = (TriangleMesh)mesh.ShallowCopy();
                UniformDataArray pointsArray = mesh.Points;
                String[] celIds = mesh.GetCellIds(typeof(Triangle));
                if (celIds.Length == 0)
                    throw new MveException(this, "No triangle cells in mesh");
                UniformDataArray trianglesArray = output.GetCells(celIds[0]);
                UniformDataArray normalsArray = output.GetPointAttrs(normalAttributteName);

                List<Point3D> points = new List<Point3D>(pointsArray.Count * 2);
                List<Normal3D> normals = new List<Normal3D>(pointsArray.Count * 2);
                List<Triangle> triangles = new List<Triangle>(trianglesArray.Count * 4);

                for (int i = 0; i < pointsArray.Count; i++)
                {
                    points.Add((Point3D)pointsArray[i]);
                    normals.Add((Normal3D)normalsArray[i]);
                }
                for (int i = 0; i < trianglesArray.Count; i++)
                    triangles.Add((Triangle)trianglesArray[i]);

                //vlastni zjemneni meshe
                for (int iteration = 0; iteration < iterations; iteration++)
                {
                    triangles = Smooth(points, normals, triangles, iteration);
                }
                //pripraveni dat pro vystup
                pointsArray = new UniformDataArray(typeof(Point3D), points.Count);
                normalsArray = new UniformDataArray(typeof(Normal3D), normals.Count);
                trianglesArray = new UniformDataArray(typeof(Triangle), triangles.Count);
                for (int i = 0; i < points.Count; i++)
                {
                    pointsArray[i] = points[i];
                    normalsArray[i] = normals[i];
                }
                for (int i = 0; i < triangles.Count; i++)
                    trianglesArray[i] = triangles[i];
                output.Points = pointsArray;
                output.RemovePointAttrs(normalAttributteName);
                output.AddPointAttrs(normalAttributteName, normalsArray);
                output.RemoveCells(celIds[0]);
                output.AddCells(celIds[0], trianglesArray);
                output.CheckConsistence();
                SetOutput("OutMesh", output, false);
            }
            else
            {
                SetOutput("OutMesh", output, true);
            }
        }
        #endregion Overriden methods
        #region Other methods
        /// <summary>
        /// This method creates new triangles and points
        /// </summary>
        private List<Triangle> Smooth(List<Point3D> points, List<Normal3D> normals, List<Triangle> triangles, int stage)
        {
            //tabulka otevrenych hran. Kazda hrana by mela byt obsazena jen ve dvou trojuhelnikach,
            //takze pokud je nalezen druhy takovy trojuhelni uz se dale neprohledava
            Hashtable openEdges = new Hashtable();
            List<Triangle> newTris = new List<Triangle>(triangles.Count * 4);

            int []newPoints = new int[3];
            Line[] tempEdges = new Line[3];
                        
            //vytvoreni novych trojuhelniku
            int progress = 0;
            foreach(Triangle t in triangles)
            {
                tempEdges[0] = new Line(t.V1, t.V2, false);
                tempEdges[1] = new Line(t.V2, t.V3, false);
                tempEdges[2] = new Line(t.V3, t.V1, false);
                for (int i = 0; i < 3; i++)
                {
                    //pokud je hrana mezi otevrenymi -> novy bod je uz vypocitan,
                    //pouzije se a hrana se odstrani
                    if (openEdges.Contains(tempEdges[i]))
                    {
                        newPoints[i] = (int)openEdges[tempEdges[i]];
                        openEdges.Remove(tempEdges[i]);
                    }
                    else
                    {
                        //jinak je vypocitan novy bod a hrana je zarazena mezi otevrene
                        newPoints[i] = points.Count;
                        openEdges.Add(tempEdges[i], points.Count);
                        CreateNewPoint(tempEdges[i].V1, tempEdges[i].V2, points, normals);
                    }
                }
                //vytvoreni novych trojuhelniku
                newTris.Add(new Triangle(t.V1, newPoints[0], newPoints[2]));
                newTris.Add(new Triangle(t.V2, newPoints[1], newPoints[0]));
                newTris.Add(new Triangle(t.V3, newPoints[2], newPoints[1]));
                newTris.Add(new Triangle(newPoints[0], newPoints[1], newPoints[2]));

                ProgressInfo((double)progress / (triangles.Count * iterations) + (double)stage / iterations);
                progress++;
            }


            return newTris;

        }
        private void CreateNewPoint(int indexA, int indexB, List<Point3D> points, List<Normal3D> normals)
        {
            //pozice bodu
            Point3D p1 = points[indexA];
            Point3D p2 = points[indexB];

            Point3D n1 = new Point3D(normals[indexA].Nx,
                                     normals[indexA].Ny,
                                     normals[indexA].Nz);

            Point3D n2 = new Point3D(normals[indexB].Nx,
                                     normals[indexB].Ny,
                                     normals[indexB].Nz);

            Point3D[] controlPoints = new Point3D[4];
            controlPoints[0] = p1;
            controlPoints[3] = p2;

            double w;
            Point3D q;

            w = n1.DotProduct(p2 - p1);
            q = 2 * p1 + p2;
            Point3D r = (q - w * n1) / 3.0f;
            r -= p1;
            r *= strength;
            controlPoints[1] = r + p1;

            w = n2.DotProduct(p1 - p2);
            q = 2 * p2 + p1;

            r = (q - w * n2) / 3.0f;
            r -= p2;
            r *= strength;
            controlPoints[2] = r + p2;

            Point3D newPoint = ComputeCurvePoint(0.5f, controlPoints);
            points.Add(newPoint);

            Point3D normal = new Point3D();
            //normala bodu
            switch (normalInterpolation)
            {
                case ENormalInterpolation.Quadratic:
                
                    Point3D vec = p2 - p1;
                    Point3D sum = n1 + n2;
                    double v = 2 * vec.DotProduct(sum) / vec.DotProduct(vec);
                    normal = sum - v * vec;
                    break;
                
                case ENormalInterpolation.Linear:
                    normal = n1 + n2;
                    break;
                case ENormalInterpolation.Local:
                    normal = ComputeNormal(indexA, indexB, points, normals, controlPoints);
                    break;
                
            }
            Normal3D newNormal = new Normal3D(normal.X, normal.Y, normal.Z);
            newNormal.Normalize();
            normals.Add(newNormal);
        }
        private Point3D ComputeNormal(int indexA, int indexB, List<Point3D> points, List<Normal3D> normals, Point3D[] controlPoints)
        {
            Normal3D n1 = normals[indexA];
            Normal3D n2 = normals[indexB];
            Point3D v1 = new Point3D((float)n1.Nx, (float)n1.Ny, (float)n1.Nz);
            Point3D v2 = new Point3D((float)n2.Nx, (float)n2.Ny, (float)n2.Nz);
            Point3D v3 = points[indexB] - points[indexA];
            v3 /= v3.Abs();
            Point3D axis;
            double[] dots = new double[3];
            dots[0] = Math.Abs(v1.DotProduct(v2));
            dots[1] = Math.Abs(v2.DotProduct(v3));
            dots[2] = Math.Abs(v3.DotProduct(v1));
            if (dots[0] > dots[1])
            {
                if (dots[1] < dots[2])
                    axis = v2.CrossProduct(v3);
                else axis = v3.CrossProduct(v1);
            }
            else
            {
                if (dots[0] < dots[2])
                    axis = v1.CrossProduct(v2);
                else axis = v3.CrossProduct(v1);
            }
            Point3D tangent = ComputeCurveDerivation(0.5f, controlPoints);
            Point3D newNormal = axis.CrossProduct(tangent);
            newNormal /= newNormal.Abs();

            v1 = new Point3D(n1.Nx, n1.Ny, n1.Nz);
            v2 = new Point3D(n2.Nx, n2.Ny, n2.Nz);        

            dots[0] = newNormal.DotProduct(v1);
            dots[1] = newNormal.DotProduct(v2);

            if (Math.Abs(dots[0]) < Math.Abs(dots[1]))
            {
                if (dots[1] < 0)
                    newNormal = -1 * newNormal;
            }
            else if (dots[0] < 0)
                newNormal = -1 * newNormal;
            return newNormal;
        }
        private Point3D ComputeCurvePoint(float t, Point3D[] controlPoints)
        {
            //vypocte bod podle zadaneho parametru
            float a = 1 - t;
            float s = a * a * a;
            float u = 3 * t * a * a;
            float v = 3 * t * t * a;
            float w = t * t * t;
            return s * controlPoints[0] + u * controlPoints[1] +
                   v * controlPoints[2] + w * controlPoints[3];
            
        }
        private Point3D ComputeCurveDerivation(float t, Point3D[] controlPoints)
        {
            Point3D res = t * t *
                        (-3 * controlPoints[0] +
                         9 * controlPoints[1] -
                         9 * controlPoints[2] +
                         3 * controlPoints[3]);
            res += t *
                   (6 * controlPoints[0] -
                   12 * controlPoints[1] +
                   6 * controlPoints[2]);
            res += -3 * controlPoints[0] + 3 * controlPoints[1];
            return res;
        }
        #endregion Other methods
    }
}
