using System;
using System.Xml;
using Zcu.Mve.Core;
using System.Collections;
using System.Diagnostics;

namespace Zcu.Mve.Visualization
{
  /// <summary>
  /// Static multidimensional homogenous data array.
  /// </summary>
  [DataObject("Static multidimensional homogenous data array.")]
  public class MveArray : Zcu.Mve.Core.DataObject, IEnumerable
  {
    /// <summary>
    /// Array of data objects.
    /// </summary>
    protected Array array;
    /// <summary>
    /// Returns the reference to array.
    /// </summary>
    public Array Array
    {
        get { return array; }
    }
    

    private Type type;
    private int[] lengths;

    /// <summary>
    /// Called before an object is assigned to specified index in array.
    /// This operation can be canceled via return value.
    /// <ul>
    /// <li>true: allow assignement</li>
    /// <li>false: forbid assignement</li>
    /// </ul>
    /// </summary>
    public delegate bool ElementSetCallBack(object element, params int[] indices);

    /// <summary>
    /// Called when an object is assigned to specified index in array
    /// </summary>
    public event ElementSetCallBack OnElementSetEvent;

    /// <summary>
    /// For XmlLoader use only. Use parametric constructor instead.
    /// </summary>
    public MveArray()
    {
      UpdateModificationTime();
    }

    /// <summary>
    /// Creates abstract array of specified type
    /// </summary>
    /// <param name="type">Type of object (only this type is accepted). Must be inherited from Mve.Core.DataObject</param>
    /// <param name="lenghts">An array of 32-bit integers that represent the size of each dimension of the Array to create.</param>
    public MveArray(Type type, params int[] lenghts)
    {
      if (!typeof(IDataObject).IsAssignableFrom(type))
      {
        throw new MveException(this, "Only IDataObject can be element of MveArray.");
      }
      this.type = type;
      this.AddLibName(Globals.GetLibraryName(type));
      this.lengths = (int[]) lenghts.Clone();

      array = Array.CreateInstance(type, lenghts);

      UpdateModificationTime();
    }

    /// <summary>
    /// Called before an object is assigned to specified index in array.
    /// This operation can be canceled via return value.
    /// </summary>
    /// <param name="indices">Index of element</param>
    /// <param name="element">value to be assigned to the index position</param>
    /// <returns>true - when element assignement is allowed <br></br>
    ///          false - when element assignement is not allowed</returns>
    protected bool InvokeElementSetEvent(object element, params int[] indices)
    {
      if (OnElementSetEvent != null)
        return OnElementSetEvent(element, indices);
      else
        return true;
    }

    /// <summary>
    /// Returns type accepted by curent array.
    /// </summary>
    public Type ElemType
    {
      get
      {
        return type;
      }
    }

    /// <summary>
    /// Get/Set element of array. ElemetnSetEvent is called and SetValue can be forbidden.
    /// </summary>
    public IDataObject this[params int[] indices]
    {
      get
      {
        return (IDataObject) array.GetValue(indices);
      }
      set
      {
        if (!this.ElemType.IsAssignableFrom(value.GetType()))
        {
          throw new MveException(this, String.Format("Cannot assign value of type \"{0}\" because MveArray contains type \"{1}\".", value.GetType().FullName, this.ElemType.FullName));
        }
        if (this.InvokeElementSetEvent(value, indices))
        {
          array.SetValue(value, indices);
        }

        UpdateModificationTime();
      }
    }

    /// <summary>
    /// Get/Set type of elemnt. For XmlLoader use only.
    /// </summary>
    protected Type Type
    {
      set
      {
        this.type = value;

        UpdateModificationTime();
      }
      get
      {
        return this.type;
      }
    }

    /// <summary>
    /// Returns size in all dimensions.
    /// </summary>
    public int[] Lengths
    {
      get
      {
        return (int[]) this.lengths.Clone();
      }
    }

    /// <summary>
    /// Gets the number of dimensions of array.
    /// </summary>
    public int Rank
    {
      get
      {
        return this.lengths.Length;
      }
    }

    /// <summary>
    /// Get array enumerator.
    /// </summary>
    /// <returns>Array enumerator.</returns>
    public IEnumerator GetEnumerator()
    {
      return array.GetEnumerator();
    }

    /// <summary>
    /// Recursive copy of multidimensional array.
    /// </summary>
    /// <param name="target"></param>
    /// <param name="dim"></param>
    /// <param name="indices"></param>
    private void CloneDimension(Array target, int dim, int[] indices)
    {
      if (dim == 1)
      {
        for (indices[this.Rank - dim] = 0; indices[this.Rank - dim] < target.GetLength(this.Rank - dim); indices[this.Rank - dim]++)
        {
          target.SetValue(((IDataObject) this.array.GetValue(indices)).DeepCopy(), indices);
        }
      }
      else
      {
        for (indices[this.Rank - dim] = 0; indices[this.Rank - dim] < target.GetLength(this.Rank - dim); indices[this.Rank - dim]++)
        {
          CloneDimension(target, dim - 1, indices);
        }
      }
    }

    /// <summary>
    /// Deep copy of data object and its subelements.
    /// </summary>
    /// <returns>same data, different memmory space.</returns>
    public override IDataObject DeepCopy()
    {
      //      MveArray mveArray = new MveArray(this.ElemType, lengths);
      object[] args = new object[this.Rank + 1];
      args[0] = this.ElemType;
      for (int i = 0; i < this.Rank; i++)
      {
        args[i + 1] = this.lengths[i];
      }
      MveArray mveArray = (MveArray) Activator.CreateInstance(this.GetType(), args);
      mveArray.array = (Array) this.array.Clone();

      // Elements are DataObjects ... lets make deep copy.
        //TODO: LVA: this is highly suspicious!!!
      if (typeof(Zcu.Mve.Core.DataObject).IsInstanceOfType(this.ElemType))
      {
        CloneDimension(mveArray.array, this.Rank, new int[this.Rank]);
      }

      return mveArray;
    }

    /// <summary>
    /// Read data from Xml file to current structure.
    /// </summary>
    /// <param name="xmlTextReader">Stream to be read.</param>
    public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
    {
      int[] indices = null;

      while (xmlTextReader.Read())
      {
        if (xmlTextReader.NodeType == XmlNodeType.Element)
        {
          switch (xmlTextReader.Name)
          {
            // --------- Header ------------
            case "mvearray":
              {
                DecodeHeader(xmlTextReader);

                indices = new int[this.Rank];
                for (int i = 0; i < this.Rank; i++)
                {
                  indices[i] = 0;
                }
                break;
              }

            // --------- Element ---------
            case "i":
              {
                string isnull = xmlTextReader.GetAttribute("isnull");
                if (isnull == null || isnull != "true")
                { /*
                if (this.ElemType.IsValueType)
                {
                  ((IDataObject)(this[indices])).ReadData(xmlTextReader);
                }
                else*/
                  {
                    IDataObject elem = (IDataObject) Activator.CreateInstance(this.ElemType);
                    elem.ReadData(xmlTextReader);
                    array.SetValue(elem, indices);
                  }
                }
                indices[this.Rank - 1] += 1;
                break;
              }

            // --------- List ------------
            case "list":
              {
                int dim = Int32.Parse(xmlTextReader.GetAttribute("dim").Trim());
                if (dim == 1)
                {
                  string indexStr = xmlTextReader.GetAttribute("index");
                  int[] newindices = ParseInt32List(indexStr);
                  if (newindices.Length != this.Rank - 1)
                  {
                    throw new MveException(this, "Wrong index rank in MveArray: " + indexStr);
                  }
                  for (int i = 0; i < this.Rank - 1; i++)
                  {
                    if (newindices[i] >= this.Lengths[i])
                    {
                      throw new MveException(this, "Index " + i + " exceeded range in MveArray: " + indexStr);
                    }
                    indices[i] = newindices[i];
                  }
                  indices[this.Rank - 1] = 0;
                }
                break;
              }

            default:
              throw new MveException(this, "Uknown element in MveArray: " + xmlTextReader.Name);
          } // switch
        } // Element

        if (xmlTextReader.NodeType == XmlNodeType.EndElement)
        {
          if (xmlTextReader.Name == "mvearray")
          {
            break;
          }
        } // EndElement
      } // while
      Debug.WriteLine("MVE Array - data read.");

      UpdateModificationTime();
    }

    /// <summary>
    /// Parse array of integer values separated by comma.
    /// </summary>
    /// <param name="intList"></param>
    /// <returns></returns>
    private int[] ParseInt32List(string intList)
    {
      string[] elems = intList.Trim().Split(',');
      int[] array = new int[elems.Length];

      for (int i = 0; i < elems.Length; i++)
      {
        array[i] = Int32.Parse(elems[i].Trim());
      }

      return array;
    }

    /// <summary>
    /// Read header and set member variables (allocate array).
    /// </summary>
    /// <param name="reader"></param>
    private void DecodeHeader(XmlTextReader reader)
    {
      string elemType = null;
      int rank = -1;

      elemType = reader.GetAttribute("elemtype").Trim();

      this.type = Globals.FindType(this.LibNames, elemType);
      if (!typeof(IDataObject).IsAssignableFrom(type))
      {
        throw new MveException(this, "Only IDataObject can be element of MveArray.");
      }
      //this.AddLibName(Globals.GetLibraryName(type));

      rank = Int32.Parse(reader.GetAttribute("rank").Trim());
      this.lengths = ParseInt32List(reader.GetAttribute("lengths"));

      if (this.Lengths.Length != rank)
      {
        throw new MveException(this, "Rank of MveArray do not agree with number of lengths value!");
      }

      array = Array.CreateInstance(type, this.Lengths);
    }

    /// <summary>
    /// Write current data to Xml file.
    /// </summary>
    /// <param name="xmlTextWriter">Stream to be read.</param>
    public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
    {
      xmlTextWriter.WriteStartElement("mvearray");
      xmlTextWriter.WriteAttributeString("elemtype", this.ElemType.FullName);
      xmlTextWriter.WriteAttributeString("rank", this.Rank.ToString(Globals.nfi));
      {
        string str = this.Lengths[0].ToString(Globals.nfi);
        for (int i = 1; i < this.Rank; i++)
        {
          str += ", " + this.Lengths[i].ToString(Globals.nfi);
        }
        xmlTextWriter.WriteAttributeString("lengths", str);
      }

      WriteDimension(xmlTextWriter, this.Rank, this.Lengths);

      xmlTextWriter.WriteEndElement();
    }

    /// <summary>
    /// Write given dimension into Xml stream.
    /// </summary>
    /// <param name="writer">Xml stream</param>
    /// <param name="dim">Dimension to be written</param>
    /// <param name="indices">Indices to be written</param>
    private void WriteDimension(XmlTextWriter writer, int dim, int[] indices)
    {
      // TODO: Osetrit vystup elementu s null hodnotout!
      if (dim < this.Rank)
      {
        writer.WriteStartElement("list");
        writer.WriteAttributeString("dim", dim.ToString(Globals.nfi));

        string str = "";

        for (int i = this.Rank - 1; i > dim; i--)
        {
          str += indices[this.Rank - i - 1].ToString(Globals.nfi) + ", ";
        }
        str += indices[this.Rank - dim - 1].ToString(Globals.nfi);

        writer.WriteAttributeString("index", str);
      }

      if (dim == 1)
      {
        for (indices[this.Rank - dim] = 0; indices[this.Rank - dim] < this.Lengths[this.Rank - dim]; indices[this.Rank - dim]++)
        {
          writer.WriteStartElement("i");
          ((IDataObject) this.array.GetValue(indices)).WriteData(writer);
          writer.WriteEndElement();
        }
      }
      else
      {
        for (indices[this.Rank - dim] = 0; indices[this.Rank - dim] < this.Lengths[this.Rank - dim]; indices[this.Rank - dim]++)
        {
          WriteDimension(writer, dim - 1, indices);
        }
      }
      if (dim < this.Rank)
      {
        writer.WriteEndElement();
      }
    }

    /// <summary>
    /// Check the consistency of current data structure.
    /// Can (and should) write remarks on Console and/or Debug or Trace console.
    /// </summary>
    /// <returns>true if structure is consistent</returns>
    public override bool CheckConsistence()
    {
      // TODO:  Add MveArray.CheckConsistence implementation
      return true;
    }

    /// <summary>
    /// Finds an element in the array and returns its index
    /// </summary>
    /// <param name="o">element to be found</param>
    /// <returns>index of the given element, or -1 if not found.</returns>
    public int IndexOf(IDataObject o)
    {
      return (Array.IndexOf(array, o));
    }

  } // class MveArray
} // namespace
