using System;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.Collections;
using System.Diagnostics;

namespace Zcu.Mve.Visualization.Processing
{
	/// <summary>
	/// Summary description for NeighborTetrahedrons.
	/// </summary>
	public class NeighborTetrahedra:Module
	{
		string inputCells = "tetrahedra";
		bool paramsChanged = true;
		UnstrGrid output = null;

		private UniformDataArray tetrahedra, neighbors;
		private Hashtable triangles;

		public string InputCells
		{
			get {return(inputCells);}
			set
			{
				if (value!="")
				{
					inputCells = value;
					paramsChanged = true;
				}
				else
				{
					throw new MveException(this, "Input cell identifier cannot be an empty string.");
				}
			}
		}

		public NeighborTetrahedra()
		{
			AddInPort("Input UnstrGrid",typeof(UnstrGrid));
			AddOutPort("Output UnstrGrid", typeof(UnstrGrid));
		}

		private struct MyTriangle
		{
			public int v1,v2,v3; //vertices
			

			public MyTriangle(int v1, int v2, int v3)
			{
				int tmp;
				if (v3<v2)
				{
					tmp = v2;
					v2=v3;
					v3=tmp;
				}
				if (v2<v1)
				{
					tmp = v1;
					v1 = v2;
					v2 = tmp;
				}
				if (v3<v2)
				{
					tmp = v2;
					v2=v3;
					v3=tmp;
				}
				this.v1 = v1;
				this.v2 = v2;
				this.v3 = v3;
			}

			public override bool Equals(object obj)
			{
				if (obj is MyTriangle)
				{
					MyTriangle mt = (MyTriangle)obj;
					if ((this.v1 == mt.v1)&&(this.v2 == mt.v2)&&(this.v3 == mt.v3))
						return(true);
					else return(false);
				}
				else
				{
					return(false);
				}													
			}	
		
			public override int GetHashCode()
			{
				return(v1^v2^v3);
			}
		}

		private void processTetrahedronFace(int i, int v1, int v2, int v3)
		{
			Tetrahedron t = (Tetrahedron)tetrahedra[i];     //take a tetrahedron
			MyTriangle mt = new MyTriangle(t[v1],t[v2],t[v3]); //take one of it's border triangles
			object nbr = triangles[mt];						//check the hashtable
			if (nbr == null)								//is it in the hashtable?
			{
				triangles.Add(mt,i);						//it's not there, add it, note index if incident tetrahedron 
			}
			else
			{
				int neighborIndex = (int)nbr;				//it's there, fetch index of incident tetrahedron
				Tetrahedron neighborTetrahedron = (Tetrahedron)tetrahedra[neighborIndex]; //fetch incident tetrahedron
				Tetrahedron nbrs = (Tetrahedron)neighbors[neighborIndex]; // fetch neighbors of incident tetrahedron
				
				// find index of the neighbor within incident tetrahedron
				// it's the one that is not equal to any of the triangle vetrices
				int j = 0;
				while ((mt.v1 == neighborTetrahedron[j])||(mt.v2 == neighborTetrahedron[j])||(mt.v3 == neighborTetrahedron[j]))
					j++;
				nbrs[j] = i; //set the neighbor to index of incident tetrahedron

				
				nbrs = (Tetrahedron)neighbors[i]; // fetch neighbors of this tetrahedron

				// find index of the neighbor within this tetrahedron
				// it's the one that is not equal to any of the triangle vetrices
				j = 0;
				while ((mt.v1 == t[j])||(mt.v2 == t[j])||(mt.v3 == t[j]))
					j++;
				nbrs[j] = neighborIndex; // set the neighbor index of this tetrahedron
				triangles.Remove(mt);
			}
		}

		public override void Execute()
		{
			if ((paramsChanged)||(!IsDataSame("Input UnstrGrid"))||(output == null))
			{
				UnstrGrid input = (UnstrGrid)GetInput("Input UnstrGrid");

				output = input.ShallowCopy();

				tetrahedra = output.GetCells(inputCells);

				if (tetrahedra.ElemType!=typeof(Tetrahedron))
				{
					throw new MveException(this, "Input cells must be tetrahedra.");
				}

				neighbors = new UniformDataArray(typeof(Tetrahedron),tetrahedra.Count);
				for (int i = 0;i<neighbors.Count;i++)
				{
					Tetrahedron t = (Tetrahedron)tetrahedra[i];
					t.V1 = -1;
					t.V2 = -1;
					t.V3 = -1;
					t.V4 = -1;
				}

				triangles = new Hashtable(tetrahedra.Count*2);

				for (int i = 0;i<tetrahedra.Count;i++)
				{
					processTetrahedronFace(i,1,2,3);
					processTetrahedronFace(i,0,2,3);
					processTetrahedronFace(i,0,1,3);
					processTetrahedronFace(i,0,1,2);
					ProgressInfo((float)i/tetrahedra.Count);
				}

				Debug.WriteLine("Number of border triangles:"+triangles.Count);

				output.AddCellAttrs(inputCells,inputCells+" neighbors",neighbors);

				triangles = null;

				SetOutput("Output UnstrGrid",output);
				paramsChanged = false;
			}
			else
			{
				SetOutput("Output UnstrGrid",output,true);
			}
		}

	}
}
