using System;
using System.ComponentModel;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;


namespace Zcu.Mve.Visualization.Processing
{
  /// <summary>
  /// Summary description for NormalComputer.
  /// </summary>  
  [ModuleInfo("Libor Va", "Module that computes vertex normals for a TriangleMesh. Normals are placed in a new point attribute of name specified in the AttrName property.", IconName = "NormalComputer.ico")]
  [PortInfo("TriangleMesh", "Input triangle mesh to be processed")]
  [PortInfo("Computed TriangleMesh", "Shallow copy of the input mesh with added attribute that represents vertex normals.")]
  public class NormalComputer : Module
  {
    private TriangleMesh output = null;
    
    private bool paramsChanged = true;

    private string attrName = "Normal3D";

    /// <summary>
    /// Gets or sets the attribute name that will contain the normals. Cannot be set to an empty string.
    /// </summary>
    [Description("Gets or sets the attribute name that will contain the normals. Cannot be set to an empty string.")]
    public string AttrName
    {
      get
      {
        return (attrName);
      }
      set
      {
        if (value != "")
        {
          attrName = value;
          paramsChanged = true;
        }
        else
        {
          throw new MveException(this, "Attribute name for normals cannot be an empty strng.");
        }
      }
    }

    private bool invertNormals = false;
    [Description("Use this option if you are importing files originally made for right-handed coordinate systems."), DefaultValue(false)]
    public bool InvertNormals
    {
      get
      {
        return invertNormals;
      }
      set
      {
        invertNormals = value;
        paramsChanged = true;
      }
    }

    /// <summary>
    /// Default constructor.
    /// </summary>
    public NormalComputer()
    {
      AddInPort("TriangleMesh", typeof(TriangleMesh));
      AddOutPort("Computed TriangleMesh", typeof(TriangleMesh));
    }

    /// <summary>
    /// Computes the normals of the mesh. If the mesh is unchanged, and no properties were changed since the last run, then the result of the last run is returned with dataSame set to true.
    /// </summary>
    public override void Execute()
    {
      if ((paramsChanged) || (!IsAllDataSame()) || (output == null))
      {
        TriangleMesh input = (TriangleMesh) GetInput("TriangleMesh");


        // LVA 27.11.2008 - I don't understand this thing, however it screws up my stuff, so I remove it
        /*if ((output != null) && (!paramsChanged) && (output.Points.ModificationTime == input.Points.ModificationTime))
        {
          if (output.GetCellIds().Count == input.GetCellIds().Count)
          {
            foreach (String s in output.GetCellIds())
            {
              if (input.GetCellIds().Contains(s))
              {
                if (((UniformDataArray) output.GetCells(s)).ModificationTime
                   ==
                   ((UniformDataArray) input.GetCells(s)).ModificationTime)
                {
                  SetOutput("Computed TriangleMesh", output, false);
                  return;
                }
              }
            }
          }
        }*/

        UniformDataArray triangles = (input.GetCells(typeof(Triangle)))[0];

        UniformDataArray normals = new UniformDataArray(typeof(Normal3D), input.Points.Count);
        for (int i = 0; i < normals.Count; i++)
          normals[i] = new Normal3D(0, 0, 0);

        int pos = 0;
        foreach (Triangle t in triangles)
        {
          Point3D v1 = (Point3D) input.Points[t.V2] - (Point3D) input.Points[t.V1];
          Point3D v2 = (Point3D) input.Points[t.V3] - (Point3D) input.Points[t.V1];
          Normal3D normal = (Normal3D) v1.CrossProduct(v2);
          normal.Normalize();
          if (!(double.IsNaN(normal[0])||double.IsNaN(normal[1])||double.IsNaN(normal[2])))
          {
              normals[t.V1] = normal + (Normal3D) normals[t.V1];
              normals[t.V2] = normal + (Normal3D) normals[t.V2];
              normals[t.V3] = normal + (Normal3D) normals[t.V3];
          }
          pos++;
          if ((pos % 1000) == 0)
            ProgressInfo((float) pos / triangles.Count);
        }

        for (int i = 0; i < normals.Count; i++)
        {
          Normal3D normal = (Normal3D) normals[i];          
          if (double.IsNaN(normal[0])||double.IsNaN(normal[1])||double.IsNaN(normal[2]))
          {
              normal = new Normal3D(1, 0, 0);
          }
          normal.Normalize();
          if (invertNormals)
          {
            Normal3D iNormal = new Normal3D(-normal.Nx, -normal.Ny, -normal.Nz);
            
            normals[i] = iNormal;
          }
          else
          {
            normals[i] = normal;
          }
        }
        output = (TriangleMesh) input.ShallowCopy();
        output.AddPointAttrs(attrName, normals);
        SetOutput("Computed TriangleMesh", output);
        paramsChanged = false;
        ProgressInfo(1.0);
      }
      else
      {
        SetOutput("Computed TriangleMesh", output, true);
      }
    }
  }
}
