using System;
using System.Globalization;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Represents a n-dimensional normal defined by N elements.
	/// </summary>
	[DataObject("Represents a n-dimensional normal defined by N elements.")]
	public struct NormalND : INormal
	{
		/// <summary>
		/// n-dimensional array of normal values.
		/// </summary>
		private double [] values;

		/// <summary>
		/// The toleration for equality operator.
		/// </summary>
		public static double Epsilon = Globals.EpsilonDoubleImplicit;

		/// <summary>
		/// Constant for 
		/// The toleration for CheckConsistence() method.
		/// The normal is consistent when its absolute value lies inside the interval: 
		/// 1 - (EPSILON / 2) ... 1 + (EPSILON / 2)
		/// </summary>
		public static double EPSILON = 0.1;

		/// <summary>
		/// Create a NormalND object.
		/// </summary>
		/// <param name="values">n-dimensional array of vector values.</param>
		public NormalND(params double[] values)
		{
			this.values = (double []) values.Clone();
		} // NormalND()

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			string[] dims = xmlTextReader.ReadString().Trim().Split(' ');
			this.values = new double[dims.Length];
			for (int i = 0; i < dims.Length; i++)
				this.values[i] = Double.Parse(dims[i], nfi);
		} // ReadData()

		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			string pom = "";
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			for (int i = 0; i < this.values.Length; i++)
				pom += this.values[i].ToString(nfi) + " ";
			pom = pom.Substring(0, pom.Length - 1);	// odstraneni prebytecne mezery na konci
			xmlTextWriter.WriteString(pom);
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public IDataObject DeepCopy()
		{
			double[] newDims = new double[this.values.Length];
			this.values.CopyTo(newDims, 0);
			return new NormalND(newDims);
		} // DeepCopy()

		/// <summary>
		/// Method check whether absolute value equals 1 (+- NormalND.EPSILON). 
		/// If no, normal is normalized.
		/// </summary>
		public void Normalize()
		{
			double absValue = NormalND.AbsValue(this.values);
			for(int i = 0; i < this.values.Length; i++)
				this.values[i] = this.values[i] / absValue;
		} // Normalize();

		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public bool CheckConsistence()
		{
			double absValue = NormalND.AbsValue(this.values);
			if (absValue < 1 - NormalND.EPSILON || absValue > 1 + NormalND.EPSILON)
			{
				Console.Write("Warning! " + this.ToString() + " was normalized -> ");
				this.Normalize();
				Console.WriteLine(this.ToString());
			}
			return true;
		} // CheckConsistence()

		/// <summary>
		/// Get the n-dimensional array of normal elements.
		/// For acces to concrete normal elements use indexer.
		/// </summary>
		public double[] Values
		{
			get
			{
				return this.values;
			}
		} // Values

		/// <summary>
		/// Get dimension of normal.
		/// </summary>
		public int Length
		{
			get
			{
				return this.values.Length;
			}
		} // Length

		/// <summary>
		/// Get/Set element of normal.
		/// </summary>
		public double this[int index]
		{
			get
			{
				if (index < this.values.Length && index >= 0)
					return this.values[index];
				else
					throw new MveException(this, "Index " + index + " is out of range <0, " + (this.values.Length - 1) + ">");
			}
			set
			{
				if (index < this.values.Length && index >= 0)
					this.values[index] = value;
				else
					throw new MveException(this, "Index " + index + " is out of range <0, " + (this.values.Length - 1) + ">");
			}
		} // this[int]

		/// <summary>
		/// Returns the absolute value of normal.
		/// </summary>
		/// <returns>The absolute value of normal.</returns>
		public double Abs()
		{
			double tmp = 0;
			foreach (double d in this.values)
				tmp += Math.Pow(d, 2);
			return Math.Sqrt(tmp);
		} // Abs()

		/// <summary>
		/// Computes absolute value of general n-dimensional vector of double values.
		/// </summary>
		/// <param name="values">n-dimensional vector of double values.</param>
		/// <returns>Absolute value of the n-dimensional vector.</returns>
		public static double AbsValue(params double[] values)
		{
			double tmp = 0;
			foreach (double d in values)
				tmp += Math.Pow(d, 2);
			return Math.Sqrt(tmp);
		} // AbsValue()

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

		// ===== Methods of the class Object ==============================================
		#region Overrided methods of the class Object

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		public override string ToString()
		{
			string pom = "NormalND (" + this.values.Length + " dimensions): [";
			foreach (double d in values)
				pom += d + "; ";
			pom = pom.Substring(0, pom.Length - 2);  // odstraneni prebytecneho stredniku a mezery na konci
			return pom + "]";
		} // ToString()

		/// <summary>
		/// Determines whether the specified Objects are equal.
		/// </summary>
		/// <param name="a">First Object.</param>
		/// <param name="b">Second Object.</param>
		/// <returns>True if objects are equal. Otherwise, false.</returns>
		public static new bool Equals(object a, object b)
		{
			if (a.GetType() != typeof(NormalND) || b.GetType() != typeof(NormalND))
				return false;

			NormalND x = (NormalND) a;
			NormalND y = (NormalND) b;
			if (x.Length != y.Length)
				return false;

			for (int i = 0; i < x.Length; i++)
				if (Math.Abs(x[i] - y[i]) > NormalND.Epsilon)
					return false;

			return true;
		}
		
		/// <summary>
		/// Determines whether the specified Object is equal to the current Object.
		/// </summary>
		/// <param name="obj">Specified Object to compare with the current Object.</param>
		/// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
		public override bool Equals(object obj)
		{
			return NormalND.Equals(this, obj);
		}
		
		/// <summary>
		/// Returns a hash code for the current Object.
		/// </summary>
		/// <returns>A hash code for the current Object.</returns>
		public override int GetHashCode()
		{
			if (this.Length == 0)
				return 0;

			if (this.Length == 1)
				return this[0].GetHashCode();

			int hashCode = this[0].GetHashCode();
			for (int i = 1; i < this.Length; i++)
				hashCode = hashCode ^ this[i].GetHashCode();

			return hashCode;
		}

		#endregion

		// ===== Conversions ==============================================================
		#region Conversion

		// ===== NormalND -> Normal ===============
		/// <summary>
		/// Explicit conversion NormalND -> Normal2D
		/// </summary>
		/// <param name="n">NormalND</param>
		/// <returns>Normal2D</returns>
		public static explicit operator Normal2D(NormalND n) 
		{
			if (n.values.Length == 1)
				return new Normal2D(n.values[0], 0);
			else
				return new Normal2D(n.values[0], n.values[1]);
		} // NormalND -> Normal2D

		/// <summary>
		/// Explicit conversion NormalND -> Normal3D
		/// </summary>
		/// <param name="n">NormalND</param>
		/// <returns>Normal3D</returns>
		public static explicit operator Normal3D(NormalND n) 
		{
			if (n.values.Length == 1)
				return new Normal3D(n.values[0], 0, 0);
			else if (n.values.Length == 2)
				return new Normal3D(n.values[0], n.values[1], 0);
			else
				return new Normal3D(n.values[0], n.values[1], n.values[2]);
		} // NormalND -> Normal3D

		// ===== NormalND -> Vector ===============
		/// <summary>
		/// Explicit conversion NormalND -> Vector2D
		/// </summary>
		/// <param name="n">NormalND</param>
		/// <returns>Vector2D</returns>
		public static explicit operator Zcu.Mve.Numerics.Vector2D(NormalND n) 
		{
			if (n.values.Length == 1)
				return new Zcu.Mve.Numerics.Vector2D(n.values[0], 0);
			else
				return new Zcu.Mve.Numerics.Vector2D(n.values[0], n.values[1]);
		} // NormalND -> Vector2D

		/// <summary>
		/// Explicit conversion NormalND -> Vector3D
		/// </summary>
		/// <param name="n">NormalND</param>
		/// <returns>Vector3D</returns>
		public static explicit operator Zcu.Mve.Numerics.Vector3D(NormalND n) 
		{
			if (n.values.Length == 1)
				return new Zcu.Mve.Numerics.Vector3D(n.values[0], 0, 0);
			else if (n.values.Length == 2)
				return new Zcu.Mve.Numerics.Vector3D(n.values[0], n.values[1], 0);
			else
				return new Zcu.Mve.Numerics.Vector3D(n.values[0], n.values[1], n.values[2]);
		} // NormalND -> Vector3D

		/// <summary>
		/// Implicit conversion NormalND -> VectorND
		/// </summary>
		/// <param name="n">NormalND</param>
		/// <returns>VectorND</returns>
		public static implicit operator Zcu.Mve.Numerics.VectorND(NormalND n) 
		{
			return new Zcu.Mve.Numerics.VectorND(((NormalND)n.DeepCopy()).values);
		} // NormalND -> VectorND

		// ===== Vector -> NormalND ===============
		/// <summary>
		/// Explicit conversion Vector2D -> NormalND
		/// </summary>
		/// <param name="v">Vector2D</param>
		/// <returns>NormalND</returns>
		public static explicit operator NormalND(Zcu.Mve.Numerics.Vector2D v) 
		{
			Normal2D tmp = (Normal2D) v;
			return new NormalND(new double[] {tmp.Nx, tmp.Ny});
		} // Vector2D -> NormalND

		/// <summary>
		/// Explicit conversion Vector3D -> NormalND
		/// </summary>
		/// <param name="v">Vector3D</param>
		/// <returns>NormalND</returns>
		public static explicit operator NormalND(Zcu.Mve.Numerics.Vector3D v) 
		{
			Normal3D tmp = (Normal3D) v;
			return new NormalND(new double[] {tmp.Nx, tmp.Ny, tmp.Nz});
		} // Vector3D -> NormalND

		/// <summary>
		/// Explicitn konverze VectorND -> NormalND
		/// </summary>
		/// <param name="v">VectorND</param>
		/// <returns>NormalND</returns>
		public static explicit operator NormalND(Zcu.Mve.Numerics.VectorND v) 
		{
			// zjisteni absolutni hodnoty vektoru
			double abs = 0;
			foreach(double d in v.Values)
				abs += Math.Pow(d, 2);
			abs = Math.Sqrt(abs);
			NormalND tmp = new NormalND(((Zcu.Mve.Numerics.VectorND)v.DeepCopy()).Values);
			// vlastni normalizace
			for(int i = 0; i < tmp.Values.Length; i++)
				tmp.Values[i] = tmp.Values[i] / abs;
			return tmp;
		} // VectorND -> NormalND

		#endregion

		// ===== Operators ================================================================
		#region Operators

		/// <summary>
		/// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>True if the values of operands are equal, false otherwise.</returns>
		public static bool operator ==(NormalND a, NormalND b)
		{
			return NormalND.Equals(a, b);
		}
		
		/// <summary>
		/// The inequality operator (!=) returns false if its operands are equal, true otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>False if the values of operands are equal, true otherwise.</returns>
		public static bool operator !=(NormalND a, NormalND b) 
		{
			return !(a == b);
		}

		#endregion

	} // class NormalND
} // namespace
