using System;
using System.Drawing;
using System.Diagnostics;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.ComponentModel;

namespace Zcu.Mve.Visualization.Loaders
{
	/// <summary>
	/// Module for loading of 2D image from a file. Supported formats (JPEG, GIF, BMP). Image is returned as RegGrid2D.
	/// </summary>
	[ModuleInfo("Miroslav Vavruka and Milan Frank",
		 "Module for loading of 2D image from a file. Supported formats (JPEG, GIF, BMP). Image is returned as RegGrid2D.",
		 Assemblied = "2005-02-05",
		 IconName = "PictureLoader.ico")]
	[PortInfo("Output", "Loaded image.")]
	public class PictureLoader : Zcu.Mve.Core.Module
	{
    RegGrid2D grid;

		/// <summary>
		/// Name of the file containing data.
		/// </summary>
		private string fileName;

    private bool reload = false;
    private bool paramsChanged = true;

    /// <summary>
    /// Determines whether or not to reload data each time the modul is asked for result. Set to true when the contens of the input files change, otherwise keep false.
    /// </summary>
    [Description("Determines whether or not to reload data each time the modul is asked for result. Set to true when the contens of the input files change, otherwise keep false.")]
    public bool Reload
    {
      get
      {
        return (reload);
      }
      set
      {
        reload = value;
      }
    }

    /// <summary>
    /// File to load picture from.
    /// </summary>
    [Editor(typeof(System.Windows.Forms.Design.FileNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
    [Description("File to load picture from.")]
    public string FileName
    {
      get
      {
        return (fileName);
      }
      set
      {
        if (fileName != value)
        {
          fileName = value;
          paramsChanged = true;
        }
      }
    }

		/// <summary>
		/// Creates module and output port.
		/// </summary>
		public PictureLoader()
		{
			AddOutPort("Output", typeof (Visualization.RegGrid2D) );
			this.fileName = "";
      this.SimulationStarted += new SimulationStartedEventHandler(PictureLoader_OnSimulationStartEvent);
    } // PictureLoader()

    void PictureLoader_OnSimulationStartEvent(object source)
    {
      paramsChanged = true;
    } 

		/// <summary>
		/// Read the configuration data from XML file.
		/// </summary>
		/// <param name="config">XmlElement containing the configuration data.</param>
		public override void ReadConfig(System.Xml.XmlElement config)
		{
			this.FileName = config.InnerText.Trim();
		} // ReadConfig()

		/// <summary>
		/// Write the configuration data in XML file.
		/// </summary>
		/// <param name="config">XmlElement containing the configuration data.</param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			config.InnerText = this.FileName;
		} // WriteConfig()

		/// <summary>
		/// Executes PictureLoader functionality - load picture from the file.
		/// </summary>
		public override void Execute()
		{
			if (this.fileName == "")		// nezadano zadne jmeno souboru
				throw new MveException(this, this.GetType().Name + " \"" + this.Name + "\" error: Pleas enter the name of file with data.");

      if (paramsChanged || reload || (grid == null))
      {
        try
        {
          Bitmap source;
          source = (Bitmap) Image.FromFile(this.fileName);
          UniformDataArray2D bmp = new UniformDataArray2D(typeof(ColorRGBA), source.Height, source.Width);

          for (int i = 0; i < bmp.Height; i++)
            for (int j = 0; j < bmp.Width; j++)
              bmp[i, j] = (ColorRGBA) source.GetPixel(j, i);

          grid = new RegGrid2D(bmp.Height, bmp.Width);
          grid.AddCellAttrs(typeof(ColorRGBA).Name, bmp);

          paramsChanged = false;
          SetOutput("Output", grid);
        } // try
        catch (OutOfMemoryException ex)
        {
          Debug.WriteLine(this.GetType().Name + " '" + this.Name + "' error:");
          Debug.WriteLine(ex);
          throw new MveException(this, this.GetType().Name + " \"" + this.Name + "\" error: Invalid file type. File must include a picture (*.bmp, *.jpg, *.gif etc.)");
        }
        catch (Exception ex)
        {
          Debug.WriteLine(this.GetType().Name + " '" + this.Name + "' error:");
          Debug.WriteLine(ex);
          throw new MveException(this, this.GetType().Name + " \"" + this.Name + "\" error: Could not find file \"" + ex.Message + "\"");
        } // catch
      }
      else
      {
        SetOutput("Output", grid, true);
      }
		} // Execute()
	} // class PictureLoader
} // namespace
