﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.Diagnostics;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization.Processing
{
	/// <summary>
	/// Traces streamlines in vector field
	/// </summary>
	class StreamTracer {
		public delegate void DProgressFunction(double progress);

		// output grid
		UnstrGrid streamlines;
		// vector grid
		RegGridND vectorGrid;
		// vectors used in interpolation
		UniformDataArrayND vectors;
		// interpolator to use
		Interpolator interpolator;
		// integration step
		double dt;
		// function to report progress to
		DProgressFunction progressFunction;
		// additional parameters for Runge-Kutta-Cash-Karp integrator
		static double SAFETY = 0.90;
		static double MAX_DIV = 10;
		static double MIN_DIV = 0.1;

		#region grid values

		Point3D min;
		Point3D max;
		Point3D cellSize;
		int dimension;
		static double Y2DHEIGHT = 1;

		#endregion

		#region user values

		int terminateCountF = 1000;
		int terminateCountB = 1000;
		double userDt = 0.01;
		double terminateSpeed;
		List<Point3D> seeds;

		int seedCount;
		EIntegratorType integratorType;
		EInterpolation interpolation;
		double errorTolerance = 1E-6;
		bool storeInputAttributes;
		string streamlinesAttributeName;
		string timeAttributeName;
		string speedAttributeName;

		#endregion

		public StreamTracer() {
			Integrator = EIntegratorType.Euler;
		}

		#region getters/setters

		public DProgressFunction ProgressFunction {
			get { return progressFunction; }
			set { progressFunction = value; }
		}

		public UnstrGrid Streamlines {
			get { return streamlines; }
		}

		public EIntegratorType Integrator {
			get { return integratorType; }
			set { integratorType = value; }
		}

		public double ErrorTolerance {
			set { errorTolerance = value; }
		}

		public string StreamlinesAttributeName {
			get { return streamlinesAttributeName; }
			set { streamlinesAttributeName = value; }
		}

		public string TimeAttributeName {
			get { return timeAttributeName; }
			set { timeAttributeName = value; }
		}

		public string SpeedAttributeName {
			get { return speedAttributeName; }
			set { speedAttributeName = value; }
		}

		#endregion

		/// <summary>
		/// Computes streamlines and stores them in UnstrGrid
		/// </summary>
		/// <param name="vectorGrid">grid with vectors</param>
		/// <param name="vectors">point attribute storing vectors</param>
		/// <param name="terminateCountF">Maximum count of forward steps in one streamline.</param>
		/// <param name="terminateCountB">Maximum count of backward steps in one streamline.</param>
		/// <param name="terminateSpeed">termination speed</param>
		/// <param name="dt">integration time step</param>
		/// <param name="seeds">seeds positions</param>
		/// <param name="storeInputAttributes">if input attributes from StructGrid should be saved</param>
		/// <param name="interpolation">interpolation type</param>
		public void ComputeStreamlines(RegGridND vectorGrid, UniformDataArrayND vectors,
			int terminateCountF, int terminateCountB, double terminateSpeed, double dt,
			List<Point3D> seeds, bool storeInputAttributes, EInterpolation interpolation) {
			this.terminateCountF = terminateCountF;
			this.terminateCountB = terminateCountB;
			this.terminateSpeed = terminateSpeed;
			this.userDt = dt;
			this.seeds = seeds;
			this.seedCount = seeds.Count;
			this.vectorGrid = vectorGrid;
			this.vectors = vectors;
			this.storeInputAttributes = storeInputAttributes;
			this.interpolation = interpolation;
			InitValues();
			GenerateGrid();
		}

		/// <summary>
		/// Loads and computes grid variables. Initializes interpolator for vector grid.
		/// </summary>
		void InitValues() {
			int len = vectorGrid.Dimensions.Length;
			int[] zeros = new int[len];
			Point3D lenghts;
			// 2d grid
			if (len == 2) {
				Point2D min2D = new Point2D(vectorGrid.GetPoint(zeros)[0], vectorGrid.GetPoint(zeros)[1]);
				Point2D max2D = new Point2D(vectorGrid.GetPoint(vectorGrid.Dimensions)[0], vectorGrid.GetPoint(vectorGrid.Dimensions)[1]);
				min = new Point3D(min2D.X, -Y2DHEIGHT / 2f, min2D.Y);
				max = new Point3D(max2D.X, Y2DHEIGHT / 2f, max2D.Y);
				lenghts = max - min;

				cellSize.X = lenghts.X / (vectorGrid.Dimensions[0]);
				cellSize.Y = 0;
				cellSize.Z = lenghts.Z / (vectorGrid.Dimensions[1]);

				Point2D cellSize2D = new Point2D(cellSize.X, cellSize.Z);

				interpolator = new Interpolator2D(vectors, min, max, cellSize2D, true, interpolation);

				dimension = len;
			}
			// 3d grid
			else {
				min = (Point3D)vectorGrid.GetPoint(zeros);
				max = (Point3D)vectorGrid.GetPoint(vectorGrid.Dimensions);
				lenghts = max - min;

				cellSize.X = lenghts.X / (vectorGrid.Dimensions[0]);
				cellSize.Y = lenghts.Y / (vectorGrid.Dimensions[1]);
				cellSize.Z = lenghts.Z / (vectorGrid.Dimensions[2]);

				interpolator = new Interpolator3D(vectors, min, max, cellSize, true, interpolation);

				dimension = len;
			}
		}

		#region integrators

		Point3D IntegratorEuler(Point3D position) {
			return (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position);
		}

		Point3D IntegratorRungeKutta4(Point3D position) {
			Point3D k1 = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position);
			Point3D k2 = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position + dt * 0.5 * k1);
			Point3D k3 = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position + dt * 0.5 * k2);
			Point3D k4 = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position + dt * k3);
			return (k1 + 2 * k2 + 2 * k3 + k4) / 6;
		}

		const double c21 = 1.0 / 5;
		const double c31 = 3.0 / 40, c32 = 9.0 / 40;
		const double c41 = 3.0 / 10, c42 = -9.0 / 10, c43 = 6.0 / 5;
		const double c51 = -11.0 / 54, c52 = 5.0 / 2, c53 = -70.0 / 27, c54 = 35.0 / 27;
		const double c61 = 1631.0 / 55296, c62 = 175.0 / 512, c63 = 575.0 / 13824, c64 = 44275.0 / 110592, c65 = 253.0 / 4096;
		const double a1 = 37.0 / 378, a3 = 250.0 / 621, a4 = 125.0 / 594, a6 = 512.0 / 1771;
		const double b1 = 2825.0 / 27648, b3 = 18575.0 / 48384, b4 = 13525.0 / 55296, b5 = 277.0 / 14336, b6 = 1.0 / 4;
		
		Point3D IntegratorRungeKutta4Adaptive(Point3D position, out double div, out double maxDif) {
			Point3D k1, k2, k3, k4, k5, k6, s1, s2;
			//double maxDif;
			k1 = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position);
			k2 = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position + dt * (c21 * k1));
			k3 = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position + dt * (c31 * k1 + c32 * k2));
			k4 = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position + dt * (c41 * k1 + c42 * k2 + c43 * k3));
			k5 = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position + dt * (c51 * k1 + c52 * k2 + c53 * k3 + c54 * k4));
			k6 = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(position + dt * (c61 * k1 + c62 * k2 + c63 * k3 + c64 * k4 + c65 * k5));
			s1 = a1 * k1 + a3 * k3 + a4 * k4 + a6 * k6;
			s2 = b1 * k1 + b3 * k3 + b4 * k4 + b5 * k5 + b6 * k6;

			// computes maximum error
			maxDif = Math.Abs(s2.X - s1.X);
			if (maxDif < Math.Abs(s2.Y - s1.Y)) {
				maxDif = Math.Abs(s2.Y - s1.Y);
			}
			if (maxDif < Math.Abs(s2.Z - s1.Z)) {
				maxDif = Math.Abs(s2.Z - s1.Z);
			}
			div = Math.Abs((maxDif*dt) / errorTolerance);

			return s1;
		}
		
		#endregion

		/// <summary>
		/// Computes one streamline and stores it in UniformDataArrays (points, speeds, times)
		/// </summary>
		/// <param name="startPoint">start point</param>
		/// <param name="points">data array for points</param>
		/// <param name="speeds">data array for speeds in points</param>
		/// <param name="times">data array for time</param>
		/// <param name="lenB">output parameter - point count in backward direction</param>
		/// <param name="lenF">output parameter - point count in forward direction</param>
		void AddStreamline(Point3D startPoint,
			Point3D[] points, Scalar[] speeds, Scalar[] times,
			out int lenB, out int lenF) {

			Point3D currPoint = startPoint;
			Point3D currSpeed = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(currPoint);

			int step = 0;
			double maxDif;
			double div;
			dt = userDt;
			double time = 0;
			int terminateCount;
			lenB = lenF = 0;

			points[terminateCountB] = currPoint;
			speeds[terminateCountB] = (Scalar)currSpeed.Abs();
			times[terminateCountB] = (Scalar)time;

			terminateCount = terminateCountF;

			// forward and backward direction
			for (int dir = 1; dir > -2; dir = dir - 2) {
				while (step < terminateCount && currSpeed.Abs() > terminateSpeed) {
					// determine speed
					switch (integratorType) {
						case EIntegratorType.Euler:
							currSpeed = IntegratorEuler(currPoint);
							time = time + dt;
							currPoint = currPoint + dt * currSpeed;
							break;
						case EIntegratorType.RungeKutta4:
							currSpeed = IntegratorRungeKutta4(currPoint);
							time = time + dt;
							currPoint = currPoint + dt * currSpeed;
							break;
						case EIntegratorType.RungeKutta4Adaptive:
							currSpeed = IntegratorRungeKutta4Adaptive(currPoint, out div, out maxDif);

							// if error < maximum error
							if (div < 1) {
								time = time + dt;
								currPoint = currPoint + dt * currSpeed;

								if (div < MIN_DIV) {
									div = MIN_DIV;
								}
								// adjust time step
								dt = dt * SAFETY * Math.Pow(1 / div, 0.2);
							}
							// if error >= maximum error
							else {
								if (div > MAX_DIV) {
									div = MAX_DIV;
								}
								// adjust time step (to smaller)
								dt = dt * SAFETY * Math.Pow(1 / div, 0.2);
								// (ignore this integration step)
								continue;
							}
							break;
					}

					if(interpolator.IsOut(currPoint)){
						break;
					}

					step++;
					points[terminateCountB + dir * step] = currPoint;
					speeds[terminateCountB + dir * step] = (Scalar)currSpeed.Abs();
					times[terminateCountB + dir * step] = (Scalar)time;
				}

				if (dir == 1) {
					lenF = step;
				}
				else {
					lenB = step;
				}
				// prepare for backward direction
				step = 0;
				terminateCount = terminateCountB;
				currPoint = startPoint;
				currSpeed = (Point3D)(Vector3D)interpolator.InterpolateAtPoint(currPoint);
				time = 0;
				dt = -userDt;
			}
		}

		/// <summary>
		/// Compacts arrays of Point3D to one UniformDataArray
		/// </summary>
		/// <param name="arrays">Point3D arrays</param>
		/// <param name="lensF">used lenghts in forward direction</param>
		/// <param name="lensB">used lenghts in backward direction</param>
		/// <param name="allCount">point count</param>
		/// <returns>UniformDataArray</returns>
		UniformDataArray CompactPoint3DArrays(Point3D[][] arrays, int[] lensF, int[] lensB, int allCount) {
			UniformDataArray compacted = new UniformDataArray(typeof(Point3D), allCount);
			int iComp = 0;
			for (int f = 0; f < arrays.Length;f++) {
				// at least one point (except from seed point) must be generated
				if (lensB[f] + lensF[f] > 0) {
					for (int i = 0; i < lensB[f]; i++) {
						compacted[iComp++] = arrays[f][terminateCountB - lensB[f] + i];
					}
					compacted[iComp++] = arrays[f][terminateCountB];
					for (int i = 0; i < lensF[f]; i++) {
						compacted[iComp++] = arrays[f][terminateCountB + 1 + i];
					}
				}
			}

			return compacted;
		}

		/// <summary>
		/// Compacts arrays of Scalar to one UniformDataArray
		/// </summary>
		/// <param name="arrays">Scalar arrays</param>
		/// <param name="lensF">used lenghts in forward direction</param>
		/// <param name="lensB">used lenghts in backward direction</param>
		/// <param name="allCount">point count</param>
		/// <returns>UniformDataArray</returns>
		UniformDataArray CompactScalarArrays(Scalar[][] arrays, int[] lensF, int[] lensB, int allCount) {
			UniformDataArray compacted = new UniformDataArray(typeof(Scalar), allCount);
			int iComp = 0;
			for (int f = 0; f < arrays.Length; f++) {
				// at least one point (except from seed point) must be generated
				if (lensB[f] + lensF[f] > 0) {
					for (int i = 0; i < lensB[f]; i++) {
						compacted[iComp++] = arrays[f][terminateCountB - lensB[f] + i];
					}
					compacted[iComp++] = arrays[f][terminateCountB];
					for (int i = 0; i < lensF[f]; i++) {
						compacted[iComp++] = arrays[f][terminateCountB + 1 + i];
					}
				}
			}

			return compacted;
		}

		/// <summary>
		/// Creates new UnstrGrid to store streamlines
		/// </summary>
		void GenerateGrid() {
			streamlines = new UnstrGrid();

			// initilaizes arrays
			Point3D[][] points = new Point3D[seedCount][];
			Scalar[][] speeds = new Scalar[seedCount][];
			Scalar[][] times = new Scalar[seedCount][];
			int[] lensF = new int[seedCount];
			int[] lensB = new int[seedCount];
			for (int i = 0; i < seedCount; i++) {
				points[i] = new Point3D[terminateCountF + terminateCountB + 1];
				speeds[i] = new Scalar[terminateCountF + terminateCountB + 1];
				times[i] = new Scalar[terminateCountF + terminateCountB + 1];
			}
			int allCount = 0;
			int streamlineCount = 0;

			// streamline generation
			for (int i = 0; i < seedCount; i++) {
				int lenB, lenF;
				AddStreamline(seeds[i], points[i], speeds[i], times[i], out lenB, out lenF);
				lensF[i] = lenF;
				lensB[i] = lenB;
				int count = lenB + lenF + 1;
				if (count > 1) {
					allCount += count;
					streamlineCount += 1;
				}

				progressFunction((double)(i + 1) / seedCount);
			}

			// if at least one streamline was generated
			if (allCount > 0) {
				// compacts arrays
				UniformDataArray allPoints = CompactPoint3DArrays(points, lensF, lensB, allCount);
				UniformDataArray allSpeeds = CompactScalarArrays(speeds, lensF, lensB, allCount);
				UniformDataArray allTimes = CompactScalarArrays(times, lensF, lensB, allCount);
				UniformDataArray lines = new UniformDataArray(typeof(PolyLine), streamlineCount);
				
				// generates polylines
				int start = 0;
				int streamlineIndex = 0;
				for (int f = 0; f < points.Length; f++) {
					// at least one point (except from seed point) must be generated
					int count = lensB[f] + lensF[f] +1;
					if (count > 1) {
						int[] indices = new int[count];
						for (int j = 0; j < count; j++) {
							indices[j] = start + j;
						}
						lines[streamlineIndex] = new PolyLine(indices);
						start += count;
						streamlineIndex++;
					}
				}

				// stores attributes to grid
				streamlines.Points = allPoints;
				streamlines.AddPointAttrs(speedAttributeName, allSpeeds);
				streamlines.AddPointAttrs(timeAttributeName, allTimes);
				streamlines.AddCells(streamlinesAttributeName, lines);

				if (storeInputAttributes) {
					StoreInputAttributes(streamlines, allPoints);
				}
			}
			else {
				throw new MveException(this, "No streamlines generated - possibly badly placed source point or large time step.");
			}
		}

		/// <summary>
		/// Saves input attributes of structured grid to unstr. grid using interpolator.
		/// </summary>
		/// <param name="grid">grid to save values to</param>
		/// <param name="points">points at which should be values interpolated</param>
		void StoreInputAttributes(UnstrGrid grid, UniformDataArray points) {
			System.Collections.ICollection col = vectorGrid.GetPointAttrIds();
			int count = col.Count;

			Interpolator inter;
			foreach (string id in col) {
				UniformDataArrayND array = (UniformDataArrayND)vectorGrid.GetPointAttrs(id);
				try {
					if (dimension == 2) {
						Point2D cellSize2D = new Point2D(cellSize.X, cellSize.Z);
						inter = new Interpolator2D(array, min, max, cellSize2D, false, interpolation);
					}
					else {
						inter = new Interpolator3D(array, min, max, cellSize, false, interpolation);
					}
					UniformDataArray interpolated = inter.Compute(points);
					grid.AddPointAttrs(id, interpolated);
				}
				// catch unsuported type exception
				catch (MveException ex) {
					Console.WriteLine(ex.Message);
				}
			}
		}
	}
}
