﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Diagnostics;

namespace Zcu.Mve.Visualization.Processing
{

	#region enums
	/// <summary>
	/// Integrator type
	/// </summary>
	/// <value name="Euler">ble</value>
	public enum EIntegratorType {
		/// <summary>
		/// Euler integrator - most simple and fastest
		/// </summary>
		Euler,
		/// <summary>
		/// Runge-Kutta 4th order - relatively accurate and fast
		/// </summary>
		RungeKutta4,
		/// <summary>
		/// 4th order Runge-Kutta integrator with adaptive step size - most accurate but slowest
		/// </summary>
		RungeKutta4Adaptive
	};
	/// <summary>
	/// Seed type
	/// </summary>
	public enum ESeedType {
		/// <summary>
		/// Seeds are generated from defined point and radius
		/// </summary>
		Point,
		/// <summary>
		/// Seeds are generated equally on the defined line
		/// </summary>
		Line
	};

	/// <summary>
	/// Interpolation type
	/// </summary>
	public enum EInterpolation { 
		/// <summary>
		/// Linear interpolation
		/// </summary>
		Linear,
		/// <summary>
		/// Cubic interpolation
		/// </summary>
		Cubic };
	#endregion

	/// <summary>
	/// Streamlines generator
	/// </summary>
	[ModuleInfo("Martin Dohnal", "Streamlines generator", IconName = "Streamlines.ico")]
	[PortInfo("vectorGrid", "Regular grid which contains vector point attribute to compute streamlines from.\nOnly RegGrid2D and three dimensional RegGridND is supported.\nVector attribute must be of type Vector2D or Vector3D.")]
	[PortInfo("streamlines", "Unstructured grid (UnstrGrid) holding streamlines and point attributes speed and time in their points.")]
	public class Streamlines : Zcu.Mve.Core.Module {
		// stream tracer for computations
		StreamTracer streamTracer;
		// random generator (for seed generation)
		Random random;
		// list of generated seed points
		List<Point3D> seeds;
		// grid bounds
		Point3D min, max;
		// smallest number for double number settings
		static double SMALLEST = 1E-12;

		#region attributes and properties

		// input
		string vectorAttributeName = "vectors";

		// integration
		double errorTolerance = 1E-6;
		int maxStepsForward = 100;
		int maxStepsBackward = 100;
		double terminateSpeed = 1E-6;
		double dt = 0.1;
		EIntegratorType integratorType = EIntegratorType.RungeKutta4;
		EInterpolation interpolation = EInterpolation.Linear;

		// seeds
		int seedCount = 50;
		bool automaticSeedPosition = true;
		ESeedType seedType = ESeedType.Line;
		Point3D sourcePoint = new Point3D(0.5, 0.5, 0.5f);
		double radius = 1;
		Point3D lineStart = new Point3D(0, 0, 0);
		Point3D lineEnd = new Point3D(1, 1, 1);
		
		// output
		string streamlinesAttributeName = "streamlines";
		string timeAttributeName = "time";
		string speedAttributeName = "speed";
		bool storeInputAttributes = true;
		bool recompute = false;

		/// <summary>
		/// Approximated maximum error of one integration step. Works only with RungeKutta4Adaptive integrator.
		/// </summary>
		[Category("Integration")]
		[Description("Approximated maximum error of one integration step. Works only with RungeKutta4Adaptive integrator.")]
		public double ErrorTolerance {
			get { return errorTolerance; }
			set { errorTolerance = Math.Max(SMALLEST, value);}
		}

		/// <summary>
		/// Integrator type. Euler integration is fastest but accumulates biggest error. Runge-Kutta 4 adaptive is slowest but offers error control and best precision.
		/// </summary>
		[Category("Integration")]
		[Description("Integrator type. Euler integration is fastest but accumulates biggest error. Runge-Kutta 4 adaptive is slowest but offers error control and best precision.")]
		public EIntegratorType IntegratorType {
			get { return integratorType; }
			set { integratorType = value; }
		}

		/// <summary>
		/// Interpolation type
		/// </summary>
		[Category("Integration")]
		[Description("Interpolation type. Linear if less acurate but lot faster.")]
		public EInterpolation Interpolation {
			get { return interpolation; }
			set { interpolation = value; }
		}

		/// <summary>
		/// Time step for integration.
		/// </summary>
		[Category("Integration")]
		[Description("Time step for integration.")]
		public double Dt {
			get { return dt; }
			set {dt = Math.Max(SMALLEST, value);}
		}

		/// <summary>
		/// Minimum speed when should be integration terminated.
		/// </summary>
		[Category("Integration")]
		[Description("Minimum speed when should be integration terminated.")]
		public double TerminateSpeed {
			get { return terminateSpeed; }
			set { terminateSpeed = Math.Max(SMALLEST, value); }
		}

		/// <summary>
		/// Maximum count of forward steps in one streamline.
		/// </summary>
		[Category("Integration")]
		[Description("Maximum count of forward steps in one streamline.")]
		public int MaxStepsForward {
			get { return maxStepsForward; }
			set { maxStepsForward = Math.Max(0,value); }
		}

		/// <summary>
		/// Maximum count of backward steps in one streamline.
		/// </summary>
		[Category("Integration")]
		[Description("Maximum count of backward steps in one streamline.")]
		public int MaxStepsBackward {
			get { return maxStepsBackward; }
			set { maxStepsBackward = Math.Max(0,value); }
		}

		/// <summary>
		/// Seed type
		/// </summary>
		[Category("Seed")]
		[Description("Seed type")]
		public ESeedType SeedType {
			get { return seedType; }
			set { seedType = value; }
		}

		/// <summary>
		/// Seed count
		/// </summary>
		[Category("Seed")]
		[Description("Seed count")]
		public int SeedCount {
			get { return seedCount; }
			set { seedCount = Math.Max(1, value); }
		}

		/// <summary>
		/// Sets source point to the center of the grid.
		/// </summary>
		[Category("Seed")]
		[Description("Sets source point to the center of the grid.")]
		public bool AutomaticSeedPosition {
			get { return automaticSeedPosition; }
			set { automaticSeedPosition = value; }
		}

		/// <summary>
		/// Maximal radius of generated seeds from specified source point.
		/// </summary>
		[Category("Seed - Point")]
		[Description("Maximal radius of generated seeds from specified source point.")]
		public double Radius {
			get { return radius; }
			set { radius = Math.Max(0,value); }
		}

		/// <summary>
		/// Source point for generating seeds.
		/// </summary>
		[Category("Seed - Point")]
		[Description("Source point for generating seeds.")]
		public Point3D SourcePoint {
			get { return sourcePoint; }
			set { sourcePoint = value; }
		}

		/// <summary>
		/// Line end point
		/// </summary>
		[Category("Seed - Line")]
		[Description("Line end point")]
		public Point3D LineEnd {
			get { return lineEnd; }
			set { lineEnd = value; }
		}

		/// <summary>
		/// Line start point
		/// </summary>
		[Category("Seed - Line")]
		[Description("Line start point")]
		public Point3D LineStart {
			get { return lineStart; }
			set { lineStart = value; }
		}

		/// <summary>
		/// Name of vector field to create streamlines from. If vector field by given name is not found, module tries to find one.
		/// </summary>
		[Category("Input")]
		[Description("Name of vector field to create streamlines from. If vector field by given name is not found, module tries to find one.")]
		public string VectorFieldName {
			get { return vectorAttributeName; }
			set { vectorAttributeName = value; }
		}

		/// <summary>
		/// If input point attributes at streamline points should be stored in output.
		/// </summary>
		[Category("Output")]
		[Description("If input point attributes at streamline points should be stored in output.")]
		public bool StoreInputAttributes {
			get { return storeInputAttributes; }
			set { storeInputAttributes = value; }
		}

		/// <summary>
		/// Time attribute name.
		/// </summary>
		[Category("Output")]
		[Description("Time attribute name.")]
		public string TimeAttributeName {
			get { return timeAttributeName; }
			set {
				if (value == null || value == "") {
					timeAttributeName = "time";
				}
				else {
					timeAttributeName = value;
				}
			}
		}

		/// <summary>
		/// Speed attribute name.
		/// </summary>
		[Category("Output")]
		[Description("Speed attribute name.")]
		public string SpeedAttributeName {
			get { return speedAttributeName; }
			set {
				if (value == null || value == "") {
					speedAttributeName = "speed";
				}
				else {
					speedAttributeName = value;
				} 
			}
		}

		/// <summary>
		/// Name of cell attribute which holds streamlines stored in output.
		/// </summary>
		[Category("Output")]
		[Description("Name of cell attribute which holds streamlines stored in output.")]
		public string StreamlinesAttributeName {
			get { return streamlinesAttributeName; }
			set {
				if (value == null || value == "") {
					streamlinesAttributeName = "streamlines";
				}
				else {
					streamlinesAttributeName = value;
				}
			}
		}

		/// <summary>
		/// If streamlines should be recomputed every execution
		/// </summary>
		[Category("Output")]
		[Description("If streamlines should be recomputed every execution")]
		public bool Recompute {
			get { return recompute; }
			set { recompute = value; }
		}

		#endregion

		/// <summary>
		/// Creates new streamlines module
		/// </summary>
		public Streamlines() {
			this.AddOutPort("streamlines", typeof(UnstrGrid));

			this.AddInPort("vectorGrid", typeof(StructGrid));

			streamTracer = new StreamTracer();
			streamTracer.ProgressFunction = ProgressInfo;

			random = new Random();
		}

		/// <summary>
		/// Computes center of the grid
		/// </summary>
		/// <param name="vectorGrid">grid</param>
		/// <returns>center of grid</returns>
		void ComputeGridBounds(RegGridND vectorGrid) {
			int len = vectorGrid.Dimensions.Length;
			int[] zeros = new int[len];
			if (len == 2) {
				Point2D min2D = new Point2D(vectorGrid.GetPoint(zeros)[0], vectorGrid.GetPoint(zeros)[1]);
				Point2D max2D = new Point2D(vectorGrid.GetPoint(vectorGrid.Dimensions)[0], vectorGrid.GetPoint(vectorGrid.Dimensions)[1]);
				min = new Point3D(min2D.X, 0, min2D.Y);
				max = new Point3D(max2D.X, 0, max2D.Y);
			}
			else {
				min = (Point3D)vectorGrid.GetPoint(zeros);
				max = (Point3D)vectorGrid.GetPoint(vectorGrid.Dimensions);
			}
		}

		/// <summary>
		/// Transforms RegGrid2D to two dimensional RegGridND
		/// </summary>
		/// <param name="grid">grid</param>
		/// <returns>two dimensional RegGridND</returns>
		RegGridND TransformToRegGridND(RegGrid2D grid) {
			// create grid with same size
			int[] sizes = new int[3];
			sizes[0] = grid.Width + 1;
			sizes[1] = grid.Height + 1;
			RegGridND transformed = new RegGridND(sizes[0] - 1, sizes[1] - 1);

			// copy data (point attributes)
			System.Collections.ICollection col = grid.GetPointAttrIds();
			int count = col.Count;

			foreach (string id in col) {
				UniformDataArray2D arr2D = grid.GetPointAttrs(id);
				UniformDataArrayND arrND = new UniformDataArrayND(arr2D.ElemType, sizes[0], sizes[1]);
				for (int x = 0; x < sizes[0]; x++) {
					for (int y = 0; y < sizes[1]; y++) {
						arrND[x, y] = arr2D[x, y];
					}
				}

				transformed.AddPointAttrs(id, arrND);
			}

			return transformed;
		}

		/// <summary>
		/// Generates random seed point with respect to radius and grid dimension
		/// </summary>
		/// <returns></returns>
		Point3D GetRandomPoint(int dimension) {
			Point3D offset = new Point3D(random.NextDouble() - 0.5, random.NextDouble() - 0.5, random.NextDouble() - 0.5);
			offset = offset / offset.Abs();
			offset = offset * random.NextDouble() * radius;
			if (dimension == 2) {
				offset.Y = 0;
			}
			return sourcePoint + offset;
		}

		/// <summary>
		/// Generates seed points
		/// </summary>
		/// <param name="vectorGrid">grid</param>
		/// <returns>list of seed points</returns>
		List<Point3D> GenerateSeeds(RegGridND vectorGrid) {
			int dimension = vectorGrid.Dimensions.Length;
			// save grid bounds to min, max points
			ComputeGridBounds(vectorGrid);
			List<Point3D> seeds = new List<Point3D>(seedCount);
			// if automatic position is true, adjust seed sources
			if (automaticSeedPosition) {
				sourcePoint = (min + max) * 0.5;
				radius = (max - min).Abs() * 0.5;
				lineStart = min;
				lineEnd = max;
			}
			// point seed
			if (seedType == ESeedType.Point) {
				for (int i = 0; i < seedCount; i++) {
					seeds.Add(GetRandomPoint(dimension));
				}
			}
			// line seed
			else {
				for (int i = 0; i < seedCount; i++) {
					seeds.Add(lineStart + (lineEnd - lineStart) * (double)i / (seedCount - 1));
				}
			}

			return seeds;
		}

		/// <summary>
		/// Computes streamlines
		/// </summary>
		public override void Execute() {
			UnstrGrid streamlines = streamTracer.Streamlines;
			if (streamlines == null || !IsAllDataSame() || !IsAllUserSettingsSame() || recompute) {
				StructGrid inputGrid = (StructGrid)GetInput("vectorGrid");
				// transform RegGrid2D to two dimensional RegGridND
				if (inputGrid is RegGrid2D) {
					RegGrid2D image = (RegGrid2D)inputGrid;
					inputGrid = TransformToRegGridND(image);
				}
				if (inputGrid is RegGridND) {
					RegGridND vectorGrid = (RegGridND)inputGrid;

					// checks grid dimension
					int dimension = vectorGrid.Dimensions.Length;
					if (dimension < 2 || dimension > 3) {
						throw new MveException(vectorGrid, "Grid dimension must be 2 or 3");
					}

					// find correct vector attribute
					System.Collections.ICollection col = vectorGrid.GetPointAttrIds();
					int count = col.Count;
					string candidate = null;
					foreach (string id in col) {
						UniformDataArrayND array = (UniformDataArrayND)vectorGrid.GetPointAttrs(id);
						if (id == vectorAttributeName) {
							if (array.ElemType == typeof(Vector3D) || array.ElemType == typeof(Vector2D)) {
								candidate = id;
							}
						}
						else {
							if (candidate == null && (array.ElemType == typeof(Vector3D) || array.ElemType == typeof(Vector2D))) {
								candidate = id;
							}
						}
					}
					if (candidate == null) {
						throw new MveException(this, "No point attribute in grid contains vectors.");
					}
					else if (candidate != vectorAttributeName) {
						if (vectorAttributeName == null || vectorAttributeName == "") {
							Console.WriteLine("Vector field not specified. Setting '" + candidate + "' as current vector field.");
							vectorAttributeName = candidate;
						}
						else {
							throw new MveException(vectorGrid, "Vector field '" + vectorAttributeName + "' not found / has bad type.");
						}
					}

					UniformDataArrayND vectors = (UniformDataArrayND)vectorGrid.GetPointAttrs(vectorAttributeName);

					// generate seeds
					seeds = GenerateSeeds(vectorGrid);
					// sets parameters for integration
					streamTracer.Integrator = integratorType;
					streamTracer.ErrorTolerance = errorTolerance;
					streamTracer.StreamlinesAttributeName = streamlinesAttributeName;
					streamTracer.TimeAttributeName = timeAttributeName;
					streamTracer.SpeedAttributeName = speedAttributeName;
					// computation
					streamTracer.ComputeStreamlines(vectorGrid, vectors, maxStepsForward, maxStepsBackward, terminateSpeed, dt,
						seeds, storeInputAttributes, interpolation);
					streamlines = streamTracer.Streamlines;
					SetOutput("streamlines", streamlines);
				}
				else {
					throw new MveException(this, "Unsupported grid type.");
				}
				// call to make sure that user settings didnt change in next call
				IsAllUserSettingsSame();
			}
			else {
				SetOutput("streamlines", streamlines, true);
			}
		}
	}
}
