using System;
using System.Diagnostics;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
  /// <summary>
  /// General projective tranformation in 4D space.
  /// </summary>
  public class Transform4D : Visualization.AbstractTransform
  {
    /// <summary>
    /// Double matrix 5 x 5 representing 4D-transformation.
    /// </summary>
    private double[,] matrix = new double[5, 5];

    /// <summary>
    /// The toleration for equality operator.
    /// </summary>
    public static double Epsilon = Globals.EpsilonFloatImplicit;

    /// <summary>
    /// Creates a Transform4D object and initialize it (set identity).
    /// </summary>
    public Transform4D()
    {
      SetIdentity();

      UpdateModificationTime();
    }

    /// <summary>
    /// Set identity transformation.
    /// </summary>
    public void SetIdentity()
    {
      for (int i = 0; i < 5; i++)
      {
        for (int j = 0; j < 5; j++)
        {
          matrix[i, j] = 0.0;
        }
      }
      for (int i = 0; i < 5; i++)
      {
        matrix[i, i] = 1.0;
      }

      UpdateModificationTime();
    }

    /// <summary>
    /// Add scaling to current transformation.
    /// TODO: Optimize it!
    /// </summary>
    /// <param name="dx">Scaling in x-axis. 1 means no scale.</param>
    /// <param name="dy">Scaling in y-axis. 1 means no scale.</param>
    /// <param name="dz">Scaling in z-axis. 1 means no scale.</param>
    /// <param name="dw">Scaling in w-axis. 1 means no scale.</param>
    public void AddScale(double dx, double dy, double dz, double dw)
    {
      Transform4D t = new Transform4D();
      t.SetScale(dx, dy, dz, dw);
      this.matrix = (this * t).matrix;

      UpdateModificationTime();
    }

    /// <summary>
    /// Set curent matrix to scaling matrix of given values.
    /// </summary>
    /// <param name="dx">Scaling in x-axis. 1 means no scale.</param>
    /// <param name="dy">Scaling in y-axis. 1 means no scale.</param>
    /// <param name="dz">Scaling in z-axis. 1 means no scale.</param>
    /// <param name="dw">Scaling in w-axis. 1 means no scale.</param>
    public void SetScale(double dx, double dy, double dz, double dw)
    {
      SetIdentity();
      matrix[0, 0] = dx;
      matrix[1, 1] = dy;
      matrix[2, 2] = dz;
      matrix[3, 3] = dw;

      UpdateModificationTime();
    }

    /// <summary>
    /// Noncumulative rotation in plane defined by two axes (0 represents the X axis, 3 represents the W axis.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void SetRotation(int axis1, int axis2, double angle)
    {
      if (axis1 == axis2)
        throw new MveException(this, "Rotation plane axes cannot be equal.");
      if ((axis1 > 3) || (axis2 > 3))
        throw new MveException(this, "Rotation axis identifier cannot be greater than 3.");
      if ((axis1 < 0) || (axis2 < 0))
        throw new MveException(this, "Rotation axis identifier cannot be smaller than 0.");

      SetIdentity();
      matrix[axis1, axis1] = matrix[axis2, axis2] = Math.Cos(angle);
      matrix[axis1, axis2] = -Math.Sin(angle);
      matrix[axis2, axis1] = -matrix[axis1, axis2];

      UpdateModificationTime();
    }

    /// <summary>
    /// Cumulative rotation in a plane defined by two axes (0 represents the X axis, 3 represents the W axis.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void AddRotation(int axis1, int axis2, double angle)
    {
      Transform4D rot = new Transform4D();
      rot.SetRotation(axis1, axis2, angle);
      this.matrix = (this * rot).matrix;

      UpdateModificationTime();
    }

    /// <summary>
    /// Noncumulative rotation in XY plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void SetRotationXY(double angle)
    {
      SetIdentity();
      matrix[0, 0] = matrix[1, 1] = Math.Cos(angle);
      matrix[0, 1] = -Math.Sin(angle);
      matrix[1, 0] = -matrix[0, 1];

      UpdateModificationTime();
    }

    /// <summary>
    /// Cumulative rotation in XY plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void AddRotationXY(double angle)
    {
      Transform4D rot = new Transform4D();
      rot.SetRotationXY(angle);
      this.matrix = (this * rot).matrix;

      UpdateModificationTime();
    }

    /// <summary>
    /// Noncumulative rotation in YZ plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void SetRotationYZ(double angle)
    {
      SetIdentity();
      matrix[1, 1] = matrix[2, 2] = Math.Cos(angle);
      matrix[1, 2] = -Math.Sin(angle);
      matrix[2, 1] = -matrix[1, 2];

      UpdateModificationTime();
    }

    /// <summary>
    /// Cumulative rotation in YZ plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void AddRotationYZ(double angle)
    {
      Transform4D rot = new Transform4D();
      rot.SetRotationYZ(angle);
      this.matrix = (this * rot).matrix;

      UpdateModificationTime();
    }

    /// <summary>
    /// Noncumulative rotation in XZ plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void SetRotationXZ(double angle)
    {
      SetIdentity();
      matrix[0, 0] = matrix[2, 2] = Math.Cos(angle);
      matrix[0, 2] = -Math.Sin(angle);
      matrix[2, 0] = -matrix[0, 2];

      UpdateModificationTime();
    }

    /// <summary>
    /// Cumulative rotation in XZ plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void AddRotationXZ(double angle)
    {
      Transform4D rot = new Transform4D();
      rot.SetRotationXZ(angle);
      this.matrix = (this * rot).matrix;

      UpdateModificationTime();
    }

    /// <summary>
    /// Noncumulative rotation in XW plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void SetRotationXW(double angle)
    {
      SetIdentity();
      matrix[0, 0] = matrix[3, 3] = Math.Cos(angle);
      matrix[0, 3] = -Math.Sin(angle);
      matrix[3, 0] = -matrix[0, 3];

      UpdateModificationTime();
    }

    /// <summary>
    /// Cumulative rotation in XW plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void AddRotationXW(double angle)
    {
      Transform4D rot = new Transform4D();
      rot.SetRotationXW(angle);
      this.matrix = (this * rot).matrix;

      UpdateModificationTime();
    }

    /// <summary>
    /// Noncumulative rotation in YW plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void SetRotationYW(double angle)
    {
      SetIdentity();
      matrix[1, 1] = matrix[3, 3] = Math.Cos(angle);
      matrix[1, 3] = -Math.Sin(angle);
      matrix[3, 1] = -matrix[1, 3];

      UpdateModificationTime();
    }

    /// <summary>
    /// Cumulative rotation in YW plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void AddRotationYW(double angle)
    {
      Transform4D rot = new Transform4D();
      rot.SetRotationYW(angle);
      this.matrix = (this * rot).matrix;

      UpdateModificationTime();
    }

    /// <summary>
    /// Noncumulative rotation in ZW plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void SetRotationZW(double angle)
    {
      SetIdentity();
      matrix[2, 2] = matrix[3, 3] = Math.Cos(angle);
      matrix[2, 3] = -Math.Sin(angle);
      matrix[3, 2] = -matrix[2, 3];

      UpdateModificationTime();
    }

    /// <summary>
    /// Cumulative rotation in ZW plane measured in radians.
    /// </summary>
    /// <param name="angle">Rotation angle masured in radians.</param>
    public void AddRotationZW(double angle)
    {
      Transform4D rot = new Transform4D();
      rot.SetRotationZW(angle);
      this.matrix = (this * rot).matrix;

      UpdateModificationTime();
    }

    /// <summary>
    /// Noncumlative translation.
    /// </summary>
    /// <param name="x">Translation in x-axis.</param>
    /// <param name="y">Translation in y-axis.</param>
    /// <param name="z">Translation in z-axis.</param>
    /// <param name="w">Translation in w-axis.</param>
    public void SetTranslation(double x, double y, double z, double w)
    {
      SetIdentity();
      matrix[0, 4] = x;
      matrix[1, 4] = y;
      matrix[2, 4] = z;
      matrix[3, 4] = w;

      UpdateModificationTime();
    }

    /// <summary>
    /// Cumulative translation
    /// </summary>
    /// <param name="x">Translation in x-axis.</param>
    /// <param name="y">Translation in y-axis.</param>
    /// <param name="z">Translation in z-axis.</param>
    /// <param name="w">Translation in w-axis.</param>
    public void AddTranslation(double x, double y, double z, double w)
    {
      Transform4D t = new Transform4D();
      t.SetTranslation(x, y, z, w);
      this.matrix = (this * t).matrix;

      UpdateModificationTime();
    }

    /// <summary>
    /// Transform the point.
    /// </summary>
    /// <param name="op">Point to be transformed. This parameter won't be changed.</param>
    /// <returns>New instance of transformed point.</returns>
    public PointND Transform(PointND op)
    {
      if (op.Length != 4)
      {
        throw new MveException(this, "Only 4D points can be transformed by Transform4D.");
      }

      PointND tp = new PointND(4);

      tp[0] = matrix[0, 0] * op[0] + matrix[0, 1] * op[1] + matrix[0, 2] * op[2] + matrix[0, 3] * op[3] + matrix[0, 4];
      tp[1] = matrix[1, 0] * op[0] + matrix[1, 1] * op[1] + matrix[1, 2] * op[2] + matrix[1, 3] * op[3] + matrix[1, 4];
      tp[2] = matrix[2, 0] * op[0] + matrix[2, 1] * op[1] + matrix[2, 2] * op[2] + matrix[2, 3] * op[3] + matrix[2, 4];
      tp[3] = matrix[3, 0] * op[0] + matrix[3, 1] * op[1] + matrix[3, 2] * op[2] + matrix[3, 3] * op[3] + matrix[3, 4];
      double w = matrix[4, 0] * op[0] + matrix[4, 1] * op[1] + matrix[4, 2] * op[2] + matrix[4, 3] * op[3] + matrix[4, 4];
      tp[0] /= w;
      tp[1] /= w;
      tp[2] /= w;
      tp[3] /= w;

      return tp;
    }

    /// <summary>
    /// Full access to particular matrix elements.
    /// </summary>
    public double this[int i, int j]
    {
      set
      {
        matrix[i, j] = value;

        UpdateModificationTime();
      }
      get
      {
        return matrix[i, j];
      }
    }

    /// <summary>
    /// Check the consistency of current data structure.
    /// </summary>
    /// <returns>True if structure is consistent.</returns>
    public override bool CheckConsistence()
    {
      for (int i = 0; i < 5; i++)
        for (int j = 0; j < 5; j++)
        {
          if (double.IsNaN(matrix[i, j]))
            Debug.WriteLine("4D transform matrix contains a NaN value.");
          if (double.IsInfinity(matrix[i, j]))
            Debug.WriteLine("4D transform matrix contains infinite value.");
        }
      return true;
    }

    /// <summary>
    /// Deep copy of data object and its subelements.
    /// </summary>
    /// <returns>Same data, different memory space.</returns>
    public override Zcu.Mve.Core.IDataObject DeepCopy()
    {
      Transform4D t = new Transform4D();

      Array.Copy(this.matrix, t.matrix, 25);

      return t;
    }

    /// <summary>
    /// Read data from Xml file to current structure.
    /// </summary>
    /// <param name="xmlTextReader">Stream to be read.</param>
    public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
    {
      string[] lines = xmlTextReader.ReadString().Trim().Split(';');

      if (lines.Length != 5)
        throw new MveException(this, "Incorrect number of lines in Transform4D.");

      for (int i = 0; i < 5; i++)
      {
        string[] elems = lines[i].Trim().Split(' ');
        if (elems.Length != 5)
          throw new MveException(this, "Incorrect number of elements in line " + i + " in Transform4D.");

        for (int j = 0; j < 5; j++)
        {
          this.matrix[i, j] = double.Parse(elems[j].Trim(), Globals.nfi);
        }
      }

      UpdateModificationTime();
    }

    /// <summary>
    /// Write current data to Xml file.
    /// </summary>
    /// <param name="xmlTextWriter">Stream to be write.</param>
    public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
    {
      string str = "";
      for (int i = 0; i < 5; i++)
      {
        for (int j = 0; j < 5; j++)
        {
          str += this.matrix[i, j].ToString(Globals.nfi);
          if (j < 4)
            str += " ";
        }
        if (i < 4)
          str += "; ";
      }
      xmlTextWriter.WriteString(str);
    }

    // ===== Methods of the class Object ==============================================
    #region Overriden methods of the class Object

    /// <summary>
    /// Returns a String that represents the current Object.
    /// </summary>
    /// <returns>String that represents the current Object.</returns>
    public override string ToString()
    {
      string str = "Transform4D:\n";
      for (int i = 0; i < 5; i++)
        str += "   [" + matrix[i, 0] + ", " + matrix[i, 1] + ", " + matrix[i, 2] + ", " + matrix[i, 3] + ", " + matrix[i, 4] + "]\r\n";
      return str.Substring(0, str.Length - 2);
    }

    /// <summary>
    /// Determines whether the specified Objects are equal.
    /// </summary>
    /// <param name="a">First Object.</param>
    /// <param name="b">Second Object.</param>
    /// <returns>True if objects are equal. Otherwise, false.</returns>
    public static new bool Equals(object a, object b)
    {
      if ((a is Transform4D) && (b is Transform4D))
      {
        Transform4D x = (Transform4D) a;
        Transform4D y = (Transform4D) b;

        for (int i = 0; i < 5; i++)
          for (int j = 0; j < 5; j++)
            if ((float) Math.Abs(x[i, j] - y[i, j]) > Transform4D.Epsilon)
              return false;

        return true;
      }
      else
      {
        return false;
      }
    }

    /// <summary>
    /// Determines whether the specified Object is equal to the current Object.
    /// </summary>
    /// <param name="obj">Specified Object to compare with the current Object.</param>
    /// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
    public override bool Equals(object obj)
    {
      return Transform4D.Equals(this, obj);
    }

    /// <summary>
    /// Returns a hash code for the current Object.
    /// </summary>
    /// <returns>A hash code for the current Object.</returns>
    public override int GetHashCode()
    {
      int hashCode = 0;

      for (int i = 0; i < 5; i++)
        for (int j = 0; j < 5; j++)
          hashCode = hashCode ^ this[i, j].GetHashCode();

      return hashCode;
    }

    #endregion

    // ===== Operators ================================================================
    #region Operators

    /// <summary>
    /// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
    /// </summary>
    /// <param name="a">First operand.</param>
    /// <param name="b">Second operand.</param>
    /// <returns>True if the values of operands are equal, false otherwise.</returns>
    public static bool operator ==(Transform4D a, Transform4D b)
    {
      return Transform4D.Equals(a, b);
    }

    /// <summary>
    /// The inequality operator (!=) returns false if its operands are equal, true otherwise.
    /// </summary>
    /// <param name="a">First operand.</param>
    /// <param name="b">Second operand.</param>
    /// <returns>False if the values of operands are equal, true otherwise.</returns>
    public static bool operator !=(Transform4D a, Transform4D b)
    {
      return !(a == b);
    }

    /// <summary>
    /// Composition of transformations. Matrix multiplication.
    /// </summary>
    /// <param name="t1"></param>
    /// <param name="t2"></param>
    /// <returns></returns>
    public static Transform4D operator *(Transform4D t1, Transform4D t2)
    {
      Transform4D t = new Transform4D();
      double sum = 0.0;

      for (int i = 0; i < 5; i++)
      {
        for (int j = 0; j < 5; j++)
        {
          sum = 0.0f;
          for (int k = 0; k < 5; k++)
          {
            sum += t1[i, k] * t2[k, j];
          }
          t[i, j] = sum;
        }
      }
      return t;
    }

    public static PointND operator *(PointND p, Transform4D t)
    {
      if (p.Length == 4)
      {
        //multiply by 4x4 submatrix
        PointND result = new PointND(4);
        for (int i = 0; i < 4; i++)
          for (int j = 0; j < 4; j++)
            result[i] += p[i] * t[i, j];

        //add translations, compute weight
        double w = 0.0;
        for (int i = 0; i < 4; i++)
        {
          result[i] += t[4, i];
          w += p[i] * t[i, 4];
        }

        //add bottomright corner to weight
        w += t[4, 4];

        //normalize to w = 1;
        for (int i = 0; i < 4; i++)
          result[i] /= w;
        return (result);

      }
      else if (p.Length == 5)
      {
        PointND result = new PointND(5);
        for (int i = 0; i < 5; i++)
          for (int j = 0; j < 5; j++)
            result[i] += p[i] * t[i, j];
        return (result);
      }
      else
      {
        throw new MveException(typeof(Transform4D), "Only 4-dimensional points or 5-dimensional homogeneous representations of them can be used in Transform4D multiplication.");
      }
    }

    #endregion

  } // class Transform4D
} // namespace