using System;
using Zcu.Mve.Core;
using System.Xml;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Structure represent topology of a Triangle. 
	/// It is defined by tree point indices.
	/// </summary>
	[DataObject("Structure represent topology of a Triangle. It is defined by tree point indices.")]
	public struct Triangle : ICell
	{
		/// <summary>
		/// Index of first vertex of the triangle.
		/// </summary>
		public int V1;
		/// <summary>
		/// Index of second vertex of the triangle.
		/// </summary>
		public int V2;
		/// <summary>
		/// Index of third vertex of the triangle.
		/// </summary>
		public int V3;

		/// <summary>
		/// Create a Triangle object.
		/// </summary>
		/// <param name="V1">Index of first vertex of the triangle.</param>
		/// <param name="V2">Index of second vertex of the triangle.</param>
		/// <param name="V3">Index of third vertex of the triangle.</param>
		public Triangle(int V1, int V2, int V3)
		{
			this.V1 = V1;
			this.V2 = V2;
			this.V3 = V3;
		} // Triangle()

		/// <summary>
		/// Returns array of three vertices of the triangle.
		/// Change of returned array do not cause change inside line object.
		/// </summary>
		/// <remarks>This access is not efficient. Use V1, V2, V3 instead.</remarks>
		[ObsoleteAttribute("This access is not efficient. Use V1, V2, V3 instead.", false)]
		public int [] Points
		{
			get
			{
				return new int [3] {V1, V2, V3};
			}
		} // PointArray

		/// <summary>
		/// Return number of points of cell.
		/// </summary>
		public int PointCount
		{
			get
			{
				return 3;
			}
		} // PointCount

		/// <summary>
		/// Returns index of vertex.
		/// </summary>
		/// <param name="index">0, 1, 2 are only acceptable vaules.</param>
		/// <returns>Index of vertex.</returns>
		/// <remarks>This access is not efficient. Use V1, V2, V3 instead.</remarks>
		[ObsoleteAttribute("This access is not efficient. Use V1, V2, V3 instead.", false)]
		public int GetPoint(int index)
		{
			switch (index)
			{
				case 0:
					return V1;
				case 1:
					return V2;
				case 2:
					return V3;
				default:
					throw new MveException(this, "Index out of range. Triangle have only three vertices.");
			}
		} // GetPoint()

		/// <summary>
		/// Get/Set index of Triangle vertex
		/// </summary>
		/// <remarks>This access is not efficient. Use V1, V2, V3 instead.</remarks>
		[ObsoleteAttribute("This access is not efficient. Use V1, V2, V3 instead.", false)]
		public int this[int index]
		{
			get
			{
				switch (index)
				{
					case 0:
						return V1;
					case 1:
						return V2;
					case 2:
						return V3;
					default:
						throw new MveException(this, "Index out of range. Triangle have only three vertices.");
				}					
			}
			set
			{
				switch (index)
				{
					case 0:
						V1 = value;
						break;
					case 1:
						V2 = value;
						break;
					case 2:
						V3 = value;
						break;
					default:
						throw new MveException(this, "Index out of range. Triangle have only three vertices.");
				}						
			}
		} // this[int]

		/// <summary>
		/// Read data from XML file.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			string [] vertices = xmlTextReader.ReadString().Trim().Split(' ', '\t');

			if (vertices.Length != 3)
				throw new MveException(this, "Incorect number of vertices in triangle.");

			this.V1 = Int32.Parse(vertices[0].Trim());
			this.V2 = Int32.Parse(vertices[1].Trim());
			this.V3 = Int32.Parse(vertices[2].Trim());
		} // ReadData()

		/// <summary>
		/// Write data to XML file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			xmlTextWriter.WriteString(V1 + " " + V2 + " " + V3);
		} // WriteData()

		/// <summary>
		/// Deep copy of curent object
		/// </summary>
		/// <returns>Deep copy (same data, different memory place).</returns>
		public IDataObject DeepCopy()
		{
			return new Triangle(this.V1, this.V2, this.V3);
		} // DeepCopy()

		/// <summary>
		/// Triangle with negative index is considered inconsistent.
		/// If two or three identical indices are detected warning is given.
		/// </summary>
		/// <returns>True if the Triangle is consistent.</returns>
		public bool CheckConsistence()
		{
			bool isOK = true;
			if (this.V1 < 0 || this.V2 < 0 || this.V3 < 0)
			{
				Console.WriteLine("Error! " + this.GetType().Name + " has negative index.");
				isOK = false;
			}
			if (this.V1 == this.V2 && this.V2 == this.V3)
				Console.WriteLine("Warning! " + this.GetType().Name + " deformation: all 3 points have identical index.");
			else if (this.V1 == this.V2 || this.V1 == this.V3 || this.V2 == this.V3)
				Console.WriteLine("Warning! " + this.GetType().Name + " deformation: some 2 points have identical index.");
			return isOK;
		} // CheckConsistence()

		/// <summary>
		/// Triangle information
		/// </summary>
		/// <returns>String that contain name of this object and indices of curent object.</returns>
		override public string ToString()
		{
			return String.Format("Triangle: [{0}, {1}, {2}]", V1, V2, V3);
		} // ToString()

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

		/// <summary>
		/// Determines equality of triangles based on vertex indexing. Differently oriented triangles are considered equal
		/// </summary>
		/// <param name="obj">Target to be compared to. Instance of Triangle class or it's descendant is expected, other data types result in false value.</param>
		/// <returns>true if vertices have same indexes (different order possible), false otherwise</returns>
		public override bool Equals(object obj)
		{
			
			if (obj is Triangle)
			{
				Triangle t = (Triangle)obj;
				if (((this.V1 == t.V1)||(this.V1 == t.V2)||(this.V1==t.V3))&&
					((this.V2 == t.V1)||(this.V2 == t.V2)||(this.V2==t.V3))&&
					((this.V3 == t.V1)||(this.V3 == t.V2)||(this.V3==t.V3)))
					return(true);
			}
			return(false);
		}

		/// <summary>
		/// Hash code generator, uses XOR of triangle vertices to obtain integer value.
		/// </summary>
		/// <returns>Generated hashcode</returns>
		public override int GetHashCode()
		{
			return(this.V1*1000^this.V2*100^this.V3);
		}


	} // class Triangle
} // namespace
