using System;
using System.Collections.Generic;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Specialization of Mesh. Can contain only Triangle cells.
	/// Attributes are not limited.
	/// </summary>
	public class TriangleMesh : Zcu.Mve.Visualization.Mesh
	{
		/// <summary>
		/// Create empty TriangleMesh. Only Triangle is eigen cell.
		/// </summary>
		public TriangleMesh()
		{
			eigCells.Clear();
			eigCells.Add(typeof(Triangle), null);

      UpdateModificationTime();
		}

		/// <summary>
		/// Returns triangle from array named "Triangle".
		/// </summary>
		/// <param name="index">Index of the triangle.</param>
		/// <returns>Triangle with given index.</returns>
		public Triangle GetTriangle(int index)
		{
			return (Triangle) GetCells("Triangle")[index];
		}

		/// <summary>
		/// Returns number of triangles. (Number of elements in field named "Triangle")
		/// </summary>
		/// <returns>Number of triangles.</returns>
		public int GetNTriangles()
		{
			return GetCells("Triangle").Count;
		}

		/// <summary>
		/// Creates shallow copy of current mesh.
		/// Only array references are coppied.
		/// </summary>
		/// <returns>Shallow copy of current mesh.</returns>
		public override UnstrGrid ShallowCopy()
		{
			TriangleMesh outMesh = new TriangleMesh();
      
			// Shallow copy
      outMesh.Transform = this.Transform;
      outMesh.Points = this.Points;
			foreach (string cellid in this.GetCellIds())
			{
				outMesh.AddCells(cellid, this.GetCells(cellid));
				foreach (string attrid in this.GetCellAttrsIds(cellid))
				{
					outMesh.AddCellAttrs(cellid, attrid, this.GetCellAttrs(cellid, attrid));
				}
			}
			foreach (string attrid in this.GetPointAttrsIds())
			{
				outMesh.AddPointAttrs(attrid, this.GetPointAttrs(attrid));
			}

            foreach (string s in this.UserTags.Keys)
            {
                outMesh.UserTags.Add(s, this.UserTags[s]);
            }

			return outMesh;
		}

        public override bool CheckConsistence()
        {
            UniformDataArray triangles = this.GetCells(typeof(Triangle))[0];
            UniformDataArray points = this.Points;

            int openEdgeCount = 0;
            int nonManifoldEdgeCount = 0;
            int unusedVertexCount = 0;
            int degenerateTriangleCount = 0;

            // compute edge incidence
            Dictionary<Line, int> edges = new Dictionary<Line,int>();
            int[] usage = new int[points.Count];

            for (int i = 0; i < triangles.Count; i++)
            {
                Triangle t = (Triangle)triangles[i];
                usage[t.V1]++;
                usage[t.V2]++;
                usage[t.V3]++;
                if ((t.V1 == t.V2)||(t.V1 == t.V3)||(t.V2==t.V3))
                {
                    System.Console.WriteLine("Degenerate triangle {0} {1} {2} found.", t.V1, t.V2, t.V3);
                    degenerateTriangleCount++;
                }
                Line l = (t.V1 < t.V2) ? (new Line(t.V1, t.V2)) : (new Line(t.V2, t.V1));
                if (edges.ContainsKey(l)) edges[l] = edges[l] + 1;
                else edges[l] = 1;
                l = (t.V2 < t.V3) ? (new Line(t.V2, t.V3)) : (new Line(t.V3, t.V2));
                if (edges.ContainsKey(l)) edges[l] = edges[l] + 1;
                else edges[l] = 1;
                l = (t.V1 < t.V3) ? (new Line(t.V1, t.V3)) : (new Line(t.V3, t.V1));
                if (edges.ContainsKey(l)) edges[l] = edges[l] + 1;
                else edges[l] = 1;
            }
                
            foreach (Line line in edges.Keys)
            {
                if (edges[line] == 2)
                    continue;
                else if (edges[line] == 1)
                {
                    System.Console.WriteLine("Open edge {0} - {1} found.", line.V1, line.V2);
                    openEdgeCount++;
                }
                else
                {
                    System.Console.WriteLine("Non-manifold edge {0} - {1} with incidence {2} found.", line.V1, line.V2, edges[line]);
                    nonManifoldEdgeCount++;
                }
            }

            for (int i = 0; i < points.Count; i++)
            {
                if (usage[i] == 0)
                {
                    System.Console.WriteLine("Unused vertex {0} found.", i);
                    unusedVertexCount++;
                }
            }

            System.Console.WriteLine("Number of vertices: {0}", points.Count);
            System.Console.WriteLine("Number of edges: {0}", edges.Keys.Count);
            System.Console.WriteLine("Number of faces: {0}", triangles.Count);
            int eulerCharacteristic = edges.Keys.Count+2-points.Count-triangles.Count;
            System.Console.WriteLine("Euler characteristic (E+2-F-V): {0}", eulerCharacteristic);
            bool pass = true;
            if (openEdgeCount > 0)
            {
                System.Console.WriteLine("Found {0} open edges.", openEdgeCount);
                pass = false;
            }
            if (nonManifoldEdgeCount > 0)
            {
                System.Console.WriteLine("Found {0} non-manifold edges.", nonManifoldEdgeCount);
                pass = false;
            }
            if (degenerateTriangleCount > 0)
            {
                System.Console.WriteLine("Found {0} degenerate triangles.", degenerateTriangleCount);
                pass = false;
            }
            if (unusedVertexCount > 0)
            {
                System.Console.WriteLine("Found {0} unused vertices.", unusedVertexCount);
                pass = false;
            }
            if (eulerCharacteristic != 0)
            {
                System.Console.WriteLine("Euler characteristic for genus 0 meshes violated");
                pass = false;
            }

            return pass;            
        }

    } // class TriangleMesh
} // namespace