using System;
using Zcu.Mve.Core;
using System.Globalization;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// This structure represents Triangle neighbors.
	/// </summary>
	public struct TriangleNeighbor : INeighbor
	{
        /// <summary>
        /// TriangleNeighbor indexer.
        /// </summary>
        /// <param name="index">Index.</param>
        /// <returns>Neighbor.</returns>
        public int this[int index] {
            get
            {
                return Neighbors[index];
            }
            set
            {
                Neighbors[index] = value;
            }
        }
        
        /// <summary>
		/// Indices to three neighboring triangles.
		/// </summary>
		public int [] Neighbors;

		/// <summary>
		/// Number of neighbors. The neighbor cannot have negative value.
		/// </summary>
        public int Count
        {
            get
            {
                int i = 0;
                foreach (int index in this.Neighbors)
                {
                    if (index >= 0)
                    {
                        i++;
                    }
                }
                return i;
            }
        }

		/// <summary>
		/// Creates an object with values .
		/// </summary>
		/// <param name="t1">Index of neighbor.</param>
		/// <param name="t2">Index of neighbor.</param>
		/// <param name="t3">Index of neighbor.</param>
		/// <param name="count">Number of neighbors.</param>
		public TriangleNeighbor(int t1, int t2, int t3, int count)
		{
			this.Neighbors = new int[] {t1,t2,t3};    
		} // TriNeighbor()

		/// <summary>
		/// Creates an object with values.
		/// </summary>
		/// <param name="t1">Index of neighbor.</param>
		/// <param name="t2">Index of neighbor.</param>
		/// <param name="t3">Index of neighbor.</param>
		public TriangleNeighbor(int t1, int t2, int t3)
		{
			this.Neighbors = new int[] {t1,t2,t3};
		} // TriNeighbor()

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			string[] split = xmlTextReader.ReadString().Trim().Split(' ', '\t');
      
			if (split.Length != 3)
				throw new MveException(this, "Invalid TriangleNeighbour data format.");

			this.Neighbors[0] = Int32.Parse(split[0], nfi);
			this.Neighbors[1] = Int32.Parse(split[1], nfi);
			this.Neighbors[2] = Int32.Parse(split[2], nfi);
		} // ReadData()

		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			xmlTextWriter.WriteString(this.Neighbors[0].ToString(nfi) + " " + 
				this.Neighbors[1].ToString(nfi) + " " + 
				this.Neighbors[2].ToString(nfi));
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public IDataObject DeepCopy()
		{
			return new TriangleNeighbor(this.Neighbors[0], this.Neighbors[1], this.Neighbors[2], this.Count);
		} // DeepCopy()

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		override public string ToString()
		{
			return "TriangleNeighbor (" + this.Count + ") : " + this.Neighbors[0] + "; " + this.Neighbors[1] +  "; " + this.Neighbors[2] + "]";
		} // ToString()

		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public bool CheckConsistence()
		{
			if (this.Neighbors == null)
			{
				Console.WriteLine("TriangleNeighbor: Neighbours array is null.");
				return false;
			}
			if (this.Neighbors.Length != 3)
			{
				Console.WriteLine("TriangleNeighbor have incorrect number of neighbours.");
				return false;
			}
			int i = 0;
			foreach (int index in this.Neighbors)
			{
				if (index >= 0)
				{
					i++;
				}
			}
			if (i != this.Count)
			{
				Console.WriteLine("Number of valid indices do not match Count value.");
				return false;
			}

			return true;
		} // CheckConsistence()

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

	} // TriangleNeighbor
} // namespace
