using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Diagnostics;
using System.ComponentModel;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Types of cells
    /// </summary>
    public enum CellTypes
    {
        /// <summary>
        /// Line.
        /// </summary>
        Line,
        /// <summary>
        /// PolyLine.
        /// </summary>
        PolyLine,
        /// <summary>
        /// Polygon.
        /// </summary>
        Polygon,
        /// <summary>
        /// Slice.
        /// </summary>
        Slice,
        /// <summary>
        /// Triangle.
        /// </summary>
        Triangle,
        /// <summary>
        /// TriangleFan.
        /// </summary>
        TriangleFan,
        /// <summary>
        /// TriangleStrip.
        /// </summary>
        TriangleStrip,
        /// <summary>
        /// Tetrahedron.
        /// </summary>
        Tetrahedron
    }
    /// <summary>
    /// Module which converts lines into tubes and points into spheres.
    /// </summary>
    [ModuleInfo("Zdenk Prokop", "Converts lines into tubes and points into spheres.", IconName = "Tube.ico")]
    [PortInfo("Input", "Input data - Points and lines.")]
    [PortInfo("Output", "Output data - triangle mesh.")]
    class Tube : Module
    {
        private bool paramsChanged = true;
        private bool addInput = false;
        private bool convertPoint = true;
        private bool convertLine = true;
        private int sphereNumberOfSides = 16;
        private int tubeNumberOfSides = 32;
        private double sphereRadius = 0.05;
        private double tubeRadius = 0.05;
        private string cellId = "";
        private string cellName = "triangles";
        private string normalName = "Normal3D";
        private CellTypes cellType = CellTypes.Line;
        private UnstrGrid result;
        private UnstrGrid input;
        private int pointCount = 0;
        private int inputPointCount = 0;
        private int inputTriangleCount = 0;
        private int spherePointCount;
        private int sphereTriangleCount;
        private int tubePointCount;
        private int tubeTriangleCount;
        private int allSpherePointCount;
        private int allSphereTriangleCount;
        private UniformDataArray points;
        private UniformDataArray newPoints;
        private UniformDataArray newTriangles;
        private UniformDataArray newNormals;
        private UniformDataArray[] pointAttrs;
        private UniformDataArray[] newPointAttrs;

        /// <summary>
        /// Constructor. Creates input and output ports.
        /// </summary>
        public Tube()
        {
            AddInPort("Input", typeof(UnstrGrid));
            AddOutPort("Output", typeof(UnstrGrid));
        }

        /// <summary>
        /// Executes module functionality.
        /// </summary>
        public override void Execute()
        {
            if (paramsChanged || !IsAllDataSame())
            {
                input = (UnstrGrid)GetInput("Input");
                UniformDataArray cells = null;

                spherePointCount = 2 * sphereNumberOfSides * (sphereNumberOfSides - 1) + 2;
                sphereTriangleCount = (sphereNumberOfSides - 1) * sphereNumberOfSides * 4;
                int linesCount = 0;
                points = input.Points;
                ICollection<string> attrsList = input.GetPointAttrsIds();
                string[] attrs = new string[attrsList.Count];
                attrsList.CopyTo(attrs, 0);
                pointAttrs = new UniformDataArray[attrs.Length];
                for (int i = 0; i < pointAttrs.Length; i++)
                    pointAttrs[i] = input.GetPointAttrs(attrs[i]);

                if (convertPoint)
                {
                    allSpherePointCount = points.Count * spherePointCount;
                    allSphereTriangleCount = points.Count * sphereTriangleCount;
                }
                else
                {
                    allSpherePointCount = 0;
                    allSphereTriangleCount = 0;
                }
                if (convertLine)
                {
                    if (cellId == "")
                    {
                        ICollection<string> cellsList = input.GetCellIds();
                        if (cellsList.Count > 0)
                        {
                            string[] cellsArray = new string[cellsList.Count];
                            cellsList.CopyTo(cellsArray, 0);
                            cells = input.GetCells(cellsArray[0]);
                        }
                        else
                        {
                            throw new MveException(this, "UnstrGrid does not contain any cells.");
                        }
                    }
                    else
                    {
                        cells = input.GetCells(cellId);
                    }
                    tubePointCount = tubeNumberOfSides * 2;
                    tubeTriangleCount = tubeNumberOfSides * 2;

                    //cells type detection
                    if (cells.ElemType.Equals(typeof(Line)))
                        cellType = CellTypes.Line;
                    else if (cells.ElemType.Equals(typeof(Polygon)))
                        cellType = CellTypes.Polygon;
                    else if (cells.ElemType.Equals(typeof(PolyLine)))
                        cellType = CellTypes.PolyLine;
                    else if (cells.ElemType.Equals(typeof(Slice)))
                        cellType = CellTypes.Slice;
                    else if (cells.ElemType.Equals(typeof(Tetrahedron)))
                        cellType = CellTypes.Tetrahedron;
                    else if (cells.ElemType.Equals(typeof(Triangle)))
                        cellType = CellTypes.Triangle;
                    else if (cells.ElemType.Equals(typeof(TriangleFan)))
                        cellType = CellTypes.TriangleFan;
                    else if (cells.ElemType.Equals(typeof(TriangleStrip)))
                        cellType = CellTypes.TriangleStrip;
                    else
                        throw new MveException(this, "Unknown type of cell.");

                    //compute linesCount
                    switch (cellType)
                    {
                        //lines
                        case CellTypes.Line:
                            linesCount += cells.Count;
                            break;
                        //polygons
                        case CellTypes.Polygon:
                            for (int i = 0; i < cells.Count; i++)
                            {
                                linesCount += ((Polygon)cells[i]).PointCount;
                            }
                            break;
                        //polylines
                        case CellTypes.PolyLine:
                            for (int i = 0; i < cells.Count; i++)
                            {
                                linesCount += ((PolyLine)cells[i]).PointCount - 1;
                            }
                            break;
                        //slices
                        case CellTypes.Slice:
                            for (int i = 0; i < cells.Count; i++)
                            {
                                Slice s = (Slice)cells[i];
                                foreach (PolyLine pl in s.PolyLineList)
                                {
                                    linesCount += pl.PointCount - 1;
                                }
                            }
                            break;
                        //tetrahedrons
                        case CellTypes.Tetrahedron:
                            linesCount = 6 * cells.Count;
                            break;
                        //triangles
                        case CellTypes.Triangle:
                            linesCount = 3 * cells.Count;
                            break;
                        //triangleFans
                        case CellTypes.TriangleFan:
                            for (int i = 0; i < cells.Count; i++)
                            {
                                int pc = ((TriangleFan)cells[i]).PointCount;
                                if (pc > 2) linesCount += 3 * (pc - 2);
                            }
                            break;
                        //triangleStrips
                        case CellTypes.TriangleStrip:
                            for (int i = 0; i < cells.Count; i++)
                            {
                                int pc = ((TriangleStrip)cells[i]).PointCount;
                                if (pc > 2) linesCount += 3 * (pc - 2);
                            }
                            break;
                    }
                }
                inputPointCount = 0;
                inputTriangleCount = 0;
                if (addInput)
                {
                    inputPointCount = points.Count;
                    if (input.ContainsCells(cellName))
                        inputTriangleCount = input.GetCells(cellName).Count;
                }
                newPoints = new UniformDataArray(typeof(Point3D), inputPointCount + allSpherePointCount + linesCount * tubePointCount);
                newTriangles = new UniformDataArray(typeof(Triangle), inputTriangleCount + allSphereTriangleCount + linesCount * tubeTriangleCount);
                newNormals = new UniformDataArray(typeof(Normal3D), newPoints.Count);
                newPointAttrs = new UniformDataArray[pointAttrs.Length];
                for (int i = 0; i < pointAttrs.Length; i++)
                {
                    newPointAttrs[i] = new UniformDataArray(pointAttrs[i].ElemType, newPoints.Count);
                }
                //create spheres
                if (convertPoint)
                {
                    for (int i = 0; i < points.Count; i++)
                    {
                        CreateSphere((Point3D)points[i], newPoints, inputPointCount + i * spherePointCount, newTriangles, inputTriangleCount + i * sphereTriangleCount, newNormals);
                        //point attributes assignment
                        for (int k = 0; k < pointAttrs.Length; k++)
                        {
                            for (int j = 0; j < spherePointCount; j++)
                                newPointAttrs[k][inputPointCount + i * spherePointCount + j] = pointAttrs[k][i];
                        }
                        ProgressInfo((double)i / points.Count, 0, 2);
                    }
                }

                //create tubes
                if (convertLine)
                {
                    switch (cellType)
                    {
                        //lines
                        case CellTypes.Line:
                            pointCount = 0;
                            for (int i = 0; i < cells.Count; i++)
                            {
                                Line l = (Line)cells[i];
                                LineToTube(l.V1, l.V2);
                                ProgressInfo((double)i / cells.Count, 1, 2);
                            }

                            break;
                        //polygons
                        case CellTypes.Polygon:
                            pointCount = 0;
                            for (int i = 0; i < cells.Count; i++)
                            {
                                Polygon p = (Polygon)cells[i];
                                if (p.PointCount > 1)
                                {
                                    LineToTube(p.GetPoint(p.PointCount - 1), p.GetPoint(0));
                                    for (int k = 1; k < p.PointCount; k++)
                                    {
                                        LineToTube(p.GetPoint(k - 1), p.GetPoint(k));
                                    }
                                }
                                ProgressInfo((double)i / cells.Count, 1, 2);
                            }
                            break;
                        //polylines
                        case CellTypes.PolyLine:
                            pointCount = 0;
                            for (int i = 0; i < cells.Count; i++)
                            {
                                PolyLine pl = (PolyLine)cells[i];
                                for (int k = 1; k < pl.PointCount; k++)
                                {
                                    LineToTube(pl.GetPoint(k - 1), pl.GetPoint(k));
                                }
                                ProgressInfo((double)i / cells.Count, 1, 2);
                            }
                            break;
                        //slices
                        case CellTypes.Slice:
                            pointCount = 0;
                            for (int i = 0; i < cells.Count; i++)
                            {
                                Slice s = (Slice)cells[i];
                                foreach (PolyLine pl in s.PolyLineList)
                                {
                                    for (int k = 1; k < pl.PointCount; k++)
                                    {
                                        LineToTube(pl.GetPoint(k - 1), pl.GetPoint(k));
                                    }
                                }
                                ProgressInfo((double)i / cells.Count, 1, 2);
                            }
                            break;
                        //tetrahedrons
                        case CellTypes.Tetrahedron:
                            pointCount = 0;
                            for (int i = 0; i < cells.Count; i++)
                            {
                                Tetrahedron t = (Tetrahedron)cells[i];
                                LineToTube(t.V1, t.V2);
                                LineToTube(t.V1, t.V3);
                                LineToTube(t.V1, t.V4);
                                LineToTube(t.V2, t.V3);
                                LineToTube(t.V2, t.V4);
                                LineToTube(t.V3, t.V4);
                                ProgressInfo((double)i / cells.Count, 1, 2);
                            }
                            break;
                        //triangles
                        case CellTypes.Triangle:
                            pointCount = 0;
                            for (int i = 0; i < cells.Count; i++)
                            {
                                Triangle t = (Triangle)cells[i];
                                LineToTube(t.V1, t.V2);
                                LineToTube(t.V1, t.V3);
                                LineToTube(t.V2, t.V3);
                                ProgressInfo((double)i / cells.Count, 1, 2);
                            }
                            break;
                        //triangleFans
                        case CellTypes.TriangleFan:
                            pointCount = 0;
                            for (int i = 0; i < cells.Count; i++)
                            {
                                TriangleFan tf = (TriangleFan)cells[i];
                                for (int k = 2; k < tf.PointCount; k++)
                                {
                                    LineToTube(tf.GetPoint(k), tf.GetPoint(0));
                                    LineToTube(tf.GetPoint(k - 1), tf.GetPoint(0));
                                    LineToTube(tf.GetPoint(k - 1), tf.GetPoint(k));
                                }
                                ProgressInfo((double)i / cells.Count, 1, 2);
                            }
                            break;
                        //triangleStrips
                        case CellTypes.TriangleStrip:
                            pointCount = 0;
                            for (int i = 0; i < cells.Count; i++)
                            {
                                TriangleStrip ts = (TriangleStrip)cells[i];
                                for (int k = 2; k < ts.PointCount; k++)
                                {
                                    LineToTube(ts.GetPoint(k - 2), ts.GetPoint(k - 1));
                                    LineToTube(ts.GetPoint(k - 1), ts.GetPoint(k));
                                    LineToTube(ts.GetPoint(k - 2), ts.GetPoint(k));
                                }
                                ProgressInfo((double)i / cells.Count, 1, 2);
                            }
                            break;
                    }
                }

                //set result
                if (convertPoint || convertLine)
                {
                    if (addInput)
                    {
                        result = input.ShallowCopy();
                        UniformDataArray uda;
                        for (int i = 0; i < inputPointCount; i++)
                        {
                            newPoints[i] = points[i];
                            for (int j = 0; j < pointAttrs.Length; j++)
                            {
                                newPointAttrs[j][i] = pointAttrs[j][i];
                            }
                        }
                        result.Points = newPoints;
                        if (input.ContainsCells(cellName))
                        {
                            uda = input.GetCells(cellName);
                            for (int i = 0; i < inputTriangleCount; i++)
                            {
                                newTriangles[i] = uda[i];
                            }
                        }
                        if (result.ContainsCells(cellName))
                            result.RemoveCells(cellName);
                        result.AddCells(cellName, newTriangles);
                        for (int i = 0; i < newPointAttrs.Length; i++)
                        {
                            if (!result.ContainsPointAttrs(attrs[i]))
                                result.AddPointAttrs(attrs[i], newPointAttrs[i]);
                        }
                        if (result.ContainsPointAttrs(normalName))
                        {
                            uda = result.GetPointAttrs(normalName);
                            for (int i = 0; i < inputPointCount; i++)
                            {
                                Normal3D n = (Normal3D) uda[i];
                                newNormals[i] = new Normal3D(n.Nx, n.Ny, n.Nz);
                            }
                            uda = null;
                            result.RemovePointAttrs(normalName);
                        }
                        result.AddPointAttrs(normalName, newNormals);
                    }
                    else
                    {
                        result = new UnstrGrid();
                        result.Points = newPoints;
                        result.AddCells(cellName, newTriangles);
                        for (int i = 0; i < newPointAttrs.Length; i++)
                            result.AddPointAttrs(attrs[i], newPointAttrs[i]);
                        if (result.ContainsPointAttrs(normalName))
                        {
                            result.RemovePointAttrs(normalName);
                        }
                        result.AddPointAttrs(normalName, newNormals);
                    }
                    //System.Console.WriteLine(string.Format("Input: {0} points, {1} lines\nOutput: {2} points, {3} triangles", input.Points.Count, linesCount, result.Points.Count, newTriangles.Count));
                }
                else
                {
                    result = input.ShallowCopy();
                }
                paramsChanged = false;
            }
            SetOutput("Output", result);
        }

        /// <summary>
        /// This method make a tube from a line
        /// </summary>
        /// <param name="p1">index of first end-point of the line</param>
        /// <param name="p2">index of second end-point of the line</param>
        private void LineToTube(int p1, int p2)
        {
            //compute startIndex
            int startIndex = inputPointCount + allSpherePointCount + pointCount * tubePointCount;
            //create tube
            CreateTube((Point3D)points[p1], (Point3D)points[p2], newPoints, startIndex, newTriangles, inputTriangleCount + allSphereTriangleCount + pointCount * tubeTriangleCount, newNormals);

            //point attributes assignment
            for (int i = 0; i < pointAttrs.Length; i++)
            {
                for (int j = 0; j < tubePointCount / 2; j++)
                    newPointAttrs[i][startIndex + j] = pointAttrs[i][p1];
                for (int j = tubePointCount / 2; j < tubePointCount; j++)
                    newPointAttrs[i][startIndex + j] = pointAttrs[i][p2];
            }
            pointCount++;
        }

        /// <summary>
        /// Create the sphere from a point
        /// </summary>
        /// <param name="centre">point</param>
        /// <param name="points">array of points for new points of this sphere</param>
        /// <param name="pointsStartIndex">start index in array of points for insert new points</param>
        /// <param name="triangles">array of triangles for new triangles of this tube</param>
        /// <param name="trianglesStartIndex">start index in array of triangles for insert new triangles</param>
        /// <param name="normals">array of normals for new points of this sphere</param>
        private void CreateSphere(Point3D centre, UniformDataArray points, int pointsStartIndex, UniformDataArray triangles, int trianglesStartIndex, UniformDataArray normals)
        {
            //angles for steps
            double phiStep = Math.PI / sphereNumberOfSides;
            double thetaStep = Math.PI / sphereNumberOfSides;
            //start angles
            double phi = 0;
            double theta = -Math.PI / 2;
            //polar coordinates of the first pole
            points[pointsStartIndex] = new Point3D(centre.X, centre.Y - sphereRadius, centre.Z);
            //normal is the direction vector
            normals[pointsStartIndex] = new Normal3D(0.0, -1.0, 0.0);

            //creation of points
            for (int x = 0; x < 2 * sphereNumberOfSides; x++)
            {
                theta = -Math.PI / 2;
                for (int y = 0; y < (sphereNumberOfSides - 1); y++)
                {
                    theta += thetaStep;
                    //polar coordinates
                    points[pointsStartIndex + 1 + x * (sphereNumberOfSides - 1) + y] = new Point3D(centre.X + (sphereRadius * Math.Cos(phi) * Math.Cos(theta)), centre.Y + (sphereRadius * Math.Sin(theta)), centre.Z + (sphereRadius * Math.Sin(phi) * Math.Cos(theta)));
                    //normal is the direction vector
                    normals[pointsStartIndex + 1 + x * (sphereNumberOfSides - 1) + y] = new Normal3D(Math.Cos(phi) * Math.Cos(theta), Math.Sin(theta), Math.Sin(phi) * Math.Cos(theta));

                }
                phi += phiStep;
            }
            //polar coordinates of the second pole
            points[pointsStartIndex + spherePointCount - 1] = new Point3D(centre.X, centre.Y + sphereRadius, centre.Z);
            //normal is the direction vector
            normals[pointsStartIndex + spherePointCount - 1] = new Normal3D(0.0, 1.0, 0.0);

            //creation of triangles
            int index = trianglesStartIndex;
            int vbindex;
            for (int x = 0; x < (2 * sphereNumberOfSides - 1); x++)
                for (int y = 0; y < sphereNumberOfSides - 2; y++)
                {
                    vbindex = pointsStartIndex + 1 + x * (sphereNumberOfSides - 1) + y;
                    triangles[index++] = new Triangle(vbindex, vbindex + sphereNumberOfSides - 1, vbindex + sphereNumberOfSides);
                    triangles[index++] = new Triangle(vbindex, vbindex + sphereNumberOfSides, vbindex + 1);
                }
            for (int y = 0; y < sphereNumberOfSides - 2; y++)
            {
                vbindex = pointsStartIndex + 1 + (2 * sphereNumberOfSides - 1) * (sphereNumberOfSides - 1) + y;
                triangles[index++] = new Triangle(vbindex, pointsStartIndex + 1 + y, pointsStartIndex + 2 + y);
                triangles[index++] = new Triangle(vbindex, pointsStartIndex + 2 + y, vbindex + 1);
            }
            for (int x = 0; x < 2 * sphereNumberOfSides - 1; x++)
            {
                vbindex = pointsStartIndex + 1 + x * (sphereNumberOfSides - 1);
                triangles[index++] = new Triangle(pointsStartIndex, vbindex + sphereNumberOfSides - 1, vbindex);
                vbindex += sphereNumberOfSides - 2;
                triangles[index++] = new Triangle(pointsStartIndex + spherePointCount - 1, vbindex, vbindex + sphereNumberOfSides - 1);
            }
            triangles[index++] = new Triangle(pointsStartIndex, pointsStartIndex + 1, pointsStartIndex + 1 + (2 * sphereNumberOfSides - 1) * (sphereNumberOfSides - 1));
            triangles[index++] = new Triangle(pointsStartIndex + spherePointCount - 1, pointsStartIndex + spherePointCount - 2, pointsStartIndex + sphereNumberOfSides - 1);
        }

        /// <summary>
        /// Create the tube from two points
        /// </summary>
        /// <param name="p1">first end-point of the line</param>
        /// <param name="p2">second end-point of the line</param>
        /// <param name="points">array of points for new points of this tube</param>
        /// <param name="pointsStartIndex">start index in array of points for insert new points</param>
        /// <param name="triangles">array of triangles for new triangles of this tube</param>
        /// <param name="trianglesStartIndex">start index in array of triangles for insert new triangles</param>
        /// <param name="normals">array of normals for new points of this tube</param>
        private void CreateTube(Point3D p1, Point3D p2, UniformDataArray points, int pointsStartIndex, UniformDataArray triangles, int trianglesStartIndex, UniformDataArray normals)
        {
            double dx = p1.X - p2.X;
            double dy = p1.Y - p2.Y;
            double dz = p1.Z - p2.Z;
            double l = Math.Sqrt(dx * dx + dz * dz);
            double alfa;
            if (l != 0)
                alfa = -Math.Asin(dx / l);
            else
                alfa = 0;

            double beta = Math.Asin(dy / Math.Sqrt(dx * dx + dy * dy + dz * dz));
            if (dz > 0)
                alfa = Math.PI - alfa;
            //transform for the first point
            Transform3D t1 = new Transform3D();
            t1.AddTranslation((float)p1.X, (float)p1.Y, (float)p1.Z);
            t1.AddRotationY((float)alfa);
            t1.AddRotationX((float)beta);
            //transform for the second point
            Transform3D t2 = new Transform3D();
            t2.AddTranslation((float)p2.X, (float)p2.Y, (float)p2.Z);
            t2.AddRotationY((float)alfa);
            t2.AddRotationX((float)beta);
            //transform for the normal
            Transform3D tNorm = new Transform3D();
            tNorm.AddRotationY((float)alfa);
            tNorm.AddRotationX((float)beta);

            double phiStep = 2 * Math.PI / tubeNumberOfSides;
            double phi = 0;
            Point3D p, n;

            //creation of points - first a circle with centre in coordinate origin is created and then the circle is rotated and translated 
            for (int i = 0; i < tubeNumberOfSides; i++)
            {
                p = new Point3D(tubeRadius * Math.Cos(phi), tubeRadius * Math.Sin(phi), 0.0);
                n = new Point3D(Math.Cos(phi), Math.Sin(phi), 0.0);
                n = tNorm.Transform(n);
                normals[pointsStartIndex + i] = new Normal3D(n.X, n.Y, n.Z);
                normals[pointsStartIndex + i + tubeNumberOfSides] = normals[pointsStartIndex + i];
                points[pointsStartIndex + i + tubeNumberOfSides] = new Point3D(p.X, p.Y, p.Z);
                points[pointsStartIndex + i] = t1.Transform(p);
                points[pointsStartIndex + i + tubeNumberOfSides] = t2.Transform((Point3D)points[pointsStartIndex + i + tubeNumberOfSides]);
                phi += phiStep;
            }

            //creation of triangles
            int lastIndex = tubeNumberOfSides - 1;
            for (int i = 0; i < tubeNumberOfSides; i++)
            {
                triangles[trianglesStartIndex + i * 2] = new Triangle(pointsStartIndex + i, pointsStartIndex + i + tubeNumberOfSides, pointsStartIndex + lastIndex);
                triangles[trianglesStartIndex + i * 2 + 1] = new Triangle(pointsStartIndex + lastIndex, pointsStartIndex + i + tubeNumberOfSides, pointsStartIndex + lastIndex + tubeNumberOfSides);
                lastIndex = i;
            }
        }


        #region Properties

        /// <summary>
        /// Property for convertPoint.
        /// </summary>
        [Category("Sphere")]
        [Description("If this property is set to true, the source points will be converted to spheres. Otherwise they won't be converted.")]
        [DefaultValue(true)]
        public bool ConvertPoints
        {
            get { return (this.convertPoint); }
            set
            {
                this.convertPoint = value;
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Property for sphereNumberOfSides variable.
        /// </summary>
        [Category("Sphere")]
        [Description("The value of this property indicates the number of faces of the meridian. The value must be greater than or equal to 2.")]
        [DefaultValue(16)]
        public int SphereNumberOfSides
        {
            get { return (this.sphereNumberOfSides); }
            set
            {
                if (value >= 2)
                {
                    this.sphereNumberOfSides = value;
                }
                else
                {
                    this.sphereNumberOfSides = 32;
                }
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Property for sphere radius variable.
        /// </summary>
        [Category("Sphere")]
        [Description("The value of this property sets the radius of the sphere.")]
        [DefaultValue(0.05)]
        public double SphereRadius
        {
            get { return (this.sphereRadius); }
            set
            {
                if (value > 0)
                {
                    this.sphereRadius = value;
                }
                else
                {
                    this.sphereRadius = tubeRadius;
                }
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Property for convertLine.
        /// </summary>
        [Category("Tube")]
        [Description("If this property is set to true, the source lines will be converted to tubes. Otherwise they won't be converted.")]
        [DefaultValue(true)]
        public bool ConvertLines
        {
            get { return (this.convertLine); }
            set
            {
                this.convertLine = value;
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Property for tubeNumberOfSides variable.
        /// </summary>
        [Category("Tube")]
        [Description("The value of this property indicates the number of faces around the circumference of the tube. The value must be greater than or equal to 3.")]
        [DefaultValue(32)]
        public int TubeNumberOfSides
        {
            get { return (this.tubeNumberOfSides); }
            set
            {
                if (value >= 3)
                {
                    this.tubeNumberOfSides = value;
                }
                else
                {
                    this.tubeNumberOfSides = 32;
                }
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Property for tube radius variable.
        /// </summary>
        [Category("Tube")]
        [Description("The value of this property sets the radius of the tube.")]
        [DefaultValue(0.05)]
        public double TubeRadius
        {
            get { return (this.tubeRadius); }
            set
            {
                if (value > 0)
                {
                    this.tubeRadius = value;
                }
                else
                {
                    this.tubeRadius = sphereRadius;
                }
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Property for cellId.
        /// </summary>
        [Category("Input")]
        [Description("The property determines name of cell which is processed.")]
        [DefaultValue("streamlines")]
        public string InputCellName
        {
            get
            {
                return cellId;
            }
            set
            {
                cellId = value;
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Property for addInput.
        /// </summary>
        [Category("Output")]
        [Description("If this property is set to true, the input data will be added to output. Otherwise output won't contain input data")]
        [DefaultValue(false)]
        public bool AddInput
        {
            get
            {
                return addInput;
            }
            set
            {
                addInput = value;
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Property for normalName.
        /// </summary>
        [Category("Output")]
        [Description("The property determines name of point attribute which contains normal in the point.")]
        [DefaultValue("Normal3D")]
        public string NormalAttributeName
        {
            get
            {
                return normalName;
            }
            set
            {
                normalName = value;
                this.paramsChanged = true;
            }
        }

        /// <summary>
        /// Property for cellName.
        /// </summary>
        [Category("Output")]
        [Description("The property determines name of cell type.")]
        [DefaultValue("triangles")]
        public string CellName
        {
            get
            {
                return cellName;
            }
            set
            {
                cellName = value;
                this.paramsChanged = true;
            }
        }

        #endregion
    }
}
