using System;
using System.Collections.Generic;
using System.Text;
using System.Globalization;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
    /// <summary>
    /// This structure represents Vertex neighbors.
    /// </summary>
    public struct VertexNeighbor : INeighbor
    {
        /// <summary>
        /// VertexNeighbor indexer.
        /// </summary>
        /// <param name="index">Index.</param>
        /// <returns>Neighbor.</returns>
        public int this[int index]
        {
            get
            {
                if (closed)
                {
                    return neighbors[index];
                }
                else
                {
                    return neighborsList[index];
                }
            }
            set
            {
                if (closed)
                {
                    neighbors[index] = value;
                }
                else
                {
                    neighborsList[index] = value;
                }
            }
        }


        /// <summary>
        /// Indices to neighboring vertices.
        /// </summary>
        public int[] Neighbors
        {
            get
            {
                if (closed)
                {
                    return neighbors;
                }
                else if (neighborsList != null)
                {
                    return neighborsList.ToArray();
                }
                else
                {
                    return new int[0];
                }
            }
        }

        /// <summary>
        /// Array of neighbors
        /// </summary>
        int[] neighbors;

        /// <summary>
        /// List of neighbors. If the list is complete, call Close() to transform
        /// the list to array and free the memory.
        /// </summary>
        List<int> neighborsList;

        /// <summary>
        /// Returns the status of VertexNeighbor list. If it is closed, it is not possible to use
        /// Add operation.
        /// </summary>
        public bool IsClosed
        {
            get { return closed; }
        }
        bool closed;

        /// <summary>
        /// Close the VertexNeighbor list, converts the list to an array and frees the memory.
        /// It will not be possible to add another neighbor.
        /// </summary>
        public void Close()
        {
            if (closed) return;
            if (neighborsList != null)
            {
                neighbors = neighborsList.ToArray();
                neighborsList.Clear();
                neighborsList = null;
            }
            else
            {
                neighbors = new int[0];
            }
            closed = true;
        }

        /// <summary>
        /// Adds a new neighbor to the list. If the list is already closed, the exception
        /// is thrown.
        /// </summary>
        /// <param name="neighbor"></param>
        public void Add(int neighbor)
        {
            if (closed) throw new MveException("The VertexNeighbor list is closed. The Add operation is not allowed.");
            if (neighborsList == null) neighborsList = new List<int>();
            neighborsList.Add(neighbor);
        }

        /// <summary>
        /// Number of neighbors.
        /// </summary>
        public int Count
        {
            get
            {
                if (closed) return neighbors.Length;
                else if (neighborsList != null) return neighborsList.Count;
                else return 0;
            }
        }

        /// <summary>
        /// Creates an object with values.
        /// </summary>
        /// <param name="neighbors">Array of neighbor indices.</param>
        public VertexNeighbor(int[] neighbors)
        {
            this.neighbors = neighbors;
            closed = true;
            this.neighborsList = null;
        }

        /// <summary>
        /// Read data from Xml file to current structure.
        /// </summary>
        /// <param name="xmlTextReader">Stream to be read.</param>
        public void ReadData(System.Xml.XmlTextReader xmlTextReader)
        {
            NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
            string[] split = xmlTextReader.ReadString().Trim().Split(' ');
            this.neighbors = new int[split.Length];
            closed = true;
            for (int i = 0; i < split.Length; i++)
                this.Neighbors[i] = Int32.Parse(split[i], nfi);
        } // ReadData()

        /// <summary>
        /// Write current data to Xml file.
        /// </summary>
        /// <param name="xmlTextWriter">Stream to be write.</param>
        public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
        {
            this.Close();
            NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
            string s = "";
            for (int i = 0; i < this.neighbors.Length; i++)
                s += this.neighbors[i] + " ";

            xmlTextWriter.WriteString(s);
        } // WriteData()

        /// <summary>
        /// Deep copy of data object and its subelements.
        /// </summary>
        /// <returns>Same data, different memory space.</returns>
        public IDataObject DeepCopy()
        {
            VertexNeighbor n = new VertexNeighbor();
            n.closed = closed;
            if (closed)
            {
                n.neighbors = new int[this.neighbors.Length];
                for (int i = 0; i < this.neighbors.Length; i++)
                    n.neighbors[i] = this.neighbors[i];
            }
            else
            {
                n.neighborsList = new List<int>();
                for (int i = 0; i < this.neighborsList.Count; i++)
                    n.neighborsList.Add(this.neighborsList[i]);
            }
            return n;
        } // DeepCopy()

        /// <summary>
        /// Returns a String that represents the current Object.
        /// </summary>
        /// <returns>String that represents the current Object.</returns>
        public override string ToString()
        {
            string s = "";
            int[] n = this.Neighbors;
            for (int i = 0; i < n.Length; i++)
                s += n[i] + "; ";
            return "VertexNeighbors: [" + s.Substring(s.Length - 2, 2) + "]";
        } // ToString()

        /// <summary>
        /// Check the consistency of current data structure.
        /// </summary>
        /// <returns>True if structure is consistent.</returns>
        public bool CheckConsistence()
        {
            return true;
        } // CheckConsistence()

        /// <summary>
        /// Returns array of assembly name.
        /// </summary>
        public string[] LibNames
        {
            get
            {
                return new string[] { Globals.GetLibraryName(this.GetType()) };
            }
        }

    } // VertexNeighbor
} // namespace

