﻿using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.Globalization;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Module for cutting by plane
    /// Module calculates distances from specified plane for all the grid.
    /// </summary>    
    [ModuleInfo("Roman Soukal", "Module for plane distance calculate used to cutting", IconName = "VolumeCut2.ico", Category = "Zcu.Mve.Gsvd")]
    [PortInfo("Input Grid", "Input RegGridND data")]
    [PortInfo("Input Point1", "First Input PointND (Size N=3 is needed) to calculate a distance (CCW orientation of points)")]
    [PortInfo("Input Point2", "Second Input PointND (Size N=3 is needed) to calculate a distance (CCW orientation of points)")]
    [PortInfo("Input Point3", "Third Input PointND (Size N=3 is needed) to calculate a distance (CCW orientation of points)")]
    [PortInfo("Output Grid", "Output RegGridND data")]
    public class PlaneCutting : Module
    {
        /// <summary>
        /// Default constructor.
        /// </summary>
        public PlaneCutting()
        {
            AddInPort("Input Grid", typeof(RegGridND));
            AddInPort("Input Point1", typeof(PointND), false);
            AddInPort("Input Point2", typeof(PointND), false);
            AddInPort("Input Point3", typeof(PointND), false);
            AddOutPort("Output Grid", typeof(RegGridND));
        }

        #region Variables
        private RegGridND data = null;
        private string pointAttributesName = "Values";
        private string cuttingAttributesName = "CuttingValues";
        private bool paramsChanged = false;
        private string plane = null;
        private double[] planeCoords = null;
        #endregion

        /// <summary>
        /// Overrided
        /// </summary> 
        public override void Execute()
        {
            if (!IsAllDataSame() || paramsChanged)
            {
                RegGridND old = (RegGridND)this.GetInput("Input Grid");
                data = old.ShallowCopy();
                //IDataObject val = null;
                //if (!data.UserTags.TryGetValue("Min", out val))
                //{
                //    old.UserTags.TryGetValue("Min", out val);
                //    data.UserTags.Add("Min", val);
                //}
                //if (!data.UserTags.TryGetValue("Max", out val))
                //{
                //    old.UserTags.TryGetValue("Max", out val);
                //    data.UserTags.Add("Max", val);
                //}
                //if (!data.UserTags.TryGetValue("Distortion", out val))
                //{
                //    old.UserTags.TryGetValue("Distortion", out val);
                //    data.UserTags.Add("Distortion", val);
                //}

                PointND point1 = new PointND();
                PointND point2 = new PointND();
                PointND point3 = new PointND();
                if (this.GetInput("Input Point1") != null)
                {
                    point1 = (PointND)this.GetInput("Input Point1");
                    if (point1.Length != 3)
                        throw new MveException(this, "Size N=3 of PointND 1 is expected!");
                }
                if (this.GetInput("Input Point2") != null)
                {
                    point2 = (PointND)this.GetInput("Input Point2");
                    if (point2.Length != 3)
                        throw new MveException(this, "Size N=3 of PointND 2 is expected!");
                }
                if (this.GetInput("Input Point3") != null)
                {
                    point3 = (PointND)this.GetInput("Input Point3");
                    if (point3.Length != 3)
                        throw new MveException(this, "Size N=3 of PointND 3 is expected!");
                }

                MveArray values = data.GetPointAttrs(this.PointAttributeName);
                MveArray newValues = new MveArray(typeof(Scalar), values.Lengths);
                if (this.GetInput("Input Point1") == null || this.GetInput("Input Point2") == null || this.GetInput("Input Point3") == null)
                {
                    if (planeCoords != null)
                    {
                        int[] sizes = values.Lengths;
                        for (int i = 0; i < sizes[0]; i++)
                        {
                            this.ProgressInfo(((float)i / sizes[0]));
                            for (int j = 0; j < sizes[1]; j++)
                                for (int k = 0; k < sizes[2]; k++)
                                {
                                    int[] coords = { i, j, k };
                                    PointND pValue = data.GetPoint(coords);
                                    double distance = planeCoords[0] * pValue.Values[0] + planeCoords[1] * pValue.Values[1] + planeCoords[2] * pValue.Values[2] + planeCoords[3];
                                    newValues.Array.SetValue(new Scalar(distance), coords);
                                }
                        }
                    }
                    else
                        throw new MveException(this, "All points or plane coordinates must be set!");
                }
                else
                {
                    if (point1.Equals(point2) || point1.Equals(point3) || point2.Equals(point3))
                        throw new MveException(this, "Any Point generated the plane cannot be the same like any other!");

                    double[] planeFunction = new double[4];
                    double detB1C2 = point2.Values[0] * point3.Values[1] - point2.Values[1] * point3.Values[0];
                    double detB1C3 = point2.Values[0] * point3.Values[2] - point2.Values[2] * point3.Values[0];
                    double detB1C4 = point2.Values[0] - point3.Values[0];
                    double detB2C3 = point2.Values[1] * point3.Values[2] - point2.Values[2] * point3.Values[1];
                    double detB2C4 = point2.Values[1] - point3.Values[1];
                    double detB3C4 = point2.Values[2] - point3.Values[2];
                    planeFunction[0] = -(point1.Values[1] * detB3C4 - point1.Values[2] * detB2C4 + detB2C3);
                    planeFunction[1] = point1.Values[0] * detB3C4 - point1.Values[2] * detB1C4 + detB1C3;
                    planeFunction[2] = -(point1.Values[0] * detB2C4 - point1.Values[1] * detB1C4 + detB1C2);
                    planeFunction[3] = point1.Values[0] * detB2C3 - point1.Values[1] * detB1C3 + point1.Values[2] * detB1C2;
                
                    int[] sizes = values.Lengths;
                    for (int i = 0; i < sizes[0]; i++)
                    {
                        this.ProgressInfo(((float)i / sizes[0]));
                        for (int j = 0; j < sizes[1]; j++)
                            for (int k = 0; k < sizes[2]; k++)
                            {
                                int[] coords = { i, j, k };
                                PointND pValue = data.GetPoint(coords);
                                /*double a11 = point1.Values[0] - pValue.Values[0];
                                double a12 = point1.Values[1] - pValue.Values[1];
                                double a13 = point1.Values[2] - pValue.Values[2];
                                double a21 = point2.Values[0] - pValue.Values[0];
                                double a22 = point2.Values[1] - pValue.Values[1];
                                double a23 = point2.Values[2] - pValue.Values[2];
                                double a31 = point3.Values[0] - pValue.Values[0];
                                double a32 = point3.Values[1] - pValue.Values[1];
                                double a33 = point3.Values[2] - pValue.Values[2];

                                double distance = a11 * (a22 * a33 - a23 * a32) - a12 * (a21 * a33 - a31 * a23) + a13
                                        * (a21 * a32 - a22 * a31);*/
                                double distance = pValue.Values[0] * planeFunction[0] + pValue.Values[1] * planeFunction[1] + pValue.Values[2] * planeFunction[2] + planeFunction[3];
                                newValues.Array.SetValue(new Scalar(distance), coords);
                            }
                    }
                }
                data.AddPointAttrs(this.cuttingAttributesName, newValues);
                this.SetOutput("Output Grid", this.data, false);
                paramsChanged = false;
            }
            else
            {
                this.SetOutput("Output Grid", data, true);
            }
        }

        #region Properties

        /// <summary>
        /// Set name of RegGridND point attribute. Default set to Values.
        /// </summary>
        [Description("Set name of RegGridND point attribute. Default set to \"Values\"")]
        public string PointAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.pointAttributesName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.pointAttributesName; }
        }

        /// <summary>
        /// Set name of RegGridND attribute with cutting information. Default set to CuttingValues.
        /// </summary>
        [Description("Set name of RegGridND attribute with cutting information. Default set to\"CuttingValues\"")]
        public string CuttingAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.cuttingAttributesName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.cuttingAttributesName; }
        }


        /// <summary>
        /// Coordinates of the plane. Should be a comma separated list of floating point numbers.
        /// The plane is needed in implicit form (a.x + b.y + c.z + d = 0). 
        /// Example: a, b, c, d where a, b, c, d are floating point numbers.
        /// </summary>
        [Description("Plane coordinates. The plane is needed in implicit form (a.x + b.y + c.z + d = 0)."+
            " Example: a,b,c,d")]
        public string Plane
        {
            get { return plane; }
            set
            {
                if (value == null || value.Equals(""))
                    return;
                try
                {
                    string[] values = value.Split(new char[] { ',' });
                    if (values.Length != 4)
                        throw new MveException(this, "The plane is needed in implicit form " +
                            "(a.x + b.y + c.z + d = 0). Example: a, b, c, d where a, b, c, d are 4 floating point numbers.");
                    planeCoords = new double[values.Length];
                    for (int i = 0; i < values.Length; i++)
                    {
                        planeCoords[i] = double.Parse(values[i].Trim(), NumberFormatInfo.InvariantInfo);
                    }
                    plane = value;
                    paramsChanged = true;
                }
                catch (FormatException)
                {
                    planeCoords = null;
                    throw new MveException(this, "Plane coordinats must be in specified implicit form"+
                        "(a.x + b.y + c.z + d = 0). Example: a, b, c, d where a, b, c, d are 4 floating point numbers.");
                }
            }
        }

        #endregion
    }
}
