using System.Collections.Generic;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Module extracts iso-surface from input volume data with Marching Cubes algorithm. Generates large amount of triangles/points to render - it would be nice to use some module 
    /// for triangle mesh simplification or another suitable post-processing. Module shall be used with 'VolLoader' module, which provides desired input data. 
    /// </summary>
    [ModuleInfo("Ji Zapletal", "Marching Cubes iso-surface extractor.\nModule shall be used with 'VolLoader' module, which provides desired input data.", IconName = "MC.ico", Category = "Gsvd")]
    [PortInfo("InThreshold", "This value defines iso-surface threshold (in percents).\nNon-required input port.")]
    [PortInfo("InRegGridND", "Input volume data.")]
    [PortInfo("Out", "Output data - triangle mesh.")]
    public class MarchingCubes : IsoSurfaceExtractor
    {
        private int cubeIndex = 0, hashIndex = 0;
        private Edge edge = new Edge();
        private Dictionary<Edge, int> dict;

        /// <summary> Constructor. Creates input and output ports. </summary>
        public MarchingCubes() : base() { } // of MarchingCubes        

        /// <summary> Executes module functionality. </summary>
        public override void Execute()
        {
            ProgressInfo(0f);
            if (!IsAllDataSame() || paramsChanged) // if I have to compute again 
            {
                ComputeIsoSurface();
                SetOutput("Out", mesh);
            }
            else
            {
                SetOutput("Out", mesh, true);
                IsoSurfaceExtractor.GCClean();
            }
        } // of Execute

        /// <summary> This method initializes variables and call cube extracting method. </summary>
        protected override void ComputeIsoSurface() { base.ComputeIsoSurface(); }

        /// <summary> Initializes variables.  </summary>
        protected override void Initialize()
        {
            dict = null;
            base.Initialize();
            indexList = new int[12];
            dict = new Dictionary<Edge, int>();
        }

        /// <summary> This method makes cubes from values in two neghbouring slices. The convention of cubes vertices numbering is shown below. 
        /// After each cube is created, it is commited to the next step - checked whether it is intersected by searched iso-surface and eventually appropriate triangles are generated.
        ///   4 ------ 5
        ///  /|       /|
        /// 7 ------ 6 |
        /// | |      | |
        /// | 0 -----| 1
        /// |/       |/ 
        /// 3 ------ 2      </summary> 
        protected override void DecomposeVolumeIntoCubes() { base.DecomposeVolumeIntoCubes(); } // of DecomposeVolumeIntoCubes        

        /// <summary> This method generates triangles for each cell of the volume. </summary>
        /// <param name="cube">Cube to process.</param>
        /// <param name="i">x-axis index</param>
        /// <param name="j">y-axis index</param>
        /// <param name="k">z-axis index</param>
        protected override void GenerateTriangles(Cube cube, int i, int j, int k)
        {
            if (valueDataType == ValueDataType.Integer)
                GenerateTrianglesI(cube);
            else if (valueDataType == ValueDataType.Scalar)
                GenerateTrianglesD(cube);
        } // of GenerateTriangles

        /// <summary> Checks whether a cube is intersected by searched iso-surface and eventually generates appropriate triangles. </summary>
        protected void GenerateTrianglesI(Cube cube)
        {
            cubeIndex = 0;

            // create index which vertices of the cube are under iso-surface
            if (cube.p[0].value < isoValue) cubeIndex |= 1;
            if (cube.p[1].value < isoValue) cubeIndex |= 2;
            if (cube.p[2].value < isoValue) cubeIndex |= 4;
            if (cube.p[3].value < isoValue) cubeIndex |= 8;
            if (cube.p[4].value < isoValue) cubeIndex |= 16;
            if (cube.p[5].value < isoValue) cubeIndex |= 32;
            if (cube.p[6].value < isoValue) cubeIndex |= 64;
            if (cube.p[7].value < isoValue) cubeIndex |= 128;

            // stop if there is no intersection 
            if (edgeTable[cubeIndex] == 0)
                return;

            // calculate normals for all cubes vertices
            for (int i = 0; i < 8; i++)
                cube.p[i].normal = CalcNormalI(cube.p[i], dataI, size);

            previousCube = cube;

            // find edges where we found intersection and calculate positions of triangles vertices
            if ((edgeTable[cubeIndex] & 1) != 0)
                GetIndex(0, cube.p[0], cube.p[1]);
            if ((edgeTable[cubeIndex] & 2) != 0)
                GetIndex(1, cube.p[1], cube.p[2]);
            if ((edgeTable[cubeIndex] & 4) != 0)
                GetIndex(2, cube.p[2], cube.p[3]);
            if ((edgeTable[cubeIndex] & 8) != 0)
                GetIndex(3, cube.p[3], cube.p[0]);
            if ((edgeTable[cubeIndex] & 16) != 0)
                GetIndex(4, cube.p[4], cube.p[5]);
            if ((edgeTable[cubeIndex] & 32) != 0)
                GetIndex(5, cube.p[5], cube.p[6]);
            if ((edgeTable[cubeIndex] & 64) != 0)
                GetIndex(6, cube.p[6], cube.p[7]);
            if ((edgeTable[cubeIndex] & 128) != 0)
                GetIndex(7, cube.p[7], cube.p[4]);
            if ((edgeTable[cubeIndex] & 256) != 0)
                GetIndex(8, cube.p[0], cube.p[4]);
            if ((edgeTable[cubeIndex] & 512) != 0)
                GetIndex(9, cube.p[1], cube.p[5]);
            if ((edgeTable[cubeIndex] & 1024) != 0)
                GetIndex(10, cube.p[2], cube.p[6]);
            if ((edgeTable[cubeIndex] & 2048) != 0)
                GetIndex(11, cube.p[3], cube.p[7]);

            // from calculated vertices create triangles
            for (int i = 0; triTable[cubeIndex, i] != -1; i += 3)
            {
                indices.Add(new Triangle(indexList[triTable[cubeIndex, i]],
                                         indexList[triTable[cubeIndex, i + 1]],
                                         indexList[triTable[cubeIndex, i + 2]]));
            }
        } // of GenerateTrianglesI

        /// <summary> Checks whether a cube is intersected by searched iso-surface and eventually generates appropriate triangles. </summary>
        protected void GenerateTrianglesD(Cube cube)
        {
            cubeIndex = 0;

            // create index which vertices of the cube are under iso-surface
            if (cube.p[0].value < isoValue) cubeIndex |= 1;
            if (cube.p[1].value < isoValue) cubeIndex |= 2;
            if (cube.p[2].value < isoValue) cubeIndex |= 4;
            if (cube.p[3].value < isoValue) cubeIndex |= 8;
            if (cube.p[4].value < isoValue) cubeIndex |= 16;
            if (cube.p[5].value < isoValue) cubeIndex |= 32;
            if (cube.p[6].value < isoValue) cubeIndex |= 64;
            if (cube.p[7].value < isoValue) cubeIndex |= 128;

            // stop if there is no intersection 
            if (edgeTable[cubeIndex] == 0)
                return;

            // calculate normals for all cubes vertices
            for (int i = 0; i < 8; i++)
                cube.p[i].normal = CalcNormalD(cube.p[i], dataD, size);

            previousCube = cube;

            // find edges where we found intersection and calculate positions of triangles vertices
            if ((edgeTable[cubeIndex] & 1) != 0)
                GetIndex(0, cube.p[0], cube.p[1]);
            if ((edgeTable[cubeIndex] & 2) != 0)
                GetIndex(1, cube.p[1], cube.p[2]);
            if ((edgeTable[cubeIndex] & 4) != 0)
                GetIndex(2, cube.p[2], cube.p[3]);
            if ((edgeTable[cubeIndex] & 8) != 0)
                GetIndex(3, cube.p[3], cube.p[0]);
            if ((edgeTable[cubeIndex] & 16) != 0)
                GetIndex(4, cube.p[4], cube.p[5]);
            if ((edgeTable[cubeIndex] & 32) != 0)
                GetIndex(5, cube.p[5], cube.p[6]);
            if ((edgeTable[cubeIndex] & 64) != 0)
                GetIndex(6, cube.p[6], cube.p[7]);
            if ((edgeTable[cubeIndex] & 128) != 0)
                GetIndex(7, cube.p[7], cube.p[4]);
            if ((edgeTable[cubeIndex] & 256) != 0)
                GetIndex(8, cube.p[0], cube.p[4]);
            if ((edgeTable[cubeIndex] & 512) != 0)
                GetIndex(9, cube.p[1], cube.p[5]);
            if ((edgeTable[cubeIndex] & 1024) != 0)
                GetIndex(10, cube.p[2], cube.p[6]);
            if ((edgeTable[cubeIndex] & 2048) != 0)
                GetIndex(11, cube.p[3], cube.p[7]);

            // from calculated vertices create triangles
            for (int i = 0; triTable[cubeIndex, i] != -1; i += 3)
            {
                indices.Add(new Triangle(indexList[triTable[cubeIndex, i]],
                                         indexList[triTable[cubeIndex, i + 1]],
                                         indexList[triTable[cubeIndex, i + 2]]));
            }
        } // of GenerateTrianglesD

        /// <summary> Gets index of the point form array of already computed points or adds new point into this array. </summary>
        /// <param name="index">Index of intersected edge.</param>
        /// <param name="a">First vertex of the edge.</param>
        /// <param name="b">Second vertex of the edge.</param>
        private void GetIndex(int index, Vertex a, Vertex b)
        {
            if (a.value < b.value)
            {
                edge.p1 = a.index;
                edge.p2 = b.index;
            }
            else
            {
                edge.p1 = b.index;
                edge.p2 = a.index;
            }

            indexList[index] = FindPointIndex(edge);

            if (indexList[index] == -1)
            {
                vertList[index] = Interpolate(a, b, isoValue, inverseDistortion, linearInterpolation);
                dict.Add(edge, hashIndex);
                pts.Add(vertList[index]);
                indexList[index] = hashIndex;
                hashIndex++;
            }
        } // of GetIndex

        /// <summary> Gets index = value from the hashtable of indices. </summary>
        /// <param name="edge">On which edge the point is.</param>
        /// <returns>Index of the point.</returns>
        private int FindPointIndex(Edge edge)
        {
            int value;
            if (dict.TryGetValue(edge, out value))
                return value;
            else
                return -1;
        } // of FindPointIndex

        /// <summary> Creates from computed data output data </summary>
        protected override void CreateOutput()
        {
            base.CreateOutput();

            hashIndex = 0;
            dict = null;
            IsoSurfaceExtractor.GCClean();

            ProgressInfo(1f);
        } // of CreateOutput

    } // of MarchingCubes class
} // of namespace Zcu.Gsvd.Extractions
