﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Collections.Generic;
using System.Threading;
using System.Text;
using System.ComponentModel;
using System.Diagnostics;
using System.Windows.Forms.Design;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{
    [ModuleInfo("Jiří Zapletal", "This module analyzes volume data in the sense of creating histogram from its values.", IconName = "Analyzer.ico", Category = "Gsvd")]
    [PortInfo("InVolData", "Input volume data.")]
    public class VolumeDataHistogram : Zcu.Mve.Core.Module
    {
        public Color oddBarColor = Color.Black, evenBarColor = Color.Red;
        public bool paramsChanged = true, showGuideLines = true, logarithmicScale = true;
        public int fontSize = 9, width = 800, height = 600, leftMargin = 90, margin = 70, yLabels = 3, xLabels = 15;

        Bitmap bmp = null;
        private string pointAttributesName = "Values";
        int[] data = null;
        private SortedList<int, int> sortedList = null;
        int minValue = int.MaxValue, maxValue = int.MinValue, minValueKey = 0, maxValueKey = 0;
        RegGridND inData = null;

        /// <summary>
        /// Constructor. Creates input and output ports.
        /// </summary>
        public VolumeDataHistogram()
        {
            AddInPort("InVolData", typeof(RegGridND));
        } // of MarchingCubes

        /// <summary>
        /// Executes module functionality.
        /// </summary>
        public override void Execute()
        {
            ProgressInfo(0f);
            if (!IsAllDataSame())
                GenerateHistogram();

            if (paramsChanged)
                this.DrawHistogram();
        } // of Execute

        /// <summary>
        /// Main function that generates histogram for volume data.
        /// </summary>
        public void GenerateHistogram()
        {
            IsoSurfaceExtractor.GCClean();

            minValue = int.MaxValue; maxValue = int.MinValue;
            minValueKey = 0; maxValueKey = 0;
            sortedList = new SortedList<int, int>();
            inData = (RegGridND)GetInput("InVolData");
            MveArray dataMveArray = inData.GetPointAttrs(pointAttributesName);
            int[] size = dataMveArray.Lengths;

            #region Convert data from MVE-2 types to .NET for faster processing
            Console.Write("Converting data...");

            int ticks = 0;
            int size2size1 = size[2] * size[1];
            data = new int[size2size1 * size[0]];

            int index1 = 0, index2 = 0;
            double size0 = size[0];

            ticks = Environment.TickCount;
            for (int i = 0; i < size[0]; i++)
            {
                index1 = i * size2size1;
                for (int j = 0; j < size[1]; j++)
                {
                    index2 = index1 + j * size[2];
                    for (int k = 0; k < size[2]; k++)
                        data[index2 + k] = (Integer)dataMveArray[i, j, k];
                }
                ProgressInfo(0.2 * (i + 1) / size0);
            }
            Console.WriteLine("\t\tDone ({0} ms)", Environment.TickCount - ticks);
            dataMveArray = null;
            inData = null;
            #endregion

            int time = Environment.TickCount;
            Console.Write("Generating histogram...");
            
            CalculateItemOccurence();
            
            DrawHistogram();
            
            Console.WriteLine("\tDone ({0} ms)", time = Environment.TickCount - time);
            ProgressInfo(1f);
            paramsChanged = false;

            IsoSurfaceExtractor.GCClean();
        }

        /// <summary>
        /// Method for calculating sorted list, representing histogram by data value on x-axis and appropriate occurence of that value in the dataset on y-axis.
        /// </summary>
        private void CalculateItemOccurence()
        {
            // for each data value, calculate its position and occurence in the sorted list
            for (int i = 0; i < data.Length; i++)
            {
                Add(data[i]);
                ProgressInfo(0.2 + 0.8 * (i + 1) / data.Length);
            }

            foreach (KeyValuePair<int, int> item in sortedList)
            {
                // recalculate min value for statistics and range of histogram graph axes
                if (item.Value < minValue)
                {
                    minValue = item.Value;
                    minValueKey = item.Key;
                }
            }
        }

        /// <summary>
        /// Method for inserting values into sorted list, representing histogram.
        /// </summary>
        /// <param name="key">Key (data value) to insert to the list.</param>
        private void Add(int key)
        {
            // if this value is already in the list, increment occurence value on its position.
            if (sortedList.ContainsKey(key))
            {
                sortedList[key]++;
                if (sortedList[key] > maxValue) // recalculate max value for statistics and range of histogram graph axes
                {
                    maxValue = sortedList[key];
                    maxValueKey = key;
                }
            }
            else // insert new value to the list
                sortedList.Add(key, 1);
        }

        /// <summary>
        /// Method that calculates layout of the histogram on the form and creates each histogram component (occurence bar for each data value, axis, labels, guide lines, colors, margins, etc.)
        /// </summary>
        public void DrawHistogram()
        {
            Brush oddBarBrush = new SolidBrush(oddBarColor), evenBarBrush = new SolidBrush(evenBarColor);
            float minValueLog = (float)Math.Log10(minValue), maxValueLog = (float)(Math.Log10(maxValue));

            bmp = new Bitmap(width, height);
            Graphics g = Graphics.FromImage(bmp);
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
            g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;
            RectangleF rect;
            Font font = new Font("Arial", fontSize, FontStyle.Regular),
                 labelFont = new Font("Arial", fontSize, FontStyle.Bold);

            int i = 0, maxColumnHeight = height - 2 * margin;
            float xpos, columnHeight, clientWidth = (width - margin - leftMargin), columnWidth = clientWidth / (float)sortedList.Count;
            bool jump = true;
            string text;

            // y axis label
            StringFormat drawFormat = new StringFormat();
            drawFormat.FormatFlags = StringFormatFlags.DirectionVertical;
            g.DrawString("[occurence]", labelFont, Brushes.Black, leftMargin - 1.5f * font.Size, margin, drawFormat);

            // x axis margin
            g.DrawString("[value]", labelFont, Brushes.Black, width - 1.5f * margin, height - margin / 1.5f);
            text = "Value:\t\tMin = " + sortedList.Keys[0] + "  /  Max = " + sortedList.Keys[sortedList.Count - 1];
            g.DrawString(text, labelFont, Brushes.Black, width / 4, margin / 4);
            text = "Occurence:\tMin = " + minValue.ToString() + " (value = " + minValueKey + ")  /  Max = " + maxValue.ToString() + " (value = " + maxValueKey + ")";
            g.DrawString(text, labelFont, Brushes.Black, width / 4, margin / 4 + 1.5f * font.Size);

            // draw guide lines and y-axis values
            for (int j = 0; j <= yLabels; j++)
            {
                if (showGuideLines)
                    g.DrawLine(Pens.Green, leftMargin, height - margin - j * maxColumnHeight / yLabels, width - margin, height - margin - j * maxColumnHeight / yLabels);

                if (logarithmicScale)
                    g.DrawString(Math.Pow(10, minValueLog + j * (maxValueLog - minValueLog) / yLabels).ToString("0"), font, Brushes.Black, leftMargin / 4, height - margin - j * maxColumnHeight / yLabels - 1.5f * font.Size);
                else
                    g.DrawString((minValue + j * (maxValue - minValue) / yLabels).ToString(), font, Brushes.Black, leftMargin / 4, height - margin - j * maxColumnHeight / yLabels - 1.5f * font.Size);
            }

            // draw occurence bars
            foreach (KeyValuePair<int, int> item in sortedList)
            {
                text = item.Key.ToString();
                xpos = leftMargin + i * columnWidth;

                if (logarithmicScale)
                    columnHeight = 2 + (maxColumnHeight * ((float)Math.Log10(item.Value) - minValueLog) / (maxValueLog - minValueLog));
                else
                    columnHeight = 2 + (maxColumnHeight * (item.Value - minValue) / (float)(maxValue - minValue));

                rect = new RectangleF(xpos, margin + maxColumnHeight - columnHeight, columnWidth, columnHeight);


                if ((sortedList.Count > xLabels) && (i % (sortedList.Count / (xLabels - 1)) == 0) || (i == sortedList.Count - 1))
                {
                    g.DrawString(text, font, Brushes.Black, xpos + columnWidth / 2 - (text.Length * font.Size / 2), height - margin / 1.15f);
                    g.DrawLine(Pens.Black, xpos + columnWidth / 2, height - margin, xpos + columnWidth / 2, height - margin + 5);
                }

                if (jump)
                    g.FillRectangle(oddBarBrush, rect);
                else
                    g.FillRectangle(evenBarBrush, rect);

                jump = !jump;
                i++;
            }

            Thread formThread = new Thread(new ThreadStart(this.RunRenderer));
            formThread.Start();
            paramsChanged = false;
        }
        
        /// <summary>
        /// Create and display histogram form
        /// </summary>
        private void RunRenderer()
        {
            HistogramForm form = new HistogramForm(bmp);
            form.ShowDialog();
        }


        #region Properties Histogram
        /// <summary>
        /// Color of odd bars in histogram.
        /// </summary>
        [Editor(typeof(ColorPalette), typeof(ColorPalette))]
        [CategoryAttribute("Histogram"), Description(" Color of odd bars in histogram.")]
        public Color OddBarColor
        {
            get { return (oddBarColor); }
            set
            {
                oddBarColor = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Color of even bars in histogram.
        /// </summary>
        [Editor(typeof(ColorPalette), typeof(ColorPalette))]
        [CategoryAttribute("Histogram"), Description(" Color of even bars in histogram.")]
        public Color EvenBarColor
        {
            get { return (evenBarColor); }
            set
            {
                evenBarColor = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Defines, whether to draw guidelines of the histogram.
        /// </summary>
        [CategoryAttribute("Histogram"), Description("Defines, whether to draw guidelines of the histogram.")]
        public bool ShowGuideLines
        {
            get { return (showGuideLines); }
            set
            {
                showGuideLines = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Defines, whether to print guidelines of the histogram.
        /// </summary>
        [CategoryAttribute("Histogram"), Description("Use logarithmic scale for the y-axis.")]
        public bool LogarithmicScale
        {
            get { return (logarithmicScale); }
            set
            {
                logarithmicScale = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Width of the form where histogram is drawn.
        /// </summary>
        [CategoryAttribute("Histogram"), Description("Width of the form where histogram is drawn.")]
        public int Width
        {
            get { return (width); }
            set
            {
                width = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Height of the form where histogram is drawn.
        /// </summary>
        [CategoryAttribute("Histogram"), Description("Height of the form where histogram is drawn.")]
        public int Height
        {
            get { return (height); }
            set
            {
                height = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Number of labels on x axis.
        /// </summary>
        [CategoryAttribute("Histogram"), Description("Number of labels on x axis.")]
        public int XLabels
        {
            get { return (xLabels); }
            set
            {
                xLabels = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Number of labels on y axis.
        /// </summary>
        [CategoryAttribute("Histogram"), Description("Number of labels on y axis.")]
        public int YLabels
        {
            get { return (yLabels); }
            set
            {
                yLabels = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Size of the font of labels.
        /// </summary>
        [CategoryAttribute("Histogram"), Description("Number of labels on y axis.")]
        public int FontSize
        {
            get { return (fontSize); }
            set
            {
                fontSize = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Left margin (between border of the form and the y axis).
        /// </summary>
        [CategoryAttribute("Histogram"), Description("Left margin (between border of the form and the y axis).")]
        public int LeftMargin
        {
            get { return (leftMargin); }
            set
            {
                leftMargin = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Margin of the border of the form.
        /// </summary>
        [CategoryAttribute("Histogram"), Description("Margin of the border of the form.")]
        public int Margin
        {
            get { return (margin); }
            set
            {
                margin = value;
                paramsChanged = true;
            }
        }
        #endregion
    }
}
